/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Access - OrderService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.business;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.exception.SolidifyFileDeleteException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.rest.Result;
import ch.unige.solidify.service.ResourceService;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.controller.AccessController;
import ch.dlcm.message.AipDownloadMessage;
import ch.dlcm.model.MetadataVisibility;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.access.Order;
import ch.dlcm.model.access.Order.OrderStatus;
import ch.dlcm.model.access.OrderArchive;
import ch.dlcm.model.access.QueryType;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.repository.OrderRepository;
import ch.dlcm.service.SearchMgmt;
import ch.dlcm.specification.OrderSpecification;

@Service
@ConditionalOnBean(AccessController.class)
public class OrderService extends ResourceService<Order> {
  private static final Logger log = LoggerFactory.getLogger(OrderService.class);

  private final DLCMProperties dlcmProperties;
  private final SearchMgmt searchMgmt;
  private final ArchivalInfoPackageService aipService;
  public OrderService(DLCMProperties dlcmProperties,
          SearchMgmt searchMgmt,
          ArchivalInfoPackageService aipService) {
    this.dlcmProperties = dlcmProperties;
    this.searchMgmt = searchMgmt;
    this.aipService = aipService;
  }

  public Order saveOrder(Order order) {
    RestCollection<OrderArchive> orderArchiveCollection;
    Pageable pageable = PageRequest.of(0, RestCollectionPage.DEFAULT_SIZE_PAGE);
    do {
      orderArchiveCollection = this.searchMgmt.search(MetadataVisibility.PUBLIC, order, true, pageable);
      for (final OrderArchive orderArchive : orderArchiveCollection.getData()) {
        final ArchivalInfoPackage archive = orderArchive.getArchive();
        if (archive.getInfo().getStatus() != PackageStatus.COMPLETED &&
                archive.getInfo().getStatus() != PackageStatus.DISPOSED) {
          log.warn("AIP ({}) not ready for Access module", archive.getResId());
          continue;
        }
        if (!this.aipService.existsById(archive.getResId())) {

          boolean isBigPackage = archive.getArchiveSize() != null
                  && archive.getArchiveSize() > this.dlcmProperties.getParameters().getFileSizeLimit().toBytes();

          archive.reset();
          this.aipService.save(archive);
          SolidifyEventPublisher.getPublisher().publishEvent(new AipDownloadMessage(archive.getResId(), isBigPackage));
        }
        order.addItem(archive);
      }
      pageable = pageable.next();
    } while (orderArchiveCollection.getPage().hasNext());
    order = this.save(order);

    order.add(linkTo(this.getClass()).slash(order.getResId()).slash(ActionName.SAVE).withSelfRel());
    order.add(linkTo(this.getClass()).slash(order.getResId()).withRel(ActionName.PARENT));
    return order;
  }

  public HttpEntity<Result> changeOrderStatus(OrderStatus newStatus, String orderId) {

    final Order order = this.findOne(orderId);

    final Result res = new Result(orderId);
    if (this.statusUpdateIsValid(order, newStatus)) {
      order.setStatus(newStatus);
      this.save(order);
      res.setStatus(Result.ActionStatus.EXECUTED);
      res.setMesssage(this.messageService.get("message.order.status"));
      return new ResponseEntity<>(res, HttpStatus.OK);
    } else {
      res.setStatus(Result.ActionStatus.NON_APPLICABLE);
      res.setMesssage(this.messageService.get("message.order.status.error"));
      return new ResponseEntity<>(res, HttpStatus.BAD_REQUEST);
    }
  }

  @Override
  public void delete(String orderId) {
    super.delete(orderId);
    final Path orderPath = Paths.get(this.dlcmProperties.getOrderLocation()).resolve(orderId);
    if (FileTool.isFolder(orderPath) && !FileTool.deleteFolder(orderPath)) {
      throw new SolidifyFileDeleteException(orderPath.toString());
    }
  }

  /**
   * Check if an order status can be manually updated to the new given status, according to the
   * order's management workflow.
   *
   * @param order The order to check
   * @param newStatus The new status to apply
   * @return True if the status update is valid
   */
  public boolean statusUpdateIsValid(Order order, Order.OrderStatus newStatus) {

    final Order.OrderStatus currentStatus = order.getStatus();

    switch (currentStatus) {
      case IN_PROGRESS:
        return newStatus == Order.OrderStatus.SUBMITTED;

      case IN_ERROR:
        return newStatus == Order.OrderStatus.IN_PROGRESS;

      case SUBMITTED:
      case IN_DISSEMINATION_PREPARATION:
      case READY:
      case IN_PREPARATION:
      case DOWNLOADING:
      default:
        return false;
    }
  }

  public Order createOrderForAIP(String aipId) {
    Order order = new Order();
    order.setName(aipId);
    order.setQueryType(QueryType.DIRECT);
    order.setQuery(aipId);
    order = this.save(order);
    return order;
  }

  public List<Order> findDirectOrderByAip(String aipId) {
    return ((OrderRepository) this.itemRepository).findDirectOrderByAip(aipId);
  }

  public List<Order> findByAip(String aipId) {
    return ((OrderRepository) this.itemRepository).findByAip(aipId);
  }

  @Override
  protected void validateItemSpecificRules(Order order, BindingResult errors) {

    if (StringTool.isNullOrEmpty(order.getQuery())) {
      errors.addError(new FieldError(order.getClass().getSimpleName(), "query", this.messageService.get("validation.order.query.empty")));
    }

    switch (order.getQueryType()) {
      case ADVANCED:
        // TODO: implement advanced query
        break;
      case DIRECT:
        // Check if AIP number
        this.checkAipNumber(order, errors, 1);
        // Check order integrity
        if (!StringTool.isNullOrEmpty(order.getQuery())) {

          if (!order.getQuery().equals(order.getName())) {
            errors.addError(new FieldError(order.getClass().getSimpleName(), "query",
                    this.messageService.get("validation.order.nameAndQuery.notMatch")));
          }

          if (order.getStatus().equals(OrderStatus.SUBMITTED)) {
            if (order.getAipPackages().isEmpty()) {
              errors.addError(new FieldError(order.getClass().getSimpleName(), "aipPackages",
                      this.messageService.get("validation.order.submittedWithoutAip")));
            } else if (!order.getQuery().equals(order.getAipPackages().get(0).getResId())) {
              errors.addError(new FieldError(order.getClass().getSimpleName(), "query",
                      this.messageService.get("validation.order.directInvalidName", new Object[] { order.getName(), order.getQuery() })));
            }
          }
        }
        break;
      case SIMPLE:
        // Check if AIP number
        this.checkAipNumber(order, errors, order.getQuery().split(SolidifyConstants.FIELD_SEP).length);
        break;
      default:
        errors.addError(new FieldError(order.getClass().getSimpleName(), "queryType",
                this.messageService.get("validation.order.invalidType", new Object[] { order.getQueryType() })));
        break;
    }
  }

  private void checkAipNumber(Order order, BindingResult errors, int expectedNumber) {
    if (!StringTool.isNullOrEmpty(order.getQuery())) {
      int aipNumber = order.getQuery().split(SolidifyConstants.FIELD_SEP).length;
      if (aipNumber != expectedNumber) {
        errors.addError(new FieldError(order.getClass().getSimpleName(), "query",
                this.messageService.get("validation.order.wrongArchiveNumber", new Object[] { expectedNumber, aipNumber })));
      }
      if (order.getStatus().equals(OrderStatus.SUBMITTED) && order.getAipPackages().size() != expectedNumber) {
        errors.addError(new FieldError(order.getClass().getSimpleName(), "query",
                this.messageService.get("validation.order.wrongArchiveNumber", new Object[] { expectedNumber, order.getAipPackages().size() })));
      }
    }
  }

  /**
   * @return the cumulated size of all AIP contained in the order
   */
  public long getSize(String resId) {
    return ((OrderRepository) this.itemRepository).getSize(resId);
  }

  @Override
  public OrderSpecification getSpecification(Order resource) {
    return new OrderSpecification(resource);
  }
}
