/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Access - MetadataController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller.access;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;
import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.methodOn;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.util.MultiValueMap;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.servlet.mvc.method.annotation.StreamingResponseBody;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.model.dto.CitationDto;
import ch.unige.solidify.rest.FacetPage;
import ch.unige.solidify.rest.FacetResult;
import ch.unige.solidify.rest.FacetResultValue;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.SearchCondition;
import ch.unige.solidify.rest.SearchConditionType;
import ch.unige.solidify.security.EveryonePermissions;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.JSONTool;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.business.OrderService;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.controller.AccessController;
import ch.dlcm.model.MetadataVisibility;
import ch.dlcm.model.dto.ArchiveStatisticsDto;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.model.settings.RatingType;
import ch.dlcm.repository.DisseminationInfoPackageRepository;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.rest.UrlPath;
import ch.dlcm.service.ArchiveUserFingerprintService;
import ch.dlcm.service.SearchService;
import ch.dlcm.service.citation.DLCMCitationGeneratorService;
import ch.dlcm.service.rest.fallback.FallbackArchiveStatisticsRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveThumbnailRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackOrganizationalUnitRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedArchiveRatingRemoteResourceService;

@EveryonePermissions
@RestController
@ConditionalOnBean(AccessController.class)
@RequestMapping(UrlPath.ACCESS_PUBLIC_METADATA)
public class MetadataController extends SearchController {
  private static final Logger log = LoggerFactory.getLogger(MetadataController.class);

  private final FallbackOrganizationalUnitRemoteResourceService organizationalUnitRemoteResourceService;
  private final FallbackArchiveThumbnailRemoteResourceService archiveThumbnailRemoteResourceService;
  private final FallbackArchiveStatisticsRemoteResourceService archiveStatisticsRemoteResourceService;
  private final TrustedArchiveRatingRemoteResourceService archiveRatingRemoteResourceService;

  public MetadataController(
          SearchService searchService,
          DLCMProperties config,
          DisseminationInfoPackageRepository dipRepository,
          FallbackOrganizationalUnitRemoteResourceService organizationalUnitRemoteResourceService,
          OrderService orderService,
          FallbackArchiveThumbnailRemoteResourceService archiveThumbnailRemoteResourceService,
          FallbackArchiveStatisticsRemoteResourceService archiveStatisticsRemoteResourceService,
          TrustedArchiveRatingRemoteResourceService archiveRatingRemoteResource,
          ArchiveUserFingerprintService archiveUserFingerprintService,
          DLCMCitationGeneratorService dlcmCitationGeneratorService) {
    super(searchService, config, dipRepository, orderService, archiveUserFingerprintService, dlcmCitationGeneratorService);
    this.organizationalUnitRemoteResourceService = organizationalUnitRemoteResourceService;
    this.archiveThumbnailRemoteResourceService = archiveThumbnailRemoteResourceService;
    this.archiveStatisticsRemoteResourceService = archiveStatisticsRemoteResourceService;
    this.archiveRatingRemoteResourceService = archiveRatingRemoteResource;
  }

  @Override
  protected MetadataVisibility getIndex() {
    return MetadataVisibility.PUBLIC;
  }

  @Override
  public HttpEntity<ArchiveMetadata> get(@PathVariable String id) {
    return super.get(id);
  }

  @Override
  public HttpEntity<List<CitationDto>> bibliographies(@PathVariable String id) {
    return super.bibliographies(id);
  }

  @Override
  public HttpEntity<List<CitationDto>> citations(@PathVariable String id) {
    return super.citations(id);
  }

  @Override
  public HttpEntity<RestCollection<ArchiveMetadata>> list(@ModelAttribute ArchiveMetadata search, Pageable pageable) {
    return super.list(search, pageable);
  }

  @Override
  public HttpEntity<RestCollection<ArchiveMetadata>> search(@RequestParam(defaultValue = "*") String query,
          @RequestParam(required = false) String conditions, Pageable pageable) {
    return super.search(query, conditions, pageable);
  }

  @Override
  public HttpEntity<RestCollection<ArchiveMetadata>> searchPost(@RequestBody MultiValueMap<String, String> params) {
    return super.searchPost(params);
  }

  @Override
  public HttpEntity<String> getDownloadStatus(@PathVariable String aipId) {
    return super.getDownloadStatus(aipId);
  }

  @GetMapping(SolidifyConstants.URL_ID_PLUS_SEP + DLCMActionName.THUMBNAIL)
  public HttpEntity<StreamingResponseBody> downloadThumbnail(@PathVariable String id) {
    // Check if archive exist
    final ArchiveMetadata archive = this.indexResourceService.findOne(MetadataVisibility.PUBLIC, id);
    if (archive == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    try {
      // Check if thumbnail exists
      Path thumbnailPath;
      if (archive.getMetadata().containsKey(DLCMConstants.AIP_THUMBNAIL)
              && archive.getMetadata().get(DLCMConstants.AIP_THUMBNAIL).equals(Boolean.TRUE)) {
        thumbnailPath = this.getThumbnailPath(id);
        // Check if thumbnail already download
        if (!FileTool.checkFile(thumbnailPath)) {
          this.archiveThumbnailRemoteResourceService.downloadArchiveThumbnail(id, thumbnailPath);
        }
      } else {
        // use orgUnit thumbnail
        String orgUnitId = archive.getMetadata().get(DLCMConstants.AIP_ORGA_UNIT).toString();
        thumbnailPath = this.getThumbnailPath(orgUnitId);
        // Check if thumbnail already download
        if (!FileTool.checkFile(thumbnailPath)) {
          this.organizationalUnitRemoteResourceService.downloadOrgUnitLogo(orgUnitId, thumbnailPath);
        }
      }
      return this.getThumbnail(thumbnailPath);
    } catch (HttpClientErrorException.NotFound e) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
  }

  @GetMapping(SolidifyConstants.URL_ID_PLUS_SEP + DLCMActionName.STATISTICS)
  public HttpEntity<ArchiveStatisticsDto> getStatistics(@PathVariable String id) {
    // Check if archive exist
    final ArchiveMetadata archive = this.indexResourceService.findOne(MetadataVisibility.PUBLIC, id);
    if (archive == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    try {
      return new ResponseEntity<>(this.archiveStatisticsRemoteResourceService.getArchiveStatistics(id), HttpStatus.OK);
    } catch (SolidifyResourceNotFoundException e) {
      // No Statistics yet
      RestCollection<RatingType> ratingTypeList = this.archiveRatingRemoteResourceService.getAllRatingType();

      ArchiveStatisticsDto stat = new ArchiveStatisticsDto();
      stat.init(ratingTypeList.getData());
      return new ResponseEntity<>(stat, HttpStatus.OK);
    } catch (SolidifyRuntimeException e) {
      return new ResponseEntity<>(HttpStatus.INTERNAL_SERVER_ERROR);
    }
  }

  @GetMapping("/" + DLCMActionName.SEARCH_DOI)
  public HttpEntity<ArchiveMetadata> searchDOI(@RequestParam("doi") String doiValue) {
    final SearchService searchService = (SearchService) this.indexResourceService;
    try {
      return new ResponseEntity<>(searchService.searchByDOI(doiValue), HttpStatus.OK);
    } catch (final SolidifyResourceNotFoundException e) {
      log.error("DOI Error", e);
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    } catch (final SolidifyRuntimeException e) {
      log.error("DOI Error", e);
      return new ResponseEntity<>(HttpStatus.INTERNAL_SERVER_ERROR);
    }
  }

  @GetMapping(SolidifyConstants.URL_ID_PLUS_SEP + ResourceName.PACKAGES)
  @PreAuthorize("@metadataPermissionService.isAllowed(#id, 'DOWNLOAD_ARCHIVE') "
          + "|| @downloadWithAclPermissionService.isAllowed(#id)")
  public HttpEntity<ArchiveMetadata> getPackages(@PathVariable String id) {
    // Check if archive exist
    final ArchiveMetadata archive = this.indexResourceService.findOne(MetadataVisibility.RESTRICTED, id);
    if (archive == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    return new ResponseEntity<>(archive, HttpStatus.OK);
  }

  @GetMapping(SolidifyConstants.URL_ID_PLUS_SEP + ResourceName.DATAFILE)
  @PreAuthorize("@metadataPermissionService.isAllowed(#id, 'DOWNLOAD_ARCHIVE') "
          + "|| @downloadWithAclPermissionService.isAllowed(#id)")
  public HttpEntity<RestCollection<ArchiveMetadata>> getDataFiles(@PathVariable String id, @RequestParam(required = false) String path,
          @RequestParam(defaultValue = "*") String file, @RequestParam(defaultValue = "0") String minSize,
          @RequestParam(required = false) String maxSize, Pageable pageable) {
    // Check if archive exist
    final ArchiveMetadata archive = this.indexResourceService.findOne(MetadataVisibility.PUBLIC, id);
    if (archive == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    // Filter on archive ID AND path AND file name
    final String query = DLCMConstants.ARCHIVE_ID_FIELD + ":" + id + " AND file.name:\"" + file + "\"";
    // Condition list
    List<SearchCondition> conditionList = new ArrayList<>();
    // Condition on type
    SearchCondition typeCondition = new SearchCondition(SearchConditionType.TERM, "type.keyword");
    typeCondition.setValue("dataFile");
    conditionList.add(typeCondition);
    // Condition on folder
    if (!StringTool.isNullOrEmpty(path)) {
      SearchCondition pathCondition = new SearchCondition(SearchConditionType.TERM, "file.path.keyword");
      pathCondition.setValue(path);
      conditionList.add(pathCondition);
    }
    // Condition on file size
    SearchCondition sizeCondition = new SearchCondition(SearchConditionType.RANGE, "file.size");
    sizeCondition.setLowerValue(minSize);
    if (!StringTool.isNullOrEmpty(maxSize)) {
      sizeCondition.setUpperValue(maxSize);
    }
    conditionList.add(sizeCondition);
    final String condition = JSONTool.convert2JsonString(conditionList);
    final FacetPage<ArchiveMetadata> itemList = this.indexResourceService.search(MetadataVisibility.RESTRICTED, query, condition,
            Collections.emptyList(), pageable);
    return this.page2Collection(itemList, pageable);
  }

  @GetMapping(SolidifyConstants.URL_ID_PLUS_SEP + ResourceName.AIP)
  @PreAuthorize("@metadataPermissionService.isAllowed(#id, 'DOWNLOAD_ARCHIVE') "
          + "|| @downloadWithAclPermissionService.isAllowed(#id)")
  public HttpEntity<RestCollection<ArchiveMetadata>> getAip(@PathVariable String id, @RequestParam(defaultValue = "0") String minSize,
          @RequestParam(required = false) String maxSize, Pageable pageable) {
    // Check if archive exist
    final ArchiveMetadata archive = this.indexResourceService.findOne(this.getIndex(), id);
    if (archive == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    // Filter on archive ID AND path AND file name
    final String query = DLCMConstants.ARCHIVE_ID_FIELD + ":" + id;
    // Condition list
    List<SearchCondition> conditionList = new ArrayList<>();
    // Condition on type
    SearchCondition typeCondition = new SearchCondition(SearchConditionType.TERM, "type.keyword");
    typeCondition.setValue("aip");
    conditionList.add(typeCondition);
    // Condition on file size
    SearchCondition sizeCondition = new SearchCondition(SearchConditionType.RANGE, "aip.size");
    sizeCondition.setLowerValue(minSize);
    if (!StringTool.isNullOrEmpty(maxSize)) {
      sizeCondition.setUpperValue(maxSize);
    }
    conditionList.add(sizeCondition);
    final String condition = JSONTool.convert2JsonString(conditionList);
    final FacetPage<ArchiveMetadata> itemList = this.indexResourceService.search(MetadataVisibility.RESTRICTED, query, condition,
            Collections.emptyList(), pageable);
    return this.page2Collection(itemList, pageable);
  }

  @GetMapping(SolidifyConstants.URL_ID_PLUS_SEP + DLCMActionName.LIST_FOLDERS)
  @PreAuthorize("@metadataPermissionService.isAllowed(#id, 'DOWNLOAD_ARCHIVE') "
          + "|| @downloadWithAclPermissionService.isAllowed(#id)")
  public HttpEntity<List<String>> listFolders(@PathVariable String id) {
    // Check if archive exist
    final ArchiveMetadata archive = this.indexResourceService.findOne(this.getIndex(), id);
    if (archive == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    // Filter on archive ID to get folder list with the facet PATH_FACET
    final String query = DLCMConstants.ARCHIVE_ID_FIELD + ":" + id;
    final FacetPage<ArchiveMetadata> itemList = this.indexResourceService.search(MetadataVisibility.RESTRICTED, query, null,
            Collections.emptyList(), PageRequest.of(0, 1));
    // Extract folders from the facet PATH_FACET
    final List<String> folderList = new ArrayList<>();
    for (FacetResult facet : itemList.getFacetResults()) {
      if (facet.getName().equals(DLCMConstants.PATH_FACET)) {
        for (FacetResultValue value : facet.getValues()) {
          folderList.add(value.getValue());
        }
        break;
      }
    }

    return new ResponseEntity<>(folderList, HttpStatus.OK);
  }

  @GetMapping(SolidifyConstants.URL_PARENT_ID + ResourceName.DATAFILE + SolidifyConstants.URL_ID)
  @PreAuthorize("@metadataPermissionService.isAllowed(#parentid, 'DOWNLOAD_ARCHIVE') "
          + "|| @downloadWithAclPermissionService.isAllowed(#parentid)")
  public HttpEntity<ArchiveMetadata> getDataFiles(@PathVariable String parentid, @PathVariable String id) {
    // Check if archive exist
    final ArchiveMetadata archive = this.indexResourceService.findOne(MetadataVisibility.PUBLIC, parentid);
    if (archive == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    // Check if archive exist
    final ArchiveMetadata datafile = this.indexResourceService.findOne(MetadataVisibility.RESTRICTED, id);
    if (datafile == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    return new ResponseEntity<>(datafile, HttpStatus.OK);
  }

  @Override
  protected <W extends RepresentationModel<W>> void addOthersLinks(W w) {
    super.addOthersLinks(w);
    w.add(linkTo(methodOn(this.getClass()).searchDOI("doiValue")).withRel(DLCMActionName.SEARCH_DOI));
  }

}
