/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Access - SearchController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller.access;

import static ch.dlcm.model.access.Order.OrderStatus.SUBMITTED;
import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;
import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.methodOn;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.NoSuchElementException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.FileSystemResource;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.util.MultiValueMap;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestParam;

import ch.unige.solidify.IndexConstants;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.model.dto.CitationDto;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.FacetPage;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.security.EveryonePermissions;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.business.OrderService;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.controller.AccessIndexResourceReadOnlyController;
import ch.dlcm.message.ArchiveStatisticMessage;
import ch.dlcm.message.ArchiveStatisticMessage.StatisticType;
import ch.dlcm.model.MetadataVisibility;
import ch.dlcm.model.access.Order;
import ch.dlcm.model.access.Order.OrderStatus;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.model.oais.DisseminationInfoPackage;
import ch.dlcm.repository.DisseminationInfoPackageRepository;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.service.ArchiveUserFingerprintService;
import ch.dlcm.service.SearchService;
import ch.dlcm.service.citation.DLCMCitationGeneratorService;

public abstract class SearchController extends AccessIndexResourceReadOnlyController<MetadataVisibility, ArchiveMetadata> {
  private static final Logger log = LoggerFactory.getLogger(SearchController.class);

  private final DisseminationInfoPackageRepository dipRepository;
  private final OrderService orderService;
  private final ArchiveUserFingerprintService archiveUserFingerprintService;
  private final DLCMCitationGeneratorService dlcmCitationGeneratorService;

  protected SearchController(
          SearchService searchService,
          DLCMProperties config,
          DisseminationInfoPackageRepository dipRepository,
          OrderService orderService,
          ArchiveUserFingerprintService archiveUserFingerprintService,
          DLCMCitationGeneratorService dlcmCitationGeneratorService) {
    super(searchService, config);
    this.dipRepository = dipRepository;
    this.orderService = orderService;
    this.archiveUserFingerprintService = archiveUserFingerprintService;
    this.dlcmCitationGeneratorService = dlcmCitationGeneratorService;
  }

  @Override
  public HttpEntity<ArchiveMetadata> get(@PathVariable String id) {
    final ArchiveMetadata metadata = this.indexResourceService.findOne(this.getIndex(), id);
    if (metadata != null) {
      // Log View Statistics
      this.logViewStatistic(metadata);
      this.addLinks(metadata);
      return new ResponseEntity<>(metadata, HttpStatus.OK);
    }
    return new ResponseEntity<>(HttpStatus.NOT_FOUND);
  }

  @GetMapping(SolidifyConstants.URL_ID_PLUS_SEP + DLCMActionName.BIBLIOGRAPHIES)
  public HttpEntity<List<CitationDto>> bibliographies(@PathVariable String id) {
    final ArchiveMetadata metadata = this.indexResourceService.findOne(this.getIndex(), id);
    if (metadata != null) {
      List<CitationDto> bibliographies = this.dlcmCitationGeneratorService.getBibliographiesDtos(metadata.getResId());
      return new ResponseEntity<>(bibliographies, HttpStatus.OK);
    }
    return new ResponseEntity<>(HttpStatus.NOT_FOUND);
  }

  @GetMapping(SolidifyConstants.URL_ID_PLUS_SEP + DLCMActionName.CITATIONS)
  public HttpEntity<List<CitationDto>> citations(@PathVariable String id) {
    final ArchiveMetadata metadata = this.indexResourceService.findOne(this.getIndex(), id);
    if (metadata != null) {
      List<CitationDto> citations = this.dlcmCitationGeneratorService.getCitationsDtos(metadata.getResId());
      return new ResponseEntity<>(citations, HttpStatus.OK);
    }
    return new ResponseEntity<>(HttpStatus.NOT_FOUND);
  }

  @GetMapping("/{dipName}/" + ActionName.DOWNLOAD)
  @PreAuthorize("@metadataPermissionService.isAllowed(#dipName, 'DOWNLOAD_ARCHIVE') "
          + "|| @downloadWithAclPermissionService.isAllowed(#dipName) "
          + "|| @downloadTokenPermissionService.isAllowed(#dipName, T(ch.dlcm.model.security.DownloadTokenType).ARCHIVE)")
  public HttpEntity<FileSystemResource> download(@PathVariable String dipName) {

    // Find the linked DIP of the archive for direct download: DIP Name = AIP Id
    final List<DisseminationInfoPackage> dipList = this.dipRepository.findByInfoName(dipName);

    if (dipList.isEmpty()) {
      throw new IllegalStateException("DIP for AIP " + dipName + " is missing");
    } else if (dipList.size() > 1) {
      throw new SolidifyRuntimeException("More than one DIP is present for AIP " + dipName);
    }

    // Download DIP of AIP
    final URI fileUri = dipList.get(0).getDownloadUri();
    // Log DL Statistics
    SolidifyEventPublisher.getPublisher().publishEvent(new ArchiveStatisticMessage(dipName, StatisticType.DOWNLOAD));
    return this.buildDownloadResponseEntity(fileUri.getPath(), dipName + SolidifyConstants.ZIP_EXT, DLCMConstants.DIP_MIME_TYPE,
            FileTool.getSize(fileUri));
  }

  @GetMapping("/{aipId}/" + DLCMActionName.DOWNLOAD_STATUS)
  public HttpEntity<String> getDownloadStatus(@PathVariable String aipId) {
    final List<Order> orderList = this.orderService.findDirectOrderByAip(aipId);
    if (orderList.isEmpty()) {
      throw new NoSuchElementException("Order not found");
    } else if (orderList.size() == 1) {
      final Order order = orderList.get(0);
      return new ResponseEntity<>(order.getStatus().toString(), HttpStatus.OK);
    } else {
      log.error("DIP download check: too many orders");
      throw new SolidifyRuntimeException("Too many orders");
    }
  }

  @GetMapping("/" + DLCMActionName.LIST_DOWNLOAD_STATUS)
  @EveryonePermissions
  public HttpEntity<OrderStatus[]> listDownloadStatus() {
    return new ResponseEntity<>(OrderStatus.values(), HttpStatus.OK);
  }

  @PostMapping("/{aipId}/" + DLCMActionName.PREPARE_DOWNLOAD)
  @PreAuthorize("@metadataPermissionService.isAllowed(#aipId, 'PREPARE_DOWNLOAD_ARCHIVE')"
          + "|| @downloadWithAclPermissionService.isAllowed(#aipId)")
  public HttpEntity<String> prepareDownload(@PathVariable String aipId) {
    final List<DisseminationInfoPackage> dipList = this.dipRepository.findByInfoName(aipId);
    if (dipList.size() > 1) {
      log.error("AIP download: too many DIPs");
      return new ResponseEntity<>("Too many DIPs found", HttpStatus.INTERNAL_SERVER_ERROR);
    } else if (dipList.isEmpty()) {
      /*
       * DIP does not exist yet -> create a DIRECT order for the AIP that will create a DIP with only one
       * AIP inside
       */

      // Create order to download AIP and create DIP
      final List<Order> orderList = this.orderService.findDirectOrderByAip(aipId);
      if (orderList.isEmpty()) {
        final Order order = this.orderService.createOrderForAIP(aipId);
        this.submitOrder(order);
      } else if (orderList.size() == 1) {
        final Order order = orderList.get(0);
        if (order.getStatus() == OrderStatus.IN_PROGRESS) {
          this.submitOrder(order);
        } else if (order.getStatus() == OrderStatus.IN_ERROR) {
          log.error("AIP download: order in error");
          return new ResponseEntity<>("Order is in error", HttpStatus.INTERNAL_SERVER_ERROR);
        }
      } else {
        log.error("AIP download: too many orders");
        return new ResponseEntity<>("Too many orders found", HttpStatus.INTERNAL_SERVER_ERROR);
      }
      return new ResponseEntity<>(HttpStatus.ACCEPTED);

    } else {
      /*
       * DIP already exists
       */
      return new ResponseEntity<>("DIP already exists", HttpStatus.BAD_REQUEST);
    }
  }

  @GetMapping("/" + ActionName.SEARCH)
  public HttpEntity<RestCollection<ArchiveMetadata>> search(@RequestParam(defaultValue = "*") String query,
          @RequestParam(required = false) String conditions, Pageable pageable) {
    // transform sort parameters from the portail to the corresponding name in the index
    if (!pageable.getSort().isEmpty() && pageable.getSort().isSorted()) {
      pageable = PageRequest.of(pageable.getPageNumber(), pageable.getPageSize(), this.adaptSortProperties(pageable.getSort()));
    }

    final FacetPage<ArchiveMetadata> listItem = this.indexResourceService.search(
            this.getIndex(),
            query,
            conditions,
            Collections.emptyList(),
            pageable);
    return this.page2Collection(listItem, pageable);
  }

  private Sort adaptSortProperties(Sort sort) {
    // create a copy from the sort since the list of orders are not modifiable.
    List<org.springframework.data.domain.Sort.Order> newOrderList = new ArrayList<>(sort.toList());
    for (Sort.Order order : sort.toList()) {
      String newNameProperty = null;
      switch (order.getProperty()) {
        case DLCMConstants.ORG_UNIT_ID_FIELD:
          newNameProperty = DLCMConstants.AIP_ORGA_UNIT_NAME + DLCMConstants.INDEXING_KEYWORD;
          break;
        case "title":
          newNameProperty = DLCMConstants.AIP_TITLE + DLCMConstants.INDEXING_KEYWORD;
          break;
        case "accessLevel":
          newNameProperty = DLCMConstants.AIP_ACCESS_LEVEL + DLCMConstants.INDEXING_KEYWORD;
          break;
        case "size":
          newNameProperty = DLCMConstants.AIP_SIZE;
          break;
        case "files":
          newNameProperty = DLCMConstants.AIP_FILE_NUMBER;
          break;
        default:
          break;
      }
      if (!StringTool.isNullOrEmpty(newNameProperty)) {
        newOrderList.remove(order);
        newOrderList.add(new Sort.Order(order.getDirection(), newNameProperty));
      }
    }
    return Sort.by(newOrderList);
  }

  @PostMapping(value = "/" + ActionName.SEARCH, consumes = MediaType.APPLICATION_FORM_URLENCODED_VALUE)
  public HttpEntity<RestCollection<ArchiveMetadata>> searchPost(@RequestBody MultiValueMap<String, String> params) {

    log.trace("Search parameters: {}", params);

    // Pageable
    int size = RestCollectionPage.DEFAULT_SIZE_PAGE;
    int page = 0;
    Sort sort = Sort.unsorted();
    if (this.getParameter(params, ActionName.SIZE) != null) {
      size = Integer.parseInt(this.getParameter(params, ActionName.SIZE));
    }
    if (this.getParameter(params, ActionName.PAGE) != null) {
      page = Integer.parseInt(this.getParameter(params, ActionName.PAGE));
    }
    if (this.getParameter(params, ActionName.SORT) != null) {
      String value = this.getParameter(params, ActionName.SORT);
      String sortValue = value.substring(0, value.indexOf(','));
      String direction = value.substring(value.indexOf(',') + 1);
      sort = Sort.by(Sort.Direction.fromString(direction), sortValue);
    }

    final Pageable pageable = PageRequest.of(page, size, sort);

    String query = this.getParameter(params, IndexConstants.SEARCH_QUERY);
    if (StringTool.isNullOrEmpty(query)) {
      query = "*";
    }

    // Search
    return this.search(query, this.getParameter(params, IndexConstants.SEARCH_CONDITIONS), pageable);
  }

  @Override
  protected void addLinks(ArchiveMetadata archiveMetadata) {
    archiveMetadata.removeLinks();
    archiveMetadata.addLinksForAccess(linkTo(this.getClass()));
  }

  @Override
  protected <W extends RepresentationModel<W>> void addOthersLinks(W w) {
    w.add(linkTo(methodOn(this.getClass()).search("*", null, null)).withRel(ActionName.SEARCH));
    w.add(linkTo(methodOn(this.getClass()).listDownloadStatus()).withRel(ActionName.VALUES));
  }

  private void submitOrder(Order order) {
    this.orderService.saveOrder(order);
    this.orderService.changeOrderStatus(SUBMITTED, order.getResId());
  }

  private void logViewStatistic(ArchiveMetadata metadata) {
    String resId = metadata.getResId();

    // Check if data file metadata
    if (metadata.getIndex().equals(MetadataVisibility.RESTRICTED.toString())
            && !metadata.getMetadata().get(DLCMConstants.TYPE_FIELD).equals(DLCMConstants.PACKAGE)) {
      resId = (String) (metadata.getMetadata().get(DLCMConstants.ARCHIVE_ID_FIELD));
    }

    if (!this.archiveUserFingerprintService.existsUnexpiredArchiveUserFingerprint(resId)) {
      SolidifyEventPublisher.getPublisher().publishEvent(new ArchiveStatisticMessage(resId, StatisticType.VIEW));
    }
  }
}
