/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Access - MetadataPermissionService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.security;

import java.util.Optional;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Service;

import ch.dlcm.controller.AccessController;
import ch.dlcm.controller.DLCMControllerAction;
import ch.dlcm.model.Access;
import ch.dlcm.model.MetadataVisibility;
import ch.dlcm.model.OrganizationalUnitAwareResource;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.model.security.Role;
import ch.dlcm.service.SearchService;
import ch.dlcm.service.rest.abstractservice.PersonRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedPersonRemoteResourceService;

@Service
@ConditionalOnBean(AccessController.class)
public class MetadataPermissionService extends AbstractPermissionWithOrgUnitService {
  private final SearchService searchService;
  private final PersonRemoteResourceService trustedPersonRemoteResourceService;
  public MetadataPermissionService(SearchService searchService, TrustedPersonRemoteResourceService personRemoteResourceService,
          PersonRemoteResourceService trustedPersonRemoteResourceService) {
    super(personRemoteResourceService);
    this.searchService = searchService;
    this.trustedPersonRemoteResourceService = trustedPersonRemoteResourceService;
  }
  @Override
  public boolean isAllowed(String targetResId, String actionString) {
    // ADMIN cannot download closed archive
    final DLCMControllerAction action = this.getControllerAction(actionString);
    if (this.isAdminRole() &&
            (action == DLCMControllerAction.DOWNLOAD_ARCHIVE || action == DLCMControllerAction.PREPARE_DOWNLOAD_ARCHIVE)) {
      final OrganizationalUnitAwareResource existingResource = this.getExistingResource(targetResId);
      if (((ArchiveMetadata) existingResource).getCurrentAccess() == Access.CLOSED) {
        return false;
      }
    }
    return super.isAllowed(targetResId, actionString);
  }
  @Override
  protected boolean isAllowedToPerformActionOnResource(String personId, OrganizationalUnitAwareResource existingResource,
          DLCMControllerAction action) {

    if (existingResource != null) {
      if (((ArchiveMetadata) existingResource).getCurrentAccess() == Access.PUBLIC) {
        if (action == DLCMControllerAction.DOWNLOAD_ARCHIVE || action == DLCMControllerAction.PREPARE_DOWNLOAD_ARCHIVE) {
          return true;
        }
      } else if (((ArchiveMetadata) existingResource).getCurrentAccess() == Access.RESTRICTED) {
        return this.isAllowedActionWithRestrictedAccess(personId, existingResource.getOrganizationalUnitId(), action);
      } else if (((ArchiveMetadata) existingResource).getCurrentAccess() == Access.CLOSED) {
        return this.isAllowedActionWithClosedAccess(personId, existingResource, action);
      }
    }
    return super.isAllowedToPerformActionOnResource(personId, existingResource, action);
  }

  @Override
  protected boolean isApproverAllowed(OrganizationalUnitAwareResource existingDeposit, DLCMControllerAction action) {
    return true;
  }

  @Override
  protected boolean isCreatorAllowed(OrganizationalUnitAwareResource existingDeposit, DLCMControllerAction action) {
    return true;
  }

  @Override
  protected boolean isManagerAllowed(OrganizationalUnitAwareResource existingDeposit, DLCMControllerAction action) {
    return true;
  }

  @Override
  protected boolean isRoleAllowed(Role role, OrganizationalUnitAwareResource existingResource, DLCMControllerAction action) {

    if (action != DLCMControllerAction.DOWNLOAD_ARCHIVE &&
            action != DLCMControllerAction.PREPARE_DOWNLOAD_ARCHIVE &&
            action != DLCMControllerAction.GET) {
      return false;
    }
    return super.checkActionByRole(role, existingResource, action);
  }

  @Override
  protected boolean isStewardAllowed(OrganizationalUnitAwareResource existingDeposit, DLCMControllerAction action) {
    return true;
  }

  @Override
  protected boolean isVisitorAllowed(OrganizationalUnitAwareResource existingDeposit, DLCMControllerAction action) {
    return true;
  }

  @Override
  public ArchiveMetadata getExistingResource(String id) {
    return this.searchService.findOne(MetadataVisibility.ALL, id);
  }

  public boolean isAllowedActionWithRestrictedAccess(String personId, String orgId, DLCMControllerAction action) {
    if (action == DLCMControllerAction.DOWNLOAD_ARCHIVE || action == DLCMControllerAction.PREPARE_DOWNLOAD_ARCHIVE) {
      final Optional<Role> role = this.trustedPersonRemoteResourceService.findOrganizationalUnitRole(personId, orgId);
      return role.isPresent();
    }
    return false;
  }

  private boolean isAllowedActionWithClosedAccess(String personId, OrganizationalUnitAwareResource existingResource, DLCMControllerAction action) {
    if (action == DLCMControllerAction.DOWNLOAD_ARCHIVE || action == DLCMControllerAction.PREPARE_DOWNLOAD_ARCHIVE) {
      final Optional<Role> role = this.trustedPersonRemoteResourceService.findOrganizationalUnitRole(personId, existingResource.getOrganizationalUnitId());
      if (role.isEmpty()) {
        return false;
      }
      return role.get().getResId().equals(Role.MANAGER_ID) || role.get().getResId().equals(Role.STEWARD_ID);
    }
    return false;
  }

}
