/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - NotificationService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.business;

import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Objects;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.ResourceService;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.controller.AdminController;
import ch.dlcm.model.dto.NotificationContributorDto;
import ch.dlcm.model.dto.NotificationsContributorDto;
import ch.dlcm.model.notification.Notification;
import ch.dlcm.model.notification.NotificationCategory;
import ch.dlcm.model.notification.NotificationStatus;
import ch.dlcm.model.notification.NotificationType;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.Person;
import ch.dlcm.repository.NotificationRepository;
import ch.dlcm.repository.UserRepository;
import ch.dlcm.specification.NotificationSpecification;

@Service
@ConditionalOnBean(AdminController.class)
public class NotificationService extends ResourceService<Notification> {

  private static final String NOTIFICATION_NO_ORGANIZATIONAL_UNIT = "validation.notification.must.have.orgunit";
  private static final String NOTIFICATION_NO_ORGANIZATIONAL_UNIT_NAME = "validation.notification.newOrgUnit.missingName";

  private final NotificationRepository notificationRepository;
  private final UserRepository userRepository;

  public NotificationService(NotificationRepository notificationRepository, UserRepository userRepository) {
    this.notificationRepository = notificationRepository;
    this.userRepository = userRepository;
  }

  public Notification getInboxNotification(String notificationId, User currentUser) {
    String personId = currentUser.getPerson().getResId();
    int userRoleLevel = currentUser.getApplicationRole().getLevel();
    final Notification inboxNotification = this.notificationRepository.findInboxNotificationById(personId, currentUser.getResId(), userRoleLevel, notificationId);
    if (inboxNotification == null) {
      throw new NoSuchElementException("No such notification");
    }
    return inboxNotification;
  }

  public Notification getSentNotification(String notificationId, User currentUser) {
    final Notification sentNotification = this.notificationRepository.findByResId(notificationId);
    if (sentNotification == null || sentNotification.getEmitter() != currentUser) {
      throw new NoSuchElementException("No such sent notification");
    }
    return sentNotification;
  }

  public Page<Notification> listInboxNotification(User currentUser, Notification search, Pageable pageable) {
    if (currentUser == null) {
      throw new SolidifyRuntimeException("The current user is null");
    }
    NotificationStatus notificationStatus = null;
    NotificationType notificationType = null;
    NotificationCategory notificationCategory = null;
    String notifiedOrgUnitId = null;
    if (search != null) {
      notificationStatus = search.getNotificationStatus();
      notificationType = search.getNotificationType();
      if (search.getNotifiedOrgUnit() != null) {
        notifiedOrgUnitId = search.getNotifiedOrgUnit().getResId();
      }
      if (notificationType != null) {
        notificationCategory = notificationType.getNotificationCategory();
        if (notificationCategory != null) {
          // if searching on category, don't specify type
          final NotificationType notificationTypetemp = notificationType;
          if (NotificationType.getAllNotificationTypes().stream().noneMatch(nt -> notificationTypetemp.getResId().equals(nt.getResId()))) {
            notificationType = null;
          }
        }
      }
    }
    Person currentPerson = currentUser.getPerson();
    if (currentPerson == null) {
      throw new SolidifyRuntimeException("The current user " + currentUser.getFirstName() + " " + currentUser.getLastName() +
              " is not associated with a person");
    }
    String personId = currentPerson.getResId();
    int userRoleLevel = currentUser.getApplicationRole().getLevel();
    return this.notificationRepository.findInboxNotificationByPersonAndStatus(personId, notificationStatus, notificationType,
            notificationCategory, notifiedOrgUnitId, userRoleLevel, pageable);
  }

  public Page<Notification> listSentNotification(Notification search, User currentUser, Pageable pageable) {
    search.setEmitter(currentUser);
    final Specification<Notification> specification = this.getSpecification(search);
    return this.findAll(specification, pageable);
  }

  @Transactional
  public void deleteAllNotificationsOnOrgUnit(String notifiedOrgUnitId) {
    this.notificationRepository.deleteNotificationsByNotifiedOrgUnitId(notifiedOrgUnitId);
  }

  @Transactional
  public void deleteAllNotificationsOnDeposit(String depositId) {
    this.notificationRepository.deleteNotificationsByObjectId(depositId);
  }

  @Override
  protected void validateItemSpecificRules(Notification item, BindingResult errors) {

    if (!NotificationType.CREATE_ORGUNIT_REQUEST.equals(item.getNotificationType()) && Objects.isNull(item.getNotifiedOrgUnit())) {
      errors.addError(
              new FieldError(item.getClass().getSimpleName(), "notifiedOrgUnit", this.messageService.get(NOTIFICATION_NO_ORGANIZATIONAL_UNIT)));
    } else if (NotificationType.CREATE_ORGUNIT_REQUEST.equals(item.getNotificationType()) && StringTool.isNullOrEmpty(item.getObjectId())) {
      errors.addError(
              new FieldError(item.getClass().getSimpleName(), "objectId", this.messageService.get(NOTIFICATION_NO_ORGANIZATIONAL_UNIT_NAME)));
    }

  }

  @Override
  public NotificationSpecification getSpecification(Notification resource) {
    return new NotificationSpecification(resource);
  }

  /**
   * Update the sentTime property of a Notification
   * Note: Transactional is required to allow the method to be called from a scheduled task context that runs outside of the main thread JPA session
   *
   * @param notificationId
   * @param sentTime
   * @return
   */
  @Transactional
  public Notification updateSentTime(String notificationId, OffsetDateTime sentTime) {
    Notification notification = this.findOne(notificationId);
    notification.setSentTime(sentTime);
    return this.save(notification);
  }

  /*
   * Verify that each contributorsId correspond to an authenticated user and the type of notification is the type of MY_INDIRECT_DEPOSIT_COMPLETED_INFO
   */
  public List<Notification> createNotificationForContributors(NotificationsContributorDto notificationsContributorsDTO) {
    List<Notification> list = new ArrayList<>();
    for (NotificationContributorDto notificationContributorDto : notificationsContributorsDTO.getListNotifications()) {
      if (!notificationContributorDto.getNotification().getNotificationType().equals(NotificationType.MY_INDIRECT_COMPLETED_DEPOSIT_INFO)) {
        throw new SolidifyRuntimeException("only valid for MY_INDIRECT_COMPLETED_DEPOSIT_INFO notification type");
      }
      User user = this.userRepository.findByPersonResId(notificationContributorDto.getContributorsId());
      if (user != null) {
        Notification notification = notificationContributorDto.getNotification();
        notification.setRecipient(user.getPerson());
        notification.setNotificationStatus(NotificationStatus.PENDING);
        Notification item = this.save(notification);
        list.add(item);
      }
    }
    return list;
  }
}
