/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - ArchiveACLController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller.admin;

import java.security.Principal;
import java.util.Map;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import ch.unige.solidify.controller.ResourceController;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.security.AdminPermissions;
import ch.unige.solidify.security.NoOnePermissions;
import ch.unige.solidify.security.UserPermissions;
import ch.unige.solidify.service.HttpRequestInfoProvider;

import ch.dlcm.business.ArchiveACLService;
import ch.dlcm.controller.AdminController;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.ArchiveACL;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.UrlPath;
import ch.dlcm.service.rest.propagate.PropagateUserRemoteResourceService;

@AdminPermissions
@RestController
@ConditionalOnBean(AdminController.class)
@RequestMapping(UrlPath.ADMIN_ARCHIVE_ACL)
public class ArchiveACLController extends ResourceController<ArchiveACL> {

  private final HttpRequestInfoProvider httpRequestInfoProvider;
  private final PropagateUserRemoteResourceService userResourceService;
  private final ArchiveACLService archiveACLService;

  public ArchiveACLController(HttpRequestInfoProvider httpRequestInfoProvider,
          PropagateUserRemoteResourceService userResourceService,
          ArchiveACLService archiveACLService) {
    this.httpRequestInfoProvider = httpRequestInfoProvider;
    this.userResourceService = userResourceService;
    this.archiveACLService = archiveACLService;
  }

  @Override
  @PreAuthorize("@aclPermissionService.isAllowedToCreate(#archiveACL)")
  public HttpEntity<ArchiveACL> create(@RequestBody ArchiveACL archiveACL) {
    return super.create(archiveACL);
  }

  @Override
  @PreAuthorize("@aclPermissionService.isAllowed(#id, 'GET')")
  public HttpEntity<ArchiveACL> get(@PathVariable String id) {
    return super.get(id);
  }

  @Override
  @PreAuthorize("@aclPermissionService.isAllowedToList(#search)")
  public HttpEntity<RestCollection<ArchiveACL>> list(@ModelAttribute ArchiveACL search, Pageable pageable) {
    return super.list(search, pageable);
  }

  @UserPermissions
  @GetMapping("/" + DLCMActionName.GET_MY_ACLS)
  public HttpEntity<RestCollection<ArchiveACL>> listMyACL(Pageable pageable) {
    User currentUser = this.getCurrentUser();
    Page<ArchiveACL> acls = this.archiveACLService.findByUser(currentUser, pageable);
    return new ResponseEntity<>(new RestCollection<>(acls, pageable), HttpStatus.OK);
  }

  @Override
  @PreAuthorize("@aclPermissionService.isAllowed(#id, 'DELETE')")
  public HttpStatus delete(@PathVariable String id) {
    return super.delete(id);
  }

  @Override
  @NoOnePermissions
  public HttpStatus deleteList(@RequestBody String[] ids) {
    return HttpStatus.FORBIDDEN;
  }

  @Override
  @PreAuthorize("@aclPermissionService.isAllowedToUpdate(#id)")
  public HttpEntity<ArchiveACL> update(@PathVariable String id, @RequestBody Map<String, Object> updateMap) {
    return super.update(id, updateMap);
  }

  private User getCurrentUser() {
    final Principal principal = this.httpRequestInfoProvider.getPrincipal();
    return userResourceService.findByExternalUid(principal.getName());
  }
}
