/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - NotificationController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller.admin;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;

import java.security.Principal;
import java.util.List;
import java.util.Map;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyUnmodifiableException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.Tool;
import ch.unige.solidify.security.AdminPermissions;
import ch.unige.solidify.security.TrustedUserPermissions;
import ch.unige.solidify.security.UserPermissions;
import ch.unige.solidify.service.HttpRequestInfoProvider;

import ch.dlcm.business.NotificationService;
import ch.dlcm.business.UserService;
import ch.dlcm.controller.AdminController;
import ch.dlcm.controller.DLCMResourceController;
import ch.dlcm.model.dto.NotificationsContributorDto;
import ch.dlcm.model.notification.Notification;
import ch.dlcm.model.notification.NotificationCategory;
import ch.dlcm.model.notification.NotificationStatus;
import ch.dlcm.model.security.User;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.UrlPath;

@AdminPermissions
@RestController
@ConditionalOnBean(AdminController.class)
@RequestMapping(UrlPath.ADMIN_NOTIFICATION)
public class NotificationController extends DLCMResourceController<Notification> {

  private final HttpRequestInfoProvider httpRequestInfoProvider;
  private final UserService userService;

  public NotificationController(HttpRequestInfoProvider httpRequestInfoProvider, UserService userService) {
    this.httpRequestInfoProvider = httpRequestInfoProvider;
    this.userService = userService;
  }

  @PreAuthorize("@notificationPermissionService.isAllowedToCreate(#notification)")
  @Override
  public HttpEntity<Notification> create(@RequestBody Notification notification) {
    if (notification.getEmitter() == null) {
      notification.setEmitter(this.getCurrentUser());
    }
    notification.setNotificationStatus(NotificationStatus.PENDING);
    return super.create(notification);
  }

  @PostMapping("/" + DLCMActionName.BY_CONTRIBUTORS)
  @TrustedUserPermissions
  public HttpEntity<List<Notification>> createForContributors(@RequestBody NotificationsContributorDto notificationsContributorsDTO) {
    List<Notification> list = ((NotificationService)this.itemService).createNotificationForContributors(notificationsContributorsDTO);
    for(Notification item: list) {
      this.addLinks(item);
    }
    return new ResponseEntity<>(list, HttpStatus.CREATED);
  }

  @Override
  public HttpEntity<Notification> get(@PathVariable String id) {
    return super.get(id);
  }

  @Override
  public HttpEntity<RestCollection<Notification>> list(@ModelAttribute Notification search, Pageable pageable) {
    return super.list(search, pageable);
  }

  @Override
  public HttpEntity<Notification> update(@PathVariable String id, @RequestBody Map<String, Object> updateMap) {
    return super.update(id, updateMap);
  }

  @Override
  public HttpStatus delete(@PathVariable String id) {
    return super.delete(id);
  }

  @Override
  public HttpStatus deleteList(@RequestBody String[] ids) {
    return super.deleteList(ids);
  }

  @DeleteMapping(DLCMActionName.DELETE_NOTIFICATIONS_ON_DEPOSIT + "/" + SolidifyConstants.URL_ID)
  public HttpStatus deleteNotificationsOnDeposit(@PathVariable String id) {
    ((NotificationService) this.itemService).deleteAllNotificationsOnDeposit(id);
    return HttpStatus.OK;
  }

  @UserPermissions
  @GetMapping("/" + DLCMActionName.INBOX + SolidifyConstants.URL_ID)
  public HttpEntity<Notification> getInboxNotification(@PathVariable String id) {
    final Notification inboxNotification = ((NotificationService) this.itemService).getInboxNotification(id, this.getCurrentUser());
    return new ResponseEntity<>(inboxNotification, HttpStatus.OK);
  }

  @UserPermissions
  @GetMapping("/" + DLCMActionName.SENT + SolidifyConstants.URL_ID)
  public HttpEntity<Notification> getSentNotification(@PathVariable String id) {
    final Notification sentNotification = ((NotificationService) this.itemService).getSentNotification(id, this.getCurrentUser());
    return new ResponseEntity<>(sentNotification, HttpStatus.OK);
  }

  @UserPermissions
  @GetMapping("/" + DLCMActionName.INBOX)
  @SuppressWarnings("squid:S4684")
  public HttpEntity<RestCollection<Notification>> listInboxNotification(@ModelAttribute Notification search, Pageable pageable) {
    final Page<Notification> listItem = ((NotificationService) (this.itemService)).listInboxNotification(this.getCurrentUser(), search,
            pageable);
    final RestCollection<Notification> collection = this.addRemainingLinks(listItem, pageable);
    return new ResponseEntity<>(collection, HttpStatus.OK);
  }

  @UserPermissions
  @GetMapping("/" + DLCMActionName.SENT)
  @SuppressWarnings("squid:S4684")
  public HttpEntity<RestCollection<Notification>> listSentNotification(@ModelAttribute Notification search, Pageable pageable) {
    final Page<Notification> listItem = ((NotificationService) (this.itemService)).listSentNotification(search, this.getCurrentUser(), pageable);
    final RestCollection<Notification> collection = this.addRemainingLinks(listItem, pageable);
    return new ResponseEntity<>(collection, HttpStatus.OK);
  }

  @PreAuthorize("@notificationPermissionService.isMyNotification(#id)")
  @PostMapping(SolidifyConstants.URL_ID + "/" + DLCMActionName.SET_PROCESSED)
  public HttpEntity<Notification> setProcessedStatus(@PathVariable String id) {
    return this.changeStatus(id, NotificationStatus.PROCESSED);
  }

  @PreAuthorize("@notificationPermissionService.isMyNotification(#id)")
  @PostMapping(SolidifyConstants.URL_ID + "/" + DLCMActionName.SET_REFUSED)
  public HttpEntity<Notification> setRefusedStatus(@PathVariable String id) {
    return this.changeStatus(id, NotificationStatus.REFUSED);
  }

  @PreAuthorize("@notificationPermissionService.isMyNotification(#id)")
  @PostMapping(SolidifyConstants.URL_ID + "/" + DLCMActionName.SET_PENDING)
  public HttpEntity<Notification> setPendingStatus(@PathVariable String id) {
    return this.changeStatus(id, NotificationStatus.PENDING);
  }

  private RestCollection<Notification> addRemainingLinks(Page<Notification> listItem, Pageable pageable) {
    final RestCollection<Notification> collection = new RestCollection<>(listItem, pageable);
    for (final Notification t : listItem) {
      this.addLinks(t);
    }
    collection.add(linkTo(this.getClass()).withSelfRel());
    collection.add(Tool.parentLink((linkTo(this.getClass())).toUriComponentsBuilder())
            .withRel(ActionName.MODULE));
    this.addSortLinks(linkTo(this.getClass()), collection);
    this.addPageLinks(linkTo(this.getClass()), collection, pageable);
    this.addOthersLinks(collection);
    return collection;
  }

  private HttpEntity<Notification> changeStatus(String id, NotificationStatus newStatus) {
    final Notification notification = ((NotificationService) (this.itemService)).getInboxNotification(id, this.getCurrentUser());
    if (notification != null) {
      if (NotificationCategory.INFO.equals(notification.getNotificationType().getNotificationCategory()) &&
              NotificationStatus.REFUSED.equals(newStatus)) {
        throw new SolidifyUnmodifiableException("Notification of category INFO cannot be refused");
      }
      notification.setNotificationStatus(newStatus);
      this.itemService.save(notification);
      return new ResponseEntity<>(notification, HttpStatus.OK);
    } else {
      throw new SolidifyResourceNotFoundException("Notification " + id + " not found");
    }
  }

  private User getCurrentUser() {
    final Principal principal = this.httpRequestInfoProvider.getPrincipal();
    return this.userService.findByExternalUid(principal.getName());
  }
}
