/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - AdminEmailProcessingService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import javax.mail.MessagingException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Service;

import ch.unige.solidify.service.EmailService;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.DLCMConstants;
import ch.dlcm.business.OrganizationalUnitService;
import ch.dlcm.business.PersonService;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.controller.AdminController;
import ch.dlcm.message.EmailMessage;

@Service
@ConditionalOnBean(AdminController.class)
@Profile("email-service")
public class AdminEmailProcessingService extends EmailProcessingService {
  private static final Logger log = LoggerFactory.getLogger(AdminEmailProcessingService.class);

  private final OrganizationalUnitService organizationalUnitService;
  private final PersonService personService;
  private static final String ORG_UNITS_CONSTANT = "organizationalUnits";

  protected AdminEmailProcessingService(MessageService messageService, EmailService emailService,
          DLCMProperties dlcmProperties, OrganizationalUnitService organizationalUnitService, PersonService personService) {
    super(messageService, emailService, dlcmProperties);
    this.organizationalUnitService = organizationalUnitService;
    this.personService = personService;
  }

  @Override
  public void processEmailMessage(EmailMessage emailMessage) {
    log.info("Reading message by Admin module {}", emailMessage);
    Map<String, Object> parameterList;
    try {
      switch (emailMessage.getTemplate()) {
        case NEW_SUBSCRIPTION:
        case REMOVE_SUBSCRIPTION:
          parameterList = this.getEmailDefaultParameters();
          parameterList.put("notificationTypes", emailMessage.getParameters().get(DLCMConstants.NOTIFICATION_IDS));
          this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(),
                  Collections.singletonList(emailMessage.getTo()), emailMessage.getTemplate().getSubject(), parameterList);
          break;
        case NEW_ORG_UNIT_TO_APPROVE:
          parameterList = this.getEmailDefaultParameters();
          parameterList.put(ORG_UNITS_CONSTANT, emailMessage.getParameters());
          this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(),
                  Collections.singletonList(emailMessage.getTo()), emailMessage.getTemplate().getSubject(), parameterList);
          break;
        case NEW_MEMBER_ORG_UNIT_TO_APPROVE:
          parameterList = this.getEmailDefaultParameters();
          parameterList.put(ORG_UNITS_CONSTANT, emailMessage.getParameters());
          this.sendEmailWithOrgUnitsName(emailMessage);
          break;
        default:
          break;
      }
    } catch (MessagingException e) {
      log.error("Cannot send email ({}): {}", emailMessage.getTemplate().getSubject(), e.getMessage(), e);
    }
  }

  private void sendEmailWithOrgUnitsName(EmailMessage emailMessage) throws MessagingException {
    Map<String, Object> parameterList;

    parameterList = this.getEmailDefaultParameters();
    parameterList.put(ORG_UNITS_CONSTANT, emailMessage.getParameters());
    List<String> orgUnitNames = new ArrayList<>();
    List<String> personNames = new ArrayList<>();
    for(String orgUnitId: emailMessage.getParameters().keySet()) {
      orgUnitNames.add(this.organizationalUnitService.findOne(orgUnitId).getName());
      Map<String, String> personRoleParams = (Map<String, String>) emailMessage.getParameters().get(orgUnitId);
      personRoleParams.keySet().forEach(pr -> personNames.add(this.personService.findOne(pr).getFullName()));
    }
    parameterList.put("orgUnitNames", orgUnitNames);
    parameterList.put("personNames", personNames);

    this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(),
            Collections.singletonList(emailMessage.getTo()), emailMessage.getTemplate().getSubject(), parameterList);
  }
}
