/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - AclPermissionService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.security;

import java.util.NoSuchElementException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Service;

import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.service.HttpRequestInfoProvider;
import ch.unige.solidify.validation.ValidationError;

import ch.dlcm.business.ArchiveACLService;
import ch.dlcm.controller.AdminController;
import ch.dlcm.controller.DLCMControllerAction;
import ch.dlcm.model.OrganizationalUnitAwareResource;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.settings.ArchiveACL;
import ch.dlcm.service.rest.abstractservice.ArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedPersonRemoteResourceService;

@Service
@ConditionalOnBean(AdminController.class)
public final class AclPermissionService extends AbstractPermissionWithOrgUnitService {
  private static final Logger log = LoggerFactory.getLogger(AclPermissionService.class);

  private final ArchivalInfoPackageRemoteResourceService aipRemoteResourceService;
  private final ArchiveACLService archiveACLService;

  public AclPermissionService(HttpRequestInfoProvider httpRequestInfoProvider,
          TrustedPersonRemoteResourceService personRemoteResourceService,
          TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService,
          ArchiveACLService archiveACLService
  ) {
    super(personRemoteResourceService);
    this.aipRemoteResourceService = aipRemoteResourceService;
    this.archiveACLService = archiveACLService;
  }

  public boolean isAllowedToList(ArchiveACL search) {
    if (this.canBypassPermissionCheck()) {
      return true;
    }
    String personId = this.getPersonId();
    return this.isAllowedToPerformActionOnResource(personId, search, DLCMControllerAction.LIST);
  }

  @Override
  public boolean isAllowed(String targetResId, String actionString) {
    if (this.canBypassPermissionCheck()) {
      return true;
    }
    OrganizationalUnitAwareResource existingResource = this.getExistingResource(targetResId);
    if (existingResource == null) {
      throw new NoSuchElementException(targetResId);
    }
    DLCMControllerAction action = this.getControllerAction(actionString);
    final String personId = this.getPersonId();
    return this.isAllowedToPerformActionOnResource(personId, existingResource, action);
  }

  @Override
  protected boolean isManagerAllowed(OrganizationalUnitAwareResource existingResource, DLCMControllerAction action) {
    return isManagerOrStewardAllowedToCreateAcl(existingResource, action);
  }

  @Override
  protected boolean isStewardAllowed(OrganizationalUnitAwareResource existingResource, DLCMControllerAction action) {
    return isManagerOrStewardAllowedToCreateAcl(existingResource, action);
  }

  private boolean isManagerOrStewardAllowedToCreateAcl(OrganizationalUnitAwareResource existingResource, DLCMControllerAction action) {
    if (existingResource instanceof ArchiveACL) {
      if (action.equals(DLCMControllerAction.CREATE)) {
        return this.checkArchiveAclCreationRequestConsistency((ArchiveACL) existingResource);
      }
      return true;
    }
    return false;
  }

  private boolean checkArchiveAclCreationRequestConsistency(ArchiveACL archiveACL) {
    final String aclOrgUnitId = archiveACL.getOrganizationalUnitId();
    final String aipId = archiveACL.getAipId();
    if (aipId == null) {
      throw new SolidifyValidationException(new ValidationError("AIP id from ACL is null."));
    }
    final ArchivalInfoPackage aip = this.aipRemoteResourceService.findOne(aipId);
    final String aipOrgUnitId = aip.getOrganizationalUnitId();
    if (!aclOrgUnitId.equals(aipOrgUnitId)) {
      log.warn("OrgUnit Id from ACL {} differs from orgUnit Id from AIP {}.", aclOrgUnitId, aipOrgUnitId);
      return false;
    }
    return true;
  }

  @Override
  protected OrganizationalUnitAwareResource getExistingResource(String resId) {
    return this.archiveACLService.findOne(resId);
  }

}
