/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Archival Storage - FixityService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service;

import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import ch.unige.solidify.ChecksumAlgorithm;
import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.business.ArchivalInfoPackageService;
import ch.dlcm.controller.ArchivalStorageController;
import ch.dlcm.exception.FixityCheckLevelNotImplementedException;
import ch.dlcm.model.ChecksumCheck;
import ch.dlcm.model.DataFileChecksum;
import ch.dlcm.model.DataFileChecksum.ChecksumOrigin;
import ch.dlcm.model.FixityCheckLevel;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.StatusHistory;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.storage.StorageService;

@Service
@ConditionalOnBean(ArchivalStorageController.class)
public class FixityService extends DLCMService {

  private static final Logger log = LoggerFactory.getLogger(FixityService.class);

  protected ArchivalInfoPackageService aipService;
  protected StorageService storageService;

  public FixityService(MessageService messageService,
          StorageService storageService, ArchivalInfoPackageService aipService) {
    super(messageService);
    this.storageService = storageService;
    this.aipService = aipService;
  }

  @Transactional
  public void checkFixity(String aipId, FixityCheckLevel fixityCheckLevel, String createdBy) {
    // Load AIP
    final ArchivalInfoPackage aip = this.aipService.findOne(aipId);
    /*
     * Check that the FixityCheckLevel is set (if not, use the one by default)
     */
    if (fixityCheckLevel == null) {
      fixityCheckLevel = this.storageService.getDefaultFixityCheckLevel();
    }
    this.checkFixityChecksum(aip, fixityCheckLevel, createdBy, false, null);

    // Check each of the items in the collection
    if (aip.isCollection()) {
      final List<PackageStatus> statusItems = new ArrayList<>();
      for (final ArchivalInfoPackage aipItem : aip.getCollection()) {
        statusItems.add(this.checkFixityChecksum(aipItem, fixityCheckLevel, createdBy, true, aip));
      }

      if (statusItems.stream().anyMatch(PackageStatus.IN_ERROR::equals)) {
        this.updateStatusToError(aip);
        final StatusHistory errorInCollection = new StatusHistory(aip.getClass().getSimpleName(), aip.getResId(),
                OffsetDateTime.now(ZoneOffset.UTC),
                PackageStatus.IN_ERROR.toString(), "Item in collection with invalid checksum verification", createdBy);

        SolidifyEventPublisher.getPublisher().publishEvent(errorInCollection);

      }

    }
  }

  public PackageStatus checkFixityChecksum(ArchivalInfoPackage aip, FixityCheckLevel fixityCheckLevel, String createdBy,
          boolean itemInCollection, ArchivalInfoPackage aipCollection) {

    final List<ChecksumAlgorithm> algorithmsToRecalculate = new ArrayList<>();
    final List<DataFileChecksum> existingChecksums = aip
            .getChecksums(aip.getInfo().getStatus().equals(PackageStatus.DISPOSED) ? ChecksumOrigin.DLCM_TOMBSTONE : ChecksumOrigin.DLCM);

    for (final DataFileChecksum existingChecksum : existingChecksums) {
      algorithmsToRecalculate.add(existingChecksum.getChecksumAlgo());
    }

    final List<ChecksumAlgorithm> validChecksums = new ArrayList<>();
    final List<ChecksumAlgorithm> invalidChecksums = new ArrayList<>();

    this.addHistoryMessage(aip, PackageStatus.CHECKING,
            StringUtils.collectionToDelimitedString(algorithmsToRecalculate, ", ") + " checksums are about to be reverified with level "
                    + fixityCheckLevel.name(),
            createdBy);

    // New StatusHistory for each item in the collection
    if (itemInCollection) {
      this.addHistoryMessage(aipCollection, PackageStatus.CHECKING,
              "Item " + aip.getResId() + " : " + StringUtils.collectionToDelimitedString(algorithmsToRecalculate, ", ")
                      + " checksums are about to be reverified with level " + fixityCheckLevel.name(),
              createdBy);
    }

    try {
      final List<String> recalculatedChecksumValues = this.storageService.checkFixity(aip, algorithmsToRecalculate, fixityCheckLevel);

      for (int i = 0; i < existingChecksums.size(); i++) {

        if (!StringTool.isNullOrEmpty(recalculatedChecksumValues.get(i))) {

          if (existingChecksums.get(i).getChecksum().equals(recalculatedChecksumValues.get(i))) {

            validChecksums.add(existingChecksums.get(i).getChecksumAlgo());
            log.info("{} checksum '{}}' verified for AIP {}", existingChecksums.get(i).getChecksumAlgo(), existingChecksums.get(i).getChecksum(),
                    aip.getResId());
          } else {

            invalidChecksums.add(existingChecksums.get(i).getChecksumAlgo());
            log.warn("{} checksum '{}' verification failed for AIP {}", existingChecksums.get(i).getChecksumAlgo(),
                    existingChecksums.get(i).getChecksum(), aip.getResId());
          }

        } else {
          log.warn("StorageService could not return a hash value for {}", existingChecksums.get(i).getChecksumAlgo());
        }
      }

      PackageStatus newStatusValue = null;
      String message = null;

      if (invalidChecksums.isEmpty() && !validChecksums.isEmpty()) {
        /*
         * At least one checksum is still valid (some other ones may not have been recalculated and
         * compared) and no one is wrong
         */
        newStatusValue = PackageStatus.COMPLETED;
        message = "All checksums that could be verified (" + StringUtils.collectionToDelimitedString(validChecksums, ", ") + ") are valid";

        this.updateSuccessfulChecksumCheck(aip);

      } else if (invalidChecksums.isEmpty() && validChecksums.isEmpty()) {
        /*
         * Something wrong happened during calculation
         */
        message = "No checksum could be verified";
        newStatusValue = PackageStatus.PRESERVATION_ERROR;
        this.updateStatusToError(aip);

      } else {
        /*
         * Some checksum(s) are not valid anymore
         */
        newStatusValue = PackageStatus.PRESERVATION_ERROR;
        message = "The following checksums verification failed: " + StringUtils.collectionToDelimitedString(invalidChecksums, ", ");

        this.updateStatusToError(aip);
      }

      this.addHistoryMessage(aip, newStatusValue, message, createdBy);

      // New StatusHistory for each item in the collection
      if (itemInCollection) {
        this.addHistoryMessage(aipCollection, newStatusValue, message, createdBy);
      }

      return newStatusValue;

    } catch (NoSuchAlgorithmException | FixityCheckLevelNotImplementedException exception) {
      // file has not been checked.
      log.warn("Unable to recalculate checksums for AIP '{}' ({})", aip.getResId(), exception.getMessage());

      this.addHistoryMessage(aip, PackageStatus.COMPLETED, "Some checksum(s) could not be verified: " + exception.getMessage(), createdBy);

      return PackageStatus.COMPLETED;

    } catch (IOException | SolidifyRuntimeException exception) {
      // an error occurred -> put AIP in error
      log.error("Unable to recalculate checksums for AIP '{}' ({})", aip.getResId(), exception.getMessage());

      this.addHistoryMessage(aip, PackageStatus.PRESERVATION_ERROR, "Some checksum(s) could not be verified: " + exception.getMessage(),
              createdBy);

      this.updateStatusToError(aip);
      return PackageStatus.PRESERVATION_ERROR;
    }
  }

  private void addHistoryMessage(ArchivalInfoPackage aip, PackageStatus status, String message, String createdBy) {

    final StatusHistory startCheckingHistory = new StatusHistory(aip.getClass().getSimpleName(), aip.getResId(),
            OffsetDateTime.now(ZoneOffset.UTC),
            status.toString(), message, createdBy);

    SolidifyEventPublisher.getPublisher().publishEvent(startCheckingHistory);
  }

  private void updateStatusToError(ArchivalInfoPackage aip) {
    try {
      aip.setPackageStatus(PackageStatus.PRESERVATION_ERROR);
      aip.setChecksumCheck(new ChecksumCheck(false));
      this.aipService.save(aip);
    } catch (final SolidifyValidationException e) {
      log.error("the new AIP status ({}) could not be saved", PackageStatus.PRESERVATION_ERROR);
    }
  }

  private void updateSuccessfulChecksumCheck(ArchivalInfoPackage aip) {
    try {
      aip.setChecksumCheck(new ChecksumCheck(true));
      this.aipService.save(aip);
    } catch (final SolidifyValidationException e) {
      log.error("the new AIP checksum check could not be saved");
    }
  }
}
