/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Archival Storage - StorageService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.storage;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.xml.bind.JAXBException;

import org.springframework.util.FileSystemUtils;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;

import ch.unige.solidify.ChecksumAlgorithm;
import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.exception.SolidifyProcessingException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.BagItTool;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.SolidifyTime;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.ZipTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.config.DLCMRepositoryDescription;
import ch.dlcm.exception.DLCMPreservationException;
import ch.dlcm.model.DataCategory;
import ch.dlcm.model.FixityCheckLevel;
import ch.dlcm.model.Package;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.model.oais.AipDataFile;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.ArchiveContainer;
import ch.dlcm.model.preservation.StoredAIP;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.ResearchDomain;
import ch.dlcm.service.DLCMService;
import ch.dlcm.service.MetadataService;
import ch.dlcm.service.rest.fallback.FallbackArchiveMetadataRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveThumbnailRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackOrganizationalUnitRemoteResourceService;

public abstract class StorageService extends DLCMService {

  private static final String AIP_ERROR_PURGE_FILES_MSG = "archival.aip.error.purge.files";

  private final String archivalLocation;
  private final String repositoryName;
  private final String repositoryInstitution;
  private final String repositoryLocation;
  private final String indexName;
  private final String privateIndexName;
  private final String[] indexingExceptions;
  private final int waitMilliseconds;
  private final int waitMaxTries;

  private final MetadataService metadataService;
  private final FallbackOrganizationalUnitRemoteResourceService orgUnitResourceService;
  private final FallbackArchiveThumbnailRemoteResourceService archiveThumbnailRemoteResourceService;
  private final FallbackArchiveMetadataRemoteResourceService archiveMetadataRemoteResourceService;

  protected StorageService(
          DLCMProperties dlcmProperties,
          DLCMRepositoryDescription repositoryDescription,
          MessageService messageService,
          MetadataService metadataService,
          FallbackOrganizationalUnitRemoteResourceService orgUnitResourceService,
          FallbackArchiveThumbnailRemoteResourceService archiveThumbnailRemoteResourceService,
          FallbackArchiveMetadataRemoteResourceService archiveMetadataRemoteResourceService) {
    super(messageService);
    this.archivalLocation = dlcmProperties.getArchivalLocation();
    this.repositoryName = repositoryDescription.getName();
    this.repositoryInstitution = repositoryDescription.getInstitution();
    this.repositoryLocation = repositoryDescription.getLocation();
    this.indexName = dlcmProperties.getIndexing().getIndexName();
    this.indexingExceptions = dlcmProperties.getIndexing().getExceptions();
    this.privateIndexName = dlcmProperties.getIndexing().getPrivateIndexName();
    this.waitMilliseconds = dlcmProperties.getWait().getMilliseconds();
    this.waitMaxTries = dlcmProperties.getWait().getMaxTries();
    this.metadataService = metadataService;
    this.orgUnitResourceService = orgUnitResourceService;
    this.archiveThumbnailRemoteResourceService = archiveThumbnailRemoteResourceService;
    this.archiveMetadataRemoteResourceService = archiveMetadataRemoteResourceService;
  }

  // **************************
  // ** Methods to implement **
  // **************************

  // Get Archive File Number
  public abstract long getAIPFileNumber(ArchiveContainer aipContainer, String aipId) throws IOException;

  // Get Archive info
  public abstract Map<String, String> getAIPInfo(String aipId) throws IOException, JAXBException;

  // Get Archive size
  public abstract long getAIPSize(String aipId) throws IOException;

  // Get Archive stream
  public abstract InputStream getAIPStream(String aipId) throws IOException;

  // Get Archive URI
  public abstract URI getArchiveLocation(String aipId) throws URISyntaxException;

  // Get the default FixityCheckLevel for the storage implementation
  public abstract FixityCheckLevel getDefaultFixityCheckLevel();

  // Get AIP status
  public abstract PackageStatus getStatus(ArchivalInfoPackage aip) throws IOException;

  // List already stored AIP
  public abstract List<StoredAIP> listStoredAip();

  // Store AIP
  public abstract String storeAIP(ArchivalInfoPackage aip);

  // Store AIP
  public abstract void replaceAIP(ArchivalInfoPackage aip, Path newPackage);

  // Submit AIP
  public abstract String submitAIP(ArchivalInfoPackage aip) throws IOException, URISyntaxException;

  // *********************
  // ** Default Methods **
  // *********************

  // Get working directory for AIP
  public Path getAIPFolder(String resId) {
    return Paths.get(this.archivalLocation).resolve(resId);
  }

  // Check Working AIP
  public void checkAIP(ArchivalInfoPackage aip) throws IOException {
    this.checkPackage(aip, true, true);
  }

  // Check Archive
  public void checkArchive(ArchivalInfoPackage aip) throws IOException {
    this.checkPackage(aip, false, true);
  }

  // Perform a fixity check on AIP file with default FixityCheckLevel
  public List<String> checkFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToRecalculate)
          throws NoSuchAlgorithmException, IOException {
    return this.checkFixity(aip, algorithmsToRecalculate, this.getDefaultFixityCheckLevel());
  }

  // Perform a fixity check on AIP with requested FixityCheckLevel
  public abstract List<String> checkFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToRecalculate,
          FixityCheckLevel fixityCheckLevel) throws NoSuchAlgorithmException, IOException;

  // Check if AIP metadata index already exists
  public boolean checkIndexing(String aipId) {
    return this.archiveMetadataRemoteResourceService.checkMetadata(aipId);
  }

  // Build archive key
  public String getArchiveId(ArchivalInfoPackage aip) {
    return aip.getInfo().getOrganizationalUnitId() + "/" + aip.getResId();
  }

  public String getArchiveId(StoredAIP aip) {
    return aip.getOrganizationalUnitId() + "/" + aip.getAipId();
  }

  // Delete AIP metadata index
  public boolean deleteIndex(String aipId) {
    int tries = 0;
    do {
      // Delete indexes
      this.archiveMetadataRemoteResourceService.deleteIndex(aipId);
      if (!this.checkIndexing(aipId)) {
        // Delete Thumbnail
        if (this.archiveThumbnailRemoteResourceService.checkArchiveThumbnail(aipId)) {
          this.archiveThumbnailRemoteResourceService.deleteArchiveThumbnail(aipId);
        }
        return true;
      }
      SolidifyTime.waitInMilliSeconds(this.waitMilliseconds);
      tries++;
    } while (tries < this.waitMaxTries);

    return false;
  }

  // Get stored AIP
  public StoredAIP getStoredAip(String aipId) throws URISyntaxException {
    final StoredAIP aip = new StoredAIP();
    aip.setResId(aipId);
    aip.setArchiveUri(this.getArchiveLocation(this.getArchiveId(aip)));
    return aip;
  }

  // Index AIP metadata
  public boolean indexAIP(ArchivalInfoPackage aip) {
    ArchiveMetadata md;
    // Get DLCM metadata
    md = this.getAipMetadata(this.getAIPFolder(aip.getResId()), this.indexName, aip);
    if (md != null) {
      // Complete DLCM metadata with AIP metadata
      for (final Entry<String, Object> metadata : aip.getAIPMetadata().entrySet()) {
        md.getMetadata().put(metadata.getKey(), metadata.getValue());
      }
      // Add file number
      md.getMetadata().put(DLCMConstants.AIP_FILE_NUMBER,
              ((long) md.getMetadata().get(DLCMConstants.AIP_TOTAL_FILE_NUMBER))
                      - ((long) md.getMetadata().get(DLCMConstants.AIP_METADATA_FILE_NUMBER)));
      // Add orgUnit Details
      final OrganizationalUnit orgUnit = this.orgUnitResourceService.findOneWithCache(aip.getInfo().getOrganizationalUnitId());
      // OrgUnit Name
      md.getMetadata().put(DLCMConstants.AIP_ORGA_UNIT_NAME, orgUnit.getName());
      // Institutions
      List<Institution> institutionList = this.orgUnitResourceService.getInstitutions(orgUnit.getResId());
      if (institutionList != null && !institutionList.isEmpty()) {
        final List<String> nameList = new ArrayList<>();
        final List<String> descList = new ArrayList<>();
        for (final Institution i : institutionList) {
          nameList.add(i.getName());
          descList.add(i.getDescription());
        }
        md.getMetadata().put(DLCMConstants.AIP_INSTITUTIONS, nameList);
        md.getMetadata().put(DLCMConstants.AIP_INSTITUTION_DESCRIPTIONS, descList);
      }
      // Research domains
      if (orgUnit.getResearchDomains() != null && !orgUnit.getResearchDomains().isEmpty()) {
        final List<String> researchDomains = new ArrayList<>();
        for (final ResearchDomain rd : orgUnit.getResearchDomains()) {
          researchDomains.add(rd.getName());
        }
        md.getMetadata().put(DLCMConstants.AIP_RESEARCH_DOMAINS, researchDomains);
      }
      List<ArchiveMetadata> metadataList = this.splitMetadata(md, aip);
      return this.archiveMetadataRemoteResourceService.indexMetadata(metadataList);
    }
    return false;
  }

  // Dispose archive
  public void disposeArchive(ArchivalInfoPackage aip) throws IOException, JAXBException {
    // Check & extract archive
    Path aipPackage = Paths.get(this.checkPackage(aip, false, false));
    // Create tombstone
    final Path tombstoneFolder = aipPackage.getParent().resolve(DLCMConstants.TOMBSTONE);
    // Clean tombstone folder if exists because of previous executions
    if (FileTool.checkFile(tombstoneFolder) && !FileTool.deleteFolder(tombstoneFolder)) {
      throw new DLCMPreservationException(this.messageService.get(AIP_ERROR_PURGE_FILES_MSG));
    }
    // Move metadata file
    if (!FileTool.ensureFolderExists(tombstoneFolder)
            || !FileTool.moveFile(aipPackage.resolve(this.getMetadataRelativePath(aip.getArchiveContainer())),
                    tombstoneFolder.resolve(Package.METADATA_FILE.getName()))) {
      throw new DLCMPreservationException(this.messageService.get("archival.disposal.error.metadata", new Object[] { tombstoneFolder }));
    }
    // Delete data files if exist
    if (!FileTool.deleteFolder(aipPackage)) {
      throw new DLCMPreservationException(this.messageService.get(AIP_ERROR_PURGE_FILES_MSG));
    }
    // Update metadata
    this.metadataService.updateMetadataForDisposal(aip, tombstoneFolder.resolve(Package.METADATA_FILE.getName()));
    // Generate tombstone archive
    Path tombstonePackage = this.generateNewPackage(aip, tombstoneFolder);
    // Save tombstone archive
    this.replaceAIP(aip, tombstonePackage);
  }

  // Update archive
  public void updateArchive(ArchivalInfoPackage aip) throws IOException, JAXBException {
    this.updateArchiveMetadata(aip);
  }

  // Update archive with updated metadata
  public void updateArchiveMetadata(ArchivalInfoPackage aip) throws IOException, JAXBException {
    // Check & extract archive
    final Path aipPackageFolder = Paths.get(this.checkPackage(aip, false, false));
    final Path aipDataFolder = this.getDataFolder(aip.getArchiveContainer(), aipPackageFolder);
    // Extract updated package
    Path newMetadataFile = null;
    if (!aip.getDataFiles().isEmpty()) {
      newMetadataFile = Paths.get(this.checkUpdatePackage(aip));
      if (!FileTool.checkFile(newMetadataFile)) {
        throw new DLCMPreservationException(
                this.messageService.get("archival.aip.error.updated-metadata", new Object[] { newMetadataFile }));
      }
    }
    // Update metadata
    this.metadataService.updateMetadata(aip, aipPackageFolder.resolve(this.getMetadataRelativePath(aip.getArchiveContainer())), newMetadataFile);
    // Generate new package
    final Path newPackageFile = this.generateNewPackage(aip, aipDataFolder);
    // Purge package folder (temporary folder for updating archive )
    if (FileTool.checkFile(aipPackageFolder) && !FileTool.deleteFolder(aipPackageFolder)) {
      throw new DLCMPreservationException(this.messageService.get(AIP_ERROR_PURGE_FILES_MSG));
    }
    // Purge update folder (temporary folder for archive files)
    final Path aipUpdateFolder = this.getAIPFolder(aip.getResId()).resolve(DLCMConstants.UPDATE_FOLDER);
    if (FileTool.checkFile(aipUpdateFolder) && !FileTool.deleteFolder(aipUpdateFolder)) {
      throw new DLCMPreservationException(this.messageService.get(AIP_ERROR_PURGE_FILES_MSG));
    }
    // Save new archive
    this.replaceAIP(aip, newPackageFile);
  }

  // *********************
  // ** Private Methods **
  // *********************

  private Path generateNewPackage(ArchivalInfoPackage aip, Path packageFolder) throws IOException {
    // Update package info
    switch (aip.getArchiveContainer()) {
      case BAG_IT:
        // Create BagIt
        BagItTool.create(packageFolder, this.getBagMetadata(aip));
        break;
      case ZIP:
        break;
      default:
        throw new SolidifyCheckingException(this.messageService.get(ERROR_AIP_CONTAINER, new Object[] { aip.getArchiveContainer() }));
    }
    // Generate new archive
    final Path newPackage = this.getAIPFolder(aip.getResId()).resolve(aip.getResId());
    final ZipTool zipPackage = new ZipTool(newPackage.toUri());
    if (!zipPackage.zipFiles(packageFolder)) {
      throw new SolidifyProcessingException(this.messageService.get("archival.disposal.error.zip"));
    }
    if (!FileSystemUtils.deleteRecursively(packageFolder)) {
      throw new SolidifyProcessingException(this.messageService.get("archival.disposal.error.purge-tombstone"));
    }
    return newPackage;
  }

  private MultiValueMap<String, String> getBagMetadata(ArchivalInfoPackage aip) {
    MultiValueMap<String, String> metadata = new LinkedMultiValueMap<>();
    metadata.add(DLCMConstants.BAGIT_SOURCE,
            this.repositoryName + ", " + this.repositoryInstitution);
    metadata.add(DLCMConstants.BAGIT_SOURCE_ADDR, this.repositoryLocation);
    metadata.add(DLCMConstants.BAGIT_METADATA_SCHEMA, aip.getInfo().getMetadataVersion().getMetsSchema());
    metadata.add(DLCMConstants.BAGIT_ARC_ID, aip.getResId());
    metadata.add(DLCMConstants.BAGIT_ARC_NAME, aip.getInfo().getName());
    metadata.add(DLCMConstants.BAGIT_ARC_RETENTION, aip.getRetention().toString());
    metadata.add(DLCMConstants.BAGIT_ARC_RETENTION_DURATION, aip.getSmartRetention());
    metadata.add(DLCMConstants.BAGIT_ARC_DISPOSITION_APPROVAL, aip.getDispositionApproval().toString());
    metadata.add(DLCMConstants.BAGIT_ORG_UNIT_ID, aip.getInfo().getOrganizationalUnitId());
    metadata.add(DLCMConstants.BAGIT_ORG_UNIT_NAME, this.orgUnitResourceService.findOne(aip.getInfo().getOrganizationalUnitId()).getName());
    metadata.add(DLCMConstants.BAGIT_PACKAGE_SIZE, aip.getSmartSize());
    metadata.add(DLCMConstants.BAGIT_PACKAGE_FILE_NUMBER, aip.getArchiveFileNumber().toString());
    return metadata;
  }

  private List<ArchiveMetadata> splitMetadata(ArchiveMetadata md, ArchivalInfoPackage aip) {
    List<ArchiveMetadata> metadataList = new ArrayList<>();
    ArchiveMetadata publicMd = this.createArchiveMetadata(md.getResId(), this.indexName);
    ArchiveMetadata privateMd = this.createArchiveMetadata(md.getResId(), this.privateIndexName);

    // Reorganize metadata fields => Public/Private/Files
    for (Entry<String, Object> entry : md.getMetadata().entrySet()) {
      if (entry.getKey().startsWith(DLCMConstants.AIP_PREFIX)
              || entry.getKey().equals(DLCMConstants.CREATION_FIELD)
              || entry.getKey().equals(DLCMConstants.METADATA_VERSION_FIELD)) {
        // Common Info
        publicMd.getMetadata().put(entry.getKey(), entry.getValue());
        privateMd.getMetadata().put(entry.getKey(), entry.getValue());
      } else if (entry.getKey().equals(DLCMConstants.DATACITE_FIELD)
              || entry.getKey().equals(DLCMConstants.DATACITE_XML_FIELD)) {
        // Public Info
        publicMd.getMetadata().put(entry.getKey(), entry.getValue());
      } else {
        // Private Info
        if (entry.getKey().equals(DLCMConstants.PREMIS_FIELD)) {
          // PREMIS information
          this.reorganizePremisMetadata(metadataList, privateMd,
                  (Map<String, Object>) (((Map<String, Object>) entry.getValue()).get(DLCMConstants.ITEMS_FIELD)), aip);
        } else {
          privateMd.getMetadata().put(entry.getKey(), entry.getValue());
        }
      }
    }
    // Add private metadata
    metadataList.add(this.mergeCustomMetadata(privateMd));
    // Add public metadata
    metadataList.add(publicMd);
    return metadataList;
  }

  private void reorganizePremisMetadata(List<ArchiveMetadata> metadataList, ArchiveMetadata privateMetadata, Map<String, Object> itemsList,
          ArchivalInfoPackage aip) {
    List<Object> itemList = (List<Object>) (itemsList.get(DLCMConstants.ITEM_FIELD));
    // For each item
    for (Object obj : itemList) {
      Map<String, Object> item = ((Map<String, Object>) obj);
      String type = (String) item.get(DLCMConstants.TYPE_FIELD);
      ArchiveMetadata metadata;
      // Test type
      if (type.equals(DLCMConstants.PACKAGE)) {
        // Packages: deposit/sip/aip
        metadata = privateMetadata;
        // Add security info
        this.addSecurtiyInfo(metadata, aip);
      } else {
        if (type.equals(DLCMConstants.AIP)) {
          // AIP file
          Map<String, Object> aipInfo = (Map<String, Object>) item.get(DLCMConstants.AIP);
          String aipId = (String) aipInfo.get(DLCMConstants.ID_FIELD);
          metadata = this.createArchiveMetadata(aip.getResId() + "-" + item.get(DLCMConstants.ID_FIELD).toString(),
                  this.privateIndexName);
          aip.getCollection().forEach(archive -> {
            if (archive.getResId().equals(aipId)) {
              metadata.getMetadata().putAll(archive.getAIPMetadata());
            }
          });
        } else {
          // Data file
          metadata = this.createArchiveMetadata((String) item.get(DLCMConstants.ID_FIELD), this.privateIndexName);
        }
        // Add archive ID (AIP)
        metadata.getMetadata().put(DLCMConstants.ARCHIVE_ID_FIELD, aip.getResId());
        // Add security info
        this.addSecurtiyInfo(metadata, aip);
        // Add new entry
        metadataList.add(metadata);
      }
      // Add different fields
      for (Entry<String, Object> metadataField : item.entrySet()) {
        if (!metadataField.getKey().equals(DLCMConstants.ID_FIELD)) {
          metadata.getMetadata().put(metadataField.getKey(), metadataField.getValue());
        }
      }
    }
  }

  private void addSecurtiyInfo(ArchiveMetadata metadata, ArchivalInfoPackage aip) {
    // OrgUnit Id
    metadata.getMetadata().put(DLCMConstants.AIP_ORGA_UNIT, aip.getInfo().getOrganizationalUnitId());
    // Access Level
    metadata.getMetadata().put(DLCMConstants.AIP_ACCESS_LEVEL, aip.getInfo().getAccess().toString());
    // Embargo
    if (aip.getInfo().hasEmbargo()) {
      metadata.getMetadata().put(DLCMConstants.AIP_EMBARGO_LEVEL, aip.getInfo().getEmbargo().getAccess().toString());
      metadata.getMetadata().put(DLCMConstants.AIP_EMBARGO_END_DATE, aip.getInfo().getEmbargo().getEndDate().toString());
    }
  }

  private ArchiveMetadata mergeCustomMetadata(ArchiveMetadata privateMd) {
    // Reorganize custom-metadata/custom-metadata => custom-metadata
    Map<String, Object> customMd = (Map<String, Object>) (privateMd.getMetadata().remove(DLCMConstants.CUSTOM_METADATA_FIELD));
    // Check if there are custom metadata
    if (customMd != null) {
      customMd = (Map<String, Object>) (customMd.remove(DLCMConstants.CUSTOM_METADATA_FIELD));
      privateMd.getMetadata().put(DLCMConstants.CUSTOM_METADATA_FIELD, customMd.get(DLCMConstants.CUSTOM_METADATA_FIELD));
      // Merge Custom metadata info & content
      List<Map<String, Object>> mdList = new ArrayList<>();
      try {
        // list of custom metadata
        mdList.addAll((List<Map<String, Object>>) (privateMd.getMetadata().get(DLCMConstants.CUSTOM_METADATA_FIELD)));
      } catch (ClassCastException e) {
        // Just one custom metadata
        mdList.add((Map<String, Object>) (privateMd.getMetadata().get(DLCMConstants.CUSTOM_METADATA_FIELD)));
      }
      // For each custom metadata
      for (Map<String, Object> md : mdList) {
        // Remove schema fields
        md.entrySet().removeIf(f -> f.getKey().startsWith("xmlns:"));
        // Move metadata content
        String name = (String) (md.get("name"));
        String content = (String) (privateMd.getMetadata().remove(name));
        md.put(DLCMConstants.CONTENT_FIELD, content);
      }
    }
    return privateMd;
  }

  private ArchiveMetadata createArchiveMetadata(String aipId, String indexName) {
    ArchiveMetadata archiveMetadata = new ArchiveMetadata();
    archiveMetadata.setIndex(indexName);
    archiveMetadata.setResId(aipId);
    return archiveMetadata;
  }

  // Check package validity with Zip integrity and BagIt conformanceAipContainer
  private String check(ArchivalInfoPackage aip, URI archiveUri, boolean deleteData) throws IOException {
    final ZipTool zip = new ZipTool(archiveUri);
    // Check if zip is correct
    if (!zip.checkZip()) {
      throw new SolidifyCheckingException(this.messageService.get("archival.aip.error.zip.check"));
    }
    // Unzip AIP
    final String aipPackage = this.getAIPFolder(aip.getResId()) + "/" + DLCMConstants.PACKAGE;
    if (!zip.unzipFiles(Paths.get(aipPackage))) {
      throw new SolidifyCheckingException(this.messageService.get("archival.aip.error.zip.unzip"));
    }

    // Check package & metadata
    switch (aip.getArchiveContainer()) {
      case BAG_IT:
        // Check if BagIt is correct
        BagItTool.check(DLCMConstants.BAGIT_PROFILE, aipPackage);
        // Check DLCM metadata
        final Path bagItMdFile = Paths.get(aipPackage, DLCMConstants.DATA_FOLDER, Package.METADATA_FILE.getName());
        this.metadataService.checkMetadataFile(aip.getInfo().getMetadataVersion(), bagItMdFile, new AipDataFile());
        // Check archive file number
        this.checkArchiveFileNumber(aip, zip.getEntryNumber() - BAGIT_INTERNAL_FILE_NUMBER);
        break;
      case ZIP:
        // Check DLCM metadata
        final Path zipMdFile = Paths.get(aipPackage, Package.METADATA_FILE.getName());
        this.metadataService.checkMetadataFile(aip.getInfo().getMetadataVersion(), zipMdFile, new AipDataFile());
        // Check archive file number
        this.checkArchiveFileNumber(aip, zip.getEntryNumber());
        break;
      default:
        throw new SolidifyCheckingException(this.messageService.get(ERROR_AIP_CONTAINER, new Object[] { aip.getArchiveContainer() }));
    }

    // Check archive size
    this.checkArchiveSize(aip, FileTool.getSize(archiveUri));

    // Remove extracted files
    if (deleteData && !FileTool.deleteFolder(Paths.get(aipPackage))) {
      throw new SolidifyCheckingException(this.messageService.get(AIP_ERROR_PURGE_FILES_MSG));
    }
    return aipPackage;
  }

  private void checkArchiveFileNumber(ArchivalInfoPackage aip, long fileNumber) {
    // For collection, add others AIPs
    if (aip.isCollection()) {
      for (final ArchivalInfoPackage aiu : aip.getCollection()) {
        fileNumber += aiu.getArchiveFileNumber();
      }
    } else if (aip.isTombstone()) {
      if (fileNumber != 1) {
        throw new SolidifyCheckingException(
                this.messageService.get("archival.aip.error.file-number", new Object[] { 1, fileNumber }));
      }
      return;
    }
    // Check file number
    if (aip.getArchiveFileNumber() == null || aip.getArchiveFileNumber() == 0 || aip.getInfo().getStatus().equals(PackageStatus.EDITING_METADATA)) {
      // Update it if not set for previous AIP
      aip.setArchiveFileNumber(fileNumber);
    } else if (aip.getArchiveFileNumber() != fileNumber) {
      throw new SolidifyCheckingException(
              this.messageService.get("archival.aip.error.file-number", new Object[] { aip.getArchiveFileNumber(), fileNumber }));
    }
  }

  private void checkArchiveSize(ArchivalInfoPackage aip, long size) {
    // For collection, add others AIPs
    if (aip.isCollection()) {
      for (final ArchivalInfoPackage aiu : aip.getCollection()) {
        if (aiu.isTombstone()) {
          size += aiu.getTombstoneSize();
        } else {
          size += aiu.getArchiveSize();
        }
      }
    }
    // Check size
    if (aip.getArchiveSize() == null || aip.getArchiveSize() == 0 || aip.getInfo().getStatus().equals(PackageStatus.EDITING_METADATA)) {
      // Update it if not set for previous AIP
      aip.setArchiveSize(size);
    } else {
      Long archiveSize = aip.getArchiveSize();
      if (aip.isTombstone()) {
        archiveSize = aip.getTombstoneSize();
      }
      if (archiveSize != size) {
        throw new SolidifyCheckingException(
                this.messageService.get("archival.aip.error.size", new Object[] { archiveSize, size }));
      }
    }
  }

  private String checkPackage(ArchivalInfoPackage aip, boolean isWorkingAip, boolean deleteData) throws IOException {
    final URI aipFile = this.getPackageFile(aip, isWorkingAip);
    final Path aipArchive = Paths.get(aipFile);
    if (!FileTool.checkFile(aipArchive) || !FileTool.isFile(aipArchive)) {
      throw new SolidifyCheckingException(this.messageService.get("checking.storage.notfound", new Object[] { aipFile }));
    }
    switch (aip.getArchiveContainer()) {
      case BAG_IT:
      case ZIP:
        return this.check(aip, aipFile, deleteData);
      default:
        throw new SolidifyCheckingException(this.messageService.get(ERROR_AIP_CONTAINER, new Object[] { aip.getArchiveContainer() }));
    }
  }

  private String checkUpdatePackage(ArchivalInfoPackage aip) {
    final URI updatedPackageFile = aip.getDataFiles().get(0).getFinalData();
    final ZipTool zip = new ZipTool(updatedPackageFile);
    // Check if zip is correct
    if (!zip.checkZip()) {
      throw new SolidifyCheckingException(this.messageService.get("archival.aip.error.zip.check"));
    }
    // Unzip AIP
    final String updatedPackage = this.getAIPFolder(aip.getResId()) + "/" + DLCMConstants.UPDATE_FOLDER;
    if (!zip.unzipFiles(Paths.get(updatedPackage))) {
      throw new SolidifyCheckingException(this.messageService.get("archival.aip.error.zip.unzip"));
    }
    // Find updated metadata file
    List<Path> updatedMetadataFiles = FileTool.readFolder(Paths.get(updatedPackage));
    if (updatedMetadataFiles.isEmpty()
            || updatedMetadataFiles.stream().anyMatch(file -> !file.getFileName().toString().startsWith(DLCMConstants.UPDATED_METADATA_FILE))) {
      throw new DLCMPreservationException(
              this.messageService.get("archival.aip.error.updated-metadata", new Object[] { updatedPackage }));
    }
    Collections.sort(updatedMetadataFiles);
    // return the last element b/c the files are ascending sorted by date
    return updatedMetadataFiles.get(updatedMetadataFiles.size() - 1).toString();
  }

  // Filter metadata
  private Map<String, Object> filterMetadata(Map<String, Object> metadata) {
    StringTool.filterMap(metadata, this.indexingExceptions);
    return metadata;
  }

  // Get AIP metadata
  private ArchiveMetadata getAipMetadata(Path workingDir, String index, ArchivalInfoPackage aip) {
    try (InputStream is = this.getArchiveStream(aip)) {
      return this.getMetadataFromInputStream(workingDir, index, aip, is);
    } catch (IOException | JAXBException e) {
      throw new SolidifyProcessingException("Problem in extracting AIP: " + e.getMessage());
    }
  }

  private InputStream getArchiveStream(ArchivalInfoPackage aip) throws IOException {
    // Check if there is a local copy
    if (aip.getDataFiles().size() == 1 &&
            aip.getDataFiles().get(0).getDataType() == DataCategory.InformationPackage &&
            FileTool.checkFile(Paths.get(aip.getDataFiles().get(0).getFinalData()))) {
      return new BufferedInputStream(Files.newInputStream(Paths.get(aip.getDataFiles().get(0).getFinalData())));
    } else {
      // Get from storage
      return this.getAIPStream(this.getArchiveId(aip));
    }
  }

  private ArchiveMetadata getMetadataFromInputStream(Path workingDir, String index, ArchivalInfoPackage aip, InputStream is)
          throws IOException, JAXBException {

    final Path destination = Paths.get(workingDir + "/" + DLCMConstants.PACKAGE);
    String metadataFile = this.getMetadataRelativePath(aip.getArchiveContainer());
    String metadataDir = this.getRelativeFolder(aip.getArchiveContainer(), Package.METADATA);
    String updatedMetadataFilesPrefix = this.getUpdatedMetadataRelativePathPrefix(aip.getArchiveContainer());
    String datasetThumbnail = this.getThumbnailRelativePath(aip.getArchiveContainer());

    final List<String> filesToExtract = new ArrayList<>();
    filesToExtract.add(metadataFile);
    filesToExtract.add(updatedMetadataFilesPrefix);
    filesToExtract.add(metadataDir);
    filesToExtract.add(datasetThumbnail);
    if (ZipTool.unzipFiles(is, destination, filesToExtract, null)
            && Paths.get(destination + "/" + metadataFile).toFile().exists()) {
      // DLCM metadata
      ArchiveMetadata metadata = this.toArchiveMetadata(index, aip.getResId(), aip.getInfo().getMetadataVersion(),
              Paths.get(destination + "/" + metadataFile));
      if (metadata == null) {
        throw new SolidifyProcessingException("Problem in loading metadata for '" + aip.getResId() + "' archive");
      }
      // Check if there is custom metadata
      if (Paths.get(destination + "/" + metadataDir).toFile().isDirectory()) {
        // Add Custom metadata
        for (final Path p : FileTool.readFolder(destination.resolve(metadataDir))) {
          metadata.getMetadata().put(p.getFileName().toString(), FileTool.toString(p));
        }
      }
      // filter metadata
      metadata.setMetadata(this.filterMetadata(metadata.getMetadata()));
      // Save dataset thumbnail
      Path thumbnail = Paths.get(destination + "/" + datasetThumbnail);
      // Delete thumbnail if exists
      if (this.archiveThumbnailRemoteResourceService.checkArchiveThumbnail(aip.getResId())
              && !this.archiveThumbnailRemoteResourceService.deleteArchiveThumbnail(aip.getResId())) {
        throw new SolidifyProcessingException("Problem in deleting thumbnail for '" + aip.getResId() + "' archive");
      }
      if (FileTool.checkFile(thumbnail)) {
        this.archiveThumbnailRemoteResourceService.createArchiveThumbnail(aip.getResId(), thumbnail);
        metadata.getMetadata().put(DLCMConstants.AIP_THUMBNAIL, Boolean.TRUE);
      } else {
        metadata.getMetadata().put(DLCMConstants.AIP_THUMBNAIL, Boolean.FALSE);
      }
      // Count metadata files : dlcm.xml + updated files
      long updatedFiles = FileTool.findFiles(destination)
              .stream()
              .filter(f -> f.getFileName().toString().startsWith(DLCMConstants.REPRESENTATION_INFO_METADATA_FILE))
              .count();
      metadata.getMetadata().put(DLCMConstants.AIP_METADATA_FILE_NUMBER, updatedFiles + 1);
      // Remove extracted files
      if (!FileTool.deleteFolder(destination)) {
        throw new SolidifyProcessingException("Problem in deleting working folder for '" + aip.getResId() + "' archive");
      }
      return metadata;
    }
    throw new SolidifyProcessingException("Problem in extracting metadata for '" + aip.getResId() + "' archive");
  }

  protected String getRelativeFolder(ArchiveContainer archiveContainer, Package subFolder) {
    if (!subFolder.isSubFolder()) {
      throw new SolidifyCheckingException(this.messageService.get("archival.aip.error.folder", new Object[] { subFolder.getName() }));
    }

    switch (archiveContainer) {
      case BAG_IT:
        return DLCMConstants.DATA_FOLDER + subFolder.getName();
      case ZIP:
        return subFolder.getName().substring(1);
      default:
        throw new SolidifyCheckingException(this.messageService.get(ERROR_AIP_CONTAINER, new Object[] { archiveContainer }));
    }
  }

  protected String getMetadataRelativePath(ArchiveContainer archiveContainer) {
    switch (archiveContainer) {
      case BAG_IT:
        return DLCMConstants.DATA_FOLDER + "/" + Package.METADATA_FILE.getName();
      case ZIP:
        return Package.METADATA_FILE.getName();
      default:
        throw new SolidifyCheckingException(this.messageService.get(ERROR_AIP_CONTAINER, new Object[] { archiveContainer }));
    }
  }

  protected String getUpdatedMetadataRelativePathPrefix(ArchiveContainer archiveContainer) {
    switch (archiveContainer) {
      case BAG_IT:
        return DLCMConstants.DATA_FOLDER + "/" + DLCMConstants.REPRESENTATION_INFO_METADATA_FILE;
      case ZIP:
        return DLCMConstants.REPRESENTATION_INFO_METADATA_FILE;
      default:
        throw new SolidifyCheckingException(this.messageService.get(ERROR_AIP_CONTAINER, new Object[] { archiveContainer }));
    }
  }

  protected String getThumbnailRelativePath(ArchiveContainer archiveContainer) {
    switch (archiveContainer) {
      case BAG_IT:
        return DLCMConstants.DATA_FOLDER + "/" + DLCMConstants.INTERNAL_FOLDER + "/"
                + Package.DATASET_THUMBNAIL.getName();
      case ZIP:
        return DLCMConstants.INTERNAL_FOLDER + "/" + Package.DATASET_THUMBNAIL.getName();
      default:
        throw new SolidifyCheckingException(this.messageService.get(ERROR_AIP_CONTAINER, new Object[] { archiveContainer }));
    }
  }

  private Path getDataFolder(ArchiveContainer archiveContainer, Path packageFolder) {
    switch (archiveContainer) {
      case BAG_IT:
        return packageFolder.resolve(DLCMConstants.DATA_FOLDER);
      case ZIP:
        return packageFolder;
      default:
        throw new SolidifyCheckingException(this.messageService.get(ERROR_AIP_CONTAINER, new Object[] { archiveContainer }));
    }
  }

  private URI getPackageFile(ArchivalInfoPackage aip, boolean isWorkingAip) {
    if (isWorkingAip) {
      return aip.getDataFiles().get(0).getFinalData();
    }
    return URI.create(aip.getArchiveId());
  }

  private ArchiveMetadata toArchiveMetadata(String index, String id, DLCMMetadataVersion version, Path dlcmXml)
          throws IOException, JAXBException {
    final ArchiveMetadata am = new ArchiveMetadata();
    am.setResId(id);
    am.setIndex(index);
    return this.metadataService.completeArchiveIndex(version, am, dlcmXml);
  }

  protected void completeAipInfoWithBagitInfo(Map<String, String> infoList, Path bagitInfoFile) {
    Map<String, String> bagInfoList = BagItTool.readBagItFile(bagitInfoFile);
    // Metadata version
    if (bagInfoList.containsKey(DLCMConstants.BAGIT_METADATA_SCHEMA)) {
      String value = bagInfoList.get(DLCMConstants.BAGIT_METADATA_SCHEMA);
      infoList.put(DLCMConstants.METADATA_VERSION_FIELD, value.substring(10, value.indexOf(".xsd")));
    }
    // Creation
    if (bagInfoList.containsKey(DLCMConstants.BAGIT_DATE)) {
      infoList.put(DLCMConstants.CREATION_FIELD, bagInfoList.get(DLCMConstants.BAGIT_DATE));
    }
    // Retention
    if (bagInfoList.containsKey(DLCMConstants.BAGIT_ARC_RETENTION)) {
      infoList.put(DLCMConstants.AIP_RETENTION, bagInfoList.get(DLCMConstants.BAGIT_ARC_RETENTION));
    }
    // Disposal approval
    if (bagInfoList.containsKey(DLCMConstants.BAGIT_ARC_DISPOSITION_APPROVAL)) {
      infoList.put(DLCMConstants.AIP_DISPOSITION_APPROVAL, bagInfoList.get(DLCMConstants.BAGIT_ARC_DISPOSITION_APPROVAL));
    }
    // Title
    if (bagInfoList.containsKey(DLCMConstants.BAGIT_ARC_NAME)) {
      infoList.put(DLCMConstants.AIP_TITLE, bagInfoList.get(DLCMConstants.BAGIT_ARC_NAME));
    }
    // Size
    String bagitSize = bagInfoList.get(BagItTool.BAGIT_PAYLOAD).substring(0, bagInfoList.get(BagItTool.BAGIT_PAYLOAD).indexOf('.'));
    String bagitNumber = bagInfoList.get(BagItTool.BAGIT_PAYLOAD).substring(bagInfoList.get(BagItTool.BAGIT_PAYLOAD).indexOf('.') + 1);
    if (bagInfoList.containsKey(DLCMConstants.BAGIT_PACKAGE_FILE_NUMBER)
            && !bagitNumber.equals(bagInfoList.get(DLCMConstants.BAGIT_PACKAGE_FILE_NUMBER))) {
      // Tombstone
      infoList.put(DLCMConstants.AIP_SIZE, Long.toString(StringTool.getSizeFromSmartSize(bagInfoList.get(DLCMConstants.BAGIT_PACKAGE_SIZE))));
      infoList.put(DLCMConstants.AIP_TOMBSTONE_SIZE, bagitSize);
      // File number
      infoList.put(DLCMConstants.AIP_FILE_NUMBER, bagInfoList.get(DLCMConstants.BAGIT_PACKAGE_FILE_NUMBER));
    } else {
      // Archive
      infoList.put(DLCMConstants.AIP_SIZE, bagitSize);
      // File number
      infoList.put(DLCMConstants.AIP_FILE_NUMBER, bagitNumber);

    }
  }

  protected void completeAipInfoWithMetadata(Map<String, String> infoList, DLCMMetadataVersion metadataVersion, Path metadataFile)
          throws IOException, JAXBException {
    infoList.putAll(this.metadataService.extractInfoFromMetadata(metadataVersion, metadataFile));
  }

}
