/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Archival Storage - S3StorageService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.storage.s3;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Service;

import com.amazonaws.AmazonServiceException;
import com.amazonaws.ClientConfiguration;
import com.amazonaws.auth.AWSCredentials;
import com.amazonaws.auth.AWSStaticCredentialsProvider;
import com.amazonaws.auth.BasicAWSCredentials;
import com.amazonaws.client.builder.AwsClientBuilder.EndpointConfiguration;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3ClientBuilder;
import com.amazonaws.services.s3.model.ObjectListing;
import com.amazonaws.services.s3.model.ObjectMetadata;
import com.amazonaws.services.s3.model.S3Object;
import com.amazonaws.services.s3.model.S3ObjectSummary;

import ch.unige.solidify.ChecksumAlgorithm;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.ChecksumTool;
import ch.unige.solidify.util.FileTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.config.DLCMRepositoryDescription;
import ch.dlcm.controller.ArchivalStorageController;
import ch.dlcm.exception.FixityCheckLevelNotImplementedException;
import ch.dlcm.model.FixityCheckLevel;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.ArchiveContainer;
import ch.dlcm.model.preservation.StoredAIP;
import ch.dlcm.service.MetadataService;
import ch.dlcm.service.rest.fallback.FallbackArchiveMetadataRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveThumbnailRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackOrganizationalUnitRemoteResourceService;
import ch.dlcm.storage.StorageService;

@Profile("arc-s3")
@Service
@ConditionalOnBean(ArchivalStorageController.class)
public class S3StorageService extends StorageService {
  private static final Logger log = LoggerFactory.getLogger(S3StorageService.class);

  private final String storageSpace;
  private final String storageAccessKey;
  private final String storageSecretKey;
  private final String storageUrl;
  private final String storageS3Signer;
  private final boolean storageIsPathStyleAccess;

  private final AmazonS3 s3Client;

  public S3StorageService(
          DLCMProperties dlcmProperties,
          DLCMRepositoryDescription repositoryDescription,
          MessageService messageService,
          MetadataService metadataService,
          FallbackOrganizationalUnitRemoteResourceService orgUnitResourceService,
          FallbackArchiveThumbnailRemoteResourceService archiveThumbnailRemoteResourceService,
          FallbackArchiveMetadataRemoteResourceService archiveMetadataRemoteResourceService) {
    super(dlcmProperties, repositoryDescription, messageService, metadataService, orgUnitResourceService,
            archiveThumbnailRemoteResourceService, archiveMetadataRemoteResourceService);
    this.storageSpace = dlcmProperties.getStorage().getSpace();
    this.storageAccessKey = dlcmProperties.getStorage().getAccessKey();
    this.storageSecretKey = dlcmProperties.getStorage().getSecretKey();
    this.storageUrl = dlcmProperties.getStorage().getUrl();
    this.storageS3Signer = dlcmProperties.getStorage().getS3Signer();
    this.storageIsPathStyleAccess = dlcmProperties.getStorage().getIsPathStyleAccess();
    this.s3Client = this.getNewS3Client();
  }

  @Override
  public List<String> checkFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToRecalculate, FixityCheckLevel fixityCheckLevel)
          throws NoSuchAlgorithmException, IOException {

    if (fixityCheckLevel == FixityCheckLevel.SIMPLE) {
      return this.checkSimpleFixity(aip, algorithmsToRecalculate);
    } else if (fixityCheckLevel == FixityCheckLevel.DEEP) {
      return this.checkDeepFixity(aip, algorithmsToRecalculate);
    }

    throw new FixityCheckLevelNotImplementedException(
            "FixityCheckLevel '" + fixityCheckLevel.name() + "' not implemented for storage of type " + this.getClass().getSimpleName());
  }

  @Override
  public long getAIPFileNumber(ArchiveContainer aipContainer, String aipId) throws IOException {
    try (S3Object s3Object = this.getS3Object(aipId)) {
      return Long.parseLong(s3Object.getObjectMetadata().getUserMetaDataOf(DLCMConstants.AIP_FILE_NUMBER));
    }
  }

  @Override
  public Map<String, String> getAIPInfo(String aipId) throws IOException {
    try (S3Object s3Object = this.getS3Object(aipId)) {
      return s3Object.getObjectMetadata().getUserMetadata();
    }
  }

  @Override
  public long getAIPSize(String aipId) throws IOException {
    try (S3Object s3Object = this.getS3Object(aipId)) {
      return s3Object.getObjectMetadata().getContentLength();
    }
  }

  @Override
  public InputStream getAIPStream(String aipId) {
    final S3Object s3Object = this.getS3Object(aipId);
    return s3Object.getObjectContent();
  }

  @Override
  public URI getArchiveLocation(String aipId) throws URISyntaxException {
    final URL url = this.getS3Client().getUrl(this.getBucket(), aipId);
    return url.toURI();
  }

  @Override
  public FixityCheckLevel getDefaultFixityCheckLevel() {
    return FixityCheckLevel.SIMPLE;
  }

  @Override
  public PackageStatus getStatus(ArchivalInfoPackage aip) throws IOException {
    // Check AIP
    final Path aipFile = Paths.get(aip.getDataFiles().get(0).getFinalData());
    if (!FileTool.checkFile(aipFile)) {
      return PackageStatus.IN_ERROR;
    }
    // Check S3 Storage
    try (S3Object s3Object = this.getS3Object(this.getArchiveId(aip))) {
      return PackageStatus.STORED;
    } catch (final AmazonServiceException e) {
      if (!e.getErrorCode().equals("NoSuchKey")) {
        throw new SolidifyCheckingException(e.getMessage());
      }
    }
    return PackageStatus.READY;
  }

  @Override
  public List<StoredAIP> listStoredAip() {
    final List<StoredAIP> list = new ArrayList<>();

    final ObjectListing s3list = this.getS3Client().listObjects(this.getBucket());
    for (final S3ObjectSummary s3obj : s3list.getObjectSummaries()) {
      if (!s3obj.getKey().endsWith("/")) {
        try {
          list.add(this.getStoredAip(s3obj.getKey()));
        } catch (final URISyntaxException e) {
          log.error("Cannot find stored AIP (" + s3obj.getKey() + ")", e);
        }
      }
    }
    return list;
  }

  @Override
  public String storeAIP(ArchivalInfoPackage aip) {
    try (BufferedInputStream in = new BufferedInputStream(Files.newInputStream(Paths.get(aip.getDataFiles().get(0).getFinalData())))) {
      // Set object metadata
      final ObjectMetadata s3md = new ObjectMetadata();
      s3md.setContentType(SolidifyConstants.ZIP_MIME_TYPE);
      for (final Entry<String, Object> metadata : aip.getAIPMetadata().entrySet()) {
        s3md.addUserMetadata(metadata.getKey(), metadata.getValue().toString());
      }
      s3md.setContentLength(aip.getDataFiles().get(0).getFileSize());
      // Upload data
      this.getS3Client().putObject(this.getBucket(), this.getArchiveId(aip), in, s3md);
      return aip.getResId();
    } catch (AmazonServiceException | IOException e) {
      log.error("Error in storing AIP in S3 container (" + aip.getResId() + ")", e);
    }
    return null;
  }

  @Override
  public void replaceAIP(ArchivalInfoPackage aip, Path newPackage) {
    try {
      // Upload data
      this.getS3Client().putObject(this.getBucket(), this.getArchiveId(aip), newPackage.toFile());
    } catch (AmazonServiceException e) {
      log.error("Error in storing tombstone AIP in S3 container (" + aip.getResId() + ")", e);
    }
  }

  @Override
  public String submitAIP(ArchivalInfoPackage aip) throws IOException, URISyntaxException {
    final Path aipFile = Paths.get(aip.getDataFiles().get(0).getFinalData());
    try (S3Object s3Object = this.getS3Object(this.getArchiveId(aip))) {
      if (FileTool.checkFile(aipFile) && FileTool.getSize(aipFile) == s3Object.getObjectMetadata().getContentLength()) {
        return this.getArchiveLocation(this.getArchiveId(aip)).toString();
      }
    }

    return null;
  }

  /**
   * Fetch the file from S3 storage and really recalculates the file hashes
   *
   * @param aip
   * @param algorithmsToCalculate
   * @return
   * @throws NoSuchAlgorithmException
   * @throws IOException
   */
  private List<String> checkDeepFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToCalculate)
          throws NoSuchAlgorithmException, IOException {

    try (InputStream stream = this.getAIPStream(this.getArchiveId(aip))) {
      return ChecksumTool.getChecksums(stream, algorithmsToCalculate);
    }
  }

  /**
   * Get the requested available file hashes by using the one provided by S3
   *
   * @param aip
   * @param algorithmsToRecalculate
   * @return the available hash values provided by S3 storage, if they are available (currently only
   * MD5 is returned)
   * @throws NoSuchAlgorithmException
   * @throws IOException
   */
  private List<String> checkSimpleFixity(ArchivalInfoPackage aip, List<ChecksumAlgorithm> algorithmsToRecalculate) {

    final List<String> hashes = new ArrayList<>();

    for (final ChecksumAlgorithm algo : algorithmsToRecalculate) {

      if (algo.name().equals("MD5")) {

        final ObjectMetadata metadata = this.getS3Client().getObjectMetadata(this.getBucket(), this.getArchiveId(aip));
        final String md5 = metadata.getETag();

        if (!md5.contains("-")) {

          /*
           * Case of MD5 of the whole file
           */
          hashes.add(md5);

        } else {

          /*
           * The file has been uploaded using multi part --> the returned MD5 is not the checksum of the
           * original file
           */
          // TODO: support fixity check for multi part files
          hashes.add(null);
        }

      } else {
        hashes.add(null);
      }
    }

    return hashes;
  }

  private String getBucket() {
    return this.storageSpace;
  }

  private AmazonS3 getNewS3Client() {
    final AWSCredentials awsCredentials = new BasicAWSCredentials(this.storageAccessKey, this.storageSecretKey);
    final EndpointConfiguration awsConfig = new EndpointConfiguration(this.storageUrl, "");
    final ClientConfiguration awsClientConfig = new ClientConfiguration().withSignerOverride(this.storageS3Signer);
    return AmazonS3ClientBuilder.standard()
            .withCredentials(new AWSStaticCredentialsProvider(awsCredentials))
            .withClientConfiguration(awsClientConfig)
            .withEndpointConfiguration(awsConfig)
            .withPathStyleAccessEnabled(this.storageIsPathStyleAccess)
            .build();
  }

  private AmazonS3 getS3Client() {
    return this.s3Client;
  }

  private S3Object getS3Object(String archiveId) {
    return this.getS3Client().getObject(this.getBucket(), archiveId);
  }

}
