/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Client - PreservationAIPClientService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.preservationplanning;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import org.springframework.core.env.Environment;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpMethod;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.service.NoSqlResourceClientService;
import ch.unige.solidify.util.CollectionTool;
import ch.unige.solidify.util.SudoRestClientTool;

import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.preservation.AipCopy;
import ch.dlcm.model.preservation.AipCopyList;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;

@Service
public class PreservationAIPClientService extends NoSqlResourceClientService<AipCopyList> {

  private static final String PRESERVATION_AIP_PATH = "/" + ResourceName.AIP;
  private static final String PRESERVATION_AIP_DELETE_CACHE_PATH = PRESERVATION_AIP_PATH + "/" + ActionName.DELETE_CACHE;

  protected PreservationAIPClientService(SudoRestClientTool restClientTool, Environment env) {
    super(AipCopyList.class, ModuleName.PRES_PLANNING, restClientTool, env);
  }

  /**
   * Clear the backend cache containing the list of AIP on all storage nodes
   */
  public void clearAipCopyListCache() {
    final RestTemplate restTemplate = this.restClientTool.getClient(this.getBaseUrl());
    restTemplate.exchange(PreservationAIPClientService.PRESERVATION_AIP_DELETE_CACHE_PATH, HttpMethod.POST, null, String.class);
  }

  /**
   * Get the list of AIP on all storage nodes
   *
   * @return
   */
  public List<AipCopyList> findAllAipCopies() {

    // clear the cache to ensure to get the current list of AIP
    this.clearAipCopyListCache();

    // get the list of replicated AIP
    final Pageable pageable = PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE);
    final String jsonResult = this.getResources(PreservationAIPClientService.PRESERVATION_AIP_PATH, pageable);
    return CollectionTool.getList(jsonResult, AipCopyList.class);
  }

  /**
   * Get the list of AIP on all storage nodes that have the COMPLETED status
   *
   * @return
   */
  public List<AipCopyList> findAllCompletedAipCopies() {
    List<AipCopyList> allAipCopies = this.findAllAipCopies();

    List<AipCopyList> completedAipCopyLists = new ArrayList<>();
    for (AipCopyList aipCopyList : allAipCopies) {
      List<AipCopy> completedCopies = aipCopyList.getCopies().stream()
              .filter(aipCopy -> aipCopy.getStatus() == PackageStatus.COMPLETED)
              .collect(Collectors.toList());
      if (!completedCopies.isEmpty()) {
        aipCopyList.setCopies(completedCopies);
        completedAipCopyLists.add(aipCopyList);
      }
    }
    return completedAipCopyLists;
  }
}
