/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - AbstractIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test;

import static ch.dlcm.DLCMConstants.NO_REPLY_PREFIX;
import static ch.dlcm.test.DLCMTestConstants.PRESERVATION_POLICY_TEST_NAME;
import static ch.dlcm.test.DLCMTestConstants.SUBMISSION_POLICY_TEST_NAME;
import static ch.unige.solidify.auth.model.AuthTestConstants.FIRST_NAME_TEST_SUFFIX;
import static ch.unige.solidify.auth.model.AuthTestConstants.LAST_NAME_TEST_SUFFIX;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.time.LocalDate;
import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadLocalRandom;
import java.util.stream.Collectors;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.core.env.Environment;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.web.client.HttpClientErrorException;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.auth.model.ApplicationRole;
import ch.unige.solidify.auth.model.AuthApplicationRole;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.util.SudoRestClientTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.AbstractDataFile;
import ch.dlcm.model.DataCategory;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.access.Order;
import ch.dlcm.model.display.OrganizationalUnitPreservationPolicyDisplay;
import ch.dlcm.model.display.OrganizationalUnitSubmissionPolicyDisplay;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.model.preingest.DepositDataFile;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.AdditionalFieldsForm;
import ch.dlcm.model.settings.FormDescriptionType;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.service.access.MetadataClientService;
import ch.dlcm.service.access.OrderClientService;
import ch.dlcm.service.admin.ArchiveTypeClientService;
import ch.dlcm.service.admin.DisseminationPolicyClientService;
import ch.dlcm.service.admin.InstitutionClientService;
import ch.dlcm.service.admin.LicenseClientService;
import ch.dlcm.service.admin.MetadataTypeClientService;
import ch.dlcm.service.admin.OrganizationalUnitClientService;
import ch.dlcm.service.admin.PersonClientService;
import ch.dlcm.service.admin.PreservationPolicyClientService;
import ch.dlcm.service.admin.ResearchDomainClientService;
import ch.dlcm.service.admin.RoleClientService;
import ch.dlcm.service.admin.SubmissionPolicyClientService;
import ch.dlcm.service.admin.UserClientService;
import ch.dlcm.service.archivalstorage.AIPClientService;
import ch.dlcm.service.datamgmt.IndexFieldAliasClientService;
import ch.dlcm.service.preingest.ContributorClientService;
import ch.dlcm.service.preingest.DepositClientService;
import ch.dlcm.test.DLCMTestConstants.OrganizationalUnitStatus;

@ActiveProfiles({ "client-sudo", "test" })
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ExtendWith(SpringExtension.class)
@SpringBootTest(properties = { "spring.cloud.bootstrap.enabled=true" })
public abstract class AbstractIT {

  protected static final String DEFAULT_DESCRIPTION = "Default description";
  protected static final String DEFAULT_TITLE = "Default title";
  protected static final String DEFAULT_TITLE_TO_UPDATE = "Archive to update";
  protected static final String DEFAULT_DESCRIPTION_TO_UPDATE = "An archive that gets updated each time integration tests are run";
  protected final static String REFERENCE_FILE = "http://www.unige.ch/rectorat/static/budget-2013.pdf";

  protected final static long REFERENCE_FILE_SIZE = 917380;
  protected final static String NO_ONE = "no_one";
  private static final Logger log = LoggerFactory.getLogger(AbstractIT.class);

  @Autowired
  protected MetadataClientService metadataService;

  @Autowired
  protected AIPClientService aipService;

  @Autowired
  protected ContributorClientService contributorService;

  @Autowired
  protected DepositClientService depositService;

  @Autowired
  protected DisseminationPolicyClientService disseminationPolicyService;

  @Autowired
  protected LicenseClientService licenseService;

  @Autowired
  protected MetadataTypeClientService metadataTypeService;

  @Autowired
  protected OrganizationalUnitClientService organizationalUnitService;

  @Autowired
  protected InstitutionClientService institutionService;

  @Autowired
  protected PersonClientService personService;

  @Autowired
  protected PreservationPolicyClientService preservationPolicyService;

  @Autowired
  protected ResearchDomainClientService researchDomainService;

  @Autowired
  protected ArchiveTypeClientService archiveTypeService;

  @Autowired
  protected RoleClientService roleService;

  @Autowired
  protected SubmissionPolicyClientService submissionPolicyService;

  @Autowired
  protected UserClientService userService;

  @Autowired
  protected IndexFieldAliasClientService indexFieldAliasService;

  @Autowired
  protected SudoRestClientTool restClientTool;

  @Autowired
  protected OrderClientService orderService;

  @Autowired
  protected Environment env;

  @BeforeEach
  protected void setup() {
    this.setUser();
    this.createFixtures();
  }

  protected void setUser() {
    this.restClientTool.sudoUser();
  }

  protected void createFixtures() {
    // Do nothing
  }

  protected void addAndVerifyPolicies(OrganizationalUnit organizationalUnit) {
    final String orgUnitId = organizationalUnit.getResId();

    final String preservationPolicyName = DLCMTestConstants.getNameWithPermanentLabel(PRESERVATION_POLICY_TEST_NAME);
    final PreservationPolicy preservationPolicy = this.preservationPolicyService.searchByProperties(Map.of("name", preservationPolicyName))
            .get(0);
    final List<OrganizationalUnitPreservationPolicyDisplay> currentPreservationPolicies = this.organizationalUnitService
            .getPreservationPolicies(orgUnitId);
    if (currentPreservationPolicies.isEmpty() ||
            currentPreservationPolicies.stream()
                    .noneMatch(orgUnitPresPolicy -> orgUnitPresPolicy.getName().equals(preservationPolicyName))) {
      this.organizationalUnitService.addPreservationPolicy(orgUnitId, preservationPolicy.getResId());
    }
    this.organizationalUnitService.setPreservationPolicyDefaultValue(orgUnitId, preservationPolicy.getResId(), true);

    final String submissionPolicyName = DLCMTestConstants.getNameWithPermanentLabel(SUBMISSION_POLICY_TEST_NAME);
    final SubmissionPolicy submissionPolicy = this.submissionPolicyService.searchByProperties(Map.of("name", submissionPolicyName)).get(0);
    final List<OrganizationalUnitSubmissionPolicyDisplay> currentSubmissionPolicies = this.organizationalUnitService
            .getSubmissionPolicies(orgUnitId);
    if (currentSubmissionPolicies.isEmpty() ||
            currentSubmissionPolicies.stream()
                    .noneMatch(orgUnitPresPolicy -> orgUnitPresPolicy.getName().equals(submissionPolicyName))) {
      this.organizationalUnitService.addSubmissionPolicy(orgUnitId, submissionPolicy.getResId());
    }
    this.organizationalUnitService.setSubmissionPolicyDefaultValue(orgUnitId, submissionPolicy.getResId(), true);
  }

  protected Deposit submitDeposit(Deposit deposit) {
    this.depositService.submit(deposit.getResId());
    return this.depositService.findOne(deposit.getResId());
  }

  protected Deposit approveDeposit(Deposit deposit) {
    this.depositService.approve(deposit.getResId());
    return this.depositService.findOne(deposit.getResId());
  }

  protected Deposit startEditingMetadata(Deposit deposit) {
    this.depositService.startEditingMetadata(deposit.getResId());
    return this.depositService.findOne(deposit.getResId());
  }

  protected void assertEqualsWithoutNanoSeconds(OffsetDateTime time1, OffsetDateTime time2) {
    assertEquals(time1.withNano(0), time2.withNano(0));
  }

  protected void assertEqualsWithoutNanoSeconds(String message, OffsetDateTime time1, OffsetDateTime time2) {
    assertEquals(time1.withNano(0), time2.withNano(0), message);
  }

  protected void enforcePersonHasRoleInOrgUnit(OrganizationalUnit organizationalUnit, Person person, String role) {
    Role effectiveRole = this.organizationalUnitService.getPersonRole(organizationalUnit.getResId(), person.getResId());
    if (effectiveRole == null) {
      this.organizationalUnitService.addPersonRole(organizationalUnit.getResId(), person.getResId(), role);
    } else if (role.equals(effectiveRole.getResId())) {
      return;
    } else {
      this.organizationalUnitService.updatePersonRole(organizationalUnit.getResId(), person.getResId(), role);
    }
  }

  protected void enforcePersonHasRoleInInstitution(Institution institution, Person person, String role) {
    Role effectiveRole = this.institutionService.getPersonRole(institution.getResId(), person.getResId());
    if (effectiveRole == null) {
      this.institutionService.addPersonRole(institution.getResId(), person.getResId(), role);
    } else if (role.equals(effectiveRole.getResId())) {
      return;
    } else {
      this.institutionService.updatePersonRole(institution.getResId(), person.getResId(), role);
    }
  }

  protected void enforceOrganizationalUnitInInstitution(Institution institution, OrganizationalUnit organizationalUnit) {
    OrganizationalUnit effectiveOrganizationalUnit = this.institutionService.getOrganizationalUnit(institution.getResId(),
            organizationalUnit.getResId());
    if (effectiveOrganizationalUnit == null) {
      this.institutionService.addOrganizationalUnit(institution.getResId(), organizationalUnit.getResId());
    }
  }

  protected void clearDepositsFixtures() {
    this.depositService.findAll().stream()
            .filter(d -> d.getTitle().startsWith(DLCMConstants.TEMPORARY_TEST_DATA_LABEL))
            .forEach(d -> {
              try {
                this.depositService.delete(d.getResId());
                log.info("Delete deposit {} ({}) with status ({})", d.getTitle(), d.getResId(), d.getStatus());
              } catch (final RuntimeException e) {
                log.info("Cannot delete deposit {} ({}) with status {}", d.getTitle(), d.getResId(), d.getStatus());
              }
            });
  }

  protected void clearDepositsFixtures(OrganizationalUnit organizationalUnit) {
    this.depositService.findAll().stream()
            .filter(d -> d.getTitle().startsWith(DLCMConstants.TEMPORARY_TEST_DATA_LABEL) &&
                    d.getOrganizationalUnitId().equals(organizationalUnit.getResId()))
            .forEach(d -> {
              try {
                this.depositService.delete(d.getResId());
              } catch (final RuntimeException e) {
                log.debug("Cannot delete deposit {} ({}) with status {}", d.getTitle(), d.getResId(), d.getStatus());
              }
            });
  }

  protected void clearLicenseFixtures() {
    this.licenseService.findAll().stream()
            .filter(license -> license.getOpenLicenseId().startsWith(DLCMConstants.TEMPORARY_TEST_DATA_LABEL))
            .forEach(license -> this.licenseService.delete(license.getResId()));
  }

  protected void clearPeopleFixtures() {
    this.personService.findAll().stream()
            .filter(person -> person.getLastName().startsWith(DLCMConstants.TEMPORARY_TEST_DATA_LABEL))
            .forEach(person -> this.personService.delete(person.getResId()));
  }

  /**
   * Create a collection of AIP
   *
   * @param deposit    the deposit used for the collection creation
   * @param num        the number of AIP in the collection
   * @param collection true => RestCollection of RestCollection, false => Collection of Unit
   * @return a list of AIP
   */
  protected List<ArchivalInfoPackage> createAIC(Deposit deposit, int num, boolean collection) {
    // Add AIU to AIC
    for (final ArchivalInfoPackage aiu : this.getAipList(num, collection)) {
      this.depositService.addAip(deposit.getResId(), aiu.getResId());
    }
    return this.depositService.getDepositAip(deposit.getResId());
  }

  protected Deposit createLocalDeposit(DLCMTestConstants.PersistenceMode persistenceMode, String title) {
    return this.createLocalDeposit(persistenceMode, title, DEFAULT_DESCRIPTION, Role.MANAGER_ID, false);
  }

  protected Deposit createLocalDeposit(DLCMTestConstants.PersistenceMode persistenceMode, String title, String description, String role,
          boolean inheritedRole) {
    final Deposit deposit = new Deposit();
    deposit.init();
    deposit.setTitle(this.getLabelPermanentOrTemporary(persistenceMode, title));
    deposit.setDescription(this.getLabelPermanentOrTemporary(persistenceMode, description));
    deposit.setStatus(DepositStatus.IN_PROGRESS);

    final OrganizationalUnit organizationalUnit = this.getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode.PERMANENT,
            OrganizationalUnitStatus.OPEN, role, inheritedRole);
    deposit.setOrganizationalUnitId(organizationalUnit.getResId());

    return deposit;
  }

  protected DepositDataFile createLocalDepositDataFile(Deposit deposit) {
    // Create a deposit data file object
    final DepositDataFile depositDataFile = new DepositDataFile(deposit);
    URI referenceFileUri = URI.create(REFERENCE_FILE);
    depositDataFile.setSourceData(referenceFileUri);
    depositDataFile.setFileSize(REFERENCE_FILE_SIZE);
    depositDataFile.setDataCategory(DataCategory.Primary);
    depositDataFile.setDataType(DataCategory.Derived);
    depositDataFile.setRelativeLocation("/tmp/dlcm/test");

    return depositDataFile;
  }

  protected Deposit createRemoteDeposit(DLCMTestConstants.PersistenceMode persistenceMode, String title, String description) {
    return this.createRemoteDeposit(persistenceMode, title, description, Role.MANAGER_ID, false);
  }

  protected Deposit createRemoteDeposit(DLCMTestConstants.PersistenceMode persistenceMode, String title, String description, String role,
          boolean inheritedRole) {
    Deposit deposit = this.createLocalDeposit(persistenceMode, title, description, role, inheritedRole);
    deposit = this.depositService.create(deposit);
    return deposit;
  }

  protected Deposit createRemoteDepositWithContributor(DLCMTestConstants.PersistenceMode persistenceMode, String title, String description) {
    final Deposit deposit = this.createRemoteDeposit(persistenceMode, title, description, Role.MANAGER_ID, false);

    final Person person = this.getOrCreateRemotePerson();
    this.depositService.addContributor(deposit.getResId(), person.getResId());
    return deposit;
  }

  protected Deposit createRemoteDepositWithData(Deposit deposit) {
    // Create the deposit on the server if needed
    Deposit remoteDeposit;
    try {
      remoteDeposit = this.depositService.findOne(deposit.getResId());
    } catch (HttpClientErrorException.NotFound e) {
      remoteDeposit = this.depositService.create(deposit);
    }

    // Create the deposit data file on the server
    for (final DepositDataFile df : deposit.getDataFiles()) {
      this.depositService.addDepositDataFile(remoteDeposit.getResId(), df);
    }
    return remoteDeposit;
  }

  /**
   * Delete fixtures created by tests
   */
  protected abstract void deleteFixtures();

  protected Person getOrCreateRemotePerson() {
    return this.getOrCreateRemotePerson(AuthApplicationRole.USER_ID);
  }

  protected Person getOrCreateRemotePerson(String applicationRole) {
    final Person person = new Person();
    person.setFirstName(DLCMTestConstants.getNameWithPermanentLabel(applicationRole + FIRST_NAME_TEST_SUFFIX));
    person.setLastName(DLCMTestConstants.getNameWithPermanentLabel(applicationRole + LAST_NAME_TEST_SUFFIX));
    // Return the existing person if already created
    for (final Person p : this.personService.findAll()) {
      if (p.getFirstName().equalsIgnoreCase(person.getFirstName()) &&
              p.getLastName().equalsIgnoreCase(person.getLastName())) {
        return p;
      }
    }
    return this.personService.create(person);
  }

  /**
   * This method will create/update an Org unit for the Person Linked to the user as role Manager.
   *
   * @param orgUnitStatus
   * @return
   */
  protected OrganizationalUnit getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode persistenceMode,
          OrganizationalUnitStatus orgUnitStatus) {
    return this.getOrCreateOrganizationalUnit(persistenceMode, orgUnitStatus, Role.MANAGER_ID, false);
  }

  /**
   * This method will create/update an Org unit for the Person Linked to the user as role Manager.
   *
   * @param orgUnitStatus
   * @return
   */
  protected OrganizationalUnit getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode persistenceMode,
          OrganizationalUnitStatus orgUnitStatus, String role) {
    return this.getOrCreateOrganizationalUnit(persistenceMode, orgUnitStatus, role, false);
  }

  /**
   * This method will create/update an Org unit for the Person Linked to the user with the role
   * specified
   *
   * @param orgUnitStatus
   * @param role
   * @param inheritedRole
   * @return
   */
  protected OrganizationalUnit getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode persistenceMode,
          OrganizationalUnitStatus orgUnitStatus, String role,
          boolean inheritedRole) {
    if (this.isRootOrAdmin()) {
      return this.getOrCreateOrganizationalUnit(persistenceMode, orgUnitStatus, role, AuthApplicationRole.ADMIN_ID, inheritedRole);
    } else {
      return this.getTestOrganizationalUnitAsUser(persistenceMode, orgUnitStatus, role, inheritedRole);
    }
  }

  protected OrganizationalUnit getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode persistenceMode,
          OrganizationalUnitStatus orgUnitStatus, String role,
          String applicationRole) {
    return this.getOrCreateOrganizationalUnit(persistenceMode, orgUnitStatus, role, applicationRole, false);
  }

  protected OrganizationalUnit getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode persistenceMode,
          OrganizationalUnitStatus orgUnitStatus, String role,
          String applicationRole, boolean inheritedRole) {
    OrganizationalUnit organizationalUnit = this.createLocalOrganizationalUnit(persistenceMode, orgUnitStatus, role,
            inheritedRole);

    final String orgUnitName = organizationalUnit.getName();

    final List<OrganizationalUnit> orgUnits = this.organizationalUnitService.searchByProperties(Map.of("name", orgUnitName));

    final Optional<OrganizationalUnit> optionalOrgUnit = orgUnits.stream().filter(item -> item.getName().equals(orgUnitName)).findFirst();

    if (optionalOrgUnit.isPresent()) {
      organizationalUnit = this.organizationalUnitService.update(optionalOrgUnit.get().getResId(), organizationalUnit);
    } else {
      organizationalUnit = this.organizationalUnitService.create(organizationalUnit);
    }
    this.addAndVerifyPolicies(organizationalUnit);

    final Person person = this.getOrCreateRemotePerson(applicationRole);

    if (inheritedRole) {
      Institution institution = this.getOrCreatePermanentInstitution(persistenceMode, role);
      this.enforceOrganizationalUnitInInstitution(institution, organizationalUnit);

      // Add person
      this.enforcePersonHasRoleInInstitution(institution, person, role);
    } else {
      // Add person
      this.enforcePersonHasRoleInOrgUnit(organizationalUnit, person, role);
    }

    return organizationalUnit;
  }

  /**
   * This method will create/update an Institution for the Person Linked to the user with the role
   * specified
   *
   * @param role
   * @return
   */
  protected Institution getOrCreatePermanentInstitution(DLCMTestConstants.PersistenceMode persistenceMode, String role) {
    if (this.isRootOrAdmin()) {
      return this.getOrCreatePermanentInstitution(persistenceMode, role, AuthApplicationRole.ADMIN_ID);
    } else {
      return this.getTestInstitutionAsUser(persistenceMode, role);
    }
  }

  protected Institution getOrCreatePermanentInstitution(DLCMTestConstants.PersistenceMode persistenceMode, String role, String applicationRole) {
    Institution institution = this.createLocalInstitution(persistenceMode, role);
    final String institutionName = institution.getName();

    final List<Institution> institutions = this.institutionService.searchByProperties(Map.of("name", institutionName));

    final Optional<Institution> optionalInstitution = institutions.stream().filter(item -> item.getName().equals(institutionName)).findFirst();

    if (optionalInstitution.isPresent()) {
      institution = this.institutionService.update(optionalInstitution.get().getResId(), institution);
    } else {
      institution = this.institutionService.create(institution);
    }

    // Add person
    final Person person = this.getOrCreateRemotePerson(applicationRole);
    this.enforcePersonHasRoleInInstitution(institution, person, role);

    institution.setOrganizationalUnit(new ArrayList<>());
    return institution;
  }

  protected OrganizationalUnit getSecondPermanentOrganizationalUnit() {
    final String orgUnitName = DLCMTestConstants.getNameWithPermanentLabel("Organizational Unit 2");
    final List<OrganizationalUnit> orgUnits = this.organizationalUnitService.searchByProperties(Map.of("name", orgUnitName));

    final Optional<OrganizationalUnit> optionalOrgUnit = orgUnits.stream().filter(item -> item.getName().equals(orgUnitName)).findFirst();

    if (optionalOrgUnit.isEmpty()) {
      final OrganizationalUnit organizationalUnit = new OrganizationalUnit();
      organizationalUnit.setName(orgUnitName);
      this.setRandomUrlOrganizationalUnit(organizationalUnit);
      this.organizationalUnitService.create(organizationalUnit);
      return organizationalUnit;
    } else {
      return optionalOrgUnit.get();
    }
  }

  protected AdditionalFieldsForm getPermanentAdditionalFieldsFormForOrganizationalUnit(String unitId) {
    AdditionalFieldsForm form = this.organizationalUnitService.getCurrentAdditionalFieldsForm(unitId);
    if (form == null) {
      form = new AdditionalFieldsForm();
      form.setName("[TEST] Additional fields form");
      form.setType(FormDescriptionType.FORMLY);
      form.setDescription(
              "[{\"key\" : \"custom_field_1\", \"type\" : \"input\", \"templateOptions\": { \"label\" : \"My 1st custom field\" }}]");

      form = this.organizationalUnitService.addAdditionalFieldsForm(unitId, form);
    }
    return form;
  }

  /***********************************/

  protected boolean isRootOrAdmin() {
    String role = this.restClientTool.getCurrentRole();
    return AuthApplicationRole.ADMIN_ID.equals(role) || AuthApplicationRole.ROOT_ID.equals(role);
  }

  protected void setRandomUrlOrganizationalUnit(OrganizationalUnit organizationalUnit) {
    organizationalUnit.setUrl(this.generateRandomUrl());
  }

  protected void setRandomUrlInstitution(Institution institution) {
    institution.setUrl(this.generateRandomUrl());
  }

  private URL generateRandomUrl() {
    URL url = null;
    try {
      url = new URL("https://unige.ch/" + ThreadLocalRandom.current().nextInt());
    } catch (final MalformedURLException e) {
      e.printStackTrace();
    }
    return url;
  }

  protected boolean waitDepositIsCompleted(String depositId) {
    return this.waitDepositIsProcessed(depositId, DepositStatus.COMPLETED);
  }

  protected boolean waitDepositIsProcessed(String depositId, DepositStatus endStatus) {
    int count = 0;
    Deposit deposit;

    do {
      deposit = this.depositService.findOne(depositId);
      DLCMTestConstants
              .failAfterNAttempt(count, "Deposit " + depositId + " should have status Completed or InError instead of " + deposit.getStatus());
      count++;
    } while (deposit.getStatus() != DepositStatus.COMPLETED && deposit.getStatus() != DepositStatus.IN_ERROR);
    return (endStatus == deposit.getStatus());
  }

  protected boolean waitDepositIsDeleted(String depositId) {
    int count = 0;
    boolean depositFound;
    do {
      try {
        this.depositService.findOne(depositId);
        depositFound = true;
      } catch (HttpClientErrorException.NotFound e) {
        depositFound = false;
      }
      DLCMTestConstants.failAfterNAttempt(count, "Deposit " + depositId + " should have been deleted");
      count++;
    } while (depositFound);

    return true;
  }

  protected boolean waitDepositWithDataIsReady(String depositId) {
    int count = 0;
    List<DepositDataFile> dataFileList;

    do {
      dataFileList = this.depositService.getDepositDataFile(depositId);
      DLCMTestConstants.failAfterNAttempt(count, "Deposit data files of deposit " + depositId + " are not all in ready status.");
      dataFileList = this.depositService.getDepositDataFile(depositId);
      count++;
    } while (!this.isAllReadyInDataFileListOrError(dataFileList));

    return this.isAllReadyInDataFileListOrError(dataFileList);
  }

  private boolean isAllReadyInDataFileListOrError(List<DepositDataFile> dataFileList) {
    return dataFileList.stream()
            .allMatch(
                    df -> df.getStatus() == AbstractDataFile.DataFileStatus.READY || df.getStatus() == AbstractDataFile.DataFileStatus.IN_ERROR);
  }

  protected boolean waitAIPIsCompletedOrError(String orgUnitId, String title) {
    int count = 0;
    List<ArchivalInfoPackage> aipList;
    do {
      aipList = this.aipService.searchByProperties(Map.of("info.organizationalUnitId", orgUnitId));
      // filter by deposit name as well
      aipList = aipList.stream().filter(a -> a.getInfo().getName().equals(title)).collect(Collectors.toList());
      DLCMTestConstants.failAfterNAttempt(count, "AIP from organizational Unit " + orgUnitId + " is not in completed status.");
      count++;

    } while (!this.isAllAIPCompletedOrError(aipList) || !(aipList.size() == 1));
    return this.isAllAIPCompletedOrError(aipList);
  }

  private boolean isAllAIPCompletedOrError(List<ArchivalInfoPackage> aipList) {
    return aipList.stream()
            .allMatch(aip -> aip.getInfo().getStatus() == PackageStatus.COMPLETED || aip.getInfo().getStatus() == PackageStatus.IN_ERROR);
  }

  protected boolean waitOrderIsReadyOrInError(String aipId) {
    int count = 0;
    List<Order> orderList;
    do {
      orderList = this.orderService.searchByProperties(Map.of("name", aipId, "query", aipId));
      DLCMTestConstants.failAfterNAttempt(count, "Order from archive " + aipId + " is not in ready status.");
      count++;

    } while (!this.isAllOrderReadyOrError(orderList));

    return this.isAllOrderReadyOrError(orderList);
  }

  protected boolean waitForPublicOrderIsReadyOrInError(String aipId) {
    int count = 0;
    List<Order> orderList;
    do {
      orderList = this.orderService.findAllPublic(Map.of("name", aipId, "query", aipId));
      DLCMTestConstants.failAfterNAttempt(count, "Order from archive " + aipId + " is not in ready status.");
      count++;

    } while (!this.isAllOrderReadyOrError(orderList));

    return this.isAllOrderReadyOrError(orderList);
  }

  private boolean isAllOrderReadyOrError(List<Order> orderList) {
    return orderList.stream().allMatch(order -> order.getStatus() == Order.OrderStatus.READY || order.getStatus() == Order.OrderStatus.IN_ERROR);
  }

  /***********************************/

  protected OrganizationalUnit createLocalOrganizationalUnit(DLCMTestConstants.PersistenceMode persistenceMode,
          OrganizationalUnitStatus orgUnitStatus,
          String role,
          boolean inheritedRole) {
    final String orgUnitName = this.getOrgUnitName(persistenceMode, role, inheritedRole);
    final OrganizationalUnit organizationalUnit = new OrganizationalUnit();
    organizationalUnit.setName(orgUnitName);
    organizationalUnit.setKeywords(Arrays.asList("chromatography", "flocculation", "ultrafiltration"));
    this.setRandomUrlOrganizationalUnit(organizationalUnit);
    switch (orgUnitStatus) {
      case CLOSED:
        organizationalUnit.setOpeningDate(LocalDate.now().minusYears(10));
        organizationalUnit.setClosingDate(LocalDate.now().minusYears(5));
        break;
      case OPEN:
        organizationalUnit.setOpeningDate(LocalDate.now().minusYears(10));
        organizationalUnit.setClosingDate(LocalDate.now().plusYears(5));
    }
    return organizationalUnit;
  }

  protected Institution createLocalInstitution(DLCMTestConstants.PersistenceMode persistenceMode, String role) {
    final String institutionName = this.getInstitutionName(persistenceMode, role);
    final Institution institution = new Institution();
    institution.setName(institutionName);
    this.setRandomUrlInstitution(institution);
    return institution;
  }

  /**
   * Get a list of AIP
   *
   * @param max          the number of AIP in the result list
   * @param isCollection true => return a list of AIP RestCollection, false => return a list of AIP
   *                     Unit
   * @return a list of AIP
   */
  private List<ArchivalInfoPackage> getAipList(int max, boolean isCollection) {
    final List<ArchivalInfoPackage> list = new ArrayList<>();
    int count = 0;
    final String aipNamePrefix;
    if (isCollection) {
      aipNamePrefix = DLCMConstants.PERMANENT_TEST_DATA_LABEL + DLCMConstants.ARCHIVAL_COLLECTION_LABEL;
    } else {
      aipNamePrefix = DLCMConstants.PERMANENT_TEST_DATA_LABEL + DLCMConstants.ARCHIVAL_UNIT_LABEL;
    }
    final List<ArchivalInfoPackage> aipList = this.aipService.searchByProperties(Map.of("info.name", aipNamePrefix));
    for (final ArchivalInfoPackage aip : aipList) {
      if (aip.getInfo().getStatus() == PackageStatus.COMPLETED) {
        if (isCollection == aip.isCollection()) {
          count++;
          list.add(aip);
          if (count == max) {
            break;
          }
        }
      }
    }
    assertEquals(max, list.size(), "Not enough completed AIP");
    return list;
  }

  private String getOrgUnitName(DLCMTestConstants.PersistenceMode persistenceMode, String role) {
    return getOrgUnitName(persistenceMode, role, false);
  }

  private OrganizationalUnit getTestOrganizationalUnitAsUser(DLCMTestConstants.PersistenceMode persistenceMode,
          OrganizationalUnitStatus orgUnitStatus, String role,
          boolean inheritedRole) {
    final OrganizationalUnit organizationalUnit = this.createLocalOrganizationalUnit(persistenceMode, orgUnitStatus, role,
            inheritedRole);
    final String orgUnitName = organizationalUnit.getName();

    final List<OrganizationalUnit> orgUnits = this.organizationalUnitService.searchByProperties(Map.of("name", orgUnitName));

    return orgUnits.stream().filter(item -> item.getName().equals(orgUnitName)).findFirst().orElse(null);
  }

  private Institution getTestInstitutionAsUser(DLCMTestConstants.PersistenceMode persistenceMode, String role) {
    final Institution institution = this.createLocalInstitution(persistenceMode, role);
    final String institutionName = institution.getName();

    final List<Institution> institutions = this.institutionService.searchByProperties(Map.of("name", institutionName));

    Institution institutionResult = institutions.stream().filter(item -> item.getName().equals(institutionName)).findFirst().orElse(null);
    institutionResult.setOrganizationalUnit(new ArrayList<>());
    return institutionResult;
  }

  protected OrganizationalUnit getOrgUnitForNoOne(DLCMTestConstants.PersistenceMode persistenceMode) {
    final List<OrganizationalUnit> orgUnits = this.organizationalUnitService
            .searchByProperties(Map.of("name", this.getOrgUnitName(persistenceMode, NO_ONE)));
    if (orgUnits.isEmpty()) {
      fail("Organizational unit for no one is missing");
    } else if (orgUnits.size() > 1) {
      fail("There is multiple organizational units for no one");
    }
    return orgUnits.get(0);
  }

  private String getOrgUnitName(DLCMTestConstants.PersistenceMode persistenceMode, String role, boolean inheritedRole) {
    String name;
    if (role == null) {
      name = "Organizational Unit";
    } else if (inheritedRole) {
      name = "Organizational Unit for inherited " + role.toLowerCase();
    } else {
      name = "Organizational Unit for " + role.toLowerCase();
    }
    return this.getLabelPermanentOrTemporary(persistenceMode, name);
  }

  private String getInstitutionName(DLCMTestConstants.PersistenceMode persistenceMode, String role) {
    String name;
    if (role == null) {
      name = "Institution";
    } else {
      name = "Institution for " + role.toLowerCase();
    }
    return this.getLabelPermanentOrTemporary(persistenceMode, name);
  }

  private String getLabelPermanentOrTemporary(DLCMTestConstants.PersistenceMode persistenceMode, String name) {
    if (persistenceMode.equals(DLCMTestConstants.PersistenceMode.PERMANENT)) {
      return DLCMTestConstants.getNameWithPermanentLabel(name);
    }
    if (persistenceMode.equals(DLCMTestConstants.PersistenceMode.TEMPORARY)) {
      return DLCMTestConstants.getRandomNameWithTemporaryLabel(name);
    }
    throw new IllegalStateException("Unmanaged persistence mode " + persistenceMode);
  }

  protected User getUserWithRole(ApplicationRole role) {
    return this.userService.searchByProperties(Map.of("email", NO_REPLY_PREFIX + role.getName().toLowerCase() + "@unige.ch")).get(0);
  }

  protected User findThirdPartyUser() {
    return this.userService.searchByProperties(Map.of("email", NO_REPLY_PREFIX + "THIRD-PARTY-USER@unige.ch")).get(0);
  }

  protected List<Map<String, Object>> getValidationErrors(HttpClientErrorException httpClientErrorException) {
    try {
      String responseBody = httpClientErrorException.getResponseBodyAsString();
      Map<String, Object> responseJsonMap = new ObjectMapper().readValue(responseBody, Map.class);
      return (List<Map<String, Object>>) responseJsonMap.get("validationErrors");
    } catch (JsonProcessingException e) {
      throw new SolidifyRuntimeException("unable to get validationErrors from HttpClientErrorException", e);
    }
  }

  protected List<String> getFieldValidationErrors(HttpClientErrorException httpClientErrorException, String fieldName) {
    List<Map<String, Object>> validationErrors = this.getValidationErrors(httpClientErrorException);
    if (validationErrors != null) {
      Optional<Map<String, Object>> fieldValidationErrorOpt = validationErrors.stream().filter(map -> map.get("fieldName").equals(fieldName))
              .findFirst();
      if (fieldValidationErrorOpt.isPresent()) {
        return (List<String>) fieldValidationErrorOpt.get().get("errorMessages");
      } else {
        return new ArrayList<>();
      }
    }
    return new ArrayList<>();
  }

  protected void assertFieldContainsValidationErrorMessage(HttpClientErrorException e, String fieldName, String expectedErrorMessage) {
    List<String> errorMessages = this.getFieldValidationErrors(e, fieldName);
    assertTrue(errorMessages.contains(expectedErrorMessage),
            errorMessages + " doesn't contain the message '" + expectedErrorMessage + "'");
  }

  protected String getNewArchiveToUpdateTitle(String currentTitle) {
    if (!currentTitle.contains("[")) {
      return currentTitle + " [0]";
    } else {
      final int currentUpdateIndex = this.getNumberBetweenSquareBrackets(currentTitle);
      return currentTitle.replace("[" + currentUpdateIndex + "]", "[" + (currentUpdateIndex + 1) + "]");
    }
  }

  protected int getNumberBetweenSquareBrackets(String text) {
    int startBracketIndex = text.lastIndexOf('[');
    int endBracketIndex = text.lastIndexOf(']');
    String updateIndexStr = text.substring(startBracketIndex + 1, endBracketIndex);
    return Integer.parseInt(updateIndexStr);
  }

  protected void runParallelCalls(Runnable runnable, int nbThreads, int nbCalls) throws InterruptedException {
    final CountDownLatch countDownLatch = new CountDownLatch(nbCalls);
    final ExecutorService executorService = Executors.newFixedThreadPool(nbThreads);
    for (int i = 0; i < nbCalls; i++) {
      executorService.submit(() -> {
        try {
          runnable.run();
          countDownLatch.countDown();
        } catch (Throwable t) {
          t.printStackTrace();
        }
      });
    }
    countDownLatch.await();
  }

  @AfterEach
  public void tearDown() {
    this.deleteFixtures();
    this.restClientTool.exitSudo();
  }
}
