/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - OrderIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.access;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.util.HashMap;
import java.util.List;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;

import ch.unige.solidify.auth.model.AuthApplicationRole;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.access.Order;
import ch.dlcm.model.access.Order.OrderStatus;
import ch.dlcm.model.access.QueryType;
import ch.dlcm.test.DLCMTestConstants;

class OrderIT extends AbstractAccessIT {

  @AfterEach
  public void cleanOrderFixture() {
    this.restClientTool.sudoAdmin();
    this.orderService.findAll().stream().filter(order -> order.getName().startsWith(DLCMConstants.TEMPORARY_TEST_DATA_LABEL))
            .forEach(order -> this.orderService.delete(order.getResId()));
    this.restClientTool.exitSudo();
  }

  @Test
  void findAllOnlyForAdminTest() {
    assertThrows(HttpClientErrorException.Forbidden.class, () -> this.orderService.findAll());

    this.restClientTool.sudoAdmin();
    List<Order> listOrderAdmin = this.orderService.findAll();
    assertNotNull(listOrderAdmin);
    this.restClientTool.exitSudo();

    this.restClientTool.sudoRoot();
    List<Order> listOrderRoot = this.orderService.findAll();
    assertNotNull(listOrderRoot);
    this.restClientTool.exitSudo();
  }

  @Test
  void findAllPublicTest() {
    List<Order> listOrderPublic = this.orderService.findAllPublic(new HashMap<>());
    assertNotNull(listOrderPublic);

    this.restClientTool.sudoAdmin();
    List<Order> listAllOrder = this.orderService.findAll();
    this.restClientTool.exitSudo();

    assertEquals(listAllOrder.stream().filter(order -> order.isPublicOrder()).toArray().length, listOrderPublic.size());
  }

  @Test
  void findAllCreatedByUserTest() {
    List<Order> listOrderCreatedByUser = this.orderService.findAllCreatedByUser();
    assertNotNull(listOrderCreatedByUser);

    this.restClientTool.sudoAdmin();
    List<Order> listAllOrder = this.orderService.findAll();

    // Only possible to list orders with role ADMIN or ROOT
    String externalUid = this.getUserWithRole(AuthApplicationRole.USER).getExternalUid();
    this.restClientTool.exitSudo();

    assertEquals(listAllOrder.stream().filter(order -> order.getCreatedBy().equals(externalUid)).toArray().length,
            listOrderCreatedByUser.size());
  }

  @Test
  void findOnePublicForEveryoneTest() {
    String orderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order");

    // Create an order
    final Order order1 = new Order();
    order1.setName(orderName);
    order1.setQueryType(QueryType.DIRECT);
    order1.setQuery(orderName);
    order1.setPublicOrder(Boolean.TRUE);
    final Order order2 = this.orderService.create(order1);

    assertNotNull(this.orderService.findOne(order2.getResId()));

    this.restClientTool.sudoThirdPartyUser();
    assertNotNull(this.orderService.findOne(order2.getResId()));
    this.restClientTool.exitSudo();
  }

  @Test
  void findOnePrivateForCreatorOnlyTest() {
    String orderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order");

    // Create an order
    final Order order1 = new Order();
    order1.setName(orderName);
    order1.setQueryType(QueryType.DIRECT);
    order1.setQuery(orderName);
    order1.setPublicOrder(Boolean.FALSE);
    final Order order2 = this.orderService.create(order1);

    assertNotNull(this.orderService.findOne(order2.getResId()));

    this.restClientTool.sudoThirdPartyUser();
    assertThrows(HttpClientErrorException.Forbidden.class, () -> this.orderService.findOne(order2.getResId()));
    this.restClientTool.exitSudo();

    this.restClientTool.sudoAdmin();
    assertNotNull(this.orderService.findOne(order2.getResId()));
    this.restClientTool.exitSudo();

    this.restClientTool.sudoRoot();
    assertNotNull(this.orderService.findOne(order2.getResId()));
    this.restClientTool.exitSudo();
  }

  @Test
  void creationTest() {
    String orderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order");

    // Create an order
    final Order order1 = new Order();
    order1.setName(orderName);
    order1.setQueryType(QueryType.DIRECT);
    order1.setQuery(orderName);
    final Order order2 = this.orderService.create(order1);

    // Test the creation return
    assertNotNull(order2, "Cannot create query order");
    assertEquals(order1.getName(), order2.getName(), "Problem with query order name");
    assertTrue(order2.isPublicOrder(), "Order should be public by default");

    // Test creation
    final Order order3 = this.orderService.findOne(order2.getResId());
    assertNotNull(order3, "Cannot find query order");
    assertEquals(order1.getName(), order3.getName(), "Problem with query order name");
    assertEquals(OrderStatus.IN_PROGRESS, order2.getStatus(), "Wrong query order status");

    // Create a private order
    final Order order4 = new Order();
    String privateOrderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Private Order");
    order4.setName(privateOrderName);
    order4.setQueryType(QueryType.DIRECT);
    order4.setQuery(privateOrderName);
    order4.setPublicOrder(false);
    final Order order5 = this.orderService.create(order4);
    assertFalse(order5.isPublicOrder(), "Order should be private");

  }

  @Test
  void deleteTest() {
    String orderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order");

    // Create a new order
    Order order = new Order();
    order.setName(orderName);
    order.setQueryType(QueryType.DIRECT);
    order.setQuery(orderName);
    order = this.orderService.create(order);

    // Delete the order
    final String resId = order.getResId();
    this.orderService.delete(resId);
    assertThrows(HttpClientErrorException.NotFound.class, () -> this.orderService.findOne(resId));
  }

  @Test
  void cantDeleteOrderOfThirdUserTest() {
    String orderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order");

    this.restClientTool.sudoThirdPartyUser();
    // Create a new order
    Order order = new Order();
    order.setName(orderName);
    order.setQueryType(QueryType.DIRECT);
    order.setQuery(orderName);
    order = this.orderService.create(order);
    this.restClientTool.exitSudo();

    // Try delete the order
    final String resId = order.getResId();
    assertThrows(HttpClientErrorException.Forbidden.class, () -> this.orderService.delete(resId));
  }

  @Test
  void submitOrderWithoutAipsTest() {
    String orderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order");

    // Create an order
    Order order = new Order();
    order.setName(orderName);
    order.setQueryType(QueryType.DIRECT);
    order.setQuery(orderName);
    order = this.orderService.create(order);

    try {
      this.orderService.setSubmitStatus(order.getResId());
      fail("A BAD REQUEST http response should be received");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode());
    }
  }

  @Test
  void updateStatusTest() {
    String orderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order");

    // Create an order
    Order order = new Order();
    order.setName(orderName);
    order.setQueryType(QueryType.DIRECT);
    order.setQuery(orderName);
    order = this.orderService.create(order);

    order.setStatus(OrderStatus.SUBMITTED);

    try {
      this.orderService.update(order.getResId(), order);
      fail("A BAD_REQUEST Http response should be received");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode());
    }
  }

  @Test
  void cantUpdateOrderOfThirdUserTest() {
    String orderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order");

    this.restClientTool.sudoThirdPartyUser();
    // Create a new order
    Order order = new Order();
    order.setName(orderName);
    order.setQueryType(QueryType.DIRECT);
    order.setQuery(orderName);
    order = this.orderService.create(order);
    this.restClientTool.exitSudo();

    order.setStatus(OrderStatus.SUBMITTED);

    try {
      this.orderService.update(order.getResId(), order);
      fail("A FORBIDDEN Http response should be received");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.FORBIDDEN, e.getStatusCode());
    }
  }

  @Test
  void updateTest() {
    // Create an order
    String orderName = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order");
    Order order = new Order();
    order.setName(orderName);
    order.setQueryType(QueryType.DIRECT);
    order.setQuery(orderName);
    order = this.orderService.create(order);

    // Update the order
    String orderName2 = DLCMTestConstants.getRandomNameWithTemporaryLabel("Order 2");
    order.setName(orderName2);
    order.setQuery(orderName2);
    final Order actualOrder = this.orderService.update(order.getResId(), order);
    final Order order2 = this.orderService.findOne(order.getResId());

    // Test the update
    assertNotNull(actualOrder, "Cannot find query order");
    assertEquals(order.getName(), order2.getName(), "Problem with query order name");
    this.assertEqualsWithoutNanoSeconds("Problem with query order creation time", order.getCreationTime(), order2.getCreationTime());
  }
}
