/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - AbstractOrganizationalUnitWithRoleIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.admin;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.junit.jupiter.api.BeforeAll;
import org.springframework.core.io.ClassPathResource;

import ch.unige.solidify.auth.model.AuthApplicationRole;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.display.OrganizationalUnitDTO;
import ch.dlcm.model.policies.DisseminationPolicy;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.test.DLCMTestConstants;

abstract class AbstractOrganizationalUnitWithRoleIT extends AbstractAdminIT {

  protected abstract String role();

  protected boolean inheritedRole() {
    return false;
  }

  @Override
  protected void setUser() {
    this.restClientTool.sudoUser();
  }

  @BeforeAll
  void setUp() {
    this.getOrCreateOrganizationalUnitForRole(DLCMTestConstants.PersistenceMode.PERMANENT);
  }

  @Override
  protected void deleteFixtures() {
    this.restClientTool.sudoAdmin();
    this.clearOrganizationalUnitFixtures();
    this.clearFundingAgencyFixtures();
    this.clearInstitutionsFixtures();
    this.restClientTool.exitSudo();
  }

  protected OrganizationalUnit getOrCreateOrganizationalUnitForRole(DLCMTestConstants.PersistenceMode persistenceMode) {
    this.restClientTool.sudoAdmin();
    OrganizationalUnit organizationalUnit = this.getOrCreateOrganizationalUnit(persistenceMode, DLCMTestConstants.OrganizationalUnitStatus.OPEN,
            this.role(), AuthApplicationRole.USER_ID, this.inheritedRole());
    this.restClientTool.exitSudo();
    return organizationalUnit;
  }

  protected OrganizationalUnit createRemoteOrgUnitWithInstitutionWithCurrentRole(DLCMTestConstants.PersistenceMode persistenceMode,
          String description) {
    Institution institution = this.getOrCreatePermanentInstitution(persistenceMode, this.role(), AuthApplicationRole.USER_ID);
    return createRemoteOrgUnit(description, institution);
  }

  protected OrganizationalUnit createRemoteOrgUnitWithInstitutionForSpecificRole(DLCMTestConstants.PersistenceMode persistenceMode,
          String description,
          String role) {
    Institution institution = this.getOrCreatePermanentInstitution(persistenceMode, role);
    return createRemoteOrgUnit(description, institution);
  }

  protected OrganizationalUnit createRemoteOrgUnit(String description) {
    return createRemoteOrgUnit(description, new ArrayList<Institution>());
  }

  protected OrganizationalUnit createRemoteOrgUnit(String description, Institution institution) {
    List<Institution> institutionList = new ArrayList<Institution>();
    institutionList.add(institution);
    return createRemoteOrgUnit(description, institutionList);
  }

  protected OrganizationalUnit createRemoteOrgUnit(String description, List<Institution> listInstitutions) {
    // Create a new OrganizationalUnit
    OrganizationalUnitDTO organizationalUnit = new OrganizationalUnitDTO();
    organizationalUnit.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel(description));
    this.setRandomUrlOrganizationalUnit(organizationalUnit);
    organizationalUnit.setDescription(description);

    organizationalUnit.setInstitutions(new ArrayList<>());
    listInstitutions.forEach(institution -> {
      organizationalUnit.addInstitution(institution);
    });

    organizationalUnit.setKeywords(Arrays.asList("magnetostatics", "photonics", "oceanography"));
    OrganizationalUnit organizationalUnitCreated = this.organizationalUnitService.create(organizationalUnit);

    return organizationalUnitCreated;
  }

  protected DisseminationPolicy createRemoteDisseminationPolicy() {
    final DisseminationPolicy policy = new DisseminationPolicy();
    policy.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("dissemination policy"));
    policy.setType(DisseminationPolicy.DisseminationPolicyType.WEB);
    policy.setDestinationServer("http://server.com");
    return this.disseminationPolicyService.create(policy);
  }

  protected PreservationPolicy createRemotePreservationPolicy() {
    final PreservationPolicy policy = new PreservationPolicy();
    policy.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Keep It For xx Years"));
    policy.setRetention(10 * DLCMConstants.DAYS_BY_YEAR);
    return this.preservationPolicyService.create(policy);
  }

  protected SubmissionPolicy createRemoteSubmissionPolicy() {
    final SubmissionPolicy policy = new SubmissionPolicy();
    policy.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("submission policy "));
    policy.setSubmissionApproval(true);
    return this.submissionPolicyService.create(policy);
  }

  void downloadLogoShouldSuccessTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit1 = this.getOrCreateOrganizationalUnitForRole(DLCMTestConstants.PersistenceMode.TEMPORARY);
    OrganizationalUnit organizationalUnit2 = this.organizationalUnitService.uploadLogo(organizationalUnit1.getResId(), logoToUpload);

    final OrganizationalUnit refetchedOrganizationalUnit = this.organizationalUnitService.findOne(organizationalUnit2.getResId());
    assertEquals(refetchedOrganizationalUnit.getImage().getFileName(), logoName);
    assertEquals(refetchedOrganizationalUnit.getImage().getFileSize(), sizeLogo);
    this.restClientTool.exitSudo();

    Path path = Paths.get(System.getProperty("java.io.tmpdir"), "download.tmp");
    this.organizationalUnitService.downloadLogo(refetchedOrganizationalUnit.getResId(), path);
    assertNotNull(path.toFile());
    assertEquals(Long.valueOf(path.toFile().length()), refetchedOrganizationalUnit.getImage().getFileSize());
  }
}
