/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - FundingAgencyIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.admin;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.Test;
import org.springframework.core.io.ClassPathResource;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;

import ch.dlcm.model.settings.FundingAgency;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.test.DLCMTestConstants;
import ch.dlcm.test.DLCMTestConstants.OrganizationalUnitStatus;

class FundingAgencyIT extends AbstractAdminIT {

  @Override
  protected void setUser() {
    this.restClientTool.sudoAdmin();
  }

  @Test
  void testAddAndRemoveOrganizationalUnit() {

    final String fundingAgencyAcronym = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA");
    final String fundingAgencyName = DLCMTestConstants.getRandomNameWithTemporaryLabel("My FundingAgency with Unit");
    final String organizationalUnitName = DLCMTestConstants.getNameWithPermanentLabel("Organizational Unit for manager");

    /*
     * Create a new FundingAgency
     */
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(fundingAgencyAcronym);
    new_agency.setName(fundingAgencyName);
    new_agency = this.fundingAgencyService.create(new_agency);

    /*
     * Create a new OrganizationalUnit and check number of org units
     */
    final OrganizationalUnit new_unit = this.getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode.PERMANENT, OrganizationalUnitStatus.OPEN);
    List<OrganizationalUnit> units = this.fundingAgencyService.getOrganizationalUnits(new_agency.getResId());
    final int sizeBefore = units.size();

    /*
     * Add the OrganizationalUnit to the FundingAgency
     */
    this.fundingAgencyService.addOrganizationalUnit(new_agency.getResId(), new_unit.getResId());

    /*
     * Check it has been added
     */

    units = this.fundingAgencyService.getOrganizationalUnits(new_agency.getResId());
    assertNotNull(units);
    assertEquals(sizeBefore + 1, units.size());
    final Optional<OrganizationalUnit> optionalU = units.stream().filter(orgUnit -> orgUnit.getName().equals(new_unit.getName())).findFirst();
    assertTrue(optionalU.isPresent());
    assertEquals(organizationalUnitName, optionalU.get().getName());

    /*
     * Remove the OrganizationalUnit
     */
    this.fundingAgencyService.removeOrganizationalUnit(new_agency.getResId(), new_unit.getResId());

    /*
     * Check it has been removed
     */
    units = this.fundingAgencyService.getOrganizationalUnits(new_agency.getResId());
    assertNotNull(units);
    assertEquals(sizeBefore, units.size());

    /*
     * Add two organizational unit and link them to the FundingAgency
     */
    final int beforeAddingFundingAgencies = this.fundingAgencyService.getOrganizationalUnits(new_agency.getResId()).size();
    this.fundingAgencyService.addOrganizationalUnit(new_agency.getResId(), new_unit.getResId());

    final OrganizationalUnit organizationalUnit2 = this.getSecondPermanentOrganizationalUnit();

    this.fundingAgencyService.addOrganizationalUnit(new_agency.getResId(), organizationalUnit2.getResId());

    /*
     * Check they are all linked correctly
     */
    units = this.fundingAgencyService.getOrganizationalUnits(new_agency.getResId());
    assertNotNull(units);
    assertEquals(beforeAddingFundingAgencies + 2, units.size());

    /*
     * Delete one specifically and check it is not in the list anymore
     */
    this.fundingAgencyService.removeOrganizationalUnit(new_agency.getResId(), organizationalUnit2.getResId());
    units = this.fundingAgencyService.getOrganizationalUnits(new_agency.getResId());
    assertNotNull(units);
    assertEquals(beforeAddingFundingAgencies + 1, units.size());
    for (final OrganizationalUnit fetched_u : units) {
      assertNotEquals(organizationalUnit2.getResId(), fetched_u.getResId());
    }
  }

  @Test
  void testCreate() {

    final String agencyAcronym = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA");
    final String agencyName = DLCMTestConstants.getRandomNameWithTemporaryLabel("My FundingAgency");

    /*
     * Create a new FundingAgency
     */
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(agencyAcronym);
    new_agency.setName(agencyName);
    new_agency = this.fundingAgencyService.create(new_agency);

    /*
     * Fetch it from server and test it has been created
     */
    final FundingAgency fetched_agency = this.fundingAgencyService.findOne(new_agency.getResId());
    assertNotNull(fetched_agency.getResId());
    assertEquals(agencyName, fetched_agency.getName());
  }

  @Test
  void getTestWithRorId() {
    // Create an institution
    final String agencyAcronym = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA");
    final String agencyName = DLCMTestConstants.getRandomNameWithTemporaryLabel("My FundingAgency");

    /*
     * Create a new FundingAgency
     */
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(agencyAcronym);
    new_agency.setName(agencyName);
    new_agency.setRorId("04p405e02");
    new_agency = this.fundingAgencyService.create(new_agency);

    /*
     * Fetch it from server and test it has been created
     */
    final FundingAgency fetched_agency = this.fundingAgencyService.findOne(new_agency.getResId());
    assertNotNull(fetched_agency.getResId());
    assertEquals(new_agency.getAcronym(), fetched_agency.getAcronym());
    assertEquals(new_agency.getName(), fetched_agency.getName());
    assertEquals(new_agency.getRorId(), fetched_agency.getRorId());

    // Get License by ROR ID
    final FundingAgency ror_agency = this.fundingAgencyService.findByRorId(new_agency.getRorId());
    // Test the get
    assertEquals(new_agency.getAcronym(), ror_agency.getAcronym());
    assertEquals(new_agency.getName(), ror_agency.getName());
    assertEquals(new_agency.getRorId(), ror_agency.getRorId());
  }

  @Test
  void testDelete() {

    final String agencyAcronym = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA");
    final String agencyName = DLCMTestConstants.getRandomNameWithTemporaryLabel("My FundingAgency");

    /*
     * Create a new FundingAgency
     */
    FundingAgency newAgency = new FundingAgency();
    newAgency.setAcronym(agencyAcronym);
    newAgency.setName(agencyName);
    newAgency = this.fundingAgencyService.create(newAgency);

    /*
     * Check it has been created
     */
    final FundingAgency fetched_agency = this.fundingAgencyService.findOne(newAgency.getResId());
    assertNotNull(fetched_agency.getResId());

    /*
     * Delete it
     */
    this.fundingAgencyService.delete(fetched_agency.getResId());

    /*
     * Fetching again returns nothing
     */
    final String newAgencyId = newAgency.getResId();
    assertThrows(HttpClientErrorException.NotFound.class, () -> this.fundingAgencyService.findOne(newAgencyId));
  }

  @Test
  void testUnicity() {

    final String agencyAcronym = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA");
    final String agencyName = DLCMTestConstants.getRandomNameWithTemporaryLabel("My FundingAgency");

    // Creating a new agency
    FundingAgency fundingAgency1 = new FundingAgency();
    fundingAgency1.setAcronym(agencyAcronym);
    fundingAgency1.setName(agencyName);
    fundingAgency1 = this.fundingAgencyService.create(fundingAgency1);

    // Creating a second agency with the same name and acronym as the first one
    FundingAgency fundingAgency2 = new FundingAgency();
    fundingAgency2.setAcronym(fundingAgency1.getAcronym());
    fundingAgency2.setName(fundingAgency1.getName());

    // Testing unicity of the name and acronym
    try {
      this.fundingAgencyService.create(fundingAgency2);
      fail("A BAD_REQUEST Http response should be received");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode(), "Unicity constraint not satisfied");
    }
  }

  @Test
  void testUpdate() {

    final String agencyAcronym1 = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA1");
    final String agencyName1 = DLCMTestConstants.getRandomNameWithTemporaryLabel("Funding Agency unit");
    final String agencyAcronym2 = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA2");
    final String agencyName2 = DLCMTestConstants.getRandomNameWithTemporaryLabel("Funding Agency updated");

    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(agencyAcronym1);
    new_agency.setName(agencyName1);
    new_agency = this.fundingAgencyService.create(new_agency);

    /*
     * Checks saving succeeded
     */
    final FundingAgency fetched_agency = this.fundingAgencyService.findOne(new_agency.getResId());
    assertEquals(fetched_agency.getName(), agencyName1, "names are equal");

    /*
     * Does the update
     */
    fetched_agency.setAcronym(agencyAcronym2);
    fetched_agency.setName(agencyName2);
    this.fundingAgencyService.update(fetched_agency.getResId(), fetched_agency);

    /*
     * Checks update succeeded
     */
    final FundingAgency refetched_agency = this.fundingAgencyService.findOne(fetched_agency.getResId());
    assertEquals(refetched_agency.getName(), agencyName2, "names are equal");
    assertEquals(refetched_agency.getAcronym(), agencyAcronym2, "acronyms are equal");
  }

  @Test
  void uploadLogoTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    /*
     * Create an fundingAgency
     */
    final FundingAgency fundingAgency1 = this.createRemoteFundingAgency("DLCM Funding Agency #LOGO");

    /*
     * Add logo
     */
    FundingAgency fetchedFundingAgency = this.fundingAgencyService.uploadLogo(fundingAgency1.getResId(), logoToUpload);

    /*
     * Check logo has been uploaded
     */
    final FundingAgency refetchedFundingAgency = this.fundingAgencyService.findOne(fetchedFundingAgency.getResId());
    assertEquals(refetchedFundingAgency.getImage().getFileName(), logoName);
    assertEquals(refetchedFundingAgency.getImage().getFileSize(), sizeLogo);
  }

  @Test
  void downloadLogoTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    /*
     * Create an fundingAgency
     */
    final FundingAgency fundingAgency1 = this.createRemoteFundingAgency("DLCM Funding Agency #LOGO");

    /*
     * Add logo
     */
    FundingAgency fundingAgency2 = this.fundingAgencyService.uploadLogo(fundingAgency1.getResId(), logoToUpload);

    /*
     * Check logo has been uploaded
     */
    final FundingAgency refetchedFundingAgency = this.fundingAgencyService.findOne(fundingAgency2.getResId());
    assertEquals(refetchedFundingAgency.getImage().getFileName(), logoName);
    assertEquals(refetchedFundingAgency.getImage().getFileSize(), sizeLogo);

    Path path = Paths.get(System.getProperty("java.io.tmpdir"), "download.tmp");
    this.fundingAgencyService.downloadLogo(refetchedFundingAgency.getResId(), path);
    assertNotNull(path.toFile());
    assertEquals(Long.valueOf(path.toFile().length()), refetchedFundingAgency.getImage().getFileSize());
  }

  @Test
  void deleteLogoTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    /*
     * Create an fundingAgency
     */
    final FundingAgency fundingAgency1 = this.createRemoteFundingAgency("DLCM Funding Agency #LOGO");

    /*
     * Add logo
     */
    FundingAgency fetchedFundingAgency = this.fundingAgencyService.uploadLogo(fundingAgency1.getResId(), logoToUpload);

    /*
     * Check logo has been uploaded
     */
    FundingAgency refetchedFundingAgency = this.fundingAgencyService.findOne(fetchedFundingAgency.getResId());
    assertEquals(refetchedFundingAgency.getImage().getFileName(), logoName);
    assertEquals(refetchedFundingAgency.getImage().getFileSize(), sizeLogo);

    this.fundingAgencyService.deleteLogo(fundingAgency1.getResId());
    refetchedFundingAgency = this.fundingAgencyService.findOne(fetchedFundingAgency.getResId());
    assertNull(refetchedFundingAgency.getImage());
  }

  /*************************************************/

  @Override
  protected void deleteFixtures() {
    this.clearFundingAgencyFixtures();
  }

  private FundingAgency createRemoteFundingAgency(String description) {
    // Create a new FundingAgency
    FundingAgency fundingAgency = new FundingAgency();
    fundingAgency.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel(description));
    fundingAgency.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel(description));
    fundingAgency = this.fundingAgencyService.create(fundingAgency);
    return fundingAgency;
  }
}
