/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - NotificationAsAdminIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.admin;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.Test;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.notification.Notification;
import ch.dlcm.model.notification.NotificationStatus;
import ch.dlcm.model.notification.NotificationType;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.test.DLCMTestConstants;
import ch.dlcm.test.DLCMTestConstants.OrganizationalUnitStatus;

class NotificationAsAdminIT extends AbstractNotificationIT {

  @Override
  protected void setUser() {
    this.restClientTool.sudoAdmin();
  }

  @Test
  void changeStatusTest() {
    final OrganizationalUnit notifiedOrgUnit = this.createOrgUnitForCurrentUser(Role.MANAGER.getResId());
    final Notification localNotification = this.createLocalNotification(NotificationType.JOIN_ORGUNIT_REQUEST, notifiedOrgUnit);
    Notification remoteNotification = this.createRemoteNotification(localNotification);
    assertEquals(NotificationStatus.PENDING, remoteNotification.getNotificationStatus(), "Initial notification status should be PENDING");
    final String notificationId = remoteNotification.getResId();

    this.notificationService.setProcessedStatus(notificationId);
    remoteNotification = this.notificationService.findOne(notificationId);
    assertEquals(NotificationStatus.PROCESSED, remoteNotification.getNotificationStatus(), "Notification status should be processed");

    this.notificationService.setRefusedStatus(notificationId);
    remoteNotification = this.notificationService.findOne(notificationId);
    assertEquals(NotificationStatus.REFUSED, remoteNotification.getNotificationStatus(), "Notification status should be refused");
  }

  @Test
  void findAllTest() {
    final OrganizationalUnit notifiedOrgUnit = this.createOrgUnitForCurrentUser(Role.APPROVER.getResId());
    final Notification localNotification = this.createLocalNotification(NotificationType.JOIN_ORGUNIT_REQUEST, notifiedOrgUnit);
    final Notification remoteNotification = this.createRemoteNotification(localNotification);
    List<Notification> notificationsList = this.notificationService.findAll();
    assertFalse(notificationsList.isEmpty());
    assertTrue(notificationsList.stream().anyMatch(notification -> notification.getResId().equals(remoteNotification.getResId())));
  }

  @Test
  void filterListTest() {
    String message = "filter me !";
    final OrganizationalUnit notifiedOrgUnit = this.createOrgUnitForCurrentUser(Role.APPROVER.getResId());
    Notification localNotification = this.createLocalNotification(NotificationType.JOIN_ORGUNIT_REQUEST, notifiedOrgUnit, message);
    this.createRemoteNotification(localNotification);
    localNotification = this.createLocalNotification(NotificationType.JOIN_ORGUNIT_REQUEST, notifiedOrgUnit);
    this.createRemoteNotification(localNotification);

    List<Notification> notificationsList = this.notificationService.searchByProperties(Map.of("message", message));
    assertEquals(1, notificationsList.size());
  }

  @Test
  void creationTest() {
    final OrganizationalUnit notifiedOrgUnit = this.createOrgUnitForCurrentUser(Role.APPROVER.getResId());
    final Notification localNotification = this.createLocalNotification(NotificationType.JOIN_ORGUNIT_REQUEST, notifiedOrgUnit);
    final Notification remoteNotification = this.createRemoteNotification(localNotification);
    assertEquals(localNotification.getEmitter().getFirstName(), remoteNotification.getEmitter().getFirstName());
    assertEquals(localNotification.getEmitter().getLastName(), remoteNotification.getEmitter().getLastName());
    assertEquals(localNotification.getNotifiedOrgUnit().getName(), remoteNotification.getNotifiedOrgUnit().getName());
    assertEquals(localNotification.getMessage(), remoteNotification.getMessage());
  }

  @Test
  void creationNotificationWithoutOrgUnitTest() {
    try {
      this.createRemoteNotificationWithoutOrgUnit(NotificationType.JOIN_ORGUNIT_REQUEST, null);
      fail("Should raise exception 400 bad request");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode());
    }

    // creating a notification without org unit doesn't fail for type CREATE_ORGUNIT_REQUEST (but needs a name in objectId)
    this.createRemoteNotificationWithoutOrgUnit(NotificationType.CREATE_ORGUNIT_REQUEST,
            DLCMConstants.TEMPORARY_TEST_DATA_LABEL + " New org unit name");
  }

  @Test
  void deletionTest() {
    final OrganizationalUnit notifiedOrgUnit = this.createOrgUnitForCurrentUser(Role.APPROVER.getResId());
    final Notification localNotification = this.createLocalNotification(NotificationType.JOIN_ORGUNIT_REQUEST, notifiedOrgUnit);
    final Notification remoteNotification = this.createRemoteNotification(localNotification);

    List<Notification> notificationsList = this.notificationService.findAll();
    assertTrue(notificationsList.stream().anyMatch(notification -> notification.getResId().equals(remoteNotification.getResId())));

    this.notificationService.delete(remoteNotification.getResId());

    notificationsList = this.notificationService.findAll();
    assertFalse(notificationsList.stream().anyMatch(notification -> notification.getResId().equals(remoteNotification.getResId())));

    assertThrows(HttpClientErrorException.NotFound.class, () -> notificationService.findOne(remoteNotification.getResId()));
  }

  @Test
  void inboxTest() {
    OrganizationalUnit managerOrgUnit = this.createOrgUnitForCurrentUser(Role.MANAGER_ID);
    OrganizationalUnit stewardOrgUnit = this.createOrgUnitForCurrentUser(Role.STEWARD_ID);
    OrganizationalUnit approverOrgUnit = this.createOrgUnitForCurrentUser(Role.APPROVER_ID);
    OrganizationalUnit creatorOrgUnit = this.createOrgUnitForCurrentUser(Role.CREATOR_ID);
    OrganizationalUnit orgUnitForNoOne = this.getOrgUnitForNoOne(DLCMTestConstants.PersistenceMode.PERMANENT);

    this.inboxTest(NotificationType.JOIN_ORGUNIT_REQUEST, managerOrgUnit, true);
    this.inboxTest(NotificationType.JOIN_ORGUNIT_REQUEST, stewardOrgUnit, true);

    this.inboxTest(NotificationType.ACCESS_DATASET_REQUEST, stewardOrgUnit, true);
    this.inboxTest(NotificationType.ACCESS_DATASET_REQUEST, approverOrgUnit, true);

    this.inboxTest(NotificationType.VALIDATE_DEPOSIT_REQUEST, approverOrgUnit, true);
    this.inboxTest(NotificationType.VALIDATE_DEPOSIT_REQUEST, creatorOrgUnit, true);

    // This message is seen because the current user is a ApplicationRole ADMIN
    this.inboxTest(NotificationType.CREATE_ORGUNIT_REQUEST, orgUnitForNoOne, DLCMConstants.TEMPORARY_TEST_DATA_LABEL + " New org unit name",
            true);

  }

  @Test
  void sentBoxTest() {
    final OrganizationalUnit notifiedOrgUnit = this.createOrgUnitForCurrentUser(Role.APPROVER.getResId());
    final Notification localNotification = this.createLocalNotification(NotificationType.JOIN_ORGUNIT_REQUEST, notifiedOrgUnit);
    final Notification remoteNotification = this.createRemoteNotification(localNotification);
    final Notification sentNotification = this.notificationService.getSentNotification(remoteNotification.getResId());
    assertNotNull(sentNotification, "Sent notification should be in sent box");
  }

  @Test
  void refuseInformationTest() {
    OrganizationalUnit notifiedOrgUnit = this.getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode.PERMANENT, OrganizationalUnitStatus.OPEN, "MANAGER");

    final Notification localNotification = this.createLocalNotification(NotificationType.CREATED_DEPOSIT_INFO, notifiedOrgUnit);
    Notification remoteNotification = this.createRemoteNotification(localNotification);
    try {
      this.notificationService.setRefusedStatus(remoteNotification.getResId());
      fail("A FORBIDDEN Http response should be received");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode());
      assertTrue(e.getMessage().contains("cannot be refused"));
    }
  }

}
