/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - OrganizationalUnitAsManagerIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.admin;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.net.URL;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ThreadLocalRandom;

import org.junit.jupiter.api.Test;
import org.springframework.core.io.ClassPathResource;
import org.springframework.web.client.HttpClientErrorException;

import ch.unige.solidify.util.StringTool;

import ch.dlcm.model.display.OrganizationalUnitDisseminationPolicyDisplay;
import ch.dlcm.model.display.OrganizationalUnitPreservationPolicyDisplay;
import ch.dlcm.model.display.OrganizationalUnitSubmissionPolicyDisplay;
import ch.dlcm.model.policies.DisseminationPolicy;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.FundingAgency;
import ch.dlcm.model.settings.Institution;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.test.DLCMTestConstants;

public class OrganizationalUnitAsManagerIT extends AbstractOrganizationalUnitWithRoleIT {
  @Override
  protected String role() {
    return Role.MANAGER_ID;
  }

  @Test
  public void updateOrgUnit() {
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Update org unit as " + this.role());
    this.restClientTool.exitSudo();

    String newName = organizationalUnit.getName() + " successfully updated";
    String newDescription = "Description successfully updated";
    this.setRandomUrlOrganizationalUnit(organizationalUnit);
    URL newUrl = organizationalUnit.getUrl();
    LocalDate newOpeningDate = LocalDate.now().minusMonths(1);
    LocalDate newClosingDate = LocalDate.now().plusMonths(1);
    List<String> newKeywords = new ArrayList<>();
    newKeywords.add("UPDATE");

    organizationalUnit.setName(newName);
    organizationalUnit.setDescription(newDescription);
    organizationalUnit.setOpeningDate(newOpeningDate);
    organizationalUnit.setClosingDate(newClosingDate);
    organizationalUnit.setKeywords(newKeywords);

    final OrganizationalUnit organizationalUnitUpdated = this.organizationalUnitService.update(organizationalUnit.getResId(),
            organizationalUnit);

    assertEquals(newName, organizationalUnitUpdated.getName());
    assertEquals(newDescription, organizationalUnitUpdated.getDescription());
    assertEquals(newUrl, organizationalUnitUpdated.getUrl());
    assertEquals(newOpeningDate, organizationalUnitUpdated.getOpeningDate());
    assertEquals(newClosingDate, organizationalUnitUpdated.getClosingDate());
    assertEquals(newKeywords.size(), organizationalUnitUpdated.getKeywords().size());
    assertEquals(newKeywords.get(0), organizationalUnitUpdated.getKeywords().get(0));
  }

  @Test
  void addAndRemoveDisseminationPolicyTest() {
    // Get test orgUnit
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Swiss Research Data Preservation #DISSEMINATION POLICY");
    this.restClientTool.exitSudo();

    List<OrganizationalUnitDisseminationPolicyDisplay> policies = this.organizationalUnitService
            .getDisseminationPolicies(organizationalUnit.getResId());
    final int initialState = policies.size();

    // Create policy
    this.restClientTool.sudoAdmin();
    final DisseminationPolicy policy = this.createRemoteDisseminationPolicy();
    this.restClientTool.exitSudo();

    // Add policy to orgUnit
    this.organizationalUnitService.addDisseminationPolicy(organizationalUnit.getResId(), policy.getResId());
    policies = this.organizationalUnitService.getDisseminationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState + 1);

    // Test add
    for (final OrganizationalUnitDisseminationPolicyDisplay policy2 : policies) {
      if (policy2.getResId().equals(policy.getResId())) {
        assertNotNull(policy2);
        assertEquals(policy.getName(), policy2.getName());
        assertEquals(policy.getType(), policy2.getType());
        assertEquals(policy.getDestinationServer(), policy2.getDestinationServer());
      }
    }

    // Remove policy
    this.organizationalUnitService.removeDisseminationPolicy(organizationalUnit.getResId(), policy.getResId());

    // Test remove
    policies = this.organizationalUnitService.getDisseminationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState);
  }

  @Test
  void addAndRemovePreservationPolicyTest() {
    // Get test orgUnit
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Swiss Research Data Preservation #PRESERVATION POLICY");
    this.restClientTool.exitSudo();

    List<OrganizationalUnitPreservationPolicyDisplay> policies = this.organizationalUnitService
            .getPreservationPolicies(organizationalUnit.getResId());
    final int initialState = policies.size();

    // Create policy
    this.restClientTool.sudoAdmin();
    final PreservationPolicy policy = this.createRemotePreservationPolicy();
    this.restClientTool.exitSudo();

    // Add policy to orgUnit
    this.organizationalUnitService.addPreservationPolicy(organizationalUnit.getResId(), policy.getResId());
    policies = this.organizationalUnitService.getPreservationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState + 1);

    // Test add
    for (final OrganizationalUnitPreservationPolicyDisplay policy2 : policies) {
      if (policy2.getResId().equals(policy.getResId())) {
        assertNotNull(policy2);
        assertEquals(policy.getName(), policy2.getName());
        assertEquals(policy.getRetention(), policy2.getRetention());
        assertEquals(policy.getDispositionApproval(), policy2.getDispositionApproval());
      }
    }

    // Remove policy
    this.organizationalUnitService.removePreservationPolicy(organizationalUnit.getResId(), policy.getResId());

    // Test remove
    policies = this.organizationalUnitService.getPreservationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState);
  }

  @Test
  void addAndRemoveSubmissionPolicyTest() {
    // Get test orgUnit
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Swiss Research Data Preservation #SUBMISSION POLICY");
    this.restClientTool.exitSudo();

    List<OrganizationalUnitSubmissionPolicyDisplay> policies = this.organizationalUnitService
            .getSubmissionPolicies(organizationalUnit.getResId());
    final int initialState = policies.size();

    // Create policy
    this.restClientTool.sudoAdmin();
    final SubmissionPolicy policy = this.createRemoteSubmissionPolicy();
    this.restClientTool.exitSudo();

    // Add policy to orgUnit
    this.organizationalUnitService.addSubmissionPolicy(organizationalUnit.getResId(), policy.getResId());
    policies = this.organizationalUnitService.getSubmissionPolicies(organizationalUnit.getResId());

    assertEquals(policies.size(), initialState + 1);

    // Test added person on the OrganizationalUnit
    for (final OrganizationalUnitSubmissionPolicyDisplay policy2 : policies) {
      if (policy2.getResId().equals(policy.getResId())) {
        assertNotNull(policy2);
        assertEquals(policy.getName(), policy2.getName());
        assertEquals(policy.getSubmissionApproval(), policy2.getSubmissionApproval());
      }
    }

    // Remove policy
    this.organizationalUnitService.removeSubmissionPolicy(organizationalUnit.getResId(), policy.getResId());

    // Test remove
    policies = this.organizationalUnitService.getSubmissionPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState);
  }

  @Test
  public void addAndRemoveInstitutionsTest() {
    /*
     * Get test orgUnit
     */
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.getOrCreateOrganizationalUnitForRole(DLCMTestConstants.PersistenceMode.TEMPORARY);
    this.restClientTool.exitSudo();

    List<Institution> list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    final int initialState = list.size();

    /*
     * Create a new Institution
     */
    this.restClientTool.sudoAdmin();
    Institution institution = new Institution();
    institution.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution"));
    institution.setDescription("description" + ThreadLocalRandom.current().nextInt());
    institution = this.institutionService.create(institution);
    this.restClientTool.exitSudo();

    /*
     * Add the Institution to the OrganizationalUnit
     */
    this.organizationalUnitService.addInstitution(organizationalUnit.getResId(), institution.getResId());
    list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    assertEquals(initialState + 1, list.size());
    final String institutionName = institution.getName();
    final Optional<Institution> optionalInstitution = list.stream().filter(ins -> ins.getName().equals(institutionName))
            .findFirst();
    assertTrue(optionalInstitution.isPresent());

    /*
     * Check it has been added
     */
    assertEquals(institution.getResId(), optionalInstitution.get().getResId());
    assertEquals(institution.getName(), optionalInstitution.get().getName());
    assertEquals(institution.getDescription(), optionalInstitution.get().getDescription());

    /*
     * Remove the Institution
     */
    this.organizationalUnitService.removeInstitution(organizationalUnit.getResId(), institution.getResId());
    /*
     * Check it has been removed
     */
    list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    assertEquals(initialState, list.size());

    /*
     * Create 3 Institutions and link them to the OrganizationalUnit
     */
    Institution institution1 = new Institution();
    institution1.setDescription(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 1"));
    institution1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 1"));

    Institution institution2 = new Institution();
    institution2.setDescription(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 2"));
    institution2.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 2"));

    Institution institution3 = new Institution();
    institution3.setDescription(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 3"));
    institution3.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Institution Sub 3"));

    final Institution[] institutions = new Institution[] { institution1, institution2, institution3 };

    String institutionsToDelete = null;

    int i = 0;
    for (Institution ins : institutions) {
      this.restClientTool.sudoAdmin();
      ins = this.institutionService.create(ins);
      this.restClientTool.exitSudo();
      this.organizationalUnitService.addInstitution(organizationalUnit.getResId(), ins.getResId());

      if (i == 1) {
        institutionsToDelete = ins.getResId();
      }

      i++;
    }

    /*
     * Check they are all linked correctly
     */
    list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    assertNotNull(list);
    assertEquals(initialState + 3, list.size());

    /*
     * Delete one specifically and check it is not in the list anymore
     */
    this.organizationalUnitService.removeInstitution(organizationalUnit.getResId(), institutionsToDelete);
    list = this.organizationalUnitService.getInstitutions(organizationalUnit.getResId());
    assertNotNull(list);
    assertEquals(initialState + 2, list.size());
    for (final Institution fetched_institution : list) {
      assertNotEquals(institutionsToDelete, fetched_institution.getResId());
    }
  }

  @Test
  void addAndRemoveFundingAgency() {
    /*
     * Get test orgUnit
     */
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Add and remove funding agency as " + this.role());
    this.restClientTool.exitSudo();

    List<FundingAgency> fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    final int initialState = fundingAgencies.size();

    /*
     * Create a new FundingAgency
     */
    this.restClientTool.sudoAdmin();
    final String fundingAgencyAcronym = DLCMTestConstants.getRandomNameWithTemporaryLabel("FA");
    final String fundingAgencyName = DLCMTestConstants.getRandomNameWithTemporaryLabel("My FundingAgency");
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(fundingAgencyAcronym);
    new_agency.setName(fundingAgencyName);
    new_agency = this.fundingAgencyService.create(new_agency);
    this.restClientTool.exitSudo();

    /*
     * Add the FundingAgency to the OrganizationalUnit
     */
    this.organizationalUnitService.addFundingAgency(organizationalUnit.getResId(), new_agency.getResId());

    /*
     * Check it has been added
     */
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState + 1, fundingAgencies.size());
    final Optional<FundingAgency> optionalFA = fundingAgencies.stream().filter(fa -> fa.getName().equals(fundingAgencyName)).findFirst();
    assertTrue(optionalFA.isPresent());
    assertEquals(fundingAgencyName, optionalFA.get().getName());

    /*
     * Remove the FundingAgency
     */
    this.organizationalUnitService.removeFundingAgency(organizationalUnit.getResId(), new_agency.getResId());

    /*
     * Check it has been removed
     */
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState, fundingAgencies.size());

    /*
     * Create 3 FundingAgencies and link them to the OrganizationalUnit
     */
    final FundingAgency subFA1 = new FundingAgency();
    subFA1.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FA 1"));
    subFA1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FundingAgency 1"));

    final FundingAgency subFA2 = new FundingAgency();
    subFA2.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FA 2"));
    subFA2.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FundingAgency 2"));

    final FundingAgency subFA3 = new FundingAgency();
    subFA3.setAcronym(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FA 3"));
    subFA3.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("Sub FundingAgency 3"));

    final FundingAgency[] subAgencies = new FundingAgency[] { subFA1, subFA2, subFA3 };

    String agencyIdToDelete = null;

    int i = 0;
    for (FundingAgency subAgency : subAgencies) {
      this.restClientTool.sudoAdmin();
      subAgency = this.fundingAgencyService.create(subAgency);
      this.restClientTool.exitSudo();
      this.organizationalUnitService.addFundingAgency(organizationalUnit.getResId(), subAgency.getResId());

      if (i == 1) {
        agencyIdToDelete = subAgency.getResId();
      }

      i++;
    }

    /*
     * Check they are all linked correctly
     */
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState + 3, fundingAgencies.size());

    /*
     * Delete one specifically and check it is not in the list anymore
     */
    this.organizationalUnitService.removeFundingAgency(organizationalUnit.getResId(), agencyIdToDelete);
    fundingAgencies = this.organizationalUnitService.getFundingAgencies(organizationalUnit.getResId());
    assertNotNull(fundingAgencies);
    assertEquals(initialState + 2, fundingAgencies.size());
    for (final FundingAgency fetched_agency : fundingAgencies) {
      assertNotEquals(agencyIdToDelete, fetched_agency.getResId());
    }

  }

  @Test
  void addPersonToTheOrganizationalUnitTest() {
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Swiss Research Data Preservation #PERSON");
    final Person person = this.createRemotePerson();
    this.restClientTool.exitSudo();

    final int peopleBeforeAdding = this.organizationalUnitService.getPeople(organizationalUnit.getResId()).size();

    // Add the person to the OrganizationalUnit with default role
    this.organizationalUnitService.addPerson(organizationalUnit.getResId(), person.getResId());
    final List<Person> people = this.organizationalUnitService.getPeople(organizationalUnit.getResId());

    // If a person is linked to the basic auth account who created the organizational unit
    // then this person is automatically added to the organizational unit.
    // So one or two people should be present in the list
    assertTrue(people.size() == peopleBeforeAdding + 1 || people.size() == peopleBeforeAdding + 2);

    // Test added person on the OrganizationalUnit
    for (final Person person2 : people) {
      if (person2.getResId().equals(person.getResId())) {
        assertNotNull(person2);
        assertEquals(person.getFirstName(), person2.getFirstName());
        assertEquals(person.getLastName(), person2.getLastName());
        assertEquals(person.getOrcid(), person2.getOrcid());
      }
    }
  }

  @Test
  void addMultipleRolesForPersonToTheOrganizationalUnitTest() {
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Swiss Research Data Preservation #ROLES");
    this.restClientTool.exitSudo();

    final Person person1 = this.createRemotePerson();
    final Person person2 = this.createRemotePerson();
    final int peopleBeforeAdding = this.organizationalUnitService.getPeople(organizationalUnit.getResId()).size();

    // Add the persons to the OrganizationalUnit with roles
    this.organizationalUnitService.addPersonRole(organizationalUnit.getResId(), person1.getResId(), Role.STEWARD_ID);
    this.organizationalUnitService.addPersonRole(organizationalUnit.getResId(), person2.getResId(), Role.CREATOR_ID);
    final List<Person> people = this.organizationalUnitService.getPeople(organizationalUnit.getResId());
    assertEquals(peopleBeforeAdding + 2, people.size());
    // Add the same role
    this.organizationalUnitService.addPersonRole(organizationalUnit.getResId(), person1.getResId(), Role.STEWARD_ID);
    // Add the new role
    assertThrows(HttpClientErrorException.BadRequest.class,
            () -> this.organizationalUnitService.addPersonRole(organizationalUnit.getResId(), person1.getResId(), Role.CREATOR_ID));
  }

  @Test
  void closeTest() {
    this.restClientTool.sudoAdmin();
    OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(DLCMTestConstants.PersistenceMode.TEMPORARY,
            "Swiss Research Data Preservation #CLOSE");
    this.restClientTool.exitSudo();

    // Close
    final String resId = organizationalUnit.getResId();
    final DateTimeFormatter formatter = DateTimeFormatter.ofPattern(StringTool.DATE_FORMAT);
    final String closingDate = LocalDate.now().plus(1, ChronoUnit.DAYS).format(formatter);

    this.organizationalUnitService.close(resId, closingDate);
    organizationalUnit = this.organizationalUnitService.findOne(resId);

    // Test close
    assertEquals(organizationalUnit.getClosingDate().toString(), closingDate);

    this.organizationalUnitService.delete(organizationalUnit.getResId());
  }

  @Test
  void uploadLogoTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    final OrganizationalUnit organizationalUnit1 = this.getOrCreateOrganizationalUnitForRole(DLCMTestConstants.PersistenceMode.TEMPORARY);
    OrganizationalUnit fetchedOrganizationalUnit = this.organizationalUnitService.uploadLogo(organizationalUnit1.getResId(), logoToUpload);

    final OrganizationalUnit refetchedOrganizationalUnit = this.organizationalUnitService.findOne(fetchedOrganizationalUnit.getResId());
    assertEquals(refetchedOrganizationalUnit.getImage().getFileName(), logoName);
    assertEquals(refetchedOrganizationalUnit.getImage().getFileSize(), sizeLogo);
  }

  @Test
  void downloadLogoTest() throws IOException {
    super.downloadLogoShouldSuccessTest();
  }

  @Test
  void deleteLogoTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    final OrganizationalUnit organizationalUnit1 = this.getOrCreateOrganizationalUnitForRole(DLCMTestConstants.PersistenceMode.TEMPORARY);
    OrganizationalUnit fetchedOrganizationalUnit = this.organizationalUnitService.uploadLogo(organizationalUnit1.getResId(), logoToUpload);

    OrganizationalUnit refetchedOrganizationalUnit = this.organizationalUnitService.findOne(fetchedOrganizationalUnit.getResId());
    assertEquals(refetchedOrganizationalUnit.getImage().getFileName(), logoName);
    assertEquals(refetchedOrganizationalUnit.getImage().getFileSize(), sizeLogo);

    this.organizationalUnitService.deleteLogo(organizationalUnit1.getResId());
    refetchedOrganizationalUnit = this.organizationalUnitService.findOne(fetchedOrganizationalUnit.getResId());
    assertNull(refetchedOrganizationalUnit.getImage());
  }
}
