/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - OrganizationalUnitAsStewardIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.admin;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.io.IOException;
import java.util.List;

import org.junit.jupiter.api.Test;
import org.springframework.core.io.ClassPathResource;
import org.springframework.web.client.HttpClientErrorException;

import ch.dlcm.model.display.OrganizationalUnitDisseminationPolicyDisplay;
import ch.dlcm.model.display.OrganizationalUnitPreservationPolicyDisplay;
import ch.dlcm.model.display.OrganizationalUnitSubmissionPolicyDisplay;
import ch.dlcm.model.policies.DisseminationPolicy;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.test.DLCMTestConstants;

public class OrganizationalUnitAsStewardIT extends AbstractOrganizationalUnitWithRoleIT {
  @Override
  protected String role() {
    return Role.STEWARD_ID;
  }

  @Test
  public void updateOrgUnitNotAllowed() {
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Update org unit as " + this.role());
    this.restClientTool.exitSudo();

    organizationalUnit.setDescription("Try to update description");
    assertThrows(HttpClientErrorException.Forbidden.class,
            () -> this.organizationalUnitService.update(organizationalUnit.getResId(), organizationalUnit));
  }

  @Test
  void addAndRemoveDisseminationPolicyNotAllowed() {
    // Get test orgUnit
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Swiss Research Data Preservation #DISSEMINATION POLICY");
    this.restClientTool.exitSudo();

    List<OrganizationalUnitDisseminationPolicyDisplay> policies = this.organizationalUnitService
            .getDisseminationPolicies(organizationalUnit.getResId());
    final int initialState = policies.size();

    // Create policy
    this.restClientTool.sudoAdmin();
    final DisseminationPolicy policy = this.createRemoteDisseminationPolicy();
    this.restClientTool.exitSudo();

    // Add policy to orgUnit
    assertThrows(HttpClientErrorException.Forbidden.class,
            () -> this.organizationalUnitService.addDisseminationPolicy(organizationalUnit.getResId(), policy.getResId()));

    this.restClientTool.sudoAdmin();
    this.organizationalUnitService.addDisseminationPolicy(organizationalUnit.getResId(), policy.getResId());
    this.restClientTool.exitSudo();

    policies = this.organizationalUnitService.getDisseminationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState + 1);

    // Test add
    for (final OrganizationalUnitDisseminationPolicyDisplay policy2 : policies) {
      if (policy2.getResId().equals(policy.getResId())) {
        assertNotNull(policy2);
        assertEquals(policy.getName(), policy2.getName());
        assertEquals(policy.getType(), policy2.getType());
        assertEquals(policy.getDestinationServer(), policy2.getDestinationServer());
      }
    }

    // Remove policy
    assertThrows(HttpClientErrorException.Forbidden.class,
            () -> this.organizationalUnitService.removeDisseminationPolicy(organizationalUnit.getResId(), policy.getResId()));
  }

  @Test
  void addAndRemoveSubmissionPolicyNotAllowed() {
    // Get test orgUnit
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Swiss Research Data Preservation #SUBMISSION POLICY");
    this.restClientTool.exitSudo();

    List<OrganizationalUnitSubmissionPolicyDisplay> policies = this.organizationalUnitService
            .getSubmissionPolicies(organizationalUnit.getResId());
    final int initialState = policies.size();

    // Create policy
    this.restClientTool.sudoAdmin();
    final SubmissionPolicy policy = this.createRemoteSubmissionPolicy();
    this.restClientTool.exitSudo();

    // Add policy to orgUnit
    assertThrows(HttpClientErrorException.Forbidden.class,
            () -> this.organizationalUnitService.addSubmissionPolicy(organizationalUnit.getResId(), policy.getResId()));

    this.restClientTool.sudoAdmin();
    this.organizationalUnitService.addSubmissionPolicy(organizationalUnit.getResId(), policy.getResId());
    this.restClientTool.exitSudo();

    policies = this.organizationalUnitService.getSubmissionPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState + 1);

    // Test added person on the OrganizationalUnit
    for (final OrganizationalUnitSubmissionPolicyDisplay policy2 : policies) {
      if (policy2.getResId().equals(policy.getResId())) {
        assertNotNull(policy2);
        assertEquals(policy.getName(), policy2.getName());
        assertEquals(policy.getSubmissionApproval(), policy2.getSubmissionApproval());
      }
    }

    // Remove policy
    assertThrows(HttpClientErrorException.Forbidden.class,
            () -> this.organizationalUnitService.removeSubmissionPolicy(organizationalUnit.getResId(), policy.getResId()));
  }

  @Test
  void addAndRemovePreservationPolicyNotAllowed() {
    // Get test orgUnit
    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit = this.createRemoteOrgUnitWithInstitutionWithCurrentRole(
            DLCMTestConstants.PersistenceMode.TEMPORARY, "Swiss Research Data Preservation #PRESERVATION POLICY");
    this.restClientTool.exitSudo();

    List<OrganizationalUnitPreservationPolicyDisplay> policies = this.organizationalUnitService
            .getPreservationPolicies(organizationalUnit.getResId());
    final int initialState = policies.size();

    // Create policy
    this.restClientTool.sudoAdmin();
    final PreservationPolicy policy = this.createRemotePreservationPolicy();
    this.restClientTool.exitSudo();

    // Add policy to orgUnit
    assertThrows(HttpClientErrorException.Forbidden.class,
            () -> this.organizationalUnitService.addPreservationPolicy(organizationalUnit.getResId(), policy.getResId()));

    this.restClientTool.sudoAdmin();
    this.organizationalUnitService.addPreservationPolicy(organizationalUnit.getResId(), policy.getResId());
    this.restClientTool.exitSudo();

    policies = this.organizationalUnitService.getPreservationPolicies(organizationalUnit.getResId());
    assertEquals(policies.size(), initialState + 1);

    // Test add
    for (final OrganizationalUnitPreservationPolicyDisplay policy2 : policies) {
      if (policy2.getResId().equals(policy.getResId())) {
        assertNotNull(policy2);
        assertEquals(policy.getName(), policy2.getName());
        assertEquals(policy.getRetention(), policy2.getRetention());
        assertEquals(policy.getDispositionApproval(), policy2.getDispositionApproval());
      }
    }

    // Remove policy
    assertThrows(HttpClientErrorException.Forbidden.class,
            () -> this.organizationalUnitService.removePreservationPolicy(organizationalUnit.getResId(), policy.getResId()));
  }

  @Test
  void uploadLogoTest() {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final OrganizationalUnit organizationalUnit1 = this.getOrCreateOrganizationalUnitForRole(DLCMTestConstants.PersistenceMode.TEMPORARY);
    assertThrows(HttpClientErrorException.Forbidden.class,
            () -> this.organizationalUnitService.uploadLogo(organizationalUnit1.getResId(), logoToUpload));
  }

  @Test
  void downloadLogoTest() throws IOException {
    this.downloadLogoShouldSuccessTest();
  }

  @Test
  void deleteLogoTest() throws IOException {
    final String logoName = "geneve.jpg";
    final ClassPathResource logoToUpload = new ClassPathResource(logoName);
    final Long sizeLogo = logoToUpload.contentLength();

    this.restClientTool.sudoAdmin();
    final OrganizationalUnit organizationalUnit1 = this.getOrCreateOrganizationalUnitForRole(DLCMTestConstants.PersistenceMode.TEMPORARY);
    OrganizationalUnit fetchedOrganizationalUnit = this.organizationalUnitService.uploadLogo(organizationalUnit1.getResId(), logoToUpload);

    OrganizationalUnit refetchedOrganizationalUnit = this.organizationalUnitService.findOne(fetchedOrganizationalUnit.getResId());
    assertEquals(refetchedOrganizationalUnit.getImage().getFileName(), logoName);
    assertEquals(refetchedOrganizationalUnit.getImage().getFileSize(), sizeLogo);
    this.restClientTool.exitSudo();

    assertThrows(HttpClientErrorException.Forbidden.class, () -> this.organizationalUnitService.deleteLogo(organizationalUnit1.getResId()));
  }
}
