/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - ResearchDomainIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.admin;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.fail;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.Test;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;

import ch.dlcm.model.settings.ResearchDomain;
import ch.dlcm.test.DLCMTestConstants;

class ResearchDomainIT extends AbstractAdminIT {

  @Override
  protected void setUser() {
    this.restClientTool.sudoAdmin();
  }

  @Test
  void creationTest() {
    // Create a research domain
    final ResearchDomain rd1 = this.createResearchDomain("test", "Z0000", "research domain ");
    final ResearchDomain rd2 = this.researchDomainService.create(rd1);

    // Test the creation
    assertEquals(rd1.getName(), rd2.getName());
    assertEquals(rd1.getCode(), rd2.getCode());
  }

  @Test
  void deleteTest() {
    // Create a new research domain
    ResearchDomain rd1 = this.createResearchDomain("test", "Z0900", "research domain ");
    rd1 = this.researchDomainService.create(rd1);

    // Delete
    final String resId = rd1.getResId();
    this.researchDomainService.delete(resId);
    assertThrows(HttpClientErrorException.NotFound.class, () -> this.researchDomainService.findOne(resId));
  }

  @Test
  void searchTest() {
    // Create a research domain
    for (int i = 0; i < 10; i++) {
      this.researchDomainService.create(this.createResearchDomain("test", "Z030" + i, "research domain "));
      if (i % 2 == 0) {
        this.researchDomainService.create(this.createResearchDomain("test", "Z040" + i, "research domain "));
      }
    }

    List<ResearchDomain> list1 = this.researchDomainService.searchByProperties(Map.of("code", "Z03"));
    assertEquals(10, list1.size(), "Category search (long list)");

    list1 = this.researchDomainService.searchByProperties(Map.of("code", "Z04"));
    assertEquals(5, list1.size(), "Category search (short list)");

    list1 = this.researchDomainService.searchByProperties(Map.of("code", "Z0301"));
    assertEquals(1, list1.size(), "Research domain search");
  }

  @Test
  void unicityTest() {
    // Create a research domain
    ResearchDomain rd1 = this.createResearchDomain("test", "Z0200", "research domain ");
    rd1 = this.researchDomainService.create(rd1);

    // Create a research domain
    ResearchDomain rd2 = this.createResearchDomain("test", "Z0200", "research domain ");

    // Testing the unicity
    try {
      this.researchDomainService.create(rd2);
      fail("A BAD_REQUEST Http response should be received");
    } catch (final HttpClientErrorException e) {
      assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode(), "Unicity constraint not satisfied");
    }
  }

  @Test
  void updateTest() {
    // Create a research domain
    ResearchDomain rd1 = this.createResearchDomain("test", "Z0100", "A ");
    rd1 = this.researchDomainService.create(rd1);

    // Update the research domain
    rd1.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel("B"));
    this.researchDomainService.update(rd1.getResId(), rd1);
    final ResearchDomain rd2 = this.researchDomainService.findOne(rd1.getResId());

    // Test the update
    assertEquals(rd1.getName(), rd2.getName());
    assertEquals(rd1.getCode(), rd2.getCode());
    this.assertEqualsWithoutNanoSeconds(rd1.getCreationTime(), rd2.getCreationTime());
  }

  @Override
  protected void deleteFixtures() {
    this.clearResearchDomainFixtures();
  }

  private ResearchDomain createResearchDomain(String source, String code, String name) {
    final ResearchDomain rd = new ResearchDomain();
    rd.setSource(source);
    rd.setCode(code);
    rd.setName(DLCMTestConstants.getRandomNameWithTemporaryLabel(name));
    return rd;
  }
}
