/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - ContributorIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.preingest;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.util.SolidifyTime;

import ch.dlcm.model.preingest.Contributor;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.test.DLCMTestConstants;
import ch.dlcm.test.DLCMTestConstants.OrganizationalUnitStatus;

class ContributorIT extends AbstractPreingestIT {

  private static final Logger log = LoggerFactory.getLogger(ContributorIT.class);

  private final String role = Role.CREATOR_ID;

  @Override
  protected void setUser() {
    this.restClientTool.sudoAdmin();
  }

  @Test
  void filterContributorsTest() {
    final Person person = this.personService.findAll().get(0);

    // Filter on first name
    int endIndexFirstName = Math.min(person.getFirstName().length(), 4);
    final Map<String, String> properties = Map.of("firstName", person.getFirstName().substring(1, endIndexFirstName));
    List<Person> personList = this.personService.searchByProperties(properties);
    List<Contributor> contributorList = this.contributorService.searchContributors(properties);
    assertEquals(contributorList.size(), personList.size());

    // Filter on last name
    int endIndexLastName = Math.min(person.getLastName().length(), 4);
    Map<String, String> newProperties = Map.of("lastName", person.getLastName().substring(1, endIndexLastName));
    personList = this.personService.searchByProperties(newProperties);
    contributorList = this.contributorService.searchContributors(newProperties);
    assertEquals(contributorList.size(), personList.size());
    for (final Contributor contributor : contributorList) {
      final Person person2 = this.personService.findOne(contributor.getResId());
      assertEquals(contributor.getFirstName(), person2.getFirstName());
      assertEquals(contributor.getLastName(), person2.getLastName());
      assertEquals(contributor.getOrcid(), person2.getOrcid());
    }
  }

  @Test
  void filterContributorsWithDepositsTest() {
    // Create a deposit with a contributor
    final String title = "deposit contributor";
    final Deposit deposit = this.createRemoteDeposit(DLCMTestConstants.PersistenceMode.TEMPORARY, title, DEFAULT_DESCRIPTION, this.role, false);
    // Create a new Person
    final Person person = this.getOrCreateRemotePerson();
    // Add the person as contributor
    this.depositService.addContributor(deposit.getResId(), person.getResId());
    final List<Person> listContributors = this.depositService.getContributors(deposit.getResId());
    assertEquals(1, listContributors.size());
    assertEquals(person.getResId(), listContributors.get(0).getResId());

    // Filter on deposit title
    final List<Deposit> depositList = this.contributorService.searchDeposits(person.getResId(), Map.of("deposit.title", title));
    assertTrue(depositList.size() > 0);
    for (final Deposit d : depositList) {
      assertTrue(d.getTitle().contains(title));
    }
  }

  @Test
  void listContributorsTest() {
    final RestCollection<Contributor> contributorList = this.contributorService.list();
    final List<Person> personList = this.personService.findAll();
    assertEquals(contributorList.getPage().getTotalItems(), personList.size());
  }

  @Test
  void listContributorsWithDepositsTest() {
    log.debug("Starting listContributorsWithDepositsTest");

    // Create a deposit with a contributor
    final Deposit deposit = this.createRemoteDeposit(DLCMTestConstants.PersistenceMode.TEMPORARY, DEFAULT_TITLE, DEFAULT_DESCRIPTION, this.role, false);
    log.debug("Deposit {} created on org unit {}", deposit.getResId(), deposit.getOrganizationalUnitId());

    // Create a new Person
    final Person person1 = this.getOrCreateRemotePerson();
    log.debug("Use person {} {} ({})", person1.getFirstName(), person1.getLastName(), person1.getResId());

    // Add the person as contributor
    this.depositService.addContributor(deposit.getResId(), person1.getResId());
    final List<Person> listContributors = this.depositService.getContributors(deposit.getResId());
    assertEquals(1, listContributors.size());
    assertEquals(person1.getResId(), listContributors.get(0).getResId());

    // List deposits of the contributor
    final RestCollection<Deposit> depositList = this.contributorService.listDeposits(person1.getResId());
    log.debug("Person {} has {} deposit", person1.getResId(), depositList.getPage().getTotalItems());

    // Test if not empty
    assertNotEquals(0, depositList.getPage().getTotalItems());

    // Test that the number of deposits of the contributor is consistent
    final RestCollection<Contributor> contributorList = this.contributorService.listWithDeposits();
    log.debug("{} contributors have at least one deposit", contributorList.getPage().getTotalItems());
    for (final Contributor contributor : contributorList.getData()) {
      if (contributor.getResId().equals(person1.getResId())) {
        // Check if deposit number is correct
        assertEquals(depositList.getPage().getTotalItems(), contributor.getDepositNumber().longValue());
      }
    }
    log.debug("Stopping listContributorsWithDepositsTest");
  }

  @Override
  protected void deleteFixtures() {
    // Wait before to delete
    SolidifyTime.waitOneSecond();

    final OrganizationalUnit organizationalUnit = this.getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode.PERMANENT, OrganizationalUnitStatus.OPEN, this.role);
    this.clearDepositsFixtures(organizationalUnit);
    this.clearPeopleFixtures();
  }
}
