/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Integration Tests - MetadataUpdateIT.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.test.preingest;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import javax.xml.parsers.ParserConfigurationException;

import org.junit.jupiter.api.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.ClassPathResource;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import ch.unige.solidify.model.dto.CitationDto;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.XMLTool;
import ch.unige.solidify.util.ZipTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.Access;
import ch.dlcm.model.EmbargoInfo;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.ArchiveContainer;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.security.Role;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.test.DLCMTestConstants;

class MetadataUpdateIT extends AbstractPreingestIT {
  private static final String TEMP_DEPOSIT_WITH_COMPLETED_AIP_TITLE = "Temporary deposit with completed AIP ";
  private static final String TEMP_DEPOSIT_TITLE = "Temporary deposit ";

  private static final Logger log = LoggerFactory.getLogger(MetadataUpdateIT.class);

  protected Deposit createDepositWithCompletedAip() {
    final OrganizationalUnit orgUnit = this.getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode.PERMANENT,
            DLCMTestConstants.OrganizationalUnitStatus.OPEN,
            Role.MANAGER_ID);
    // Create and complete a deposit
    this.restClientTool.sudoAdmin();
    final Deposit deposit = this.createRemoteDepositWithContributor(DLCMTestConstants.PersistenceMode.TEMPORARY,
            TEMP_DEPOSIT_WITH_COMPLETED_AIP_TITLE, DEFAULT_DESCRIPTION);
    deposit.addItem(this.createLocalDepositDataFile(deposit));
    final Deposit deposit1 = this.createRemoteDepositWithData(deposit);
    this.waitDepositWithDataIsReady(deposit1.getResId());
    this.approveDeposit(deposit1);
    this.waitDepositIsCompleted(deposit1.getResId());
    this.waitAIPIsCompletedOrError(orgUnit.getResId(), deposit.getTitle());
    this.restClientTool.exitSudo();
    log.info("Creating a Deposit : {} ({}) with status in : {}", deposit1.getTitle(), deposit1.getResId(), deposit1.getStatus());
    return this.depositService.searchByProperties(Map.of("title", TEMP_DEPOSIT_WITH_COMPLETED_AIP_TITLE)).get(0);
  }

  @Test
  void cannotUpdateCompletedDepositTest() {
    Deposit deposit = this.createDepositWithCompletedAip();
    deposit.setTitle(DLCMTestConstants.getRandomNameWithTemporaryLabel("Deposit v1"));
    deposit.setDescription("Deposit corrected (version 1)");

    String depositId = deposit.getResId();
    HttpClientErrorException e = assertThrows(HttpClientErrorException.class, () -> this.depositService.update(depositId, deposit));
    assertEquals(HttpStatus.FORBIDDEN, e.getStatusCode());
  }

  @Test
  void updateEditingMetadataDepositTest() {
    Deposit deposit = this.createDepositWithCompletedAip();
    deposit = this.startEditingMetadata(deposit);
    String title = DLCMTestConstants.getRandomNameWithTemporaryLabel("Deposit v1");
    deposit.setTitle(title);
    deposit.setDescription("Deposit corrected (version 1)");

    this.depositService.update(deposit.getResId(), deposit);
    Deposit depositChanged = this.depositService.searchByProperties(Map.of("title", title)).get(0);

    assertEquals(deposit.getTitle(), depositChanged.getTitle());
    assertEquals(deposit.getDescription(), depositChanged.getDescription());
  }

  @Test
  void updateAndValidateArchive() throws IOException, ParserConfigurationException, SAXException {
    Deposit deposit = this.depositService.searchByProperties(Map.of("title", DEFAULT_TITLE_TO_UPDATE)).get(0);
    deposit = this.startEditingMetadata(deposit);
    String newTitle = this.getNewArchiveToUpdateTitle(deposit.getTitle());
    deposit.setTitle(newTitle);

    this.depositService.update(deposit.getResId(), deposit);
    this.approveDeposit(deposit);
    this.waitDepositIsCompleted(deposit.getResId());
    this.waitAIPIsCompletedOrError(deposit.getOrganizationalUnitId(), newTitle);

    // Check that the AIP has the new title
    List<ArchivalInfoPackage> aipsList = this.aipService.searchByProperties(Map.of("info.name", newTitle));
    assertEquals(1, aipsList.size());

    // check if order has been already created
    ArchivalInfoPackage aip = aipsList.get(0);
    String aipId = aip.getResId();
    try {
      this.metadataService.getDownloadStatus(aipId);
    } catch (final HttpClientErrorException e) {
      this.metadataService.prepareDownload(aipId);
      this.waitForPublicOrderIsReadyOrInError(aipId);
    }

    // Download updated archive from Access module
    Path packagedArchivePath = Paths.get(System.getProperty("java.io.tmpdir"), "archiveToUpdate.tmp");
    this.metadataService.downloadArchive(aipId, packagedArchivePath);
    assertNotNull(packagedArchivePath.toFile());

    // Extract archive
    Path extractedArchivePath = Paths.get(System.getProperty("java.io.tmpdir"), "archiveToUpdate");
    ZipTool zipAip = new ZipTool(packagedArchivePath.toString());
    if (aip.getArchiveContainer() == ArchiveContainer.ZIP) {
      zipAip.unzipFiles(extractedArchivePath);
    } else if (aip.getArchiveContainer() == ArchiveContainer.BAG_IT) {
      zipAip.unzipFolder(extractedArchivePath, DLCMConstants.DATA_FOLDER + "/");
    }

    // Check that the dlcm.xml file in the archive contains the new updated title
    Path dlcmXmlPath = Paths.get(extractedArchivePath + "/" + aipId + "/" + DLCMConstants.METADATA_FILE);
    Document dlcmDoc = XMLTool.parseXML(FileTool.toString(dlcmXmlPath));
    String titleFromArchive = XMLTool.getFirstTextContent(dlcmDoc, "//datacite:titles/datacite:title",
            Map.of("datacite", DLCMConstants.DATACITE_NAMESPACE_4));
    assertEquals(newTitle, titleFromArchive);

    // Check that the number of update metadata files matches the number of updates
    int numberOfUpdates = this.getNumberBetweenSquareBrackets(newTitle);
    Path folderPath = Paths.get(extractedArchivePath + "/" + aipId);
    long filesNumber = FileTool.findFiles(folderPath).stream()
            .filter(path -> path.getFileName().toString().contains(DLCMConstants.REPRESENTATION_INFO_METADATA_FILE)).count();
    assertEquals(numberOfUpdates, filesNumber);

    // Clean files
    FileTool.deleteFile(packagedArchivePath);
    FileTool.deleteFolder(extractedArchivePath);

    // Check that citations have been modified as well
    List<CitationDto> citations = this.metadataService.getBibliographies(aipId);
    assertNotNull(citations);
    if (!citations.isEmpty()) {
      CitationDto citationDto = citations.get(0);
      assertTrue(citationDto.getText().contains(newTitle), "'" + citationDto.getText() + "' does not contain '" + newTitle + "'");
    }
  }

  @Test
  void cannotUpdateDepositInErrorTest() {
    Deposit deposit = this.createDepositWithCompletedAip();

    this.restClientTool.sudoRoot();
    this.depositService.putDepositInError(deposit.getResId());
    this.restClientTool.exitSudo();

    String depositId = deposit.getResId();
    HttpClientErrorException e = assertThrows(HttpClientErrorException.class, () -> this.depositService.update(depositId, deposit));
    assertEquals(HttpStatus.FORBIDDEN, e.getStatusCode());
  }

  @Test
  void cannotStartEditingDepositInErrorTest() {
    this.restClientTool.sudoRoot();
    final Deposit deposit = this.createRemoteDeposit(DLCMTestConstants.PersistenceMode.TEMPORARY, TEMP_DEPOSIT_TITLE, DEFAULT_DESCRIPTION);
    this.depositService.putDepositInError(deposit.getResId());
    Deposit depositInError = this.depositService.findOne(deposit.getResId());
    this.restClientTool.exitSudo();

    HttpClientErrorException e = assertThrows(HttpClientErrorException.class, () -> this.startEditingMetadata(depositInError));
    assertEquals(HttpStatus.FORBIDDEN, e.getStatusCode());
  }

  @Test
  void cannotStartEditingDepositInProgressTest() {
    this.restClientTool.sudoAdmin();
    final Deposit deposit = this.createRemoteDeposit(DLCMTestConstants.PersistenceMode.TEMPORARY, TEMP_DEPOSIT_TITLE, DEFAULT_DESCRIPTION);
    this.restClientTool.exitSudo();

    HttpClientErrorException e = assertThrows(HttpClientErrorException.class, () -> this.startEditingMetadata(deposit));
    assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode());
  }

  @Test
  void updateFromDataciteTest() {
    Deposit deposit = this.createDepositWithCompletedAip();

    Deposit depositEditing = this.startEditingMetadata(deposit);

    // upload data from a datacite xml file
    this.depositService.uploadMetadataFile(depositEditing.getResId(), new ClassPathResource("datacite.xml"));

    this.waitDepositWithDataIsReady(depositEditing.getResId());
    this.restClientTool.sudoAdmin();
    this.approveDeposit(depositEditing);
    this.waitDepositIsCompleted(depositEditing.getResId());
    this.restClientTool.exitSudo();

    Deposit depositUploaded = this.depositService.searchByProperties(Map.of("title", TEMP_DEPOSIT_WITH_COMPLETED_AIP_TITLE)).get(0);
    // title is not updated from datacite file content (Deposit properties have precedence over datacite file)
    assertEquals(deposit.getTitle(), depositUploaded.getTitle());
    assertEquals(Deposit.DepositStatus.COMPLETED, depositUploaded.getStatus());
    assertEquals(deposit.getDataFileNumber() + 1, depositUploaded.getDataFileNumber());
  }

  @Test
  void updateFromDataciteThenDepositTest() {
    final OrganizationalUnit orgUnit = this.getOrCreateOrganizationalUnit(DLCMTestConstants.PersistenceMode.PERMANENT,
            DLCMTestConstants.OrganizationalUnitStatus.OPEN,
            Role.MANAGER_ID);
    Deposit deposit = this.createDepositWithCompletedAip();

    Deposit depositEditing = this.startEditingMetadata(deposit);

    // upload data from a datacite xml file and wait until the file is ready
    this.depositService.uploadMetadataFile(depositEditing.getResId(), new ClassPathResource("datacite.xml"));
    this.waitDepositWithDataIsReady(depositEditing.getResId());
    this.restClientTool.sudoAdmin();
    this.approveDeposit(depositEditing);
    this.waitDepositIsCompleted(deposit.getResId());
    this.waitAIPIsCompletedOrError(orgUnit.getResId(), deposit.getTitle());
    this.restClientTool.exitSudo();

    Deposit depositEditing2 = this.startEditingMetadata(depositEditing);

    depositEditing2.setTitle(DLCMTestConstants.getRandomNameWithTemporaryLabel("Deposit title changed"));
    depositEditing2.setDescription("Deposit title changed (version 1)");

    Deposit depositChanged = this.depositService.update(depositEditing2.getResId(), depositEditing2);

    this.restClientTool.sudoAdmin();
    this.approveDeposit(depositChanged);
    this.waitDepositIsCompleted(depositChanged.getResId());
    this.restClientTool.exitSudo();

    Deposit depositUploaded2 = this.depositService.searchByProperties(Map.of("title", "Deposit title changed")).get(0);

    assertEquals(depositEditing2.getTitle(), depositUploaded2.getTitle());
    assertEquals(depositEditing2.getDescription(), depositUploaded2.getDescription());
  }

  @Test
  void updateDepositKeywordsTest() {
    Deposit deposit = this.createDepositWithCompletedAip();
    deposit = this.startEditingMetadata(deposit);
    deposit.setKeywords(Arrays.asList("Science", "Computer Science"));

    this.depositService.update(deposit.getResId(), deposit);
    Deposit depositChanged = this.depositService.searchByProperties(Map.of("title", TEMP_DEPOSIT_WITH_COMPLETED_AIP_TITLE)).get(0);

    assertEquals(deposit.getTitle(), depositChanged.getTitle());
    assertEquals(deposit.getKeywords(), depositChanged.getKeywords());
  }

  @Test
  void updateDepositAccessLevelTest() {
    Deposit deposit = this.createDepositWithCompletedAip();
    deposit = this.startEditingMetadata(deposit);
    deposit.setAccess(Access.CLOSED);

    this.depositService.update(deposit.getResId(), deposit);
    Deposit depositChanged = this.depositService.findOne(deposit.getResId());

    assertEquals(deposit.getTitle(), depositChanged.getTitle());
    assertEquals(deposit.getAccess(), depositChanged.getAccess());
  }

  @Test
  void updateDepositEmbargoTest() {
    Deposit deposit = this.createDepositWithCompletedAip();
    deposit = this.startEditingMetadata(deposit);
    EmbargoInfo embargo = new EmbargoInfo();
    embargo.setAccess(Access.CLOSED);
    embargo.setMonths(10);
    deposit.setEmbargo(embargo);

    this.depositService.update(deposit.getResId(), deposit);
    Deposit depositChanged = this.depositService.findOne(deposit.getResId());

    assertEquals(deposit.getEmbargo(), depositChanged.getEmbargo());
    assertEquals(deposit.getEmbargoAccess(), depositChanged.getEmbargoAccess());
    assertEquals(deposit.getEmbargoMonths(), depositChanged.getEmbargoMonths());
  }

  @Test
  void cannotUpdateFromCorruptedDataciteTest() {
    Deposit deposit = this.createDepositWithCompletedAip();

    Deposit depositEditing = this.startEditingMetadata(deposit);

    // upload corrupted datacite xml file
    this.depositService.uploadMetadataFile(depositEditing.getResId(), new ClassPathResource("datacite-corrupted.xml"));

    Deposit deposit1 = this.depositService.searchByProperties(Map.of("title", TEMP_DEPOSIT_WITH_COMPLETED_AIP_TITLE)).get(0);

    // Approving the deposit fails as the corrupted Datacite datafile is in error
    HttpClientErrorException e = assertThrows(HttpClientErrorException.class, () -> this.approveDeposit(deposit1));
    assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode());
    this.assertFieldContainsValidationErrorMessage(e, "status", "Data files of deposit are not ready or not cleaned");
  }

  @Test
  void cannotUpdateFromEmptyDataciteTest() {
    Deposit deposit = this.createDepositWithCompletedAip();

    Deposit depositEditing = this.startEditingMetadata(deposit);

    String depositId = depositEditing.getResId();

    HttpClientErrorException e = assertThrows(HttpClientErrorException.class,
            () -> this.depositService.uploadMetadataFile(depositId, null));
    assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode());
    assertNotNull(e.getMessage());
    assertTrue(e.getMessage().contains("Required request part 'file' is not present"));
  }

  @Test
  void cancelMetadataEditingTest() {
    Deposit deposit = this.createDepositWithCompletedAip();
    String originalTitle = deposit.getTitle();
    String originalDescription = deposit.getDescription();
    this.waitDepositIsCompleted(deposit.getResId());
    this.waitAIPIsCompletedOrError(deposit.getOrganizationalUnitId(), originalTitle);

    deposit = this.startEditingMetadata(deposit);
    String title = DLCMTestConstants.getRandomNameWithTemporaryLabel("Deposit v1");
    deposit.setTitle(title);
    deposit.setDescription("Deposit corrected (version 1)");

    this.depositService.update(deposit.getResId(), deposit);
    Deposit depositChanged = this.depositService.searchByProperties(Map.of("title", title)).get(0);

    assertEquals(deposit.getTitle(), depositChanged.getTitle());
    assertEquals(deposit.getDescription(), depositChanged.getDescription());

    this.depositService.cancelEditingMetadata(deposit.getResId());
    this.waitDepositIsCompleted(deposit.getResId());
    depositChanged = this.depositService.searchByProperties(Map.of("title", originalTitle)).get(0);

    assertEquals(originalTitle, depositChanged.getTitle());
    assertEquals(originalDescription, depositChanged.getDescription());
  }

  @Override
  protected void deleteFixtures() {
    this.restClientTool.sudoRoot();
    this.clearDepositsFixtures();
    this.restClientTool.exitSudo();
  }
}
