/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - RegistryListener.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.listener;

import java.util.List;
import javax.persistence.PostPersist;
import javax.persistence.PostRemove;
import javax.persistence.PostUpdate;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;

import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.message.RegistryMessage;
import ch.dlcm.model.AbstractDataFile;
import ch.dlcm.model.AbstractDataFile.DataFileStatus;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.access.Order;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.DisseminationInfoPackage;
import ch.dlcm.model.oais.SubmissionInfoPackage;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.model.preservation.PreservationJob;
import ch.dlcm.model.settings.OrganizationalUnit;

public class RegistryListener {
  private static final Logger log = LoggerFactory.getLogger(RegistryListener.class);

  @Value("${dlcm.parameters.registryEnabled}")
  private boolean isRegistryEnabled;

  public void register(Object object) {
    final Resource res = (Resource) object;
    final RegistryMessage regMessage = new RegistryMessage(res.getResId(), res.getClass().getSimpleName());
    regMessage.setCreation(res.getCreation());
    regMessage.setLastUpdate(res.getLastUpdate());
    this.completeMessage(res, regMessage);
    this.logMessage(regMessage);
    SolidifyEventPublisher.getPublisher().publishEvent(regMessage);
  }

  private void addOrganizationalUnit(RegistryMessage regMessage, String organizationalUnitId) {
    regMessage.getMetadata().put("organizationalUnit", organizationalUnitId);
  }

  private void addSize(RegistryMessage regMessage, List<? extends AbstractDataFile> datafiles) {
    regMessage.getMetadata().put("size", datafiles.stream().mapToLong(AbstractDataFile::getFileSize).sum());
  }

  private void completeMessage(Resource res, RegistryMessage regMessage) {
    if (res instanceof OrganizationalUnit) {
      final OrganizationalUnit p = (OrganizationalUnit) res;
      regMessage.getMetadata().put("name", p.getName());
      regMessage.getMetadata().put("description", p.getDescription());
      regMessage.getMetadata().put("open", p.getOpeningDate());
      if (p.getClosingDate() != null) {
        regMessage.getMetadata().put("closed", p.getClosingDate());
      }
    } else if (res instanceof Deposit) {
      final Deposit d = (Deposit) res;
      this.addOrganizationalUnit(regMessage, d.getOrganizationalUnitId());
      this.addSize(regMessage, d.getDataFiles());
      if (d.getStatus() == DepositStatus.COMPLETED) {
        regMessage.getMetadata().put("fileNumber", d.getDataFiles().size());
      }
    } else if (res instanceof AbstractDataFile) {
      final AbstractDataFile df = (AbstractDataFile) res;
      if (df.getStatus() == DataFileStatus.READY) {
        regMessage.getMetadata().put(DLCMConstants.DB_PACKAGE_ID, df.getInfoPackage().getResId());
        regMessage.getMetadata().put("size", df.getFileSize());
      }
    } else if (res instanceof SubmissionInfoPackage) {
      final SubmissionInfoPackage sip = (SubmissionInfoPackage) res;
      this.addOrganizationalUnit(regMessage, sip.getInfo().getOrganizationalUnitId());
      this.addSize(regMessage, sip.getDataFiles());
      regMessage.getMetadata().put("deposit", sip.getDepositId());
      if (sip.getInfo().getStatus() == PackageStatus.COMPLETED) {
        regMessage.getMetadata().put("aip", sip.getAipId());
      }
    } else if (res instanceof ArchivalInfoPackage) {
      final ArchivalInfoPackage aip = (ArchivalInfoPackage) res;
      this.addOrganizationalUnit(regMessage, aip.getInfo().getOrganizationalUnitId());
      regMessage.getMetadata().put("size", aip.getFileSize());
      regMessage.getMetadata().put("unit", aip.getArchivalUnit());
      regMessage.getMetadata().put("retention", aip.getSmartRetention());
      regMessage.getMetadata().put("retentionEnd", aip.getRetentionEnd());
      if (aip.getInfo().getStatus() == PackageStatus.COMPLETED && aip.getSipIds() != null) {
        regMessage.getMetadata().put("sip", aip.getSipIds().toString());
      }
    } else if (res instanceof DisseminationInfoPackage) {
      final DisseminationInfoPackage dip = (DisseminationInfoPackage) res;
      this.addOrganizationalUnit(regMessage, dip.getInfo().getOrganizationalUnitId());
      if (dip.getInfo().getStatus() == PackageStatus.COMPLETED) {
        regMessage.getMetadata().put("aip", dip.getAipIds().toString());
      }
    } else if (res instanceof PreservationJob) {
      final PreservationJob job = (PreservationJob) res;
      regMessage.getMetadata().put("jobName", job.getName());
      // TODO To check
      // if (job.getStatus() == JobStatus.COMPLETED) {
      // regMessage.getMetadata().put("begin", job.getStartDate());
      // regMessage.getMetadata().put("end", job.getEndDate());
      // regMessage.getMetadata().put("items", job.getProcessedItems());
      // regMessage.getMetadata().put("itemsInError", job.getInErrorItems());
      // regMessage.getMetadata().put("itemsIgnored", job.getIgnoredItems());
      // }
    } else if (res instanceof Order) {
      final Order o = (Order) res;
      regMessage.getMetadata().put("queryName", o.getName());
      regMessage.getMetadata().put("queryType", o.getQueryType());
      if (!StringTool.isNullOrEmpty(o.getQuery())) {
        regMessage.getMetadata().put("query", o.getQuery());
      }
    }
  }

  private void logMessage(RegistryMessage regMessage) {
    log.trace("Sending message {}", regMessage);
  }

  @PostPersist
  private void registerCreation(Object object) {
    if(!this.isRegistryEnabled) {
      return;
    }
    final Resource res = (Resource) object;
    final RegistryMessage regMessage = new RegistryMessage(res.getResId(), res.getClass().getSimpleName());
    regMessage.setCreation(res.getCreation());
    this.completeMessage(res, regMessage);
    this.logMessage(regMessage);
    SolidifyEventPublisher.getPublisher().publishEvent(regMessage);
  }

  @PostRemove
  private void registerDeletion(Object object) {
    if(!this.isRegistryEnabled) {
      return;
    }
    final Resource res = (Resource) object;
    final RegistryMessage regMessage = new RegistryMessage(res.getResId(), res.getClass().getSimpleName());
    regMessage.setDisposition(res.getLastUpdate());
    this.logMessage(regMessage);
    SolidifyEventPublisher.getPublisher().publishEvent(regMessage);
  }

  @PostUpdate
  private void registerUpdate(Object object) {
    if(!this.isRegistryEnabled) {
      return;
    }
    final Resource res = (Resource) object;
    final RegistryMessage regMessage = new RegistryMessage(res.getResId(), res.getClass().getSimpleName());
    regMessage.setLastUpdate(res.getLastUpdate());
    this.completeMessage(res, regMessage);
    this.logMessage(regMessage);
    SolidifyEventPublisher.getPublisher().publishEvent(regMessage);
  }
}
