/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - ArchiveMetadata.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.index;

import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.model.index.IndexMetadata;
import ch.unige.solidify.model.oai.OAIRecord;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.Access;
import ch.dlcm.model.DataTag;
import ch.dlcm.model.OrganizationalUnitAwareResource;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ResourceName;

@JsonDeserialize(using = ArchiveMetadataDeserializer.class)
public class ArchiveMetadata extends IndexMetadata implements OrganizationalUnitAwareResource, OAIRecord {

  public void addLinksForAccess(WebMvcLinkBuilder linkBuilder) {
    this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.DOWNLOAD_STATUS).withRel(DLCMActionName.DOWNLOAD_STATUS));
    this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.PREPARE_DOWNLOAD).withRel(DLCMActionName.PREPARE_DOWNLOAD));
    this.add(linkBuilder.slash(this.getResId()).slash(ActionName.DOWNLOAD).withRel(ActionName.DOWNLOAD));
    this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.THUMBNAIL).withRel(DLCMActionName.THUMBNAIL));
    this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.STATISTICS).withRel(DLCMActionName.STATISTICS));
    this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.PACKAGES).withRel(ResourceName.PACKAGES));
    this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.DATAFILE).withRel(ResourceName.DATAFILE));
    this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.AIP).withRel(ResourceName.AIP));
    this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.LIST_FOLDERS).withRel(DLCMActionName.LIST_FOLDERS));
    this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.RATING).withRel(DLCMActionName.RATING));
    this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.BIBLIOGRAPHIES).withRel(DLCMActionName.BIBLIOGRAPHIES));
    this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.CITATIONS).withRel(DLCMActionName.CITATIONS));
    super.addLinks(linkBuilder);
  }

  @JsonProperty(access = JsonProperty.Access.READ_ONLY)
  public Access getCurrentAccess() {
    if (this.getIndex().endsWith(DLCMConstants.REGISTRY)) {
      return null;
    }
    final Optional<Access> embargoAccessLevel = this.getEmbargoAccessLevel();
    if (embargoAccessLevel.isPresent()) {
      final OffsetDateTime embargoEndDate = this.getEmbargoEndDate().orElseThrow(() -> new RuntimeException(
              "ArchiveMetadata " + this.getResId() + " has an embargo AccessLevel but does not contain any embargo end date"));
      if (embargoEndDate.isBefore(OffsetDateTime.now())) {
        // embargo is over
        return this.getFinalAccessLevel();
      } else {
        // embargo is still running
        return embargoAccessLevel.get();
      }
    } else {
      // no embargo
      return this.getFinalAccessLevel();
    }
  }

  @JsonIgnore
  public Optional<Access> getEmbargoAccessLevel() {
    return this.getAccessLevel(DLCMConstants.AIP_EMBARGO_LEVEL);
  }

  @JsonIgnore
  public Optional<OffsetDateTime> getEmbargoEndDate() {
    final String endEmbargo = (String) this.getMetadata(DLCMConstants.AIP_EMBARGO_END_DATE);
    if (!StringTool.isNullOrEmpty(endEmbargo)) {
      return Optional.of(OffsetDateTime.parse(endEmbargo));
    }
    return Optional.empty();
  }

  @JsonIgnore
  public Access getFinalAccessLevel() {
    return this.getAccessLevel(DLCMConstants.AIP_ACCESS_LEVEL)
            .orElseThrow(() -> new RuntimeException("Final access level is unknown for ArchiveMetadata " + this.getResId()));
  }

  @JsonIgnore
  public DataTag getDataTag() {
    final String dataTag = (String) this.getMetadata(DLCMConstants.AIP_DATA_TAG);
    if (!StringTool.isNullOrEmpty(dataTag)) {
      return DataTag.valueOf(dataTag);
    }
    return DataTag.UNDEFINED;
  }

  @Override
  @JsonIgnore
  public String getOrganizationalUnitId() {
    return (String) this.getMetadata(DLCMConstants.AIP_ORGA_UNIT);
  }

  @JsonIgnore
  public String getMetadataVersion() {
    return (String) this.getMetadata(DLCMConstants.METADATA_VERSION_FIELD);
  }

  @Override
  @JsonIgnore
  public String getOAIMetadata() {
    return (String) this.getMetadata(DLCMConstants.DATACITE_XML_FIELD);
  }

  private Optional<Access> getAccessLevel(String accessField) {
    final String accessLevel = (String) this.getMetadata(accessField);
    if (!StringTool.isNullOrEmpty(accessLevel)) {
      return Optional.of(Access.valueOf(accessLevel));
    }
    return Optional.empty();
  }

  @Override
  @JsonIgnore
  public OffsetDateTime getCreationDate() {
    final String creation = (String) this.getMetadata(DLCMConstants.AIP_LAST_UPDATE.split(SolidifyConstants.REGEX_FIELD_PATH_SEP));
    return OffsetDateTime.parse(creation).toInstant().atOffset(ZoneOffset.UTC).withNano(0);
  }

  @Override
  @JsonIgnore
  public String getOaiId() {
    return this.getResId();
  }

  @JsonIgnore
  public String getDoi() {
    if (((String) this.getMetadata(DLCMConstants.IDENTIFIER_TYPE_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP))).equals(
            DLCMConstants.DOI)) {
      return (String) this.getMetadata(DLCMConstants.IDENTIFIER_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP));
    }
    return null;
  }

  @JsonIgnore
  public String getResourceTypeGeneral() {
    return this.getMetadata(DLCMConstants.RESOURCE_TYPE_GENERAL_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)).toString();
  }

  @JsonIgnore
  public String getResourceType() {
    return this.getMetadata(DLCMConstants.RESOURCE_TYPE_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)).toString();
  }

  @JsonIgnore
  public String getTitle() {
    return this.getMetadata(DLCMConstants.TITLE_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)).toString();
  }

  @JsonIgnore
  public String getAbstract() {
    if (this.getMetadata(DLCMConstants.DESCRIPTION_TYPE_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)).equals(DLCMConstants.ABSTRACT)) {
      return (String) this.getMetadata(DLCMConstants.DESCRIPTION_CONTENT_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP));
    }
    return null;
  }

  @JsonIgnore
  public String getAbstractLanguage() {
    if (this.getMetadata(DLCMConstants.DESCRIPTION_TYPE_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)).equals(DLCMConstants.ABSTRACT)) {
      return (String) this.getMetadata(DLCMConstants.DESCRIPTION_LANGUAGE_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP));
    }
    return null;
  }

  @JsonIgnore
  public List<String> getKeywords() {
    if (((Map<String, Object>) this.getMetadata(DLCMConstants.DATACITE_RESOURCE_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)))
            .containsKey(DLCMConstants.DATACITE_SUBJECTS_NAME)
            && this.getMetadata(DLCMConstants.SUBJECT_SCHEME_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP))
            .equals(DLCMConstants.KEYWORDS)) {
      String keywords = (String) this.getMetadata(DLCMConstants.SUBJECT_CONTENT_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP));
      return Arrays.asList(keywords.split(SolidifyConstants.FIELD_SEP));
    }
    return Collections.emptyList();
  }

  @JsonIgnore
  public List<Map<String, Object>> getRelations() {
    if (((Map<String, Object>) this.getMetadata(DLCMConstants.DATACITE_RESOURCE_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)))
            .containsKey(DLCMConstants.DATACITE_RELATIONS_NAME)
            &&
            ((Map<String, Object>) this.getMetadata(DLCMConstants.RELATIONS_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)))
                    .containsKey(DLCMConstants.DATACITE_RELATION_NAME)) {
      return this.getField(DLCMConstants.RELATION_FIELD);
    }
    return Collections.emptyList();
  }

  @JsonIgnore
  public List<Map<String, Object>> getRights() {
    return this.getField(DLCMConstants.RIGHTS_FIELD);
  }

  @JsonIgnore
  public String getPublisher() {
    return this.getMetadata(DLCMConstants.PUBLISHER_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)).toString();
  }

  @JsonIgnore
  public List<Map<String, String>> getCreators() {
    List<Map<String, Object>> creators = this.getField(DLCMConstants.CREATOR_FIELD);
    List<Map<String, String>> creatorsList = new ArrayList<>();
    for (Map<String, Object> creator : creators) {
      Map<String, String> creatorMap = new HashMap<>();

      if (creator.containsKey(DLCMConstants.DATACITE_FAMILY_NAME)) {
        String familyName = creator.get(DLCMConstants.DATACITE_FAMILY_NAME).toString();
        creatorMap.put(DLCMConstants.FAMILY_NAME, familyName);
      }
      if (creator.containsKey(DLCMConstants.DATACITE_FAMILY_NAME)) {
        String givenName = creator.get(DLCMConstants.DATACITE_GIVEN_NAME).toString();
        creatorMap.put(DLCMConstants.GIVEN_NAME, givenName);
      }
      if (creator.containsKey(DLCMConstants.DATACITE_IDENTIFIER_NAME)) {
        String orcid = ((Map<String, Object>) creator.get(DLCMConstants.DATACITE_IDENTIFIER_NAME)).get(DLCMConstants.CONTENT_FIELD).toString();
        creatorMap.put(DLCMConstants.PERSON_ORCID, orcid);
      }

      if (!creatorMap.isEmpty()) {
        creatorsList.add(creatorMap);
      }
    }
    return creatorsList;
  }

  @JsonIgnore
  public OffsetDateTime getIssuedDate() {
    return this.getDate(DLCMConstants.DATE_ISSUED_ATTRIBUTE);
  }

  @JsonIgnore
  public OffsetDateTime getCreatedDate() {
    return this.getDate(DLCMConstants.DATE_CREATED_ATTRIBUTE);
  }

  @JsonIgnore
  public OffsetDateTime getUpdatedDate() {
    return this.getDate(DLCMConstants.DATE_UPDATED_ATTRIBUTE);
  }

  @JsonIgnore
  public OffsetDateTime getAcceptedDate() {
    return this.getDate(DLCMConstants.DATE_ACCEPTED_ATTRIBUTE);
  }

  @JsonIgnore
  public OffsetDateTime getAvailableDate() {
    return this.getDate(DLCMConstants.DATE_AVAILABLE_ATTRIBUTE);
  }

  @JsonIgnore
  public OffsetDateTime getCollectedStartDate() {
    return this.getDate(DLCMConstants.DATE_COLLECTED_START_ATTRIBUTE);
  }

  @JsonIgnore
  public OffsetDateTime getCollectedEndDate() {
    return this.getDate(DLCMConstants.DATE_COLLECTED_END_ATTRIBUTE);
  }

  @JsonIgnore
  public String getRetention() {
    return this.getMetadata(DLCMConstants.AIP_RETENTION.split(SolidifyConstants.REGEX_FIELD_PATH_SEP)).toString();
  }

  private OffsetDateTime getDate(String dateType) {
    List<Map<String, String>> dates = (List<Map<String, String>>) this
            .getMetadata(DLCMConstants.DATES_FIELD.split(SolidifyConstants.REGEX_FIELD_PATH_SEP));
    for (Map<String, String> date : dates) {
      if (date.get(DLCMConstants.DATE_TYPE).equals(dateType)) {
        String dateStr = date.get(DLCMConstants.CONTENT_FIELD);
        return OffsetDateTime.parse(dateStr);
      }
    }
    return null;
  }

  private List<Map<String, Object>> getField(String fieldName) {
    Object fieldObject = this.getMetadata(fieldName.split(SolidifyConstants.REGEX_FIELD_PATH_SEP));
    if (fieldObject instanceof ArrayList) {
      return (ArrayList<Map<String, Object>>) fieldObject;
    } else if (fieldObject instanceof Map) {
      return new ArrayList<>(Arrays.asList((Map<String, Object>) fieldObject));
    }
    return Collections.emptyList();
  }

}
