/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - Deposit.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.preingest;

import static ch.unige.solidify.SolidifyConstants.DB_DATE_LENGTH;
import static ch.unige.solidify.SolidifyConstants.DB_ID_LENGTH;

import java.net.URI;
import java.time.LocalDate;
import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import javax.persistence.CascadeType;
import javax.persistence.CollectionTable;
import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.ElementCollection;
import javax.persistence.Embedded;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.JoinColumn;
import javax.persistence.Lob;
import javax.persistence.OneToMany;
import javax.persistence.OrderBy;
import javax.persistence.PrePersist;
import javax.persistence.PreUpdate;
import javax.persistence.Transient;
import javax.persistence.UniqueConstraint;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import com.fasterxml.jackson.annotation.JsonFormat;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.converter.KeywordsConverter;
import ch.unige.solidify.model.Language;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.util.CollectionTool;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.validation.AssertThatAnotherFieldHasValue;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.listener.HistoryListener;
import ch.dlcm.listener.OrganizationalUnitDataListener;
import ch.dlcm.listener.RegistryListener;
import ch.dlcm.model.Access;
import ch.dlcm.model.ComplianceLevel;
import ch.dlcm.model.DataFileOwnerInterface;
import ch.dlcm.model.DataTag;
import ch.dlcm.model.EmbargoInfo;
import ch.dlcm.model.OffsetDateTimeDeserializer;
import ch.dlcm.model.OrganizationalUnitAwareResource;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.MetadataVersionAware;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.settings.ArchiveType;
import ch.dlcm.model.settings.License;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.model.settings.SubmissionAgreement;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;

@Entity
@EntityListeners({ HistoryListener.class, RegistryListener.class, OrganizationalUnitDataListener.class })
@AssertThatAnotherFieldHasValue.List({
        @AssertThatAnotherFieldHasValue(fieldName = "agreement", fieldValue = "true", dependFieldName = "submissionAgreementId"),
        @AssertThatAnotherFieldHasValue(fieldName = "access", fieldValue = "PUBLIC", dependFieldName = "licenseId") })
public class Deposit extends ResourceNormalized implements SubResourceContainer, DataFileOwnerInterface<DepositDataFile>,
        OrganizationalUnitAwareResource, MetadataVersionAware {

  public enum DepositStatus {
    APPROVED, CHECKED, CLEANED, CLEANING, COMPLETED, DELETING, IN_ERROR, IN_PROGRESS, IN_VALIDATION, REJECTED, SUBMITTED, EDITING_METADATA, CHECKING_COMPLIANCE, CHECKING_COMPLIANCE_CLEANED, COMPLIANCE_ERROR, EDITING_ERROR, CANCEL_EDITING_METADATA
  }

  @Enumerated(EnumType.STRING)
  private Access access;

  @Enumerated(EnumType.STRING)
  private DataTag dataSensitivity;

  private Boolean agreement;

  @Transient
  private int dataFileNumber;

  @Transient
  private ComplianceLevel complianceLevel;

  @ElementCollection
  @Column(name = DLCMConstants.DB_AIP_ID, length = DB_ID_LENGTH)
  @CollectionTable(name = "depositAip", joinColumns = {
          @JoinColumn(name = DLCMConstants.DB_COLLECTION_ID, referencedColumnName = DLCMConstants.DB_RES_ID) }, uniqueConstraints = @UniqueConstraint(columnNames = {
                  DLCMConstants.DB_COLLECTION_ID, DLCMConstants.DB_AIP_ID }))
  @JsonIgnore
  private List<String> collection = new ArrayList<>();

  @JsonFormat(shape = JsonFormat.Shape.STRING, pattern = StringTool.DATE_TIME_FORMAT)
  @JsonDeserialize(using = OffsetDateTimeDeserializer.class)
  @Column(length = DB_DATE_LENGTH)
  private OffsetDateTime collectionBegin;

  @JsonFormat(shape = JsonFormat.Shape.STRING, pattern = StringTool.DATE_TIME_FORMAT)
  @JsonDeserialize(using = OffsetDateTimeDeserializer.class)
  @Column(length = DB_DATE_LENGTH)
  private OffsetDateTime collectionEnd;

  @JsonIgnore
  @OneToMany(mappedBy = "compositeResId.deposit", cascade = CascadeType.ALL, orphanRemoval = true)
  @OrderBy("position ASC")
  private List<DepositContributor> contributors = new ArrayList<>();

  @JsonIgnore
  @OneToMany(mappedBy = "infoPackage", cascade = CascadeType.ALL, orphanRemoval = true)
  private List<DepositDataFile> dataFiles = new ArrayList<>();

  @NotNull
  @Size(min = 1, max = SolidifyConstants.DB_LARGE_STRING_LENGTH)
  private String description;

  @Column(unique = true)
  private String doi;

  @Embedded
  @Valid
  private EmbargoInfo embargo = new EmbargoInfo();

  @Convert(converter = KeywordsConverter.class)
  private List<String> keywords;

  @Transient
  private Language language;

  @Size(max = DB_ID_LENGTH)
  @Column(length = DB_ID_LENGTH)
  private String languageId;

  private String licenseId;

  private String archiveTypeId;

  @Transient
  private ArchiveType archiveType;

  @NotNull
  private DLCMMetadataVersion metadataVersion;

  @Transient
  private OrganizationalUnit organizationalUnit;

  @NotNull
  @Size(min = 1)
  @Column(name = DLCMConstants.DB_ORG_UNIT_ID, length = DB_ID_LENGTH)
  private String organizationalUnitId;

  @Size(max = DB_ID_LENGTH)
  @Column(length = DB_ID_LENGTH)
  private String preparationId;

  @Transient
  private PreservationPolicy preservationPolicy;

  @Size(max = DB_ID_LENGTH)
  @Column(length = DB_ID_LENGTH)
  private String preservationPolicyId;

  @NotNull
  @JsonFormat(shape = JsonFormat.Shape.STRING, pattern = StringTool.DATE_FORMAT)
  private LocalDate publicationDate;

  @JsonIgnore
  private URI sip;

  @Size(max = DB_ID_LENGTH)
  @Column(length = DB_ID_LENGTH)
  private String sipId;

  @Enumerated(EnumType.STRING)
  private DepositStatus status;

  @Size(max = SolidifyConstants.DB_LONG_STRING_LENGTH)
  private String statusMessage;

  @Size(max = DB_ID_LENGTH)
  @Column(length = DB_ID_LENGTH)
  private String submissionAgreementId;

  @Transient
  private SubmissionPolicy submissionPolicy;

  @Size(max = DB_ID_LENGTH)
  @Column(length = DB_ID_LENGTH)
  private String submissionPolicyId;

  @NotNull
  @Size(min = 1)
  private String title;

  private boolean containsUpdatedMetadata = false;

  private String isIdenticalTo;

  private String isObsoletedBy;

  private String isReferencedBy;

  @Override
  public <T> boolean addItem(T t) {

    if (t instanceof Person) {
      return this.addContributor((Person) t);
    } else if (t instanceof DepositDataFile) {
      return this.addDataFile((DepositDataFile) t);
    } else if (t instanceof ArchivalInfoPackage) {
      return this.addAip((ArchivalInfoPackage) t);
    } else {

      if (t instanceof Language) {
        this.setLanguageId(((Language) t).getResId());
      } else if (t instanceof SubmissionAgreement) {
        this.setSubmissionAgreementId(((SubmissionAgreement) t).getResId());
      } else if (t instanceof License) {
        this.setLicenseId(((License) t).getResId());
      } else if (t instanceof OrganizationalUnit) {
        this.setOrganizationalUnitId(((OrganizationalUnit) t).getResId());
      } else if (t instanceof SubmissionPolicy) {
        this.setSubmissionAgreementId(((SubmissionPolicy) t).getResId());
      } else if (t instanceof PreservationPolicy) {
        this.setPreservationPolicyId(((PreservationPolicy) t).getResId());
      } else if (t instanceof ArchiveType) {
        this.setArchiveTypeId(((ArchiveType) t).getResId());
      } else {
        throw new UnsupportedOperationException("Missing implementation 'addItem' method (" + t.getClass().getSimpleName() + ")");
      }

      return true;
    }
  }

  @Lob
  private String additionalFieldsValues;

  @Size(max = DB_ID_LENGTH)
  @Column(length = DB_ID_LENGTH)
  private String additionalFieldsFormId;

  @JsonIgnore
  @Transient
  private boolean isStatusDirty = false;

  /****************************************************************/

  @Override
  public void addLinks(WebMvcLinkBuilder linkBuilder, boolean mainRes, boolean subResOnly) {
    super.addLinks(linkBuilder, mainRes, subResOnly);
    if (mainRes) {
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.CONTRIBUTOR).withRel(ResourceName.CONTRIBUTOR));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.DATAFILE).withRel(ResourceName.DATAFILE));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.AIP).withRel(ResourceName.AIP));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.SUBMISSION_AGREEMENT).withRel(ResourceName.SUBMISSION_AGREEMENT));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.LICENSE).withRel(ResourceName.LICENSE));
      if (this.getStatus() == DepositStatus.COMPLETED) {
        this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.SIP_FILE).withRel(ResourceName.SIP_FILE));
        this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.CLEAN).withRel(DLCMActionName.CLEAN));
      }
      this.add(linkBuilder.slash(this.getResId()).slash(ActionName.UPLOAD).withRel(ActionName.UPLOAD));
      this.add(linkBuilder.slash(this.getResId()).slash(ActionName.UPLOAD_ARCHIVE).withRel(ActionName.UPLOAD_ARCHIVE));
      this.add(linkBuilder.slash(this.getResId()).slash(ActionName.DOWNLOAD).withRel(ActionName.DOWNLOAD));
      this.add(linkBuilder.slash(this.getResId()).slash(ActionName.HISTORY).withRel(ActionName.HISTORY));

      switch (this.getStatus()) {
        case IN_PROGRESS:
          if (this.getSubmissionPolicy() != null) {
            if (Boolean.TRUE.equals(this.getSubmissionPolicy().getSubmissionApproval())) {
              this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.SUBMIT_FOR_APPROVAL).withRel(DLCMActionName.SUBMIT_FOR_APPROVAL));
            } else {
              this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.APPROVE).withRel(DLCMActionName.APPROVE));
            }
          }
          break;
        case IN_VALIDATION:
          this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.APPROVE).withRel(DLCMActionName.APPROVE));
          this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.REJECT).withRel(DLCMActionName.REJECT));
          this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.ENABLE_REVISION).withRel(DLCMActionName.ENABLE_REVISION));
          break;
        case IN_ERROR:
        case REJECTED:
          this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.ENABLE_REVISION).withRel(DLCMActionName.ENABLE_REVISION));
          break;
        case COMPLETED:
        case CLEANED:
          this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.CHECK_COMPLIANCE).withRel(DLCMActionName.CHECK_COMPLIANCE));
          break;
        default:
          break;
      }
    }
  }

  public Access getAccess() {
    return this.access;
  }

  public DataTag getDataSensitivity() {
    return this.dataSensitivity;
  }

  public Boolean getAgreement() {
    return this.agreement;
  }

  @JsonIgnore
  public List<String> getCollection() {
    return this.collection;
  }

  public OffsetDateTime getCollectionBegin() {
    return this.collectionBegin;
  }

  public OffsetDateTime getCollectionEnd() {
    return this.collectionEnd;
  }

  @JsonProperty(access = JsonProperty.Access.READ_ONLY)
  public int getCollectionSize() {
    return this.getCollection().size();
  }

  public ComplianceLevel getComplianceLevel() {
    return this.complianceLevel;
  }

  public void setComplianceLevel(ComplianceLevel complianceLevel) {
    this.complianceLevel = complianceLevel;
  }

  @JsonIgnore
  public List<String> getContributorIds() {
    return this.getContributors().stream().sorted(Comparator.comparing(DepositContributor::getPosition))
            .map(DepositContributor::getPersonId)
            .collect(Collectors.toList());
  }

  @JsonIgnore
  public List<DepositContributor> getContributors() {
    return this.contributors;
  }

  public int getDataFileNumber() {
    return this.dataFileNumber;
  }

  public void setDataFileNumber(int dataFileNumber) {
    this.dataFileNumber = dataFileNumber;
  }

  @Override
  public List<DepositDataFile> getDataFiles() {
    return this.dataFiles;
  }

  /****************************************************************/

  @JsonIgnore
  public DepositStatus getDefaultStatus() {
    return DepositStatus.IN_PROGRESS;
  }

  public String getDescription() {
    return this.description;
  }

  public String getDoi() {
    return this.doi;
  }

  public EmbargoInfo getEmbargo() {
    return this.embargo;
  }

  @JsonIgnore
  public Access getEmbargoAccess() {
    return this.getEmbargo().getAccess();
  }

  @JsonIgnore
  public Integer getEmbargoMonths() {
    return this.getEmbargo().getMonths();
  }

  @JsonIgnore
  public OffsetDateTime getEmbargoStartDate() {
    return this.getEmbargo().getStartDate();
  }

  @Override
  @JsonIgnore
  public List<Class<? extends Resource>> getEmbeddedResourceTypes() {
    final List<Class<? extends Resource>> embeddedResources = new ArrayList<>();

    embeddedResources.add(Language.class);
    embeddedResources.add(OrganizationalUnit.class);
    embeddedResources.add(SubmissionPolicy.class);
    embeddedResources.add(PreservationPolicy.class);
    embeddedResources.add(ArchiveType.class);

    return embeddedResources;
  }

  public boolean getHasEmbargo() {
    return this.hasEmbargo();
  }

  public List<String> getKeywords() {
    return this.keywords;
  }

  public Language getLanguage() {
    return this.language;
  }

  /****************************************************************
   * Embedded resources
   */

  public String getLanguageId() {
    return this.languageId;
  }

  /*****/

  public String getLicenseId() {
    return this.licenseId;
  }

  @Override
  public DLCMMetadataVersion getMetadataVersion() {
    return this.metadataVersion;
  }

  public OrganizationalUnit getOrganizationalUnit() {
    return this.organizationalUnit;
  }

  /*****/

  @Override
  public String getOrganizationalUnitId() {
    return this.organizationalUnitId;
  }

  public String getPreparationId() {
    return this.preparationId;
  }

  public String getArchiveTypeId() {
    return this.archiveTypeId;
  }

  public ArchiveType getArchiveType() {
    return this.archiveType;
  }

  /*****/

  @JsonProperty(access = JsonProperty.Access.READ_ONLY)
  public PreservationPolicy getPreservationPolicy() {
    if (this.getPreservationPolicyId() == null) {
      if (this.getOrganizationalUnit() != null) {
        return this.getOrganizationalUnit().getDefaultPreservationPolicy();
      } else {
        return null;
      }
    }
    return this.preservationPolicy;
  }

  public String getPreservationPolicyId() {
    return this.preservationPolicyId;
  }

  public LocalDate getPublicationDate() {
    return this.publicationDate;
  }

  public URI getSip() {
    return this.sip;
  }

  public String getSipId() {
    return this.sipId;
  }

  public DepositStatus getStatus() {
    return this.status;
  }

  public String getStatusMessage() {
    return this.statusMessage;
  }

  /****************************************************************
   * Linked resources
   */

  public String getSubmissionAgreementId() {
    return this.submissionAgreementId;
  }

  /*****/

  @JsonProperty(access = JsonProperty.Access.READ_ONLY)
  public SubmissionPolicy getSubmissionPolicy() {
    if (this.getSubmissionPolicyId() == null) {
      if (this.getOrganizationalUnit() != null) {
        return this.getOrganizationalUnit().getDefaultSubmissionPolicy();
      } else {
        return null;
      }
    }
    return this.submissionPolicy;
  }

  public String getSubmissionPolicyId() {
    return this.submissionPolicyId;
  }

  @Override
  public <T> List<String> getSubResourceIds(Class<T> type) {

    List<String> subResourceIds = Collections.emptyList();

    if (type == Person.class && this.getContributors() != null) {
      subResourceIds = this.getContributors().stream()
              .map(DepositContributor::getPersonId)
              .collect(Collectors.toList());
    } else if (type == Language.class && this.getLanguageId() != null) {
      subResourceIds = CollectionTool.initList(this.getLanguageId());
    } else if (type == SubmissionAgreement.class && this.getSubmissionAgreementId() != null) {
      subResourceIds = CollectionTool.initList(this.getSubmissionAgreementId());
    } else if (type == License.class && this.getLicenseId() != null) {
      subResourceIds = CollectionTool.initList(this.getLicenseId());
    } else if (type == OrganizationalUnit.class && this.getOrganizationalUnitId() != null) {
      subResourceIds = CollectionTool.initList(this.getOrganizationalUnitId());
    } else if (type == ArchivalInfoPackage.class && this.getCollection() != null) {
      subResourceIds = this.getCollection();
    } else if (type == SubmissionPolicy.class && this.getSubmissionPolicyId() != null) {
      subResourceIds = CollectionTool.initList(this.getSubmissionPolicyId());
    } else if (type == PreservationPolicy.class && this.getPreservationPolicyId() != null) {
      subResourceIds = CollectionTool.initList(this.getPreservationPolicyId());
    } else if (type == ArchiveType.class && this.getArchiveTypeId() != null) {
      subResourceIds = CollectionTool.initList(this.getArchiveTypeId());
    } else if (type == null) {
      throw new IllegalArgumentException("type should not be null");
    }

    return subResourceIds;
  }

  /****************************************************************/

  public String getTitle() {
    return this.title;
  }

  public boolean hasEmbargo() {
    return (this.embargo != null && this.embargo.hasEmbargoDefined());
  }

  public String getAdditionalFieldsValues() {
    return this.additionalFieldsValues;
  }

  public void setAdditionalFieldsValues(String additionalFieldsValues) {
    this.additionalFieldsValues = additionalFieldsValues;
  }

  public String getAdditionalFieldsFormId() {
    return this.additionalFieldsFormId;
  }

  public void setAdditionalFieldsFormId(String additionalFieldsFormId) {
    this.additionalFieldsFormId = additionalFieldsFormId;
  }

  public String getIsIdenticalTo() {
    return this.isIdenticalTo;
  }

  public String getIsReferencedBy() {
    return this.isReferencedBy;
  }

  public String getIsObsoletedBy() {
    return this.isObsoletedBy;
  }

  /****************************************************************/

  @PrePersist
  @PreUpdate
  @Override
  public void init() {
    if (this.status == null) {
      this.status = this.getDefaultStatus();
    }
    if (this.publicationDate == null) {
      this.publicationDate = LocalDate.now();
    }
    if (this.access == null) {
      this.access = Access.PUBLIC;
    }
    if (this.dataSensitivity == null) {
      this.dataSensitivity = DataTag.BLUE;
    }

    if (this.agreement == null) {
      this.agreement = false;
    }
    if (this.metadataVersion == null) {
      this.metadataVersion = DLCMMetadataVersion.getDefaultVersion();
    }
  }

  public boolean isCollection() {
    return (!this.collection.isEmpty());
  }

  @Override
  public boolean isDeletable() {
    return (this.getStatus() == DepositStatus.REJECTED || this.getStatus() == DepositStatus.IN_ERROR
            || this.getStatus() == DepositStatus.IN_PROGRESS || this.getStatus() == DepositStatus.IN_VALIDATION
            || this.getStatus() == DepositStatus.COMPLETED || this.getStatus() == DepositStatus.CLEANED);
  }

  @Override
  @JsonIgnore
  public boolean isModifiable() {
    return (this.getStatus() != DepositStatus.APPROVED && this.getStatus() != DepositStatus.REJECTED
            && this.getStatus() != DepositStatus.COMPLETED && this.getStatus() != DepositStatus.CLEANED);
  }

  public boolean getContainsUpdatedMetadata() {
    return this.containsUpdatedMetadata;
  }

  @Override
  public String managedBy() {
    return ModuleName.PREINGEST;
  }

  @Override
  public <T> boolean removeItem(T t) {

    if (t instanceof Person) {
      return this.removeContributor((Person) t);
    } else if (t instanceof DepositDataFile) {
      return this.removeDataFile((DepositDataFile) t);
    } else if (t instanceof ArchivalInfoPackage) {
      return this.removeAip((ArchivalInfoPackage) t);
    } else {

      if (t instanceof Language) {
        this.setLanguageId(null);
      } else if (t instanceof SubmissionAgreement) {
        this.setSubmissionAgreementId(null);
      } else if (t instanceof License) {
        this.setLicenseId(null);
      } else if (t instanceof OrganizationalUnit) {
        this.setOrganizationalUnitId(null);
      } else if (t instanceof SubmissionPolicy) {
        this.setSubmissionPolicyId(null);
      } else if (t instanceof PreservationPolicy) {
        this.setPreservationPolicyId(null);
      } else if (t instanceof ArchiveType) {
        this.setArchiveTypeId(null);
      } else {
        throw new UnsupportedOperationException("Missing implementation 'removeItem' method (" + t.getClass().getSimpleName() + ")");
      }

      return true;
    }

  }

  public void setAccess(Access access) {
    this.access = access;
  }

  public void setDataSensitivity(DataTag dataSensitivity) {
    this.dataSensitivity = dataSensitivity;
  }

  public void setAgreement(Boolean agreement) {
    this.agreement = agreement;
  }

  public void setCollectionBegin(OffsetDateTime collectionBegin) {
    this.collectionBegin = collectionBegin;
  }

  public void setCollectionEnd(OffsetDateTime collectionEnd) {
    this.collectionEnd = collectionEnd;
  }

  public void setDescription(String description) {
    this.description = description;
  }

  public void setDoi(String doi) {
    this.doi = doi;
  }

  public void setEmbargo(EmbargoInfo embargo) {
    this.embargo = embargo;
  }

  public void setEmbargoAccess(Access access) {
    this.getEmbargo().setAccess(access);
  }

  public void setEmbargoMonths(Integer months) {
    this.getEmbargo().setMonths(months);
  }

  public void setEmbargoStartDate(OffsetDateTime startDate) {
    this.getEmbargo().setStartDate(startDate);
  }

  public void setErrorStatusWithMessage(String statusMessage) {
    this.setStatus(DepositStatus.IN_ERROR);
    this.setStatusMessage(statusMessage);
  }

  public void setKeywords(List<String> keywords) {
    this.keywords = keywords;
  }

  public void setLanguage(Language language) {
    this.language = language;
  }

  public void setLanguageId(String languageId) {
    this.languageId = languageId;
  }

  public void setLicenseId(String licenseId) {
    this.licenseId = licenseId;
  }

  public void setArchiveTypeId(String archiveTypeId) {
    this.archiveTypeId = archiveTypeId;
  }

  public void setArchiveType(ArchiveType archiveType) {
    this.archiveType = archiveType;
  }

  public void setMetadataVersion(DLCMMetadataVersion metadataVersion) {
    this.metadataVersion = metadataVersion;
  }

  public void setOrganizationalUnit(OrganizationalUnit organizationalUnit) {
    this.organizationalUnit = organizationalUnit;
  }

  public void setOrganizationalUnitId(String organizationalUnitId) {
    this.organizationalUnitId = organizationalUnitId;
  }

  public void setPreparationId(String preparationId) {
    this.preparationId = preparationId;
  }

  @JsonIgnore
  public void setPreservationPolicy(PreservationPolicy preservationPolicy) {
    this.preservationPolicy = preservationPolicy;
  }

  public void setPreservationPolicyId(String preservationPolicyId) {
    this.preservationPolicyId = preservationPolicyId;
  }

  public void setPublicationDate(LocalDate publicationDate) {
    this.publicationDate = publicationDate;
  }

  public void setSip(URI sip) {
    this.sip = sip;
  }

  public void setSipId(String sipId) {
    this.sipId = sipId;
  }

  public void setStatus(DepositStatus status) {
    if (this.getStatus() != status) {
      this.status = status;
      this.isStatusDirty = true;
    }
  }

  /**
   * Return true if the status has been modified since the object has been created
   *
   * @return
   */
  @JsonIgnore
  public boolean isStatusDirty() {
    return this.isStatusDirty;
  }

  public void setStatusMessage(String message) {
    this.statusMessage = StringTool.truncateWithEllipsis(message, SolidifyConstants.DB_LONG_STRING_LENGTH);
  }

  public void setSubmissionAgreementId(String submissionAgreementId) {
    this.submissionAgreementId = submissionAgreementId;
  }

  @JsonIgnore
  public void setSubmissionPolicy(SubmissionPolicy submissionPolicy) {
    this.submissionPolicy = submissionPolicy;
  }

  public void setSubmissionPolicyId(String submissionPolicyId) {
    this.submissionPolicyId = submissionPolicyId;
  }

  public void setTitle(String title) {
    this.title = title;
  }

  public void setContainsUpdatedMetadata(boolean containsUpdatedMetadataMetadata) {
    this.containsUpdatedMetadata = containsUpdatedMetadataMetadata;
  }

  public void setIsIdenticalTo(String doi) {
    this.isIdenticalTo = doi;
  }

  public void setIsReferencedBy(String doi) {
    this.isReferencedBy = doi;
  }

  public void setIsObsoletedBy(String doi) {
    this.isObsoletedBy = doi;
  }

  private boolean addAip(ArchivalInfoPackage aip) {
    this.collection.add(aip.getResId());
    return true;
  }

  private boolean addContributor(Person person) {
    final DepositContributor contributor = new DepositContributor();
    contributor.setDeposit(this);
    contributor.setPersonId(person.getResId());
    contributor.setPosition(this.contributors.size() * 10);
    return this.contributors.add(contributor);
  }

  @Override
  public boolean addDataFile(DepositDataFile depositDataFile) {
    if (this.getStatus() == DepositStatus.IN_PROGRESS ||
            this.getStatus() == DepositStatus.EDITING_METADATA) {
      return this.dataFiles.add(depositDataFile);
    } else {
      return false;
    }
  }

  private boolean removeAip(ArchivalInfoPackage t) {
    return !this.collection.contains(t.getResId()) || this.collection.remove(t.getResId());
  }

  private boolean removeContributor(Person person) {
    return this.contributors.removeIf(depositContributor -> depositContributor.getPersonId().equals(person.getResId()));
  }

  private boolean removeDataFile(DepositDataFile t) {
    if (this.getStatus() == DepositStatus.IN_PROGRESS) {
      return this.dataFiles.remove(t);
    } else {
      return false;
    }
  }

  @Override
  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (!(o instanceof Deposit))
      return false;
    if (!super.equals(o))
      return false;
    Deposit deposit = (Deposit) o;
    return Objects.equals(this.dataFiles, deposit.dataFiles) &&
            Objects.equals(this.preparationId, deposit.preparationId) &&
            Objects.equals(this.publicationDate, deposit.publicationDate) &&
            Objects.equals(this.sip, deposit.sip) &&
            Objects.equals(this.sipId, deposit.sipId) &&
            this.status == deposit.status &&
            Objects.equals(this.title, deposit.title);
  }

  @Override
  public int hashCode() {
    return Objects.hash(this.additionalFieldsValues, this.additionalFieldsFormId, this.access, this.dataSensitivity, this.agreement,
            this.collectionBegin, this.collectionEnd, this.description, this.doi, this.embargo, this.keywords, this.languageId,
            this.licenseId, this.metadataVersion, this.organizationalUnitId, this.preparationId, this.preservationPolicyId, this.publicationDate,
            this.sip, this.sipId, this.status, this.statusMessage, this.submissionAgreementId, this.submissionPolicyId, this.title,
            this.dataFiles);
  }
}
