/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - ArchiveACL.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.settings;

import static ch.dlcm.DLCMRestFields.AIP_ID_FIELD;
import static ch.dlcm.DLCMRestFields.ORGANIZATIONAL_UNIT_FIELD;
import static ch.dlcm.DLCMRestFields.RES_ID_FIELD;
import static ch.dlcm.DLCMRestFields.USER_FIELD;
import static ch.unige.solidify.SolidifyConstants.DB_DATE_LENGTH;
import static ch.unige.solidify.SolidifyConstants.DB_ID_LENGTH;

import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.Arrays;
import java.util.List;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.Table;
import javax.persistence.UniqueConstraint;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import com.fasterxml.jackson.annotation.JsonFormat;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonProperty.Access;

import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.OrganizationalUnitAwareResource;
import ch.dlcm.model.security.User;
import ch.dlcm.rest.ModuleName;

@Entity
@Table(name = "orgUnitArchiveAcl", uniqueConstraints = { @UniqueConstraint(columnNames = { "aipId", DLCMConstants.DB_USER_ID }) })
public class ArchiveACL extends ResourceNormalized implements OrganizationalUnitAwareResource {

  @NotNull
  @Size(max = DB_ID_LENGTH)
  @Column(length = DB_ID_LENGTH)
  private String aipId;

  @NotNull
  @ManyToOne(targetEntity = OrganizationalUnit.class)
  @JoinColumn(name = DLCMConstants.DB_ORG_UNIT_ID, referencedColumnName = DLCMConstants.DB_RES_ID)
  private OrganizationalUnit organizationalUnit;

  @NotNull
  @ManyToOne(targetEntity = User.class)
  @JoinColumn(name = DLCMConstants.DB_USER_ID, referencedColumnName = DLCMConstants.DB_RES_ID)
  private User user;

  @NotNull
  @JsonFormat(shape = JsonFormat.Shape.STRING, pattern = StringTool.DATE_TIME_FORMAT)
  @Column(length = DB_DATE_LENGTH) // Used by MySQL5ExtendedDialect
  private OffsetDateTime expiration;

  public String getAipId() {
    return this.aipId;
  }

  public void setAipId(String aipId) {
    this.aipId = aipId;
  }

  public OrganizationalUnit getOrganizationalUnit() {
    return this.organizationalUnit;
  }

  public void setOrganizationalUnit(OrganizationalUnit organizationalUnit) {
    this.organizationalUnit = organizationalUnit;
  }

  public User getUser() {
    return this.user;
  }

  public void setUser(User user) {
    this.user = user;
  }

  public OffsetDateTime getExpiration() {
    return this.expiration;
  }

  public void setExpiration(OffsetDateTime expiration) {
    this.expiration = expiration;
  }

  @JsonProperty(access = Access.READ_ONLY)
  public boolean isExpired() {
    if (this.getExpiration() == null) {
      return false;
    }
    return (this.getExpiration().isBefore(OffsetDateTime.now(ZoneOffset.UTC)));
  }

  @Override
  public void init() {
    if (this.expiration == null) {
      // Be default 3 month access
      this.expiration = OffsetDateTime.now(ZoneOffset.UTC).plusMonths(3);
    }
  }

  @Override
  public String managedBy() {
    return ModuleName.ADMIN;
  }

  @Override
  public List<String> getNonUpdatableFields() {
    return Arrays.asList(RES_ID_FIELD, AIP_ID_FIELD, USER_FIELD, ORGANIZATIONAL_UNIT_FIELD);
  }

  @Override
  @JsonIgnore
  public String getOrganizationalUnitId() {
    return this.getOrganizationalUnit().getResId();
  }

}
