/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - Institution.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.settings;

import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.Entity;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToMany;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.converter.WeakManyToManyConverter;
import ch.unige.solidify.model.Image;
import ch.unige.solidify.model.ImageInterface;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.rest.Tool;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.IdentifierType;
import ch.dlcm.model.RorInterface;
import ch.dlcm.model.security.Role;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;

@Entity
public class Institution extends ResourceNormalized implements SubResourceContainer, ImageInterface, RorInterface {
  @Size(max = SolidifyConstants.DB_LONG_STRING_LENGTH)
  private String description;

  @Convert(converter = WeakManyToManyConverter.class)
  private List<String> emailSuffixes = new ArrayList<>();

  @NotNull
  @Column(unique = true)
  @Size(min = 1)
  private String name;

  @ManyToMany(mappedBy = "institutions")
  @JsonIgnore
  private List<OrganizationalUnit> organizationalUnits;

  @JsonIgnore
  @OneToMany(mappedBy = InstitutionPersonRole.PATH_TO_INSTITUTION, cascade = CascadeType.ALL, orphanRemoval = true)
  private List<InstitutionPersonRole> personRoles = new ArrayList<>();

  @ManyToMany(mappedBy = "institutions")
  @JsonIgnore
  private List<Person> people;

  private URL url;

  @Column(unique = true)
  private String rorId;

  @OneToOne(cascade = CascadeType.ALL, orphanRemoval = true)
  @JoinColumn(name = DLCMConstants.DB_LOGO_ID, referencedColumnName = DLCMConstants.DB_RES_ID)
  private InstitutionLogo logo;

  @Transient
  private Map<String, String> identifiers = new HashMap<>();

  @Override
  public <T> boolean addItem(T t) {
    if (t instanceof Person) {
      return this.addPerson((Person) t);
    }
    if (t instanceof OrganizationalUnit) {
      return this.addOrganizationalUnit((OrganizationalUnit) t);
    }
    return false;
  }

  @Override
  public void addLinks(WebMvcLinkBuilder linkBuilder, boolean mainRes, boolean subResOnly) {
    super.addLinks(linkBuilder, mainRes, subResOnly);
    if (mainRes) {
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.PERSON).withRel(ResourceName.PERSON));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.ORG_UNIT).withRel(ResourceName.ORG_UNIT));
      this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.UPLOAD_LOGO).withRel(DLCMActionName.UPLOAD_LOGO));
      this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.DOWNLOAD_LOGO).withRel(DLCMActionName.DOWNLOAD_LOGO));
      this.add(Tool.filter(
                      linkBuilder.slash(this.getRorId()).toUriComponentsBuilder(),
                      DLCMConstants.IDENTIFIER_TYPE_PARAM,
                      IdentifierType.NAME.toString())
              .withRel(IdentifierType.NAME.toString().toLowerCase()));
      if (!StringTool.isNullOrEmpty(this.getRorId())) {
        this.add(
                Tool.filter(
                                linkBuilder.slash(this.getRorId()).toUriComponentsBuilder(),
                                DLCMConstants.IDENTIFIER_TYPE_PARAM,
                                IdentifierType.ROR_ID.toString())
                        .withRel(IdentifierType.ROR_ID.toString().toLowerCase()));
      }
    }
  }

  public void setOrganizationalUnit(List<OrganizationalUnit> organizationalUnits) {
    this.organizationalUnits = organizationalUnits;
  }

  public boolean addOrganizationalUnit(OrganizationalUnit organizationalUnit) {
    final boolean result = this.organizationalUnits.add(organizationalUnit);
    if (!organizationalUnit.getInstitutions().contains(this)) {
      organizationalUnit.addInstitution(this);
    }
    return result;
  }

  public boolean addPerson(Person p) {
    final boolean result = this.people.add(p);
    if (!p.getInstitutions().contains(this)) {
      p.addInstitution(this);
    }
    return result;
  }

  public String getDescription() {
    return this.description;
  }

  public List<String> getEmailSuffixes() {
    return this.emailSuffixes;
  }

  public String getName() {
    return this.name;
  }

  public List<OrganizationalUnit> getOrganizationalUnits() {
    return this.organizationalUnits;
  }

  public List<Person> getPeople() {
    return this.people;
  }

  public URL getUrl() {
    return this.url;
  }

  @Override
  public String getRorId() {
    return this.rorId;
  }

  @Override
  public void init() {
    // Do nothing
  }

  @Override
  public String managedBy() {
    return ModuleName.ADMIN;
  }

  @Override
  public <T> boolean removeItem(T t) {
    if (t instanceof Person) {
      return this.removePerson((Person) t);
    }
    if (t instanceof OrganizationalUnit) {
      return this.removeOrganizationalUnit((OrganizationalUnit) t);
    }
    return false;
  }

  public boolean removeOrganizationalUnit(OrganizationalUnit organizationalUnit) {
    final boolean result = this.organizationalUnits.remove(organizationalUnit);
    if (organizationalUnit.getInstitutions().contains(this)) {
      organizationalUnit.removeInstitution(this);
    }
    return result;
  }

  public boolean removePerson(Person p) {
    final boolean result = this.people.remove(p);
    if (p.getInstitutions().contains(this)) {
      p.removeInstitution(this);
    }
    return result;
  }

  public void setDescription(String description) {
    this.description = description;
  }

  public void setEmailSuffixes(List<String> emailSuffixes) {
    this.emailSuffixes = emailSuffixes;
  }

  public void setName(String name) {
    this.name = name;
  }

  public void setUrl(URL url) {
    this.url = url;
  }

  @Override
  public void setRorId(String rorId) {
    this.rorId = rorId;
  }

  public InstitutionLogo getLogo() {
    return this.logo;
  }

  public void setLogo(InstitutionLogo logo) {
    this.logo = logo;
  }

  @Override
  @JsonIgnore
  public Image getNewIfNeededImage() {
    if (this.getLogo() == null) {
      this.setLogo(new InstitutionLogo());
    }
    return this.getLogo();
  }

  @Override
  @JsonIgnore
  public Image getImage() {
    return this.getLogo();
  }

  @Override
  public void setImage(Image logo) {
    this.setLogo((InstitutionLogo) logo);
  }

  @Override
  public Map<String, String> getIdentifiers() {
    return this.identifiers;
  }

  public List<InstitutionPersonRole> getPersonRoles() {
    return personRoles;
  }

  public boolean addInstitutionPersonRole(Person person, Role role) {
    final InstitutionPersonRole institutionPersonRole = new InstitutionPersonRole();
    institutionPersonRole.setInstitution(this);
    institutionPersonRole.setRole(role);
    institutionPersonRole.setPerson(person);
    return this.personRoles.add(institutionPersonRole);
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + Objects.hash(description, emailSuffixes, name, rorId, url);
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (!super.equals(obj))
      return false;
    if (getClass() != obj.getClass())
      return false;
    Institution other = (Institution) obj;
    return Objects.equals(this.description, other.description) && Objects.equals(this.emailSuffixes, other.emailSuffixes)
            && Objects.equals(this.name, other.name) && Objects.equals(this.rorId, other.rorId) && Objects.equals(this.url, other.url);
  }

}
