/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - License.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.settings;

import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.JoinColumn;
import javax.persistence.OneToOne;
import javax.validation.constraints.NotNull;

import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import com.fasterxml.jackson.annotation.JsonAnySetter;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonValue;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.model.Image;
import ch.unige.solidify.model.ImageInterface;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.rest.Tool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.model.IdentifierType;
import ch.dlcm.rest.ModuleName;

/**
 * Structure based on open licenses. http://licenses.opendefinition.org
 */
@Entity
public class License extends ResourceNormalized implements ImageInterface {

  public enum ConformanceStatus {
    APPROVED("approved"), EMPTY(""), NOT_REVIEWED("not reviewed"), REJECTED("rejected");

    private String value;

    ConformanceStatus(String value) {
      this.value = value;
    }

    @JsonValue
    public String getValue() {
      return this.value;
    }
  }

  // ======
  // end internals types
  // ======
  public enum LicenseStatus {
    ACTIVE("active"), RETIRED("retired"), SUPERSEDED("superseded");

    private String value;

    LicenseStatus(String value) {
      this.value = value;
    }

    @JsonValue
    public String getValue() {
      return this.value;
    }
  }

  public static class OpenLicenseList {
    private Map<String, OpenLicense> openLicenses = new HashMap<>();

    public List<Map<String, Object>> getOpenLicensesMaps() {
      List<Map<String, Object>> openLicensesMaps = new ArrayList<>();
      for (OpenLicense openLicense : this.openLicenses.values()) {
        openLicensesMaps.add(openLicense.convertToMap());
      }
      return openLicensesMaps;
    }

    @JsonAnySetter
    public void setOpenLicenses(String name, OpenLicense openLicense) {
      this.openLicenses.put(name, openLicense);
    }
  }

  /**
   * Open License proxy structure. Used to import all.json format data available on open license
   * foundation
   */
  public static class OpenLicense extends License {

    @Override
    @JsonProperty("domain_content")
    public Boolean getDomainContent() {
      return super.getDomainContent();
    }

    @Override
    @JsonProperty("domain_data")
    public Boolean getDomainData() {
      return super.getDomainData();
    }

    @Override
    @JsonProperty("domain_software")
    public Boolean getDomainSoftware() {
      return super.getDomainSoftware();
    }

    @Override
    @JsonProperty("is_generic")
    public Boolean getIsGeneric() {
      return super.getIsGeneric();
    }

    @Override
    @JsonProperty("od_conformance")
    public ConformanceStatus getOdConformance() {
      return super.getOdConformance();
    }

    @Override
    @JsonProperty("id")
    public String getOpenLicenseId() {
      return super.getOpenLicenseId();
    }

    @Override
    @JsonProperty("osd_conformance")
    public ConformanceStatus getOsdConformance() {
      return super.getOsdConformance();
    }

    public Map<String, Object> convertToMap() {
      final Map<String, Object> licenseMap = new ObjectMapper().convertValue(this, Map.class);
      licenseMap.put(DLCMConstants.OPEN_LICENSE_ID_FIELD, this.getOpenLicenseId());
      licenseMap.put(DLCMConstants.DOMAIN_CONTENT_FIELD, this.getDomainContent());
      licenseMap.put(DLCMConstants.DOMAIN_DATA_FIELD, this.getDomainData());
      licenseMap.put(DLCMConstants.DOMAIN_SOFTWARE_FIELD, this.getDomainSoftware());
      licenseMap.put(DLCMConstants.IS_GENERIC_FIELD, this.getIsGeneric());
      licenseMap.put(DLCMConstants.OD_CONFORMANCE_FIELD, this.getOdConformance());
      licenseMap.put(DLCMConstants.OSD_CONFORMANCE_FIELD, this.getOsdConformance());
      return licenseMap;
    }
  }

  private Boolean domainContent;
  private Boolean domainData;
  private Boolean domainSoftware;
  private String family;
  private Boolean isGeneric;
  private String maintainer;

  @Enumerated(EnumType.STRING)
  private ConformanceStatus odConformance;

  @NotNull
  @Column(unique = true, nullable = false)
  private String openLicenseId;

  @Enumerated(EnumType.STRING)
  private ConformanceStatus osdConformance;

  @Enumerated(EnumType.STRING)
  private LicenseStatus status;

  @NotNull
  private String title;

  private URL url;

  @OneToOne(cascade = CascadeType.ALL, orphanRemoval = true)
  @JoinColumn(name = DLCMConstants.DB_LOGO_ID, referencedColumnName = DLCMConstants.DB_RES_ID)
  private LicenseLogo logo;

  public String getDescription() {
    return this.getOpenLicenseId() + " (" + this.getTitle() + ")";
  }

  public Boolean getDomainContent() {
    return this.domainContent;
  }

  public Boolean getDomainData() {
    return this.domainData;
  }

  public Boolean getDomainSoftware() {
    return this.domainSoftware;
  }

  public String getFamily() {
    return this.family;
  }

  public Boolean getIsGeneric() {
    return this.isGeneric;
  }

  public String getMaintainer() {
    return this.maintainer;
  }

  public ConformanceStatus getOdConformance() {
    return this.odConformance;
  }

  /***********************************************************/

  public String getOpenLicenseId() {
    return this.openLicenseId;
  }

  public ConformanceStatus getOsdConformance() {
    return this.osdConformance;
  }

  public LicenseStatus getStatus() {
    return this.status;
  }

  public String getTitle() {
    return this.title;
  }

  public URL getUrl() {
    return this.url;
  }

  public LicenseLogo getLogo() {
    return this.logo;
  }

  @Override
  public void init() {
    // Do nothing
  }

  @Override
  public String managedBy() {
    return ModuleName.ADMIN;
  }

  public void setDomainContent(Boolean domainContent) {
    this.domainContent = domainContent;
  }

  public void setDomainData(Boolean domainData) {
    this.domainData = domainData;
  }

  public void setDomainSoftware(Boolean domainSoftware) {
    this.domainSoftware = domainSoftware;
  }

  public void setFamily(String family) {
    this.family = family;
  }

  public void setIsGeneric(Boolean isGeneric) {
    this.isGeneric = isGeneric;
  }

  public void setMaintainer(String maintainer) {
    this.maintainer = maintainer;
  }

  public void setOdConformance(ConformanceStatus odConformance) {
    this.odConformance = odConformance;
  }

  public void setOpenLicenseId(String openLicenseId) {
    this.openLicenseId = openLicenseId;
  }

  public void setOsdConformance(ConformanceStatus osdConformance) {
    this.osdConformance = osdConformance;
  }

  // ==============
  // Internals types
  // ==============

  public void setStatus(LicenseStatus status) {
    this.status = status;
  }

  public void setTitle(String title) {
    this.title = title;
  }

  public void setUrl(URL url) {
    this.url = url;
  }

  public void setLogo(LicenseLogo logo) {
    this.logo = logo;
  }

  @Override
  @JsonIgnore
  public Image getNewIfNeededImage() {
    if (this.getLogo() == null) {
      this.setLogo(new LicenseLogo());
    }
    return this.getLogo();
  }

  @Override
  @JsonIgnore
  public Image getImage() {
    return this.getLogo();
  }

  @Override
  public void setImage(Image logo) {
    this.setLogo((LicenseLogo) logo);
  }

  @Override
  public void addLinks(WebMvcLinkBuilder linkBuilder, boolean mainRes, boolean subResOnly) {
    super.addLinks(linkBuilder, mainRes, subResOnly);
    if (mainRes) {
      this.add(
              Tool.filter(
                      linkBuilder.slash(this.getOpenLicenseId()).toUriComponentsBuilder(),
                      DLCMConstants.IDENTIFIER_TYPE_PARAM,
                      IdentifierType.SPDX_ID.toString())
                      .withRel(IdentifierType.SPDX_ID.toString().toLowerCase()));
    }
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + Objects.hash(domainContent, domainData, domainSoftware, family, isGeneric, maintainer, odConformance,
            openLicenseId, osdConformance, status, title, url);
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (!super.equals(obj))
      return false;
    if (getClass() != obj.getClass())
      return false;
    License other = (License) obj;
    return Objects.equals(this.domainContent, other.domainContent) && Objects.equals(this.domainData, other.domainData)
            && Objects.equals(this.domainSoftware, other.domainSoftware) && Objects.equals(this.family, other.family)
            && Objects.equals(this.isGeneric, other.isGeneric) && Objects.equals(this.maintainer, other.maintainer)
            && this.odConformance == other.odConformance && Objects.equals(this.openLicenseId, other.openLicenseId)
            && this.osdConformance == other.osdConformance && this.status == other.status && Objects.equals(this.title, other.title)
            && Objects.equals(this.url, other.url);
  }

}
