/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Model - OrganizationalUnit.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.model.settings;

import java.net.URL;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.JoinColumn;
import javax.persistence.JoinTable;
import javax.persistence.ManyToMany;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.PrePersist;
import javax.persistence.PreUpdate;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.persistence.UniqueConstraint;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;

import com.fasterxml.jackson.annotation.JsonFormat;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.converter.KeywordsConverter;
import ch.unige.solidify.model.Image;
import ch.unige.solidify.model.ImageInterface;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.listener.RegistryListener;
import ch.dlcm.model.OrganizationalUnitAwareResource;
import ch.dlcm.model.policies.DisseminationPolicy;
import ch.dlcm.model.policies.OrganizationalUnitDisseminationPolicy;
import ch.dlcm.model.policies.OrganizationalUnitPreservationPolicy;
import ch.dlcm.model.policies.OrganizationalUnitSubmissionPolicy;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.security.Role;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ModuleName;
import ch.dlcm.rest.ResourceName;

@Entity
@Table(name = "orgUnit")
@EntityListeners({ RegistryListener.class })
public class OrganizationalUnit extends ResourceNormalized
        implements SubResourceContainer, ImageInterface, OrganizationalUnitAwareResource {

  @JsonFormat(shape = JsonFormat.Shape.STRING, pattern = StringTool.DATE_FORMAT)
  private LocalDate closingDate;

  @Transient
  private DisseminationPolicy defaultDisseminationPolicy;

  @ManyToOne(targetEntity = License.class)
  private License defaultLicense;

  @Transient
  private PreservationPolicy defaultPreservationPolicy;

  @Transient
  private SubmissionPolicy defaultSubmissionPolicy;

  @Size(max = SolidifyConstants.DB_LONG_STRING_LENGTH)
  private String description;

  @JsonIgnore
  @OneToMany(mappedBy = OrganizationalUnitDisseminationPolicy.PATH_TO_ORG_UNIT, cascade = CascadeType.ALL, orphanRemoval = true)
  private List<OrganizationalUnitDisseminationPolicy> disseminationPolicies = new ArrayList<>();

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "orgUnitFundingAgencies",
             joinColumns = { @JoinColumn(name = DLCMConstants.DB_ORG_UNIT_ID) },
             inverseJoinColumns = { @JoinColumn(name = DLCMConstants.DB_AGENCY_ID) },
             uniqueConstraints = { @UniqueConstraint(columnNames = { DLCMConstants.DB_ORG_UNIT_ID, DLCMConstants.DB_AGENCY_ID }) })
  private List<FundingAgency> fundingAgencies;

  @JsonProperty(access = JsonProperty.Access.WRITE_ONLY)
  @ManyToMany
  @JoinTable(name = "orgUnitInstitutions",
             joinColumns = { @JoinColumn(name = DLCMConstants.DB_ORG_UNIT_ID) },
             inverseJoinColumns = { @JoinColumn(name = DLCMConstants.DB_INSTITUTION_ID) },
             uniqueConstraints = { @UniqueConstraint(columnNames = { DLCMConstants.DB_ORG_UNIT_ID, DLCMConstants.DB_INSTITUTION_ID }) })
  private List<Institution> institutions;

  private Boolean isEmpty;

  @Convert(converter = KeywordsConverter.class)
  private List<String> keywords;

  @NotNull
  @Column(unique = true)
  @Size(min = 1)
  private String name;

  @JsonFormat(shape = JsonFormat.Shape.STRING, pattern = StringTool.DATE_FORMAT)
  @NotNull
  private LocalDate openingDate;

  @JsonIgnore
  @OneToMany(mappedBy = OrganizationalUnitPersonRole.PATH_TO_ORG_UNIT, cascade = CascadeType.ALL, orphanRemoval = true)
  private List<OrganizationalUnitPersonRole> personRoles = new ArrayList<>();

  @JsonIgnore
  @OneToMany(mappedBy = OrganizationalUnitPreservationPolicy.PATH_TO_ORG_UNIT, cascade = CascadeType.ALL, orphanRemoval = true)
  private List<OrganizationalUnitPreservationPolicy> preservationPolicies = new ArrayList<>();

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "orgUnitResearchDomains",
             joinColumns = { @JoinColumn(name = DLCMConstants.DB_ORG_UNIT_ID) },
             inverseJoinColumns = { @JoinColumn(name = DLCMConstants.DB_RESEARCH_DOMAIN_ID) },
             uniqueConstraints = { @UniqueConstraint(columnNames = { DLCMConstants.DB_ORG_UNIT_ID, DLCMConstants.DB_RESEARCH_DOMAIN_ID }) })
  private List<ResearchDomain> researchDomains;

  @JsonIgnore
  @OneToMany(mappedBy = OrganizationalUnitSubmissionPolicy.PATH_TO_ORG_UNIT, cascade = CascadeType.ALL, orphanRemoval = true)
  private List<OrganizationalUnitSubmissionPolicy> submissionPolicies = new ArrayList<>();

  private URL url;

  @OneToOne(cascade = CascadeType.ALL, orphanRemoval = true)
  @JoinColumn(name = DLCMConstants.DB_LOGO_ID, referencedColumnName = DLCMConstants.DB_RES_ID)
  private OrganizationalUnitLogo logo;

  @OneToMany(mappedBy = "organizationalUnit", cascade = CascadeType.ALL, orphanRemoval = true)
  private List<AdditionalFieldsForm> additionalFieldsForms = new ArrayList<>();

  @JsonIgnore
  @OneToMany(mappedBy = "organizationalUnit", cascade = CascadeType.ALL, orphanRemoval = true)
  private List<ArchiveACL> archiveACLs = new ArrayList<>();

  public boolean addFundingAgency(FundingAgency fundingAgency) {
    final boolean result = this.fundingAgencies.add(fundingAgency);
    if (!fundingAgency.getOrganizationalUnits().contains(this)) {
      fundingAgency.addOrganizationalUnit(this);
    }
    return result;
  }

  public boolean addInstitution(Institution institution) {
    final boolean result = this.institutions.add(institution);
    if (!institution.getOrganizationalUnits().contains(this)) {
      institution.addOrganizationalUnit(this);
    }
    return result;
  }

  public boolean addResearchDomain(ResearchDomain researchDomain) {
    final boolean result = this.researchDomains.add(researchDomain);
    if (!researchDomain.getOrganizationalUnits().contains(this)) {
      researchDomain.addOrganizationalUnit(this);
    }
    return result;
  }

  @Override
  public <T> boolean addItem(T t) {
    if (t instanceof ResearchDomain) {
      return this.addResearchDomain((ResearchDomain) t);
    }
    if (t instanceof Institution) {
      return this.addInstitution((Institution) t);
    }
    if (t instanceof FundingAgency) {
      return this.addFundingAgency((FundingAgency) t);
    }
    if (t instanceof PreservationPolicy) {
      return this.addPreservationPolicy((PreservationPolicy) t);
    }
    if (t instanceof SubmissionPolicy) {
      return this.addSubmissionPolicy((SubmissionPolicy) t);
    }
    if (t instanceof DisseminationPolicy) {
      return this.addDisseminationPolicy((DisseminationPolicy) t);
    }
    if (t instanceof AdditionalFieldsForm) {
      return this.addAdditionalFieldsForm((AdditionalFieldsForm) t);
    }
    if (t instanceof ArchiveACL) {
      return this.addArchiveACL((ArchiveACL) t);
    }
    return false;
  }

  @Override
  public void addLinks(WebMvcLinkBuilder linkBuilder, boolean mainRes, boolean subResOnly) {
    super.addLinks(linkBuilder, mainRes, subResOnly);
    if (mainRes) {
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.INSTITUTION).withRel(ResourceName.INSTITUTION));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.PRES_POLICY).withRel(ResourceName.PRES_POLICY));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.SUB_POLICY).withRel(ResourceName.SUB_POLICY));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.DIS_POLICY).withRel(ResourceName.DIS_POLICY));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.PERSON).withRel(ResourceName.PERSON));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.FUNDING_AGENCY).withRel(ResourceName.FUNDING_AGENCY));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.ADDITIONAL_FIELDS_FORMS).withRel(ResourceName.ADDITIONAL_FIELDS_FORMS));
      this.add(linkBuilder.slash(this.getResId()).slash(ResourceName.ARCHIVE_ACL).withRel(ResourceName.ARCHIVE_ACL));
      this.add(linkBuilder.slash(this.getResId()).slash(ActionName.CLOSE).withRel(ActionName.CLOSE));
      this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.UPLOAD_LOGO).withRel(DLCMActionName.UPLOAD_LOGO));
      this.add(linkBuilder.slash(this.getResId()).slash(DLCMActionName.DOWNLOAD_LOGO).withRel(DLCMActionName.DOWNLOAD_LOGO));

    }
  }

  @Override
  public boolean equals(final Object other) {
    if (!(other instanceof OrganizationalUnit)) {
      return false;
    }
    if (!super.equals(other)) {
      return false;
    }
    final OrganizationalUnit castOther = (OrganizationalUnit) other;
    return Objects.equals(this.name, castOther.name) && Objects.equals(this.description, castOther.description) && Objects
            .equals(this.institutions, castOther.institutions) && Objects.equals(this.personRoles, castOther.personRoles);
  }

  public LocalDate getClosingDate() {
    return this.closingDate;
  }

  @JsonIgnore
  public String getDefault() {
    return Role.DEFAULT_ROLE_ID;
  }

  public DisseminationPolicy getDefaultDisseminationPolicy() {
    if (this.defaultDisseminationPolicy == null) {
      for (final OrganizationalUnitDisseminationPolicy policy : this.getDisseminationPolicies()) {
        if (policy.getDefaultPolicy()) {
          this.defaultDisseminationPolicy = policy.getDisseminationPolicy();
        }
      }
    }
    return this.defaultDisseminationPolicy;
  }

  public License getDefaultLicense() {
    return this.defaultLicense;
  }

  public PreservationPolicy getDefaultPreservationPolicy() {
    if (this.defaultPreservationPolicy == null) {
      for (final OrganizationalUnitPreservationPolicy policy : this.getPreservationPolicies()) {
        if (policy.getDefaultPolicy()) {
          this.defaultPreservationPolicy = policy.getPreservationPolicy();
        }
      }
    }
    return this.defaultPreservationPolicy;
  }

  public SubmissionPolicy getDefaultSubmissionPolicy() {
    if (this.defaultSubmissionPolicy == null) {
      for (final OrganizationalUnitSubmissionPolicy policy : this.getSubmissionPolicies()) {
        if (policy.getDefaultPolicy()) {
          this.defaultSubmissionPolicy = policy.getSubmissionPolicy();
        }
      }
    }
    return this.defaultSubmissionPolicy;
  }

  public String getDescription() {
    return this.description;
  }

  @JsonIgnore
  public List<OrganizationalUnitDisseminationPolicy> getDisseminationPolicies() {
    return this.disseminationPolicies;
  }

  public List<FundingAgency> getFundingAgencies() {
    return this.fundingAgencies;
  }

  public List<Institution> getInstitutions() {
    return this.institutions;
  }

  public Boolean getIsEmpty() {
    return this.isEmpty;
  }

  public List<String> getKeywords() {
    return this.keywords;
  }

  public String getName() {
    return this.name;
  }

  public LocalDate getOpeningDate() {
    return this.openingDate;
  }

  public List<OrganizationalUnitPersonRole> getPersonRoles() {
    return this.personRoles;
  }

  @JsonIgnore
  public List<OrganizationalUnitPreservationPolicy> getPreservationPolicies() {
    return this.preservationPolicies;
  }

  public List<ResearchDomain> getResearchDomains() {
    return this.researchDomains;
  }

  public OrganizationalUnitLogo getLogo() {
    return this.logo;
  }

  @JsonIgnore
  public List<OrganizationalUnitSubmissionPolicy> getSubmissionPolicies() {
    return this.submissionPolicies;
  }

  public URL getUrl() {
    return this.url;
  }

  @Override
  public int hashCode() {
    return Objects.hash(super.hashCode(), this.name, this.description, this.institutions, this.personRoles);
  }

  @PrePersist
  @PreUpdate
  @Override
  public void init() {
    if (this.openingDate == null) {
      this.openingDate = LocalDate.now();
    }
    if (this.isEmpty == null) {
      this.isEmpty = true;
    }
  }

  public boolean isOpen() {
    final LocalDate now = LocalDate.now();
    return (this.openingDate != null && !this.openingDate.isAfter(now) && (this.closingDate == null || this.closingDate.isAfter(now)));
  }

  @Override
  @JsonIgnore
  public boolean isDeletable() {
    return this.isEmpty;
  }

  @Override
  public String managedBy() {
    return ModuleName.ADMIN;
  }

  public boolean removeFundingAgency(FundingAgency fundingAgency) {
    final boolean result = this.fundingAgencies.remove(fundingAgency);
    if (fundingAgency.getOrganizationalUnits().contains(this)) {
      fundingAgency.removeOrganizationalUnit(this);
    }
    return result;
  }

  public boolean removeArchiveACL(ArchiveACL archiveACL) {
    return this.archiveACLs.remove(archiveACL);
  }

  @Override
  public <T> boolean removeItem(T t) {

    if (t instanceof Person) {
      return this.removePerson((Person) t);
    }

    if (t instanceof Institution) {
      return this.removeInstitution((Institution) t);
    }

    if (t instanceof FundingAgency) {
      return this.removeFundingAgency((FundingAgency) t);
    }
    if (t instanceof PreservationPolicy) {
      return this.removePreservationPolicy((PreservationPolicy) t);
    }
    if (t instanceof OrganizationalUnitPreservationPolicy) {
      return this.removePreservationPolicy(((OrganizationalUnitPreservationPolicy) t).getPreservationPolicy());
    }
    if (t instanceof SubmissionPolicy) {
      return this.removeSubmissionPolicy((SubmissionPolicy) t);
    }
    if (t instanceof OrganizationalUnitSubmissionPolicy) {
      return this.removeSubmissionPolicy(((OrganizationalUnitSubmissionPolicy) t).getSubmissionPolicy());
    }
    if (t instanceof DisseminationPolicy) {
      return this.removeDisseminationPolicy((DisseminationPolicy) t);
    }
    if (t instanceof OrganizationalUnitDisseminationPolicy) {
      return this.removeDisseminationPolicy(((OrganizationalUnitDisseminationPolicy) t).getDisseminationPolicy());
    }
    if (t instanceof ResearchDomain) {
      return this.removeResearchDomain(((ResearchDomain) t));
    }
    if (t instanceof ArchiveACL) {
      return this.removeArchiveACL(((ArchiveACL) t));
    }
    return false;
  }

  public void setLogo(OrganizationalUnitLogo logo) {
    // Don't set dirty field, logo is updated through dedicated endpoints
    this.logo = logo;
  }

  public void setClosingDate(LocalDate closingDate) {
    this.closingDate = closingDate;
  }

  public void setDefaultDisseminationPolicy(DisseminationPolicy defaultPolicy) {
    this.defaultDisseminationPolicy = defaultPolicy;
  }

  public void setDefaultLicense(License defaultLicense) {
    this.defaultLicense = defaultLicense;
  }

  public void setDefaultPreservationPolicy(PreservationPolicy defaultPolicy) {
    this.defaultPreservationPolicy = defaultPolicy;
  }

  public void setDefaultSubmissionPolicy(SubmissionPolicy defaultPolicy) {
    this.defaultSubmissionPolicy = defaultPolicy;
  }

  public void setDescription(String description) {
    this.description = description;
  }

  public void setInstitutions(List<Institution> institutions) {
    this.institutions = institutions;
  }

  public void setIsEmpty(Boolean isEmpty) {
    this.isEmpty = isEmpty;
  }

  public void setKeywords(List<String> keywords) {
    this.keywords = keywords;
  }

  public void setName(String name) {
    this.name = name;
  }

  public void setOpeningDate(LocalDate openingDate) {
    this.openingDate = openingDate;
  }

  public void setUrl(URL url) {
    this.url = url;
  }

  @Override
  public String toString() {
    return super.toString() + " (name=" + this.getName() + ")";
  }

  protected boolean addDisseminationPolicy(DisseminationPolicy policy) {
    final OrganizationalUnitDisseminationPolicy orgUnitDisseminationPolicy = new OrganizationalUnitDisseminationPolicy();
    orgUnitDisseminationPolicy.setOrganizationalUnit(this);
    orgUnitDisseminationPolicy.setDisseminationPolicy(policy);
    if (this.disseminationPolicies.isEmpty()) {
      orgUnitDisseminationPolicy.setDefaultPolicy(true);
    }
    return this.disseminationPolicies.add(orgUnitDisseminationPolicy);
  }

  protected boolean addPreservationPolicy(PreservationPolicy policy) {
    final OrganizationalUnitPreservationPolicy orgUnitPreservationPolicy = new OrganizationalUnitPreservationPolicy();
    orgUnitPreservationPolicy.setOrganizationalUnit(this);
    orgUnitPreservationPolicy.setPreservationPolicy(policy);
    if (this.preservationPolicies.isEmpty()) {
      orgUnitPreservationPolicy.setDefaultPolicy(true);
    }
    return this.preservationPolicies.add(orgUnitPreservationPolicy);
  }

  protected boolean addAdditionalFieldsForm(AdditionalFieldsForm additionalFieldsForm) {
    additionalFieldsForm.setOrganizationalUnit(this);
    return this.additionalFieldsForms.add(additionalFieldsForm);
  }

  protected boolean addArchiveACL(ArchiveACL archiveACL) {
    archiveACL.init();
    archiveACL.setOrganizationalUnit(this);
    return this.archiveACLs.add(archiveACL);
  }

  protected boolean addSubmissionPolicy(SubmissionPolicy policy) {
    final OrganizationalUnitSubmissionPolicy orgUnitSubmissionPolicy = new OrganizationalUnitSubmissionPolicy();
    orgUnitSubmissionPolicy.setOrganizationalUnit(this);
    orgUnitSubmissionPolicy.setSubmissionPolicy(policy);
    if (this.submissionPolicies.isEmpty()) {
      orgUnitSubmissionPolicy.setDefaultPolicy(true);
    }
    return this.submissionPolicies.add(orgUnitSubmissionPolicy);
  }

  protected boolean removeDisseminationPolicy(DisseminationPolicy policy) {
    final DisseminationPolicy policyFinal = policy;
    return this.disseminationPolicies.removeIf(oupPredicate -> oupPredicate
            .getCompositeKey().getDisseminationPolicy() == policyFinal);
  }

  protected boolean removeInstitution(Institution institution) {
    final boolean result = this.institutions.remove(institution);
    if (institution.getOrganizationalUnits().contains(this)) {
      institution.removeOrganizationalUnit(this);
    }
    return result;
  }

  protected boolean removePerson(Person person) {
    final Person personFinal = person;
    return this.personRoles.removeIf(ouprPredicate -> ouprPredicate.getCompositeKey()
            .getPerson() == personFinal);
  }

  protected boolean removePreservationPolicy(PreservationPolicy policy) {
    final PreservationPolicy policyFinal = policy;
    return this.preservationPolicies.removeIf(oupPredicate -> oupPredicate
            .getCompositeKey().getPreservationPolicy() == policyFinal);
  }

  protected boolean removeSubmissionPolicy(SubmissionPolicy policy) {
    final SubmissionPolicy policyFinal = policy;
    return this.submissionPolicies.removeIf(oupPredicate -> oupPredicate
            .getCompositeKey()
            .getSubmissionPolicy() == policyFinal);
  }

  protected boolean removeResearchDomain(ResearchDomain researchDomain) {
    final boolean result = this.researchDomains.remove(researchDomain);
    if (researchDomain.getOrganizationalUnits().contains(this)) {
      researchDomain.removeOrganizationalUnit(this);
    }
    return result;
  }

  public boolean addPersonRole(Person person, Role role) {
    final OrganizationalUnitPersonRole organizationalUnitPersonRole = new OrganizationalUnitPersonRole();
    organizationalUnitPersonRole.setOrganizationalUnit(this);
    organizationalUnitPersonRole.setRole(role);
    organizationalUnitPersonRole.setPerson(person);
    return this.personRoles.add(organizationalUnitPersonRole);
  }

  public boolean removePersonRole(Person person, Role role) {
    return this.personRoles.removeIf(oupr -> oupr.getOrganizationalUnit().equals(this) &&
            oupr.getPerson().equals(person) &&
            oupr.getRole().equals(role));
  }

  @Override
  @JsonIgnore
  public String getOrganizationalUnitId() {
    return this.getResId();
  }

  @Override
  @JsonIgnore
  public Image getNewIfNeededImage() {
    if (this.getLogo() == null) {
      this.setLogo(new OrganizationalUnitLogo());
    }
    return this.getLogo();
  }

  @Override
  public void setImage(Image logo) {
    this.setLogo((OrganizationalUnitLogo) logo);
  }

  @Override
  @JsonIgnore
  public Image getImage() {
    return this.getLogo();
  }

  @Override
  public List<String> getPatchableSubResources() {
    return List.of("logo");
  }

}
