/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM PreIngest - DepositService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.business;

import static ch.dlcm.DLCMRestFields.STATUS_FIELD;
import static ch.dlcm.model.preingest.Deposit.DepositStatus.IN_VALIDATION;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;

import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.exception.SolidifyFileDeleteException;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.exception.SolidifyUnmodifiableException;
import ch.unige.solidify.model.Language;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.service.CompositeResourceService;
import ch.unige.solidify.service.HttpRequestInfoProvider;
import ch.unige.solidify.service.RemoteResourceService;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.JSONTool;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.ValidationTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.controller.PreIngestController;
import ch.dlcm.message.DepositMessage;
import ch.dlcm.model.Access;
import ch.dlcm.model.ComplianceLevel;
import ch.dlcm.model.DataCategory;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.dto.DepositDto;
import ch.dlcm.model.notification.NotificationType;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.policies.PreservationPolicy;
import ch.dlcm.model.policies.SubmissionPolicy;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.ArchiveType;
import ch.dlcm.model.settings.License;
import ch.dlcm.model.settings.OrganizationalUnit;
import ch.dlcm.model.settings.Person;
import ch.dlcm.model.settings.SubmissionAgreement;
import ch.dlcm.model.xml.dlcm.v3.mets.RelationType;
import ch.dlcm.repository.DepositRepository;
import ch.dlcm.service.HistoryService;
import ch.dlcm.service.rest.fallback.FallbackAdditionalFieldsFormRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveTypeRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackLanguageRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackLicenseRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackOrganizationalUnitRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackPersonRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackPreservationPolicyRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackSubmissionAgreementRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackSubmissionPolicyRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackUserRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedArchiveMetadataRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedNotificationRemoteResourceService;
import ch.dlcm.specification.DepositSpecification;
import ch.dlcm.specification.dto.DepositDtoSpecification;

@Service
@ConditionalOnBean(PreIngestController.class)
public class DepositService extends CompositeResourceService<Deposit> {

  private static final Logger log = LoggerFactory.getLogger(DepositService.class);

  private static final String AIP_BAD_ORGANIZATIONAL_UNIT = "validation.deposit.collection.aip.badOrganizationalUnit";
  private static final String VALIDATION_NOTIFICATION_DEPOSIT = "validation.notification.deposit";
  private static final String NOTIFICATION_DEPOSIT_APPROVED = "preingest.notification.deposit.approved";
  private static final String NOTIFICATION_DEPOSIT_CREATED = "preingest.notification.deposit.created";
  private static final String NOTIFICATION_DEPOSIT_ERROR = "preingest.notification.deposit.error";
  private static final String BAD_EMBARGO = "validation.access.badEmbargoAccess";
  private static final String INVALID_EMBARGO = "validation.embargo.invalid";
  private static final String COLLECTION_BEGIN_NOT_BEFORE_COLLECTION_END = "validation.dates.collectionBeginMustBeBeforeCollectionEnd";
  private static final String DATA_SENSITIVITY = "validation.deposit.dataSensitivity";
  private static final String ARCHIVE_TYPE_INCONSISTANT = "validation.deposit.archiveType.inconsistant";
  private static final String ARCHIVE_TYPE_WRONG_CONFIG = "validation.deposit.archiveType.wrongConfig";
  private static final String DATA_FILES_NOT_READY_OR_NOT_CLEANED = "validation.deposit.datafile.notReadyOrNotCleaned";
  private static final String HAS_DATA_FILES_BEFORE_VALIDATION = "validation.deposit.status.containsFilesBeforeInValidation";
  private static final String NOT_EXIST = "validation.resource.notexist";
  private static final String UNIT_CLOSED = "validation.organizationalUnit.closed";
  private static final String INVALID_ADDITIONAL_FIELDS_VALUES = "validation.deposit.additionalFieldsValues.invalidJson";
  private static final String ADDITIONAL_FIELDS_FORM_REQUIRED_WHEN_ADDITIONAL_FIELDS_VALUES_PRESENT = "validation.deposit.additionalFieldsValues.formIdMissing";
  private static final String NOT_SUPPORTED = "validation.deposit.notSupported";

  private final DLCMProperties dlcmProperties;
  private final SpecificationPermissionsFilter<Deposit> specificationPermissionFilter;
  private final HttpRequestInfoProvider httpRequestInfoProvider;
  private final FallbackArchivalInfoPackageRemoteResourceService aipService;
  private final FallbackLanguageRemoteResourceService languageService;
  private final FallbackLicenseRemoteResourceService licenseService;
  private final FallbackOrganizationalUnitRemoteResourceService organizationalUnitService;
  private final FallbackPersonRemoteResourceService personService;
  private final FallbackPreservationPolicyRemoteResourceService preservationPolicyService;
  private final FallbackSubmissionAgreementRemoteResourceService submissionAgreementService;
  private final FallbackSubmissionPolicyRemoteResourceService submissionPolicyService;
  private final FallbackUserRemoteResourceService userService;
  private final FallbackOrganizationalUnitRemoteResourceService orgUnitService;
  private final FallbackAdditionalFieldsFormRemoteResourceService additionalFieldsFormService;
  private final FallbackArchiveTypeRemoteResourceService archiveTypeService;
  private final TrustedNotificationRemoteResourceService trustedNotificationRemoteResourceService;
  private final TrustedArchiveMetadataRemoteResourceService trustedArchiveMetadataRemoteService;
  private final HistoryService historyService;

  public DepositService(DLCMProperties dlcmProperties,
          SpecificationPermissionsFilter<Deposit> specificationPermissionFilter,
          HttpRequestInfoProvider httpRequestInfoProvider,
          FallbackLanguageRemoteResourceService languageService,
          FallbackArchivalInfoPackageRemoteResourceService aipService,
          FallbackLicenseRemoteResourceService licenseService,
          FallbackOrganizationalUnitRemoteResourceService organizationalUnitService,
          FallbackPersonRemoteResourceService personService,
          FallbackPreservationPolicyRemoteResourceService preservationPolicyService,
          FallbackUserRemoteResourceService userService,
          FallbackSubmissionAgreementRemoteResourceService submissionAgreementService,
          FallbackSubmissionPolicyRemoteResourceService submissionPolicyService,
          FallbackOrganizationalUnitRemoteResourceService trustedOrganizationalUnitRemoteResourceService,
          FallbackAdditionalFieldsFormRemoteResourceService additionalFieldsFormService,
          FallbackArchiveTypeRemoteResourceService archiveTypeService,
          TrustedNotificationRemoteResourceService trustedNotificationRemoteResourceService,
          TrustedArchiveMetadataRemoteResourceService trustedArchiveMetadataRemoteService,
          HistoryService historyService) {
    this.dlcmProperties = dlcmProperties;
    this.specificationPermissionFilter = specificationPermissionFilter;
    this.httpRequestInfoProvider = httpRequestInfoProvider;
    this.languageService = languageService;
    this.aipService = aipService;
    this.licenseService = licenseService;
    this.organizationalUnitService = organizationalUnitService;
    this.personService = personService;
    this.preservationPolicyService = preservationPolicyService;
    this.userService = userService;
    this.submissionAgreementService = submissionAgreementService;
    this.submissionPolicyService = submissionPolicyService;
    this.orgUnitService = trustedOrganizationalUnitRemoteResourceService;
    this.additionalFieldsFormService = additionalFieldsFormService;
    this.archiveTypeService = archiveTypeService;
    this.trustedNotificationRemoteResourceService = trustedNotificationRemoteResourceService;
    this.trustedArchiveMetadataRemoteService = trustedArchiveMetadataRemoteService;
    this.historyService = historyService;
  }

  /**
   * Delete Deposit from database
   * <p>
   * The use of transaction with propagation to Propagation.REQUIRES_NEW ensure the associated
   * DepositDataFile are deleted as well when the call to super.delete() is over
   * <p>
   * Note: As associated DepositDatafiles are deleted from database and disk as well, this process may
   * be time-consuming. Therefore deleting a Deposit should better be done by setting its status to
   * DELETING, which then trigger the call to the delete() method asynchronously.
   *
   * @param depositId
   */
  @Override
  @Transactional(propagation = Propagation.REQUIRES_NEW)
  public void delete(String depositId) {
    super.delete(depositId);
  }

  public void checkArchiveIsEditable(String depositId) {
    final String cannotEditMessage = this.messageService.get("preingest.deposit.error.cannot_edit_deposit", new Object[] { depositId });
    try {
      final String aipId = this.trustedArchiveMetadataRemoteService.getAipIdFromDepositId(depositId);
      final ArchivalInfoPackage aip = this.aipService.findOne(aipId);
      if (aip.getInfo().getStatus() != PackageStatus.COMPLETED) {
        throw new SolidifyUnmodifiableException(cannotEditMessage);
      }
    } catch (SolidifyResourceNotFoundException e) {
      throw new SolidifyUnmodifiableException(cannotEditMessage);
    }
  }

  public void putInMetadataEditingCollections(String depositId) {
    try {
      String aipId = this.trustedArchiveMetadataRemoteService.getAipIdFromDepositId(depositId);
      List<String> collections = this.trustedArchiveMetadataRemoteService.getCollections(aipId, true);
      if (!collections.isEmpty()) {
        for (String collectionId : collections) {
          Deposit collectionDeposit = this.findOne(this.trustedArchiveMetadataRemoteService.getDepositId(collectionId));
          if (this.statusUpdateIsValid(collectionDeposit, DepositStatus.EDITING_METADATA)) {
            collectionDeposit.setStatus(DepositStatus.EDITING_METADATA);
            this.save(collectionDeposit);
            log.info("Editing metadata for collection deposit: resId={} status={}", collectionDeposit.getResId(), collectionDeposit.getStatus());
          }
        }
      }
    } catch (Exception e) {
      throw new SolidifyRuntimeException(this.messageService.get("preingest.deposit.error.cannot_edit_collection", new Object[] { depositId }), e);
    }
  }

  public boolean checkIfCancelApplicableForCollection(String depositId) {
    try {
      Deposit deposit = this.findOne(depositId);
      if (!deposit.isCollection()) {
        return true;
      }
      for (final String aipId : deposit.getCollection()) {
        Deposit aipDeposit = this.findOne(this.trustedArchiveMetadataRemoteService.getDepositId(aipId));
        if (aipDeposit.getStatus() != DepositStatus.COMPLETED && aipDeposit.getStatus() != DepositStatus.CLEANED) {
          return false;
        }
      }
      return true;
    } catch (Exception e) {
      throw new SolidifyRuntimeException(this.messageService.get("preingest.deposit.error.cannot_check_collection", new Object[] { depositId }), e);
    }
  }

  public Deposit deleteDepositSip(String depositId) {
    try {
      Deposit deposit = this.findOne(depositId);
      if (deposit.getSip() != null) {
        FileTool.deleteFile(deposit.getSip());
        deposit.setSip(null);
        deposit = this.save(deposit);
        log.info("SIP package purged for deposit: resId={} status={}", deposit.getResId(), deposit.getStatus());
      }
      return deposit;
    } catch (IOException e) {
      throw new SolidifyRuntimeException(this.messageService.get("preingest.deposit.error.cannotpurged", new Object[] { depositId }), e);
    }
  }

  /**
   * Delete the folder containing deposit's files if it exists
   *
   * @param depositId
   */
  public void deleteDepositFolderFromDisk(String depositId) {
    final Path depositPath = Paths.get(this.dlcmProperties.getPreingestLocation()).resolve(depositId);
    if (FileTool.isFolder(depositPath)) {
      if (FileTool.deleteFolder(depositPath)) {
        log.info("Deposit's folder deleted: {}", depositPath);
      } else {
        throw new SolidifyFileDeleteException("Cannot delete deposit's folder '" + depositPath + "'");
      }
    } else {
      log.info("Deposit's folder cannot be deleted as it does not exist: {}", depositPath);
    }
  }

  /**
   * Delete notifications related to this Deposit on admin module
   *
   * @param depositId
   */
  public void deleteDepositNotifications(String depositId) {
    try {
      log.info("Will delete notifications for deposit {}", depositId);
      this.trustedNotificationRemoteResourceService.deleteNotificationsOnDeposit(depositId);
    } catch (SolidifyResourceNotFoundException e) {
      log.info("No notification for deposit {} to delete", depositId);
    }
  }

  @Override
  public Deposit save(Deposit deposit) {

    // If not set use keywords from organizational unit
    if (deposit.getKeywords() == null && deposit.getOrganizationalUnitId() != null) {
      final List<String> orgUnitKeywords = this.orgUnitService.findOne(deposit.getOrganizationalUnitId())
              .getKeywords();
      deposit.setKeywords(orgUnitKeywords);
    }

    final Deposit result = super.save(deposit);

    if (deposit.isStatusDirty()) {
      User updaterUser = this.userService.findByExternalUid(deposit.getLastUpdate().getWho());
      String message;
      switch (deposit.getStatus()) {
        case IN_VALIDATION:
          if (Boolean.TRUE.equals(deposit.getSubmissionPolicy().getSubmissionApproval())) {
            message = this.messageService.get(VALIDATION_NOTIFICATION_DEPOSIT);
            this.trustedNotificationRemoteResourceService.createNotification(updaterUser, message, NotificationType.VALIDATE_DEPOSIT_REQUEST,
                    deposit.getOrganizationalUnit(), deposit.getResId());
          }
          break;
        case APPROVED:
          message = this.messageService.get(NOTIFICATION_DEPOSIT_APPROVED);
          this.trustedNotificationRemoteResourceService.createdDepositApprovedNotificationForCreators(deposit, updaterUser, message);
          break;
        case IN_ERROR:
          message = this.messageService.get(NOTIFICATION_DEPOSIT_ERROR, new Object[] { deposit.getResId() });
          this.trustedNotificationRemoteResourceService.createNotification(updaterUser, message, NotificationType.IN_ERROR_DEPOSIT_INFO,
                  deposit.getOrganizationalUnit(), deposit.getResId());
          break;
        case COMPLETED:
          // Check if the deposit was already completed to avoid recreating notifications and resending emails
          if (!this.historyService.hasBeenCompleted(deposit.getResId())
                  || this.historyService.hasBeenUpdatingMetadata(deposit.getResId())) {
            message = this.messageService.get(NOTIFICATION_DEPOSIT_CREATED, new Object[] { deposit.getResId() });
            this.trustedNotificationRemoteResourceService.createDepositCompletedNotifications(deposit, updaterUser, message);
          }
          break;
        default:
          break;
      }
    }

    final boolean statusHasChangedAndNeedsProcessing = deposit.isStatusDirty() &&
            (deposit.getStatus().equals(DepositStatus.APPROVED) ||
                    deposit.getStatus().equals(DepositStatus.CLEANING) ||
                    deposit.getStatus().equals(DepositStatus.DELETING) ||
                    deposit.getStatus().equals(DepositStatus.CANCEL_EDITING_METADATA) ||
                    deposit.getStatus().equals(DepositStatus.CHECKING_COMPLIANCE) ||
                    deposit.getStatus().equals(DepositStatus.CHECKING_COMPLIANCE_CLEANED) ||
                    (deposit.getStatus().equals(DepositStatus.CHECKED) && deposit.getPreparationId() != null));

    if (statusHasChangedAndNeedsProcessing) {
      log.info("DepositService will send a DepositMessage for deposit '{}' with status '{}'", deposit.getResId(), deposit.getStatus());
      boolean bigPackage = this.getSize(deposit.getResId()) > this.dlcmProperties.getParameters().getFileSizeLimit().toBytes();
      SolidifyEventPublisher.getPublisher().publishEvent(new DepositMessage(deposit.getResId(), bigPackage));
    }
    return result;
  }

  /**
   * Check if a deposit status can be manually updated to the new given status, according to the
   * deposit's management workflow. Some status transitions (e.g. from APPROVED to CHECK) are
   * considered here as invalid, because even if they exist in the workflow, they are automatically
   * done, and should not be forced manually.
   *
   * @param deposit The deposit to check
   * @param newStatus The new status to apply
   * @return True if the status update is valid
   */
  public boolean statusUpdateIsValid(Deposit deposit, DepositStatus newStatus) {

    final DepositStatus currentStatus = deposit.getStatus();

    switch (currentStatus) {
      case IN_PROGRESS:
        return (newStatus == IN_VALIDATION && this.needsValidation(deposit))
                || (newStatus == DepositStatus.APPROVED && !this.needsValidation(deposit));

      case IN_VALIDATION:
        return newStatus == DepositStatus.IN_PROGRESS
                || newStatus == DepositStatus.EDITING_METADATA
                || newStatus == DepositStatus.APPROVED
                || newStatus == DepositStatus.REJECTED;

      case IN_ERROR:
      case REJECTED:
        return newStatus == DepositStatus.IN_PROGRESS
                || newStatus == DepositStatus.EDITING_METADATA;

      case COMPLETED:
      case CLEANED:
        return newStatus == DepositStatus.EDITING_METADATA;

      case EDITING_METADATA:
        return newStatus == DepositStatus.CLEANED
                || newStatus == DepositStatus.CANCEL_EDITING_METADATA
                || (newStatus == IN_VALIDATION && this.needsValidation(deposit))
                || (newStatus == DepositStatus.APPROVED && !this.needsValidation(deposit));

      case APPROVED:
      case SUBMITTED:
      case CHECKED:
      default:
        return false;
    }
  }

  @Override
  public void validateLinkedResources(Deposit item, BindingResult errors) {

    /*
     * OrganizationalUnit
     */
    if (!StringTool.isNullOrEmpty(item.getOrganizationalUnitId())) {
      try {
        // Need a trusted remote service because this method can be called by a Job
        final OrganizationalUnit subitem = this.orgUnitService.findOne(item.getOrganizationalUnitId());
        if (!subitem.isOpen()) {
          errors.addError(new FieldError(item.getClass().getSimpleName(), DLCMConstants.ORG_UNIT_ID_FIELD,
                  this.messageService.get(UNIT_CLOSED, new Object[] { item.getOrganizationalUnitId() })));
        }
      } catch (final SolidifyResourceNotFoundException e) {
        errors.addError(new FieldError(item.getClass().getSimpleName(), DLCMConstants.ORG_UNIT_ID_FIELD,
                this.messageService.get(NOT_EXIST, new Object[] { item.getOrganizationalUnitId() })));
      }
    }

    /*
     * Contributors
     */
    if (!item.getContributorIds().isEmpty()) {
      for (final String contributorId : item.getContributorIds()) {
        if (!StringTool.isNullOrEmpty(contributorId)) {
          try {
            this.personService.findOne(contributorId);
          } catch (final SolidifyResourceNotFoundException e) {
            errors.addError(new FieldError(item.getClass().getSimpleName(), "person",
                    this.messageService.get(NOT_EXIST, new Object[] { contributorId })));
          }
        }
      }
    }

    /*
     * SubmissionAgreement
     */
    if (!StringTool.isNullOrEmpty(item.getSubmissionAgreementId())) {
      try {
        this.submissionAgreementService.findOne(item.getSubmissionAgreementId());
      } catch (final SolidifyResourceNotFoundException e) {
        errors.addError(new FieldError(item.getClass().getSimpleName(), "submissionAgreementId",
                this.messageService.get(NOT_EXIST, new Object[] { item.getSubmissionAgreementId() })));
      }
    }

    /*
     * License
     */
    if (!StringTool.isNullOrEmpty(item.getLicenseId())) {
      try {
        this.licenseService.findOne(item.getLicenseId());
      } catch (final SolidifyResourceNotFoundException e) {
        errors.addError(new FieldError(item.getClass().getSimpleName(), "licenseId",
                this.messageService.get(NOT_EXIST, new Object[] { item.getLicenseId() })));
      }
    }

    /*
     * Language
     */
    if (!StringTool.isNullOrEmpty(item.getLanguageId())) {
      try {
        this.languageService.findOne(item.getLanguageId());
      } catch (final SolidifyResourceNotFoundException e) {
        errors.addError(new FieldError(item.getClass().getSimpleName(), "languageId",
                this.messageService.get(NOT_EXIST, new Object[] { item.getLanguageId() })));
      }
    }

    /*
     * Archive Type
     */
    if (!StringTool.isNullOrEmpty(item.getArchiveTypeId())) {
      try {
        this.archiveTypeService.findOne(item.getArchiveTypeId());
      } catch (final SolidifyResourceNotFoundException e) {
        errors.addError(new FieldError(item.getClass().getSimpleName(), "archiveTypeId",
                this.messageService.get(NOT_EXIST, new Object[] { item.getArchiveTypeId() })));
      }
    }

    /*
     * AdditionalFieldsForm
     */
    if (!StringTool.isNullOrEmpty(item.getAdditionalFieldsFormId())) {
      try {
        this.additionalFieldsFormService
                .findOneWithCache(item.getOrganizationalUnitId(), item.getAdditionalFieldsFormId());
      } catch (final SolidifyResourceNotFoundException e) {
        errors.addError(new FieldError(item.getClass().getSimpleName(), "additionalFieldsFormId",
                this.messageService.get(NOT_EXIST, new Object[] { item.getAdditionalFieldsFormId() })));
      }
    }
  }

  @Override
  protected Specification<Deposit> addSpecificationFilter(Specification<Deposit> spec) {
    return this.specificationPermissionFilter.addPermissionsFilter(spec, this.httpRequestInfoProvider.getPrincipal());
  }

  @Override
  protected void beforeValidate(Deposit deposit) {
    if (deposit.getAccess() == Access.PUBLIC && StringTool.isNullOrEmpty(deposit.getLicenseId())) {
      OrganizationalUnit orgUnit = this.organizationalUnitService.findOne(deposit.getOrganizationalUnitId());
      License orgUnitDefaultLicense = orgUnit.getDefaultLicense();
      if (orgUnitDefaultLicense != null) {
        deposit.setLicenseId(orgUnitDefaultLicense.getResId());
      } else {
        deposit.setLicenseId(this.dlcmProperties.getParameters().getDefaultLicense());
      }
    }
    if (deposit.getAdditionalFieldsValues() != null && deposit.getAdditionalFieldsValues().isEmpty()) {
      deposit.setAdditionalFieldsValues(null);
    }
    if (deposit.getAdditionalFieldsValues() == null
            || (deposit.getAdditionalFieldsFormId() != null && deposit.getAdditionalFieldsFormId().isEmpty())) {
      deposit.setAdditionalFieldsFormId(null);
    }
  }

  /*****************************************/

  @Override
  protected <S extends Resource> RemoteResourceService<S> getResourceService(Class<S> type) {

    if (type == Language.class) {
      return (RemoteResourceService<S>) this.languageService;
    } else if (type == Person.class) {
      return (RemoteResourceService<S>) this.personService;
    } else if (type == OrganizationalUnit.class) {
      return (RemoteResourceService<S>) this.organizationalUnitService;
    } else if (type == SubmissionAgreement.class) {
      return (RemoteResourceService<S>) this.submissionAgreementService;
    } else if (type == License.class) {
      return (RemoteResourceService<S>) this.licenseService;
    } else if (type == ArchivalInfoPackage.class) {
      return (RemoteResourceService<S>) this.aipService;
    } else if (type == SubmissionPolicy.class) {
      return (RemoteResourceService<S>) this.submissionPolicyService;
    } else if (type == PreservationPolicy.class) {
      return (RemoteResourceService<S>) this.preservationPolicyService;
    } else if (type == ArchiveType.class) {
      return (RemoteResourceService<S>) this.archiveTypeService;
    }

    throw new UnsupportedOperationException("Missing resource service for type " + type.getName());
  }

  @Override
  protected void validateItemSpecificRules(Deposit deposit, BindingResult errors) {

    /*
     * Check embargo definition
     */
    if (deposit.getEmbargo() != null) {
      if (deposit.getEmbargo().getAccess() != null && (deposit.getEmbargo().getMonths() == null || deposit.getEmbargo().getMonths() == 0)) {
        errors.addError(new FieldError(deposit.getClass().getSimpleName(), "embargo.months", this.messageService.get(INVALID_EMBARGO)));
      }
      if (deposit.getEmbargo().getMonths() != null && deposit.getEmbargo().getMonths() > 0 && deposit.getEmbargo().getAccess() == null) {
        errors.addError(new FieldError(deposit.getClass().getSimpleName(), "embargo.access", this.messageService.get(INVALID_EMBARGO)));
      }
    }

    /*
     * Check embargo values
     */
    if (deposit.hasEmbargo() && ((deposit.getAccess() == deposit.getEmbargoAccess()) || (deposit.getAccess() == Access.RESTRICTED
            && deposit.getEmbargoAccess() == Access.PUBLIC) || (deposit.getAccess() == Access.CLOSED && deposit.getEmbargo() != null))) {

      errors.addError(new FieldError(deposit.getClass().getSimpleName(), "access",
              this.messageService.get(BAD_EMBARGO, new Object[] { deposit.getAccess().name(), deposit.getEmbargoAccess().name() })));
    }

    /*
     * Check collection dates
     */
    if (deposit.getCollectionBegin() != null && deposit.getCollectionEnd() != null
            && deposit.getCollectionBegin().isAfter(deposit.getCollectionEnd())) {

      errors.addError(new FieldError(deposit.getClass().getSimpleName(), "collectionEnd", this.messageService
              .get(COLLECTION_BEGIN_NOT_BEFORE_COLLECTION_END, new Object[] { deposit.getCollectionBegin(), deposit.getCollectionEnd() })));
    }

    /*
     * Check if data sensitivity (DataTag) is compatible with access level
     */
    if (!deposit.getDataSensitivity().isAccessLevelAllowed(deposit.getAccess())) {
      errors.addError(new FieldError(deposit.getClass().getSimpleName(), "dataSensitivity",
              this.messageService.get(DATA_SENSITIVITY, new Object[] { deposit.getDataSensitivity(), deposit.getAccess() })));

    }
    /*
     * Check if archive type is compatible with deposit only if there are datafiles/aip
     */
    if (!deposit.getDataFiles().isEmpty() || !deposit.getCollection().isEmpty()) {
      // Default
      if (StringTool.isNullOrEmpty(deposit.getArchiveTypeId())) {
        String archiveTypeId = this.dlcmProperties.getParameters().getDefaultMasterArchiveType();
        if (deposit.isCollection()) {
          archiveTypeId = ArchiveType.COLLECTION_ID;
        }
        List<ArchiveType> archiveTypes = this.archiveTypeService.get(archiveTypeId).getData();
        if (archiveTypes.isEmpty()) {
          errors.addError(new FieldError(deposit.getClass().getSimpleName(), "archiveTypeId",
                  this.messageService.get(ARCHIVE_TYPE_WRONG_CONFIG)));
        } else {
          deposit.setArchiveTypeId(archiveTypes.get(0).getResId());
          deposit.setArchiveType(archiveTypes.get(0));
        }
      } else {
        // Check if collection compatible
        deposit.setArchiveType(this.archiveTypeService.findOneWithCache(deposit.getArchiveTypeId()));
        String masterTypeId = deposit.getArchiveType().getMasterType().getResId();
        if ((deposit.isCollection() && !masterTypeId.equals(ArchiveType.COLLECTION_ID))
                || (!deposit.isCollection() && masterTypeId.equals(ArchiveType.COLLECTION_ID))) {
          errors.addError(new FieldError(deposit.getClass().getSimpleName(), "archiveTypeId",
                  this.messageService.get(ARCHIVE_TYPE_INCONSISTANT, new Object[] { deposit.getArchiveType().getTypeName() })));
        }
      }
    }

    /*
     * Check that all AIPs in collection belong to the same Organizational Unit as the Deposit itself
     * (collections with AIPs from multiple Organizational Unit are not allowed)
     */
    if (!deposit.getCollection().isEmpty()) {
      for (final String aipId : deposit.getCollection()) {
        final ArchivalInfoPackage aip = this.aipService.findOne(aipId);
        if (!aip.getInfo().getOrganizationalUnitId().equals(deposit.getOrganizationalUnitId())) {
          final OrganizationalUnit badOrgUnit = this.organizationalUnitService.findOne(aip.getInfo().getOrganizationalUnitId());
          errors.addError(new FieldError(deposit.getClass().getSimpleName(), "collection",
                  this.messageService.get(AIP_BAD_ORGANIZATIONAL_UNIT, new Object[] { aip.getInfo().getName(), badOrgUnit.getName() })));
        }
      }
    }

    /*
     * Check that deposit contains data files before sending to IN_VALIDATION status
     */
    if (deposit.getStatus() == DepositStatus.APPROVED && ((!deposit.isCollection() && !this.hasDataFiles(deposit.getResId())) ||
            (deposit.isCollection() && deposit.getCollection().isEmpty()))) {

      errors.addError(
              new FieldError(deposit.getClass().getSimpleName(), STATUS_FIELD, this.messageService.get(HAS_DATA_FILES_BEFORE_VALIDATION)));
    }

    /*
     * Check that deposit data files are ready before sending to IN_VALIDATION status
     */
    if (deposit.getStatus() == DepositStatus.APPROVED && !deposit.isCollection() && !this.hasAllDataFileReadyOrCleaned(deposit.getResId())) {
      errors.addError(new FieldError(deposit.getClass().getSimpleName(), STATUS_FIELD, this.messageService.get(
              DATA_FILES_NOT_READY_OR_NOT_CLEANED)));
    }

    /*
     * Check that if additional fields values are present an additional fields form is also present
     */
    if (!StringTool.isNullOrEmpty(deposit.getAdditionalFieldsValues()) && StringTool.isNullOrEmpty(deposit.getAdditionalFieldsFormId())) {
      errors.addError(new FieldError(deposit.getClass().getSimpleName(), "additionalFieldsFormId",
              this.messageService.get(ADDITIONAL_FIELDS_FORM_REQUIRED_WHEN_ADDITIONAL_FIELDS_VALUES_PRESENT)));
    }

    /*
     * Check that if a value is given for custom metadata, it is valid JSON
     */
    if (!StringTool.isNullOrEmpty(deposit.getAdditionalFieldsValues())) {
      try {
        JSONTool.wellformed(deposit.getAdditionalFieldsValues());
      } catch (SolidifyCheckingException e) {
        errors.addError(new FieldError(deposit.getClass().getSimpleName(), "customMetadata", this.messageService.get(
                INVALID_ADDITIONAL_FIELDS_VALUES)));
      }
    }

    /*
     * Check DOI compatibility with metadata version
     */
    if (!StringTool.isNullOrEmpty(deposit.getIsObsoletedBy())
            && (deposit.getMetadataVersion() == DLCMMetadataVersion.V1_0
                    || deposit.getMetadataVersion() == DLCMMetadataVersion.V1_1)) {
      errors.addError(new FieldError(deposit.getClass().getSimpleName(), "isObsoletedBy", this.messageService.get(
              NOT_SUPPORTED, new Object[] { deposit.getMetadataVersion().getVersion() })));
    }
    /*
     * Check that if a value is given for DOI isdIdenticalTo, isReferencedBy and isObsoletedBy, is a
     * valid DOI
     */
    List<RelationType> relationTypeList = Arrays.asList(RelationType.IS_IDENTICAL_TO, RelationType.IS_REFERENCED_BY,
            RelationType.IS_OBSOLETED_BY);
    for (RelationType rType : relationTypeList) {
      this.validateDoiRelationType(deposit, rType, errors);
    }

  }

  private void validateDoiRelationType(Deposit deposit, RelationType relationType, BindingResult errors) {
    String doiRef;
    String fieldName;
    switch (relationType) {
      case IS_IDENTICAL_TO:
        doiRef = deposit.getIsIdenticalTo();
        fieldName = "isIdenticalTo";
        break;
      case IS_REFERENCED_BY:
        doiRef = deposit.getIsReferencedBy();
        fieldName = "isReferencedBy";
        break;
      case IS_OBSOLETED_BY:
        doiRef = deposit.getIsObsoletedBy();
        fieldName = "isObsoletedBy";
        break;
      default:
        throw new IllegalArgumentException("RelationType not supported : " + relationType);
    }
    if (!StringTool.isNullOrEmpty(doiRef) && !ValidationTool.isValidDOI(doiRef)) {
      final String errorMessage = this.messageService.get("validation.deposit.doi.invalid", new Object[] { doiRef });
      errors.addError(new FieldError(deposit.getClass().getSimpleName(), fieldName, errorMessage));
    }
  }

  private boolean needsValidation(Deposit deposit) {
    final SubmissionPolicy submissionPolicy = deposit.getSubmissionPolicy();
    if (submissionPolicy != null) {
      return submissionPolicy.getSubmissionApproval();
    } else {
      throw new SolidifyCheckingException(this.messageService.get("checking.policy.submission.missing"));
    }
  }

  public long getSize(String depositId) {
    return ((DepositRepository) this.itemRepository).getSize(depositId);
  }

  @Override
  public DepositSpecification getSpecification(Deposit resource) {
    if (resource instanceof DepositDto) {
      return new DepositDtoSpecification((DepositDto) resource);
    } else {
      return new DepositSpecification(resource);
    }
  }

  public int getDataFileNumber(String depositId) {
    return ((DepositRepository) this.itemRepository).countDataFiles(depositId);
  }

  /**
   * Return the minimum compliance level value found amongst deposit datafiles
   *
   * @param depositId
   * @return
   */
  public ComplianceLevel getComplianceLevel(String depositId) {
    Integer minComplianceLevel = ((DepositRepository) this.itemRepository).getMinimumDataFilesComplianceLevel(depositId);
    if (minComplianceLevel != null) {
      return ComplianceLevel.fromValue(minComplianceLevel);
    }
    return null;
  }

  public boolean hasDataFiles(String depositId) {
    return this.getDataFileNumber(depositId) > 0;
  }

  public boolean hasDataFileInError(String depositId) {
    return ((DepositRepository) this.itemRepository).countDataFilesInError(depositId) > 0;
  }

  public boolean hasAllDataFileCleaned(String depositId) {
    final int nbDataFilesNotCleaned = ((DepositRepository) this.itemRepository).countDataFilesNotCleaned(depositId);
    return nbDataFilesNotCleaned == 0;
  }

  public boolean hasAllDataFileReadyOrCleaned(String depositId) {
    final int nbDataFilesNotCleanedOrReady = ((DepositRepository) this.itemRepository).countDataFilesNotCleanedOrReady(depositId);
    return nbDataFilesNotCleanedOrReady == 0;
  }

  public boolean hasAllDataFileComplianceLevelCheckedOrInError(String depositId) {
    final int nbDataFilesNotCheckedOrInError = ((DepositRepository) this.itemRepository).countDataFilesNotComplianceLevelCheckedOrInError(
            depositId);
    return nbDataFilesNotCheckedOrInError == 0;
  }

  public int countDataFilesInError(String depositId) {
    return ((DepositRepository) this.itemRepository).countDataFilesInError(depositId);
  }

  public int countDataFilesWithDataCategory(String depositId, DataCategory dataCategory) {
    return ((DepositRepository) this.itemRepository).countDataFilesWithDataCategory(depositId, dataCategory);
  }

  public int countDataFilesWithDataCategoryAndDataType(String depositId, DataCategory dataCategory, DataCategory datatype) {
    return ((DepositRepository) this.itemRepository).countDataFilesWithDataCategoryAndDataType(depositId, dataCategory, datatype);
  }

  public int countContentDataFiles(String depositId) {
    return ((DepositRepository) this.itemRepository).countContentDataFiles(depositId);
  }

  @Override
  public Page<Deposit> findAll(Pageable pageable) {
    Page<Deposit> depositPage = super.findAll(pageable);
    return this.addComputedValues(depositPage);
  }

  @Override
  public Page<Deposit> findAll(Specification<Deposit> spec, Pageable pageable) {
    Page<Deposit> depositPage = super.findAll(spec, pageable);
    return this.addComputedValues(depositPage);
  }

  @Override
  public Deposit findOne(String id) {
    Deposit deposit = super.findOne(id);
    return this.addComputedValues(deposit);
  }

  public void setReadyOrCleaningToCleaned(String depositId) {
    ((DepositRepository) this.itemRepository).setReadyOrCleaningToCleaned(depositId);
  }

  private Deposit addDataFileNumber(Deposit deposit) {
    final int dataFileNumber = this.getDataFileNumber(deposit.getResId());
    deposit.setDataFileNumber(dataFileNumber);
    return deposit;
  }

  private Deposit addComplianceLevel(Deposit deposit) {
    final ComplianceLevel complianceLevel = this.getComplianceLevel(deposit.getResId());
    deposit.setComplianceLevel(complianceLevel);
    return deposit;
  }

  private Deposit addComputedValues(Deposit deposit) {
    this.addDataFileNumber(deposit);
    this.addComplianceLevel(deposit);
    return deposit;
  }

  private Page<Deposit> addComputedValues(Page<Deposit> depositPage) {
    for (Deposit deposit : depositPage) {
      this.addComputedValues(deposit);
    }
    return depositPage;
  }

}
