/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM PreIngest - PreparationService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.preparation;

import java.net.URI;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.OffsetDateTime;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.exception.SolidifyFileDeleteException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.FileTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.business.DepositService;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.model.Access;
import ch.dlcm.model.DataCategory;
import ch.dlcm.model.DataTag;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.service.DLCMService;
import ch.dlcm.service.HistoryService;
import ch.dlcm.service.rest.fallback.FallbackArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackArchiveMetadataRemoteResourceService;
import ch.dlcm.service.rest.fallback.FallbackSubmissionInfoPackageRemoteResourceService;

public abstract class PreparationService extends DLCMService {

  private final String preingestLocation;
  private final long maxFileNumber;

  protected final HistoryService historyService;
  private final DepositService depositService;
  private final FallbackSubmissionInfoPackageRemoteResourceService sipRemoteService;
  private final FallbackArchivalInfoPackageRemoteResourceService aipRemoteService;
  private final FallbackArchiveMetadataRemoteResourceService archiveMetadataRemoteService;

  protected PreparationService(
          DLCMProperties dlcmProperties,
          MessageService messageService,
          HistoryService historyService,
          DepositService depositService,
          FallbackSubmissionInfoPackageRemoteResourceService sipRemoteService,
          FallbackArchivalInfoPackageRemoteResourceService aipRemoteService,
          FallbackArchiveMetadataRemoteResourceService archiveMetadataRemoteService) {
    super(messageService);
    this.preingestLocation = dlcmProperties.getPreingestLocation();
    this.maxFileNumber = dlcmProperties.getParameters().getMaxFileNumber();
    this.historyService = historyService;
    this.depositService = depositService;
    this.sipRemoteService = sipRemoteService;
    this.aipRemoteService = aipRemoteService;
    this.archiveMetadataRemoteService = archiveMetadataRemoteService;
  }

  // **************************
  // ** Methods to implement **
  // **************************

  // Transform Deposit => SIP
  public abstract String generateSIP(Deposit deposit);

  // Transform Deposit => SIP for update
  public abstract String generateSIPForUpdate(Deposit deposit);

  // Get status of Information Package
  public abstract DepositStatus getStatus(Deposit deposit);

  // *********************
  // ** Default Methods **
  // *********************

  // Verify deposit
  public void checkDeposit(Deposit deposit) {
    // Check if there are submission policies
    if (deposit.getSubmissionPolicy() == null) {
      throw new SolidifyCheckingException(this.messageService.get("checking.policy.submission.missing"));
    } else {
      // Check if submission need an approval. It means that there is validation before approbation
      if (Boolean.TRUE.equals(deposit.getSubmissionPolicy().getSubmissionApproval())) {
        // Verify history status
        this.historyService.checkLastEvents(deposit.getResId(), this.messageService.get("checking.policy.submission.novalidation"),
                DepositStatus.CHECKED.toString(), DepositStatus.APPROVED.toString(), DepositStatus.IN_VALIDATION.toString());
      }
    }

    // Check if there are preservation policies
    if (deposit.getPreservationPolicy() == null) {
      throw new SolidifyCheckingException(this.messageService.get("checking.policy.preservation.missing"));
    }

    // Check if there is research data
    final long metadataCount = this.getDepositMetadataCategoryCount(deposit, DataCategory.Package, DataCategory.Metadata);

    if (metadataCount == 0) {
      throw new SolidifyCheckingException(this.messageService.get("checking.metadata.missing"));
    }
    if (metadataCount > 1) {
      throw new SolidifyCheckingException(this.messageService.get("checking.metadata.multiple"));
    }
    // For collection
    if (deposit.isCollection()) {
      this.checkCollectionDeposit(deposit);
    } else {
      final long primaryCount = this.getDepositMetadataCategoryCount(deposit, DataCategory.Primary, null);
      // Check if there is at least a research data file
      if (primaryCount < 1) {
        throw new SolidifyCheckingException(this.messageService.get("checking.data.missing"));
      }
      // Check if the file number max limit is exceeded
      final int fileNumber = this.depositService.countContentDataFiles(deposit.getResId());
      if (fileNumber > this.maxFileNumber) {
        throw new SolidifyCheckingException(this.messageService.get("checking.data.max", new Object[] { fileNumber, this.maxFileNumber }));
      }
    }
    // Check if there is no or one deposit thumbnail
    final long thumbnailCount = this.getDepositMetadataCategoryCount(deposit, DataCategory.Internal, DataCategory.DatasetThumbnail);
    if (thumbnailCount > 1) {
      throw new SolidifyCheckingException(this.messageService.get("checking.dataset.thumbnail.multiple"));
    }
  }

  public boolean depositHasMetadata(Deposit deposit) {
    final long metadataCount = this.getDepositMetadataCategoryCount(deposit, DataCategory.Package, DataCategory.Metadata);

    return (metadataCount > 0);
  }

  // Get working directory for Deposit
  public Path getDepositFolder(String depositId) {
    return Paths.get(this.preingestLocation).resolve(depositId);
  }

  // Get working directory for Deposit for updating metadata
  public Path getDepositFolderForUpdate(String depositId) {
    return Paths.get(this.preingestLocation).resolve(depositId).resolve(DLCMConstants.UPDATE_FOLDER);
  }

  // Get URI of SIP
  public URI getSIPUri(String depositId) {
    return Paths.get(this.preingestLocation).resolve(depositId + SolidifyConstants.ZIP_EXT).toUri();
  }

  // Send SIP to Ingest module
  public String sendSIP(Deposit deposit) {
    return this.sipRemoteService.sendSIP(deposit);
  }

  // Send updated metadata to Ingest module
  public void sendSIPForUpdate(Deposit deposit) {
    this.sipRemoteService.sendSIPForUpdate(deposit);
  }

  // Purge deposit files
  public void deleteAllFiles(Deposit deposit) {
    // Check if deposit folder exists
    // and Delete file on file system
    if (FileTool.checkFile(this.getDepositFolder(deposit.getResId())) && !FileTool.deleteFolder(this.getDepositFolder(deposit.getResId()))) {
      throw new SolidifyFileDeleteException("Cannot delete deposit folder ('" + deposit.getResId() + "')");
    }
  }

  private long getDepositMetadataCategoryCount(Deposit deposit, DataCategory category, DataCategory datatype) {
    if (datatype != null) {
      return this.depositService.countDataFilesWithDataCategoryAndDataType(deposit.getResId(), category, datatype);
    } else {
      return this.depositService.countDataFilesWithDataCategory(deposit.getResId(), category);
    }
  }

  private void checkCollectionDeposit(Deposit deposit) {
    Access maxAipAccessLevel = Access.PUBLIC;
    DataTag maxAipDataTag = DataTag.UNDEFINED;
    boolean aipEmbargo = false;
    Access maxAipEmbargoLevel = Access.PUBLIC;
    OffsetDateTime maxAipEmbargoEnd = OffsetDateTime.now();
    // For each AIP, find max value for Access leve, data tag & emabrgo
    for (final String aipId : deposit.getCollection()) {
      // Check if deposit status is completed or cleaned
      Deposit aipDeposit = this.depositService.findOne(this.archiveMetadataRemoteService.getDepositId(aipId));
      if (aipDeposit.getStatus() != DepositStatus.COMPLETED && aipDeposit.getStatus() != DepositStatus.CLEANED) {
        throw new SolidifyCheckingException("Deposit status of " + aipDeposit.getResId() + " is " + aipDeposit.getStatus()
                + " it should be COMPLETED or CLEANED");
      }
      // Check if AIP status is completed
      ArchivalInfoPackage aip = this.aipRemoteService.getAIP(aipId);
      if (aip.getInfo().getStatus() != PackageStatus.COMPLETED) {
        throw new SolidifyCheckingException(this.messageService.get("checking.storage.notready", new Object[] { aipId }));
      }
      // Keep more restricted access level
      if (aip.getInfo().getAccess().value() > maxAipAccessLevel.value()) {
        maxAipAccessLevel = aip.getInfo().getAccess();
      }
      // Keep more sensitive data tag
      if (aip.getInfo().getDataSensitivity().value() > maxAipDataTag.value()) {
        maxAipDataTag = aip.getInfo().getDataSensitivity();
      }
      // Keep embargo info
      if (aip.getInfo().hasEmbargo()) {
        aipEmbargo = true;
        // Keep more restricted embargo level
        if (aip.getInfo().getEmbargo().getAccess().value() > maxAipEmbargoLevel.value()) {
          maxAipEmbargoLevel = aip.getInfo().getEmbargo().getAccess();
        }
        // Keep more embargo end
        if (aip.getInfo().getEmbargo().getEndDate().isAfter(maxAipEmbargoEnd)) {
          maxAipEmbargoEnd = aip.getInfo().getEmbargo().getEndDate();
        }
      }
    }
    // Check Access levels compatibility ( >= max AIP access level)
    if (deposit.getAccess().value() < maxAipAccessLevel.value()) {
      throw new SolidifyCheckingException(
              this.messageService.get("checking.access.notcompatible", new Object[] { deposit.getAccess(), maxAipAccessLevel }));
    }
    // Check Data Tags compatibility ( >= max AIP access level)
    if (deposit.getDataSensitivity().value() < maxAipDataTag.value()) {
      throw new SolidifyCheckingException(
              this.messageService.get("checking.datatag.notcompatible", new Object[] { deposit.getDataSensitivity(), maxAipDataTag }));
    }

    // Check if there is an active embargo
    if (aipEmbargo && maxAipEmbargoEnd.isAfter(OffsetDateTime.now())) {
      // Check Embargo
      if (deposit.hasEmbargo() != aipEmbargo) {
        throw new SolidifyCheckingException(this.messageService.get("checking.embargo.notcompatible"));
      }
      // Check embargo levels compatibility ( >= max AIP access level)
      if (deposit.getEmbargoAccess().value() < maxAipEmbargoLevel.value()) {
        throw new SolidifyCheckingException(
                this.messageService.get("checking.embargo.access.notcompatible",
                        new Object[] { deposit.getEmbargoAccess(), maxAipEmbargoLevel }));
      }
      // Check embargo end compatibility
      if (deposit.getEmbargo().getEndDate().isBefore(maxAipEmbargoEnd)) {
        throw new SolidifyCheckingException(
                this.messageService.get("checking.embargo.duration.notcompatible",
                        new Object[] { deposit.getEmbargo().getEndDate(), maxAipEmbargoEnd }));
      }
    }
  }

}
