/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM PreIngest - DepositPermissionService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.security;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Service;

import ch.dlcm.business.DepositService;
import ch.dlcm.controller.DLCMControllerAction;
import ch.dlcm.controller.PreIngestController;
import ch.dlcm.model.OrganizationalUnitAwareResource;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.service.rest.trusted.TrustedPersonRemoteResourceService;

@Service
@ConditionalOnBean(PreIngestController.class)
public class DepositPermissionService extends AbstractPermissionWithOrgUnitService {
  private static final Logger log = LoggerFactory.getLogger(DepositPermissionService.class);


  private final DepositService depositService;

  public DepositPermissionService(DepositService depositService, TrustedPersonRemoteResourceService personRemoteResourceService) {
    super(personRemoteResourceService);
    this.depositService = depositService;
  }

  @Override
  protected boolean isManagerAllowed(OrganizationalUnitAwareResource existingResource, DLCMControllerAction action) {

    final DepositStatus status = this.getCurrentStatus(existingResource);
    log.info("Action {} required on deposit '{}' with status {}", action.name(), existingResource.getResId(), status);
    if (status == DepositStatus.IN_PROGRESS || status == DepositStatus.IN_VALIDATION) {
      return true;
    } else if (status == DepositStatus.IN_ERROR || status == DepositStatus.REJECTED) {
      return action == DLCMControllerAction.ENABLE_REVISION;
    } else if (status == DepositStatus.COMPLETED || status == DepositStatus.CLEANED) {
      return action == DLCMControllerAction.START_METADATA_EDITING;
    } else if (status == DepositStatus.EDITING_METADATA) {
      return action == DLCMControllerAction.CANCEL_METADATA_EDITING || action == DLCMControllerAction.UPDATE
              || action == DLCMControllerAction.SUBMIT_FOR_APPROVAL || action == DLCMControllerAction.APPROVE
              || action == DLCMControllerAction.UPLOAD_FILE;
    }

    return false;
  }

  @Override
  protected boolean isStewardAllowed(OrganizationalUnitAwareResource existingResource, DLCMControllerAction action) {
    final DepositStatus status = this.getCurrentStatus(existingResource);

    if (status == DepositStatus.COMPLETED || status == DepositStatus.CLEANED) {
      return action == DLCMControllerAction.START_METADATA_EDITING;
    } else if (status == DepositStatus.EDITING_METADATA) {
      return action == DLCMControllerAction.CANCEL_METADATA_EDITING || action == DLCMControllerAction.UPDATE
              || action == DLCMControllerAction.SUBMIT_FOR_APPROVAL || action == DLCMControllerAction.APPROVE
              || action == DLCMControllerAction.UPLOAD_FILE;
    }
    return this.isApproverAllowed(existingResource, action);
  }

  @Override
  protected boolean isApproverAllowed(OrganizationalUnitAwareResource existingResource, DLCMControllerAction action) {

    final DepositStatus status = this.getCurrentStatus(existingResource);

    if (status == DepositStatus.IN_PROGRESS) {
      return this.isActionAllowedWhenInProgress(action);
    } else if (status == DepositStatus.IN_VALIDATION) {
      if (action == DLCMControllerAction.UPLOAD_FILE ||
              action == DLCMControllerAction.RESUME_FILE ||
              action == DLCMControllerAction.VALIDATE_FILE ||
              action == DLCMControllerAction.DELETE_FOLDER ||
              action == DLCMControllerAction.DELETE_FILE ||
              action == DLCMControllerAction.RESERVE_DOI ||
              action == DLCMControllerAction.DELETE ||
              action == DLCMControllerAction.ENABLE_REVISION ||
              action == DLCMControllerAction.APPROVE ||
              action == DLCMControllerAction.REJECT) {
        return true;
      }
      return false;
    } else if (status == DepositStatus.IN_ERROR || status == DepositStatus.REJECTED) {
      return action == DLCMControllerAction.ENABLE_REVISION;
    } else if (status == DepositStatus.CLEANED || status == DepositStatus.COMPLETED) {
      return action == DLCMControllerAction.CHECK_COMPLIANCE;
    } else if (status == DepositStatus.EDITING_METADATA) {
      return action == DLCMControllerAction.APPROVE
              || action == DLCMControllerAction.SUBMIT_FOR_APPROVAL;
    }

    return false;
  }

  @Override
  protected boolean isCreatorAllowed(OrganizationalUnitAwareResource existingResource, DLCMControllerAction action) {

    final DepositStatus status = this.getCurrentStatus(existingResource);

    if (status == DepositStatus.IN_PROGRESS) {
      return this.isActionAllowedWhenInProgress(action);
    } else if (status == DepositStatus.IN_VALIDATION) {
      return action == DLCMControllerAction.ENABLE_REVISION;
    }

    return false;
  }

  private DepositStatus getCurrentStatus(OrganizationalUnitAwareResource existingResource) {
    final DepositStatus depositStatus = ((Deposit) existingResource).getStatus();
    if(depositStatus == null) {
      return ((Deposit) existingResource).getDefaultStatus();
    } else {
      return depositStatus;
    }
  }

  private boolean isActionAllowedWhenInProgress(DLCMControllerAction action) {
    if (action == DLCMControllerAction.CREATE ||
            action == DLCMControllerAction.UPDATE ||
            action == DLCMControllerAction.UPLOAD_FILE ||
            action == DLCMControllerAction.RESUME_FILE ||
            action == DLCMControllerAction.VALIDATE_FILE ||
            action == DLCMControllerAction.DELETE_FILE ||
            action == DLCMControllerAction.DELETE_FOLDER ||
            action == DLCMControllerAction.UPDATE_FILE ||
            action == DLCMControllerAction.RESERVE_DOI ||
            action == DLCMControllerAction.SUBMIT_FOR_APPROVAL ||
            action == DLCMControllerAction.APPROVE ||
            action == DLCMControllerAction.DELETE) {
      return true;
    }
    return false;
  }

  @Override
  protected Deposit getExistingResource(String resId) {
    return this.depositService.findOne(resId);
  }

}
