/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Preservation Planning - ApplyAipActionJob.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.job;

import org.springframework.data.domain.Pageable;

import ch.unige.solidify.exception.SolidifyProcessingException;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.Result;
import ch.unige.solidify.rest.Result.ActionStatus;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.preservation.PreservationJob;
import ch.dlcm.service.JobProcessingService;
import ch.dlcm.service.rest.trusted.TrustedArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedArchiveDownloadRemoteResourceService;

public class ApplyAipActionJob extends AbstractJob {

  private final TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService;
  private final TrustedArchiveDownloadRemoteResourceService trustedArchiveDownloadRemoteResourceService;

  public ApplyAipActionJob(
          DLCMProperties configService,
          MessageService messageService,
          TrustedArchiveDownloadRemoteResourceService trustedArchiveDownloadRemoteResourceService,
          JobProcessingService jobProcessingService,
          TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService) {
    super(configService, messageService, jobProcessingService);
    this.aipRemoteResourceService = aipRemoteResourceService;
    this.trustedArchiveDownloadRemoteResourceService = trustedArchiveDownloadRemoteResourceService;
  }

  @Override
  public void execute(PreservationJob job, String executionId) {
    // Determine if job applicable on all node or not
    final String[] storageList = this.defineStorageList(job);
    // Compute item total on all storage node
    final long total = this.getTotal(storageList);
    // For each storage node, run action for each aip
    for (final String storageUrl : storageList) {
      job = this.applyAction(job, executionId, total, storageUrl);
    }
  }

  private PreservationJob applyAction(PreservationJob job, String executionId, long total, String storageUrl) {
    Pageable pageable = this.getPageRequest();
    RestCollection<ArchivalInfoPackage> collection;
    do {
      // List AIPs
      collection = this.aipRemoteResourceService.getAipList(storageUrl, pageable);
      pageable = pageable.next();
      for (final ArchivalInfoPackage aip : collection.getData()) {
        try {
          Result res;
          switch (job.getJobType()) {
            case ARCHIVE_CHECK:
              res = this.aipRemoteResourceService.checkAIP(storageUrl, aip.getResId());
              break;
            case REINDEX:
              res = this.aipRemoteResourceService.reindexAIP(storageUrl, aip.getResId());
              break;
            case ARCHIVE_PRELOAD_SMALL:
              res = this.preloadArchive(0L, this.configService.getParameters().getFileSizeLimit().toBytes(), aip);
              break;
            case ARCHIVE_PRELOAD_BIG:
              res = this.preloadArchive(this.configService.getParameters().getFileSizeLimit().toBytes(), Long.MAX_VALUE, aip);
              break;
            default:
              throw new SolidifyProcessingException("Job type " + job.getJobType().getLabel() + " not applicable");
          }

          switch (res.getStatus()) {
            case EXECUTED:
              job = this.addProcessedItem(job, executionId, aip.getResId(), this.getAipPublicUrl(storageUrl, aip.getResId()));
              break;
            case NON_APPLICABLE:
              job = this.addIgnoredItem(job, executionId, aip.getResId(), this.getAipPublicUrl(storageUrl, aip.getResId()), res.getMessage());
              break;
            case NOT_EXECUTED:
            default:
              job = this.addInErrorItem(job, executionId, aip.getResId(), this.getAipPublicUrl(storageUrl, aip.getResId()), res.getMessage());
              break;
          }
        } catch (final Exception e) {
          job = this.addInErrorItem(job, executionId, aip.getResId(), this.getAipPublicUrl(storageUrl, aip.getResId()), e.getMessage());
        }
        // Save intermediate step by 10%
        final long num = job.getLastExecution().getLastExecutionReport().getProcessedItems()
                + job.getLastExecution().getLastExecutionReport().getIgnoredItems()
                + job.getLastExecution().getLastExecutionReport().getInErrorItems();
        job = this.jobProcessingService.saveStep(job, executionId, num, total);
      }
    } while (collection.getPage().hasNext());
    return job;
  }

  private Result preloadArchive(long minSize, long maxSize, ArchivalInfoPackage aip) {
    Result result = new Result(aip.getResId());
    // Check if AIP is completed
    if (!aip.getInfo().getStatus().equals(PackageStatus.COMPLETED)) {
      result.setMesssage(this.messageService.get("job.aip.not_completed", new Object[] { aip.getInfo().getStatus().toString() }));
      return result;
    }
    // Check AIP size
    if (aip.getArchiveSize() < minSize || aip.getArchiveSize() > maxSize) {
      result.setMesssage(this.messageService.get("job.aip.wrong_size",
              new Object[] { aip.getSmartSize(), StringTool.formatSmartSize(minSize), StringTool.formatSmartSize(maxSize) }));
      return result;
    }
    // Check if AIP already downloaded
    try {
      try {
        this.trustedArchiveDownloadRemoteResourceService.getArchiveDownloadStatus(aip.getResId());
        result.setMesssage(this.messageService.get("job.aip.already_preload"));
      } catch (SolidifyResourceNotFoundException e) {
        // Trigger download
        this.trustedArchiveDownloadRemoteResourceService.prepareArchiveDownload(aip.getResId());
        result.setStatus(ActionStatus.EXECUTED);
      }
    } catch (Exception e) {
      result.setStatus(ActionStatus.NOT_EXECUTED);
      result.setMesssage(e.getMessage());
    }

    return result;
  }

  private String[] defineStorageList(PreservationJob job) {
    switch (job.getJobType()) {
      case ARCHIVE_CHECK:
        return this.configService.getArchivalStorageUrls();
      case REINDEX:
      case ARCHIVE_PRELOAD_SMALL:
      case ARCHIVE_PRELOAD_BIG:
        return new String[] { this.configService.getDefaultArchivalStorageUrl() };
      default:
        throw new SolidifyProcessingException("Job type " + job.getJobType().getLabel() + " not applicable");
    }
  }

  private long getTotal(String[] storageList) {
    long total = 0;
    for (final String storageUrl : storageList) {
      final RestCollection<ArchivalInfoPackage> collection = this.aipRemoteResourceService.getAipList(storageUrl, this.getPageRequestForTotal());
      total += collection.getPage().getTotalItems();
    }
    return total;
  }
}
