/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Preservation Planning - CheckFixityJob.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.job;

import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Pageable;

import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.Result;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.exception.DLCMReplicationException;
import ch.dlcm.model.MetadataVisibility;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.preservation.PreservationJob;
import ch.dlcm.service.AipService;
import ch.dlcm.service.JobProcessingService;
import ch.dlcm.service.rest.trusted.TrustedArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedArchiveMetadataRemoteResourceService;

public class CheckFixityJob extends AbstractJob {

  private static final Logger log = LoggerFactory.getLogger(CheckFixityJob.class);

  private final AipService aipService;
  private final TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService;
  private final TrustedArchiveMetadataRemoteResourceService archiveMetadataRemoteResourceService;

  public CheckFixityJob(
          DLCMProperties configService,
          MessageService messageService,
          JobProcessingService jobProcessingService,
          AipService aipService,
          TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService,
          TrustedArchiveMetadataRemoteResourceService archiveMetadataRemoteResourceService) {
    super(configService, messageService, jobProcessingService);
    this.aipService = aipService;
    this.aipRemoteResourceService = aipRemoteResourceService;
    this.archiveMetadataRemoteResourceService = archiveMetadataRemoteResourceService;
  }

  @Override
  public void execute(PreservationJob job, String executionId) {
    Pageable pageable = this.getPageRequest();
    RestCollection<ArchiveMetadata> collection;
    long num = 0;

    // delete eventual AIP infos in cache to get up to date data
    this.aipService.deleteCache();

    do {
      // List AIPs from index
      collection = this.archiveMetadataRemoteResourceService.getMetadataList(MetadataVisibility.PUBLIC, pageable);
      pageable = pageable.next();
      // For each archive
      for (final ArchiveMetadata md : collection.getData()) {
        num++;
        final String aipId = md.getResId();

        try {
          // check that at least one COMPLETED AIP copy exists
          // (throws a DLCMReplicationException if none is found)
          this.aipService.findFirstCompletedCopy(aipId, false);

          // get all COMPLETED copies of the AIP
          // (we don't check copies that are not complete)
          final Map<String, ArchivalInfoPackage> aipList = this.aipService.loadCompletedAips(aipId, false);

          for (final Map.Entry<String, ArchivalInfoPackage> entry : aipList.entrySet()) {
            final String storageUrl = entry.getKey();

            log.info("AIP checksum verification will start on Storage node ({}) for AIP '{}'", storageUrl, aipId);

            try {
              final Result result = this.aipRemoteResourceService.checkAIPFixity(storageUrl, aipId);

              if (result.getStatus() == Result.ActionStatus.EXECUTED) {
                job = this.addProcessedItem(job, executionId, md.getResId(), this.getAipPublicUrl(storageUrl, md.getResId()));
              } else if (result.getStatus() == Result.ActionStatus.NOT_EXECUTED) {
                job = this
                        .addIgnoredItem(job, executionId, md.getResId(), this.getAipPublicUrl(storageUrl, md.getResId()), "Action not executed");
              }
            } catch (final Exception e) {
              job = this.addInErrorItem(job, executionId, md.getResId(), this.getAipPublicUrl(storageUrl, md.getResId()), e.getMessage());
            }
          }
        } catch (final DLCMReplicationException e) {
          job = this.addInErrorItem(job, executionId, md.getResId(),
                  this.getAipPublicUrl(this.configService.getDefaultArchivalStorageUrl(), md.getResId()), e.getMessage());
        }

        job = this.jobProcessingService.saveStep(job, executionId, num, collection.getPage().getTotalItems());
      }
    } while (collection.getPage().hasNext());
  }
}
