/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Preservation Planning - CleanSubmissionJob.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.job;

import java.time.OffsetDateTime;
import java.time.ZoneOffset;

import org.springframework.data.domain.Pageable;

import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.model.PackageStatus;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.SubmissionInfoPackage;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preingest.Deposit.DepositStatus;
import ch.dlcm.model.preservation.PreservationJob;
import ch.dlcm.service.JobProcessingService;
import ch.dlcm.service.rest.trusted.TrustedArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedDepositRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedSubmissionInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedSubmissionPolicyRemoteResourceService;

public class CleanSubmissionJob extends AbstractJob {

  private final TrustedSubmissionPolicyRemoteResourceService submissionPolicyRemoteResourceService;
  private final TrustedDepositRemoteResourceService depositRemoteResourceService;
  private final TrustedSubmissionInfoPackageRemoteResourceService sipRemoteResourceService;
  private final TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService;

  public CleanSubmissionJob(
          DLCMProperties configService,
          MessageService messageService,
          JobProcessingService jobProcessingService,
          TrustedSubmissionPolicyRemoteResourceService submissionPolicyRemoteResourceService,
          TrustedDepositRemoteResourceService depositRemoteResourceService,
          TrustedSubmissionInfoPackageRemoteResourceService sipRemoteResourceService,
          TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService) {
    super(configService, messageService, jobProcessingService);
    this.submissionPolicyRemoteResourceService = submissionPolicyRemoteResourceService;
    this.depositRemoteResourceService = depositRemoteResourceService;
    this.sipRemoteResourceService = sipRemoteResourceService;
    this.aipRemoteResourceService = aipRemoteResourceService;
  }

  @Override
  public void execute(PreservationJob job, String executionId) {
    final String[] archivalStorageUrls = this.configService.getArchivalStorageUrls();
    int count = 0;
    // Loop on completed SIP to delete working SIP associated to deposit
    Pageable sipPageable = this.getPageRequest();
    RestCollection<SubmissionInfoPackage> sipList;
    do {
      // Find all completed SIP
      sipList = this.sipRemoteResourceService.getSIPList(sipPageable);
      // Go to next page
      sipPageable = sipPageable.next();
      for (final SubmissionInfoPackage sipPackage : sipList.getData()) {
        count++;
        // Check if SIP completed
        if (!sipPackage.isCompleted()) {
          job = this.addIgnoredItem(job, executionId, sipPackage.getResId(), this.getSipPublicUrl(sipPackage.getResId()),
                  "Package in-progress");
          continue;
        }
        // Check if submission policy is defined
        if (sipPackage.getSubmissionPolicyId() == null) {
          job = this.addIgnoredItem(job, executionId, sipPackage.getResId(), this.getSipPublicUrl(sipPackage.getResId()),
                  "Missing submission policy");
          continue;
        }
        // Check submission policy retention time
        final OffsetDateTime endTimeToKeep = sipPackage.getCreation()
                .getWhen()
                .plusDays(
                        this.submissionPolicyRemoteResourceService.getSubmissionPolicy(
                                sipPackage.getSubmissionPolicyId())
                                .getTimeToKeep());

        if (endTimeToKeep.isAfter(OffsetDateTime.now(ZoneOffset.UTC))) {
          job = this.addIgnoredItem(job, executionId, sipPackage.getResId(), this.getSipPublicUrl(sipPackage.getResId()),
                  "Too early to clean package");
          continue;
        }

        int replicaNumber = 0;
        for (final String url : archivalStorageUrls) {
          try {
            final ArchivalInfoPackage aip = this.aipRemoteResourceService.getAIP(url, sipPackage.getAipId());
            if (aip.isCompleted()) {
              replicaNumber++;
            }
          } catch (final SolidifyResourceNotFoundException e) {
            // Do not count => missing a replica
          } catch (final Exception e) {
            throw new SolidifyRuntimeException("Error in submission cleaning: " + e.getMessage(), e);
          }
        }
        if (replicaNumber == archivalStorageUrls.length) {
          final Deposit deposit = this.depositRemoteResourceService.getDeposit(sipPackage.getDepositId());
          // Check if SIP cleaned
          if (sipPackage.getInfo().getStatus().equals(PackageStatus.CLEANED) && deposit.getStatus().equals(DepositStatus.CLEANED)) {
            job = this.addIgnoredItem(job, executionId, sipPackage.getResId(), this.getSipPublicUrl(sipPackage.getResId()),
                    "Package already cleaned");
            continue;
          }

          try {
            // update deposit with the CLEANED status
            this.sipRemoteResourceService.cleanSIPDatafile(sipPackage);
            job = this.addProcessedItem(job, executionId, sipPackage.getResId(), this.getSipPublicUrl(sipPackage.getResId()));

            this.depositRemoteResourceService.cleanDepositDatafile(deposit);
            job = this.addProcessedItem(job, executionId, deposit.getResId(), this.getDepositPublicUrl(deposit.getResId()));

          } catch (final Exception e) {
            job = this.addInErrorItem(job, executionId, sipPackage.getResId(), this.getSipPublicUrl(sipPackage.getResId()), e.getMessage());
          }
        } else {
          job = this.addIgnoredItem(job, executionId, sipPackage.getResId(), this.getSipPublicUrl(sipPackage.getResId()),
                  "Missing a copy in replication");
        }
        this.jobProcessingService.saveStep(job, executionId, count, sipList.getPage().getTotalItems());
      }
    } while (sipList.getPage().hasNext());
  }
}
