/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - PurgeOrderJob.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.job;

import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.ArrayList;

import org.springframework.data.domain.Pageable;

import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.model.access.Order;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.DisseminationInfoPackage;
import ch.dlcm.model.preservation.PreservationJob;
import ch.dlcm.service.JobProcessingService;
import ch.dlcm.service.rest.trusted.TrustedAIPDownloadRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedDisseminationInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedOrderRemoteResourceService;

public class PurgeOrderJob extends AbstractJob {

  private final TrustedOrderRemoteResourceService orderRemoteResourceService;
  private final TrustedDisseminationInfoPackageRemoteResourceService disseminationInfoPackageRemoteResourceService;
  private final TrustedAIPDownloadRemoteResourceService aipDownloadRemoteResourceService;
  private long count = 0;
  private ArrayList<String> dipIds;
  private ArrayList<String> aipDownloadIds;
  private final int timeToKeep;

  public PurgeOrderJob(
          DLCMProperties configService,
          MessageService messageService,
          JobProcessingService jobProcessingService,
          TrustedOrderRemoteResourceService orderRemoteResourceService,
          TrustedDisseminationInfoPackageRemoteResourceService disseminationInfoPackageRemoteResourceService,
          TrustedAIPDownloadRemoteResourceService aipDownloadRemoteResourceService) {
    super(configService, messageService, jobProcessingService);
    this.orderRemoteResourceService = orderRemoteResourceService;
    this.disseminationInfoPackageRemoteResourceService = disseminationInfoPackageRemoteResourceService;
    this.aipDownloadRemoteResourceService = aipDownloadRemoteResourceService;
    this.timeToKeep = configService.getPreservationPlanning().getTimeToKeepOrders();
  }

  @Override
  protected void execute(PreservationJob job, String executionId) {
    // Compute item total
    final long total = this.getTotal();
    this.count = 0;
    this.dipIds = new ArrayList<>();
    this.aipDownloadIds = new ArrayList<>();
    job = this.deleteCompletedOrder(job, executionId, total);
    this.deleteDIPAndAIPLinkedToOrders(job, executionId, total);
  }

  private PreservationJob deleteCompletedOrder(PreservationJob job, String executionId, long total) {
    Pageable pageable = this.getPageRequest();
    RestCollection<Order> collection;
    do {
      collection = this.orderRemoteResourceService.getCompletedOrder(pageable);
      // check time to keep
      pageable = pageable.next();
      for (final Order order : collection.getData()) {
        this.count++;
        // Check if order creation time
        if (order.getCreation().getWhen().plusDays(this.timeToKeep).isAfter(OffsetDateTime.now(ZoneOffset.UTC))) {
          job = this.addIgnoredItem(job, executionId, order.getResId(), this.getOrderPublicUrl(order.getResId()),
                  "Too early to remove the order");
          continue;
        }
        // Get all dip and aip ids to preserve them since it is needed to delete first the order
        Pageable dipPageable = this.getPageRequest();
        RestCollection<DisseminationInfoPackage> dipList;
        do {
          dipList = this.disseminationInfoPackageRemoteResourceService.getDIPByOrder(order.getResId(), dipPageable);
          dipPageable = dipPageable.next();
          dipList.getData().forEach(d -> {
            if (!this.dipIds.contains(d.getResId())) {
              this.dipIds.add(d.getResId());
            }
          });
        } while (dipList.getPage().hasNext());

        Pageable aipPageable = this.getPageRequest();
        RestCollection<ArchivalInfoPackage> aipList;
        do {
          aipList = this.aipDownloadRemoteResourceService.getAipsByOrder(order.getResId(), aipPageable);
          aipPageable = dipPageable.next();
          aipList.getData().forEach(d -> {
            if (!this.aipDownloadIds.contains(d.getResId())) {
              this.aipDownloadIds.add(d.getResId());
            }
          });
        } while (aipList.getPage().hasNext());

        try {
          this.orderRemoteResourceService.deleteOrder(order.getResId());
          job = this.addProcessedItem(job, executionId, order.getResId(), null);

        } catch (final Exception e) {
          job = this.addInErrorItem(job, executionId, order.getResId(), this.getOrderPublicUrl(order.getResId()), e.getMessage());
        }
        this.jobProcessingService.saveStep(job, executionId, this.count, total);
      }
    } while (collection.getPage().hasNext());

    return job;
  }

  private void deleteDIPAndAIPLinkedToOrders(PreservationJob job, String executionId, long total) {
    for (String dipId : this.dipIds) {
      try {
        this.count++;
        this.disseminationInfoPackageRemoteResourceService.deleteDIP(dipId);
        job = this.addProcessedItem(job, executionId, dipId, null);
      } catch (final Exception e) {
        job = this.addInErrorItem(job, executionId, dipId, this.getDIPPublicUrl(dipId), e.getMessage());
      }
      this.jobProcessingService.saveStep(job, executionId, this.count, total);
    }

    // Once all the orders have been removed, we can delete all AIP Download
    for (String aipId : this.aipDownloadIds) {
      try {
        this.count++;
        this.aipDownloadRemoteResourceService.deleteAIPDownload(aipId);
        job = this.addProcessedItem(job, executionId, aipId, null);
      } catch (final SolidifyRestException e) {
        // Add as ignore all the AIPs that cannot be removed since there are linked to an existing order not
        // removed by this job
        // (i.e. the ones created by direct download)
        job = this.addIgnoredItem(job, executionId, aipId, this.getAIPDownloadPublicUrl(aipId), e.getMessage());
      } catch (final Exception e) {
        job = this.addInErrorItem(job, executionId, aipId, this.getAIPDownloadPublicUrl(aipId), e.getMessage());
      }
      this.jobProcessingService.saveStep(job, executionId, this.count, total);
    }
  }

  private long getTotal() {
    final RestCollection<Order> orderList = this.orderRemoteResourceService.getCompletedOrder(this.getPageRequestForTotal());
    final int totalDIP = this.disseminationInfoPackageRemoteResourceService.getCountDIPOfCompletedOrders();
    final int totalAip = this.aipDownloadRemoteResourceService.getCountAIPDownloadOfCompletedOrders();
    return orderList.getPage().getTotalItems() + totalAip + totalDIP;
  }
}
