/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Preservation Planning - RebuildRegistryJob.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.job;

import org.springframework.data.domain.Pageable;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.service.MessageService;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.listener.RegistryListener;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.SubmissionInfoPackage;
import ch.dlcm.model.preingest.Deposit;
import ch.dlcm.model.preservation.PreservationJob;
import ch.dlcm.service.JobProcessingService;
import ch.dlcm.service.rest.trusted.TrustedArchivalInfoPackageRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedDepositRemoteResourceService;
import ch.dlcm.service.rest.trusted.TrustedSubmissionInfoPackageRemoteResourceService;

public class RebuildRegistryJob extends AbstractJob {

  private final TrustedDepositRemoteResourceService depositRemoteResourceService;
  private final TrustedSubmissionInfoPackageRemoteResourceService sipRemoteResourceService;
  private final TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService;
  private long count = 0;

  public RebuildRegistryJob(
          DLCMProperties configService,
          MessageService messageService,
          JobProcessingService jobProcessingService,
          TrustedDepositRemoteResourceService depositRemoteResourceService,
          TrustedSubmissionInfoPackageRemoteResourceService sipRemoteResourceService,
          TrustedArchivalInfoPackageRemoteResourceService aipRemoteResourceService) {
    super(configService, messageService, jobProcessingService);
    this.depositRemoteResourceService = depositRemoteResourceService;
    this.sipRemoteResourceService = sipRemoteResourceService;
    this.aipRemoteResourceService = aipRemoteResourceService;
  }

  @Override
  public void execute(PreservationJob job, String executionId) {

    this.count = 0;

    // List AIPs
    job = this.processItems(job, executionId, ArchivalInfoPackage.class);

    // List SIPs
    job = this.processItems(job, executionId, SubmissionInfoPackage.class);

    // List deposits
    this.processItems(job, executionId, Deposit.class);
  }

  private PreservationJob processDataFiles(PreservationJob job, String executionId, String itemId, Class<?> itemClass) {
    long num = 0;
    final RegistryListener registerService = new RegistryListener();
    Pageable pageable = this.getPageRequest();
    RestCollection<? extends Resource> collection;
    do {
      String loggedUrl = null;
      if (itemClass.equals(ArchivalInfoPackage.class)) {
        collection = this.aipRemoteResourceService.getAipDataFileList(this.configService.getDefaultArchivalStorageUrl(), itemId, pageable);
        loggedUrl = this.configService.getDefaultArchivalStoragePublicUrl();
      } else if (itemClass.equals(SubmissionInfoPackage.class)) {
        collection = this.sipRemoteResourceService.getSipDataFileList(itemId, pageable);
        loggedUrl = this.configService.getModule().getIngest().getPublicUrl();
      } else if (itemClass.equals(Deposit.class)) {
        collection = this.depositRemoteResourceService.getDepositDataFileList(itemId, pageable);
        loggedUrl = this.configService.getModule().getPreingest().getPublicUrl();
      } else {
        throw new SolidifyRuntimeException("Unable to process items of type " + itemClass.getName());
      }
      pageable = pageable.next();
      for (final Resource datafile : collection.getData()) {
        num++;
        registerService.register(datafile);
        job = this.addProcessedItem(job, executionId, datafile.getResId(), loggedUrl);
        // Save intermediate step by 10%
        job = this.jobProcessingService.saveStep(job, executionId, num, collection.getPage().getTotalItems());
      }
    } while (collection.getPage().hasNext());
    this.count += num;
    return job;
  }

  private PreservationJob processItems(PreservationJob job, String executionId, Class<?> itemClass) {
    final RegistryListener registerService = new RegistryListener();
    Pageable pageable = this.getPageRequest();
    RestCollection<? extends Resource> collection;

    do {
      String loggedUrl;
      if (itemClass.equals(ArchivalInfoPackage.class)) {
        collection = this.aipRemoteResourceService.getAipList(this.configService.getDefaultArchivalStorageUrl(), pageable);
        loggedUrl = this.configService.getDefaultArchivalStoragePublicUrl();
      } else if (itemClass.equals(SubmissionInfoPackage.class)) {
        collection = this.sipRemoteResourceService.getSIPList(pageable);
        loggedUrl = this.configService.getModule().getIngest().getPublicUrl();
      } else if (itemClass.equals(Deposit.class)) {
        collection = this.depositRemoteResourceService.getDepositList(pageable);
        loggedUrl = this.configService.getModule().getPreingest().getPublicUrl();
      } else {
        throw new SolidifyRuntimeException("Unable to process items of type " + itemClass.getName());
      }
      pageable = pageable.next();
      for (final Resource item : collection.getData()) {
        this.count++;
        registerService.register(item);
        job = this.addProcessedItem(job, executionId, item.getResId(), loggedUrl);
        // List data files
        job = this.processDataFiles(job, executionId, item.getResId(), itemClass);
        // Save intermediate step by 10%
        job = this.jobProcessingService.saveStep(job, executionId, this.count, collection.getPage().getTotalItems());
      }
    } while (collection.getPage().hasNext());

    return job;
  }
}
