/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - AbstractPackageController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.controller;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.springframework.core.io.ClassPathResource;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.servlet.mvc.method.annotation.StreamingResponseBody;

import ch.unige.solidify.ChecksumAlgorithm;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.security.EveryonePermissions;
import ch.unige.solidify.security.UserPermissions;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.DLCMMetadataVersion;
import ch.dlcm.business.DataFileService;
import ch.dlcm.model.AbstractDataFile;
import ch.dlcm.model.DataFileChecksum;
import ch.dlcm.model.DataFileOwnerInterface;
import ch.dlcm.model.dto.FileUploadDto;
import ch.dlcm.model.oais.MetadataVersionAware;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ResourceName;
import ch.dlcm.service.MetadataService;
import ch.dlcm.service.rest.propagate.PropagateMetadataTypeRemoteResourceService;

/**
 * Abstract controller for packages (Deposit, SIP, AIP, ...) with data file(s)
 *
 * @param <S> The type of the data file (DepositDataFile, SIPDataFile, AipDataFile, ...)
 * @param <T> The type of the package (Deposit, SubmissionInfoPackage, ArchivalInfoPackage, ...)
 */
@UserPermissions
public abstract class AbstractPackageController<S extends AbstractDataFile, T extends Resource & MetadataVersionAware & DataFileOwnerInterface<S>>
        extends DLCMResourceController<T> {

  public enum UploadMode {
    STANDARD, UPLOAD_METADATA
  }

  protected final MetadataService metadataService;
  protected final PropagateMetadataTypeRemoteResourceService metadataTypeService;
  protected final DataFileService<S> dataFileService;

  protected AbstractPackageController(PropagateMetadataTypeRemoteResourceService metadataTypeService, MetadataService metadataService,
          DataFileService<S> dataFileService) {
    this.metadataTypeService = metadataTypeService;
    this.metadataService = metadataService;
    this.dataFileService = dataFileService;
  }

  @EveryonePermissions
  @GetMapping("/" + DLCMActionName.LIST_VERSION)
  public HttpEntity<List<String>> listVersions() {
    return new ResponseEntity<>(
            Stream.of(DLCMMetadataVersion.values())
                    .map(DLCMMetadataVersion::getVersion)
                    .collect(Collectors.toList()),
            HttpStatus.OK);
  }

  @EveryonePermissions
  @GetMapping("/" + ResourceName.PROFILE)
  public HttpEntity<StreamingResponseBody> profile(@RequestParam(value = "version", required = false) String version) {
    final DLCMMetadataVersion metadataVersion = this.getVersion(version);
    final ClassPathResource profile = new ClassPathResource(
            SolidifyConstants.SCHEMA_HOME + "/" + metadataVersion.getMetsProfile());
    try {
      final InputStream is = profile.getInputStream();
      return this.buildDownloadResponseEntity(is, profile.getFilename(), SolidifyConstants.XML_MIME_TYPE, profile.contentLength());
    } catch (final IOException e) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
  }

  protected HttpEntity<StreamingResponseBody> getSchemaStream(String version, AbstractDataFile<?, ?> df) {
    final DLCMMetadataVersion metadataVersion = this.getVersion(version);
    final ClassPathResource xsd = this.metadataService.getResourceSchema(metadataVersion, df);
    try {
      final InputStream is = xsd.getInputStream();
      return this.buildDownloadResponseEntity(is, xsd.getFilename(), SolidifyConstants.XML_MIME_TYPE, xsd.contentLength());
    } catch (final IOException e) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
  }

  protected HttpEntity<S> upload(String id, FileUploadDto<S> fileUploadDto, String tmpLocation, UploadMode uploadMode) throws IOException {
    // Check parent resource
    final T item = this.itemService.findOne(id);
    if (!(item.isModifiable() || uploadMode.equals(UploadMode.UPLOAD_METADATA))) {
      throw new IllegalStateException(this.messageService.get("validation.resource.unmodifiable",
              new Object[] { item.getClass().getSimpleName() + ": " + item.getResId() }));
    }

    // Check temporary folder
    String tmpDir = tmpLocation + "/" + id;
    if (!FileTool.ensureFolderExists(Paths.get(tmpDir))) {
      throw new IOException(this.messageService.get("message.upload.folder.notcreated", new Object[] { tmpDir }));
    }

    // Check relative location
    String relativeLocation = fileUploadDto.getRelLocation();

    if (!StringTool.isNullOrEmpty(relativeLocation)) {
      // Check format
      relativeLocation = AbstractDataFile.checkRelativeLocation(relativeLocation);
      // Check temporary folder
      tmpDir += relativeLocation;
      if (!FileTool.ensureFolderExists(Paths.get(tmpDir))) {
        throw new IOException(this.messageService.get("message.upload.folder.notcreated", new Object[] { tmpDir }));
      }
    } else {
      relativeLocation = "/";
    }

    final String fileName;
    switch (uploadMode) {
      case STANDARD:
        fileName = fileUploadDto.getOriginalFileName();
        break;
      case UPLOAD_METADATA:
        fileName = DLCMConstants.UPDATED_METADATA_FILE;
        break;
      default:
        throw new IllegalStateException(this.messageService.get("message.upload.mode.unknown", new Object[] { uploadMode }));
    }

    // check if duplicate data files
    if (this.dataFileService.findByPackageIdAndRelativeLocationAndFileName(item.getResId(), relativeLocation, fileName).isPresent()) {
      String relativeFullName = relativeLocation + "/" + fileName;
      if (relativeLocation.equals("/")) {
        relativeFullName = "/" + fileName;
      }
      throw new IOException(this.messageService.get("message.upload.file.duplicate", new Object[] { relativeFullName }));
    }

    Files.copy(fileUploadDto.getInputStream(), Paths.get(tmpDir).resolve(fileName));

    S dataFile = this.getDataFile(item, Paths.get(tmpDir).resolve(fileName), fileUploadDto);
    this.dataFileService.save(dataFile);

    return new ResponseEntity<>(dataFile, HttpStatus.OK);
  }

  private S getDataFile(T dataFileOwner, Path path, FileUploadDto<S> fileUploadDto) {
    final S dataFile = fileUploadDto.getDataFile();

    dataFile.setSourceData(path.toUri());
    dataFile.setDataCategory(fileUploadDto.getDataCategory());
    dataFile.setDataType(fileUploadDto.getDataType());
    dataFile.setRelativeLocation(fileUploadDto.getRelLocation());
    if (fileUploadDto.getMetadataType() != null) {
      dataFile.setMetadataType(this.metadataTypeService.findOneWithCache(fileUploadDto.getMetadataType()));
    }
    dataFile.setFileSize(FileTool.getSize(path));

    // Set package
    dataFile.setInfoPackage(dataFileOwner);

    this.generateDataFileChecksums(fileUploadDto, dataFile);

    this.setAuthenticatedUserProperties(dataFile);

    return dataFile;
  }

  private DLCMMetadataVersion getVersion(String version) {
    if (!StringTool.isNullOrEmpty(version)) {
      return DLCMMetadataVersion.fromVersion(version);
    }
    return DLCMMetadataVersion.getDefaultVersion();
  }

  private void generateDataFileChecksums(FileUploadDto<S> fileUploadDto, S dataFile) {
    List<DataFileChecksum> listDataFileChecksum = dataFile.getChecksums();
    if (fileUploadDto.getChecksums() != null) {
      for (Map.Entry<ChecksumAlgorithm, String> checksumEntry : fileUploadDto.getChecksums().entrySet()) {
        var dataFileChecksum = new DataFileChecksum();
        dataFileChecksum.init();
        dataFileChecksum.setChecksum(checksumEntry.getValue());
        dataFileChecksum.setChecksumType(DataFileChecksum.ChecksumType.COMPLETE);
        dataFileChecksum.setChecksumAlgo(checksumEntry.getKey());

        DataFileChecksum.ChecksumOrigin checksumOrigin = fileUploadDto.getChecksumsOrigin().get(checksumEntry.getKey());
        if (checksumOrigin == null || !checksumOrigin.equals(DataFileChecksum.ChecksumOrigin.PORTAL)) {
          checksumOrigin = DataFileChecksum.ChecksumOrigin.USER;
        }
        dataFileChecksum.setChecksumOrigin(checksumOrigin);

        listDataFileChecksum.add(dataFileChecksum);
      }
    }
  }

  protected void setAuthenticatedUserProperties(AbstractDataFile<?, ?> resource) {
    final String userId = this.getAuthenticatedUserId();
    if (userId != null) {
      resource.setUpdatedBy(userId);
      if (resource.getCreatedBy() == null) {
        resource.setCreatedBy(userId);
      }
    }
  }

}
