/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - CacheService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service;

import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.CacheManager;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.context.annotation.Lazy;
import org.springframework.jms.annotation.JmsListener;
import org.springframework.stereotype.Service;

import ch.unige.solidify.message.CacheMessage;
import ch.unige.solidify.message.ResourceCacheMessage;
import ch.unige.solidify.rest.CacheNames;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.message.cache.InstitutionPersonRoleCacheMessage;
import ch.dlcm.message.cache.OrganizationalUnitPersonRoleCacheMessage;
import ch.dlcm.model.dto.AverageRating;
import ch.dlcm.model.security.User;
import ch.dlcm.model.settings.OrganizationalUnit;

@Service
public class CacheService {

  private static final Logger logger = LoggerFactory.getLogger(CacheService.class);

  /**
   * Autowiring itself allows cache to work on a method inside its own class
   */
  @Autowired
  @Lazy
  protected CacheService self;

  @Autowired
  CacheManager cacheManager;

  private final Map<String, List<OrganizationalUnit>> cachedOrganizationalUnits = new ConcurrentHashMap<>();

  private final Map<String, Date> cachedOrganizationalUnitsDates = new ConcurrentHashMap<>();

  /**
   * Cache for ArchiveUserRating where is stored: archiveId and list of AverageRating: ratingType and
   * value(average).
   */
  private final Map<String, List<AverageRating>> cachedArchiveAverageRating = new ConcurrentHashMap<>();

  private final Map<String, Date> cachedArchiveAverageRatingDates = new ConcurrentHashMap<>();

  @Autowired
  private DLCMProperties config;

  /**
   * Clear entry in OrganizationalUnitRoles cache identified by a personId and an organizational unit id
   *
   * @param personId
   * @param organizationalUnitId
   */
  @CacheEvict(cacheNames = DLCMCacheNames.ORGANIZATIONAL_UNIT_ROLES)
  public void cleanCacheOrganizationalUnitRoles(String personId, String organizationalUnitId) {
    CacheService.logger
            .trace("'OrganizationalUnitRoles' cache cleared for: personId={}, organizationalUnitId={}", personId, organizationalUnitId);
    this.cachedOrganizationalUnits.clear();
    this.cachedOrganizationalUnitsDates.clear();
  }

  /**
   * Clear entry in InheritedOrganizationalUnitRoles cache identified by a personId and an organizational unit id
   *
   * @param personId
   * @param organizationalUnitId
   */
  @CacheEvict(cacheNames = DLCMCacheNames.INHERITED_ORGANIZATIONAL_UNIT_ROLE)
  public void cleanCacheInheritedOrganizationalUnitRoles(String personId, String organizationalUnitId) {
    // Clean all cache because:
    // 1) Changing institutional roles happens rarely
    // 2) The JMS message InstitutionPersonRoleCacheMessage is sent when an IPR relation is changed and the cache
    //    INHERITED_ORGANIZATIONAL_UNIT_ROLE store role for personId and orgUnitId which belongs to an institution, clearing individual entry
    //    implies to send REST calls to obtain all org units belonging to an institution
    this.cacheManager.getCache(DLCMCacheNames.INHERITED_ORGANIZATIONAL_UNIT_ROLE).clear();
    CacheService.logger.trace("'InheritedOrganizationalUnitRoles' cache cleared");
  }

  /**
   * Clear entry in RESTResources cache identified by its resId
   *
   * @param resId
   */
  @CacheEvict(cacheNames = { CacheNames.RESOURCES, DLCMCacheNames.SUB_RESOURCES })
  public void cleanCacheResources(String resId) {
    CacheService.logger.trace("'Resources' cache cleared for resId: {}", resId);
  }

  /**
   * Clear entry in search cache identified by query
   */
  @CacheEvict(cacheNames = DLCMCacheNames.SEARCH, allEntries = true)
  public void cleanCacheSearch() {
    CacheService.logger.trace("'search' cache cleared ");
  }

  /**
   * Clear entry in UserExternalUid cache identified by externalUid
   *
   * @param externalUid
   */
  @CacheEvict(cacheNames = DLCMCacheNames.USER_EXTERNAL_UID)
  public void cleanCacheUserExternalUid(String externalUid) {
    CacheService.logger.trace("'UserExternalUid' cache cleared for externalUid: {}", externalUid);
  }

  public List<OrganizationalUnit> getCachedAuthorizedOrganizationalUnits(String cacheKey) {

    if (!StringTool.isNullOrEmpty(cacheKey)) {
      final Date cacheDate = this.cachedOrganizationalUnitsDates.get(cacheKey);

      if (cacheDate != null && (System.currentTimeMillis() - cacheDate.getTime()) < (this.config.getParameters().getAuthorizedUnitsCacheTime()
              * 1000)) {
        /*
         * organizationalUnits taken from the cache
         */
        CacheService.logger.trace("organizational units taken from cache for {}", cacheKey);

        return this.cachedOrganizationalUnits.get(cacheKey);
      } else {
        /*
         * organizationalUnits have not been cached yet or cache is too old
         */
        this.cachedOrganizationalUnits.remove(cacheKey);
        return null;
      }
    } else {
      return null;
    }
  }

  public List<AverageRating> getCachedArchiveAverageRating(String cacheKey) {
    if (!StringTool.isNullOrEmpty(cacheKey)) {
      final Date cacheDate = this.cachedArchiveAverageRatingDates.get(cacheKey);

      if (cacheDate != null && (System.currentTimeMillis() - cacheDate.getTime()) < (this.config.getParameters().getArchiveRatingCacheTime()
              * 1000)) {
        /*
         * AverageRating for this Archive taken from the cache
         */
        CacheService.logger.trace("Average Rating taken from cache for {}", cacheKey);

        return this.cachedArchiveAverageRating.get(cacheKey);
      } else {
        /*
         * RatingGlobals for this Archive have not been cached yet or cache is too old
         */
        this.cachedArchiveAverageRating.remove(cacheKey);
        return null;
      }
    } else {
      return null;
    }
  }

  @JmsListener(destination = "${dlcm.topic.cache}", containerFactory = "topicListenerFactory")
  public void processCacheMessage(CacheMessage message) {

    if (message instanceof ResourceCacheMessage) {

      this.self.cleanCacheResources(((ResourceCacheMessage) message).getResId());

      if (message.hasOtherProperty("search")) {
        this.self.cleanCacheSearch();
      }

      if (((ResourceCacheMessage) message).getResourceClass() == User.class) {
        this.self.cleanCacheUserExternalUid(message.getOtherProperty("externalUid").toString());
      }

    } else if (message instanceof OrganizationalUnitPersonRoleCacheMessage) {
      OrganizationalUnitPersonRoleCacheMessage ouprCacheMessage = (OrganizationalUnitPersonRoleCacheMessage) message;
      this.self.cleanCacheOrganizationalUnitRoles(ouprCacheMessage.getPersonId(), ouprCacheMessage.getOrganizationalUnitId());
    } else if (message instanceof InstitutionPersonRoleCacheMessage) {
      InstitutionPersonRoleCacheMessage iprCacheMessage = (InstitutionPersonRoleCacheMessage) message;
      this.self.cleanCacheInheritedOrganizationalUnitRoles(iprCacheMessage.getPersonId(), iprCacheMessage.getInstitutionId());
    }
  }

  public void storeAuthorizedOrganizationalUnits(String cacheKey, List<OrganizationalUnit> organizationalUnits) {
    if (!StringTool.isNullOrEmpty(cacheKey)) {
      CacheService.logger.trace("organizational units put in cache for {}", cacheKey);
      this.cachedOrganizationalUnits.put(cacheKey, organizationalUnits);
      this.cachedOrganizationalUnitsDates.put(cacheKey, new Date());
    }
  }

  public void storeArchiveAverageRating(String cacheKey, List<AverageRating> averageRatings) {
    if (!StringTool.isNullOrEmpty(cacheKey)) {
      CacheService.logger.trace("Average Rating put in cache for {}", cacheKey);
      this.cachedArchiveAverageRating.put(cacheKey, averageRatings);
      this.cachedArchiveAverageRatingDates.put(cacheKey, new Date());
    }
  }

}
