/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - ArchiveMetadataRemoteResourceService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.rest.abstractservice;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.web.util.UriComponentsBuilder;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.service.SolidifyRestClientService;
import ch.unige.solidify.service.rest.abstractservice.IndexMetadataRemoteResourceService;
import ch.unige.solidify.util.StringTool;

import ch.dlcm.DLCMConstants;
import ch.dlcm.config.DLCMProperties;
import ch.dlcm.model.MetadataVisibility;
import ch.dlcm.model.index.ArchiveMetadata;
import ch.dlcm.rest.ResourceName;

public abstract class ArchiveMetadataRemoteResourceService extends IndexMetadataRemoteResourceService<MetadataVisibility, ArchiveMetadata> {
  private static final Logger log = LoggerFactory.getLogger(ArchiveMetadataRemoteResourceService.class);

  private static final String DEPOSIT_ID = "packages.deposits.id";
  private static final String SIP_ID = "packages.sip.id";

  private final DLCMProperties config;

  protected ArchiveMetadataRemoteResourceService(DLCMProperties config, SolidifyRestClientService dlcmRestClientService) {
    super(ArchiveMetadata.class, dlcmRestClientService);
    this.config = config;
  }

  public ArchiveMetadata getMetadata(MetadataVisibility mdVisibility, String aipId) {
    return this.getIndexMetadata(mdVisibility, aipId);
  }

  public ArchiveMetadata getMetadataByDoi(String aipDoi) {
    String valueField = DLCMConstants.IDENTIFIER_FIELD;
    String identifierTypeField = DLCMConstants.IDENTIFIER_TYPE_FIELD;

    valueField = this.formatField(valueField);
    identifierTypeField = this.formatField(identifierTypeField);

    final String query = valueField + ":" + "\"" + aipDoi + "\"" + " AND " + identifierTypeField + ":DOI";

    final RestCollection<ArchiveMetadata> collection = this.getMetadataList(MetadataVisibility.PUBLIC, query,
            null, null, null, null, null, null);
    if (collection.getPage().getTotalItems() == 1) {
      return collection.getData().get(0);
    } else if (collection.getPage().getTotalItems() == 0) {
      throw new SolidifyResourceNotFoundException("Archive with DOI '" + aipDoi + "' not found");
    } else {
      throw new SolidifyRuntimeException("more than one archive correspond to the same DOI '" + aipDoi + "'");
    }
  }

  public RestCollection<ArchiveMetadata> getMetadataList(MetadataVisibility mdVisibility, Map<String, Object> parameters, Pageable pageable) {
    String url = this.buildUrl(mdVisibility, parameters);
    return this.getIndexMetadataListWithUrl(url, pageable);
  }

  public RestCollection<ArchiveMetadata> getMetadataList(MetadataVisibility mdVisibility, Pageable pageable) {
    return this.getMetadataList(mdVisibility, new HashMap<>(), pageable);
  }

  public RestCollection<ArchiveMetadata> getMetadataList(MetadataVisibility mdVisibility, String query, String field, String from, String to,
          String set, String orgUnit, Pageable pageable) {

    HashMap<String, Object> parameters = this.createParameterMap(query, field, from, to, set);
    parameters.put(DLCMConstants.SEARCH_ORG_UNIT, orgUnit);

    return this.getMetadataList(mdVisibility, parameters, pageable);
  }

  public RestCollection<ArchiveMetadata> getMetadataListWithConditions(MetadataVisibility mdVisibility, String query, String conditions,
          Pageable pageable) {
    return this.getIndexMetadataListWithConditions(mdVisibility, query, conditions, pageable);
  }

  public RestCollection<ArchiveMetadata> getMetadataList(MetadataVisibility mdVisibility, String query, String set, Pageable pageable) {
    return this.getIndexMetadataList(mdVisibility, query, set, pageable);
  }

  public String getDepositId(String aipId) {
    ArchiveMetadata aip = this.getMetadata(MetadataVisibility.RESTRICTED, aipId);
    return (String) aip.getMetadata(DEPOSIT_ID.split(SolidifyConstants.REGEX_FIELD_PATH_SEP));
  }

  public String getSipId(String aipId) {
    ArchiveMetadata aip = this.getMetadata(MetadataVisibility.RESTRICTED, aipId);
    return (String) aip.getMetadata(SIP_ID.split(SolidifyConstants.REGEX_FIELD_PATH_SEP));
  }

  public String getAipIdFromDepositId(String depositId) {
    ArchiveMetadata aip = this.getPackage(DEPOSIT_ID, depositId);
    return aip.getResId();
  }

  public String getAipIdFromSipId(String sipId) {
    ArchiveMetadata aip = this.getPackage(SIP_ID, sipId);
    return aip.getResId();
  }

  public List<String> getCollections(String aipId, boolean allParents) {
    List<String> aipList = new ArrayList<>();
    Pageable pageable = PageRequest.of(0, RestCollectionPage.DEFAULT_SIZE_PAGE);
    RestCollection<ArchiveMetadata> archiveMetadataCollection;
    do {
      // List collection of the AIP
      archiveMetadataCollection = this.getArchiveMetadataCollections(aipId, pageable);
      // If no collection, return
      if (archiveMetadataCollection.getData().isEmpty()) {
        return aipList;
      }
      pageable = pageable.next();
      for (final ArchiveMetadata archive : archiveMetadataCollection.getData()) {
        final String collectionId = (String) archive.getMetadata("archiveId");
        aipList.add(collectionId);
        // Retrieve all parent collections recursively
        if (allParents) {
          aipList.addAll(this.getCollections(collectionId, allParents));
        }
      }
    } while (archiveMetadataCollection.getPage().hasNext());
    return aipList;

  }

  protected ArchiveMetadata getPackage(String fieldToSearch, String valueToSearch) {
    final String query = "type:package AND " + fieldToSearch + ":\"" + valueToSearch +"\"";
    RestCollection<ArchiveMetadata> collection = this.getIndexMetadataList(MetadataVisibility.RESTRICTED, query, null, null, null, null, null);
    if (collection.getPage().getTotalItems() == 1) {
      return collection.getData().get(0);
    } else if (collection.getPage().getTotalItems() == 0) {
      throw new SolidifyResourceNotFoundException("Archive not found: " + fieldToSearch + " = " + valueToSearch);
    } else {
      throw new SolidifyRuntimeException("Too many archives: " + query);
    }
  }

  protected RestCollection<ArchiveMetadata> getArchiveMetadataCollections(String aipId, Pageable pageable) {
    final String query = "type:aip AND aip.id:\"" + aipId + "\"";
    return this.getIndexMetadataList(MetadataVisibility.RESTRICTED, query, null, null, null, null, pageable);
  }

  @Override
  protected String buildUrl(MetadataVisibility mdVisibility, Map<String, Object> parameters) {
    String url = this.getIndexUrl(mdVisibility) + "/" + ActionName.SEARCH;
    UriComponentsBuilder uriBuilder = UriComponentsBuilder.fromUriString(url);

    if (parameters != null) {

      this.completeUri(uriBuilder, parameters);

      if (!StringTool.isNullOrEmpty((String) parameters.get(DLCMConstants.SEARCH_ORG_UNIT))) {
        uriBuilder.queryParam(DLCMConstants.SEARCH_ORG_UNIT, parameters.get(DLCMConstants.SEARCH_ORG_UNIT));
      }
    }

    return uriBuilder.toUriString();
  }

  public boolean indexMetadata(List<ArchiveMetadata> metadataList) {
    boolean result = true;
    long count = 0;
    for (ArchiveMetadata metadata : metadataList) {
      String metadataType = ResourceName.DI_PRIVATE;
      if (metadata.getIndex().equals(this.config.getIndexing().getIndexName())) {
        metadataType = ResourceName.DI_PUBLIC;
      }
      log.debug("Creating index: {}) {} {}", ++count, metadataType, metadata.getResId());
      result = result && this.createIndex(metadataType, metadata.getResId(), metadata);
    }
    return result;
  }

  public boolean checkMetadata(String aipId) {
    return (this.headIndex(ResourceName.DI_PUBLIC, aipId) == HttpStatus.OK && this.headIndex(ResourceName.DI_PRIVATE, aipId) == HttpStatus.OK);
  }

  public boolean indexItem(String aipId, ArchiveMetadata md) {
    return this.createIndex(ResourceName.ITEM, aipId, md);
  }

  public boolean checkItem(String aipId) {
    return this.getIndex(ResourceName.ITEM, aipId);
  }

  public boolean updateItem(String aipId, ArchiveMetadata md) {
    return this.updateIndex(ResourceName.ITEM, aipId, md);
  }

  public boolean deleteIndex(String aipId) {
    try {
      this.deleteIndexMetadata(MetadataVisibility.PUBLIC, aipId);
      this.deleteIndexMetadata(MetadataVisibility.RESTRICTED, aipId);
      this.deleteDataFileIndex(aipId);
    } catch (SolidifyResourceNotFoundException | SolidifyRestException e) {
      log.error("Error in deleting metadata index (" + aipId + ")", e);
      return false;
    }
    return true;
  }

  private void deleteDataFileIndex(String aipId) {
    final Pageable page = PageRequest.of(0, RestCollectionPage.MAX_SIZE_PAGE);
    final String url = this.getIndexUrl(MetadataVisibility.RESTRICTED) + "/" + ActionName.SEARCH + "?query=" + DLCMConstants.ARCHIVE_ID_FIELD +
            ":" + aipId;
    RestCollection<ArchiveMetadata> dataFiles = this.restClientService.getResourceList(url, ArchiveMetadata.class, page);
    while (dataFiles.getPage().getTotalItems() > 0) {
      // List IDs
      final List<String> dataFilesIds = dataFiles.getData().stream().map(df -> df.getResId()).collect(Collectors.toList());
      if (!dataFilesIds.isEmpty()) {
        // Delete IDs
        this.restClientService.deleteResourceList(this.getIndexUrl(MetadataVisibility.RESTRICTED), dataFilesIds);
      }
      // List all data files => always first page because of deletion
      dataFiles = this.restClientService.getResourceList(url, ArchiveMetadata.class, page);
    }
  }

  // Build index web service URL
  @Override
  protected String getIndexUrl(MetadataVisibility mdVisibility) {
    String url = this.config.getModule().getDataMgmt().getUrl() + "/";
    switch (mdVisibility) {
      case ALL:
        url += ResourceName.DI;
        break;
      case RESTRICTED:
        url += ResourceName.DI_PRIVATE;
        break;
      case PUBLIC:
      default:
        url += ResourceName.DI_PUBLIC;
        break;

    }
    return url;
  }

  private boolean createIndex(String resource, String aipId, ArchiveMetadata md) {
    try {
      ArchiveMetadata res = this.restClientService.postResource(
              this.config.getModule().getDataMgmt().getUrl() + "/" + resource,
              md, ArchiveMetadata.class);
      if (res == null) {
        return false;
      }

    } catch (SolidifyResourceNotFoundException | SolidifyRestException e) {
      log.error("Error in indexing (" + aipId + ")", e);
      return false;
    }
    return true;
  }

  private HttpStatus headIndex(String resource, String aipId) {
    String url = this.config.getModule().getDataMgmt().getUrl() + "/" + resource + "/" + aipId;
    return this.restClientService.headResource(url);
  }

  private boolean getIndex(String resource, String aipId) {
    this.restClientService.checkResId(aipId);
    try {
      ArchiveMetadata res = this.restClientService.getResource(this.config.getModule().getDataMgmt().getUrl() + "/" + resource + "/" + aipId,
              ArchiveMetadata.class);
      if (res == null) {
        return false;
      }
    } catch (SolidifyResourceNotFoundException e) {
      return false;
    } catch (SolidifyRestException e) {
      log.error("Error in getting " + resource + " (" + aipId + ")", e);
      return false;
    }
    return true;
  }

  private boolean updateIndex(String resource, String aipId, ArchiveMetadata md) {
    try {
      this.restClientService.patchResource(this.config.getModule().getDataMgmt().getUrl() + "/" + resource + "/" + aipId, md,
              ArchiveMetadata.class);
    } catch (SolidifyResourceNotFoundException | SolidifyRestException e) {
      log.error("Error in updating index (" + aipId + ")", e);
      return false;
    }
    return true;
  }

  private String formatField(String query) {
    return query.replace(":", "\\:");
  }

}
