/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Resource Server Common - ArchiveThumbnailRemoteResourceService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.dlcm.service.rest.abstractservice;

import java.nio.file.Path;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.FileSystemResource;
import org.springframework.http.HttpStatus;

import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.security.TokenUsage;
import ch.unige.solidify.service.RemoteResourceService;
import ch.unige.solidify.service.SolidifyRestClientService;

import ch.dlcm.config.DLCMProperties;
import ch.dlcm.model.index.ArchiveThumbnail;
import ch.dlcm.rest.DLCMActionName;
import ch.dlcm.rest.ResourceName;

public abstract class ArchiveThumbnailRemoteResourceService extends RemoteResourceService<ArchiveThumbnail> {
  private static final Logger log = LoggerFactory.getLogger(ArchiveThumbnailRemoteResourceService.class);

  private final String dataMgmtUrl;

  public ArchiveThumbnailRemoteResourceService(DLCMProperties dlcmProperties, SolidifyRestClientService restClientService) {
    super(restClientService);
    this.dataMgmtUrl = dlcmProperties.getModule().getDataMgmt().getUrl();
  }

  public void createArchiveThumbnail(String resId, Path thumbnail) {
    // Create SIP
    ArchiveThumbnail archiveThumbnail = new ArchiveThumbnail();
    archiveThumbnail.setResId(resId);

    try {
      String url = this.dataMgmtUrl + "/" + ResourceName.ARCHIVE_THUMBNAIL;
      // Create object
      this.restClientService.postResource(url, archiveThumbnail, ArchiveThumbnail.class);
      // Send image
      this.restClientService.uploadContent(url + "/" + resId + "/" + DLCMActionName.UPLOAD_THUMBNAIL,
              new FileSystemResource(thumbnail), "file", ArchiveThumbnail.class);
    } catch (Exception e) {
      log.error("Error in creating archive thumbnail ({})", resId);
      throw e;
    }
  }

  public boolean deleteArchiveThumbnail(String resId) {
    try {
      this.restClientService.deleteResource(this.dataMgmtUrl + "/" + ResourceName.ARCHIVE_THUMBNAIL + "/" + resId);
    } catch (SolidifyResourceNotFoundException | SolidifyRestException e) {
      log.error("Error in deleting archive thumbnail(" + resId + ")", e);
      return false;
    }
    return true;
  }

  public boolean checkArchiveThumbnail(String resId) {
    String url = this.dataMgmtUrl + "/" + ResourceName.ARCHIVE_THUMBNAIL + "/" + resId;
    return (this.restClientService.headResource(url) == HttpStatus.OK);
  }

  public void downloadArchiveThumbnail(String resId, Path path) {
    String url = this.dataMgmtUrl + "/" + ResourceName.ARCHIVE_THUMBNAIL + "/" + resId + "/" + DLCMActionName.DOWNLOAD_THUMBNAIL;
    this.restClientService.downloadContent(url, path, TokenUsage.WITH_TOKEN);
  }

  @Override
  protected Class<ArchiveThumbnail> getResourceClass() {
    return ArchiveThumbnail.class;
  }

  @Override
  protected String getResourceUrl() {
    return this.dataMgmtUrl + "/" + ResourceName.ARCHIVE_THUMBNAIL;
  }

}
